/********************************************************************/
/*  o_dir.c - NSR support routines for directory operations         */
/*                                                                  */
/*      Unpublished Confidential Information of Hewlett-Packard     */
/*      Company. Do not disclose.  Copyright (c) Hewlett-Packard    */
/*      Company 1996. All rights reserved                           */
/*                                                                  */
/* The contents of this software are proprietary and confidential   */
/* the Hewlett- Packard Company, and are limited in distribution to */
/* those with a direct need to know.  Individuals having access to  */
/* this software are responsible for maintaining the                */
/* confidentiality of the content and for keeping the software      */
/* secure when not in use.  Transfer to any party is strictly       */
/* forbidden other than as expressly permitted in writing by        */
/* Hewlett-Packard Company.  Unauthorized transfer to or possession */
/* by any unauthorized party may be a criminal offense.             */
/*                                                                  */
/*                   RESTRICTED RIGHTS LEGEND                       */
/*                                                                  */
/*      Use,  duplication,  or disclosure by the                    */
/*      U. S. Government is subject to restrictions as set forth    */
/*      in subparagraph (c) (1) (ii) in the Rights in Technical     */
/*      Data and Computer Software clause in DFARS 252.227-7013     */
/*      or any other successor clause.                              */
/*                                                                  */
/*                       HEWLETT-PACKARD COMPANY                    */
/*                       3000 Hanover St.                           */
/*                       Palo Alto, CA  94304 U.S.A.                */
/*                                                                  */
/*      Rights for non-DOD U.S. Government Departments and Agencies */
/*      are as set forth in FAR 52.227-19 (c) (1,2) or any other    */
/*      successor clause.                                           */
/*                                                                  */
/*  Notes                                                           */
/*      Procedure Filltag is duplicated from mkvol.c                */
/*      ot_getFileID and filename are essentially stubbed and are   */
/*         either not general enough or don't work at all           */
/*                                                                  */
/*  Functions in this module:                                       */
/*      ot_addDirentry      ot_dirempty     ot_dirlook              */
/*      ot_deleteDirentry   ot_direntry     ot_getDirentry          */
/*      ot_writeDirentry    ot_reorgDirectory  ot_dirpad            */
/*                                                                  */
/*  Modification history:                                           */
/*      16.Mar.92   ECB     converted to ANSI function prototypes   */
/*      16.Sep.92   ECB     Changed to ECMA 167 directory format    */
/*                          added ot_dirpad and ot_reorgDirectory   */
/*                          and ot_writeDirentry                    */
/*      03.Dec.92   ECB     Moved FillTag to o_struct.c             */
/*                                                                  */
/*                                                                  */
/********************************************************************/
/*
 $Source: /projects/HMS/nsr/udfOld/src/nsrCore/o_dir.c,v $
 $Revision: 1.28 $      $Author: hmgr $
 $State: Exp $          $Locker:  $
 $Date: 96/05/30 14:52:55 $

 ********************************************************************/

#include <string.h>
#include "Debug.h"
#include "../nsrHdrs/def.h"
#include "../nsrHdrs/o_vndops.h"
#include "../nsrHdrs/nsr.h"
#include "../nsrHdrs/o_node.h"
#include "../nsrHdrs/o_proto.h"
#include "../nsrHdrs/nsrerrno.h"

#if SUN
extern INT32 otfs_debug; /* to enable debugging */
#include "o_fs.h"
#endif

#pragma page
/********************************************************************/
/*  ot_writeDirentry                                                */
/*      write a directory entry at a specified location             */
/*                                                                  */
/*  input parameters:                                               */
/*      fsp:        the otfs structure for the volume               */
/*      directory:  the otnode for the directory file               */
/*      dirbuff:    pointer to a buffer containing the full direc-  */
/*                      tory entry                                  */
/*      offset:     byte location in the file to write the entry    */
/*                                                                  */
/*  returns:                                                        */
/*      SUCCESS:    Directory entry written successfully            */
/*            ?:    I/O Error                                       */
/*                                                                  */
/*  notes:                                                          */
/*      ot_write directory writes a pre-formed directory entry      */
/*        at location offset in the directory file.  No             */
/*        checking for improper overwrites is done...this routine   */
/*        is dumb..it just writes at a location if it can.  The     */
/*        tag field of the directory entry is filled in by          */
/*        ot_writeDirentry just before the write occurs, which is   */
/*        the reason for this routine in the first place            */
/*                                                                  */
/*  Modifcation History:                                            */
/*      17.Aug.92  ECB    Created to support nasty directory tags   */
/*                                                                  */
/********************************************************************/
INT32 ot_writeDirentry(struct otfs *fsp,
                       struct otnode *directory,
                       caddr_t dirbuff,
                       UINT32 offset)

{     
    struct space      *addrList;     /* space list directory append */
    struct space      *nextSct;      /* intermediate variable in write loop */
    struct FileIDDesc *direntry;     /* constant len part of the directory */
    int               dirlen;        /* length of directory entry */
    UINT32            dirsize;       /* temporary copy of dirlen */
    struct uio        uio;           /* structures for I/O */
    struct iovec      iovec;
    UINT32            startBlk;      /* block numbers (from start of file) */
    UINT32            endBlk;        /* of start and end of write */
    caddr_t           writeptr;      /* intermediate buffer pointer */
    UINT32            numbytes;      /* number of bytes remaining to write */
    struct gen_AD     *ADp;          /* pointer to ADs */
    UINT32            ADIndex;       /* 0 based index into ADL for desired extent */
    int               found;         /* TRUE if overlapping AD found */
    UINT32            begExt;        /* Sector offset into file at beginning of AD */
    UINT32            XtentSize;     /* extent size in bytes */
    UINT32            physec;        /* physical sector address of a read */
    UINT32            amtRead;       /* value (in bytes) returned from ot_read() */
    int               FirstWrite;    /* TRUE for first write only of do/while */
    UINT32            xferlen;       /* transfer length (in BYTES) for a write */
    UINT32            lboffset;      /* byte offset within startBlk to begin the transfer */
    INT32             error = 0;

    /* some initialization */

    FirstWrite = TRUE;

    /* put together a uio structure to transfer the data */ 
    /* but first, calculate the length of the entry */

    direntry = (struct FileIDDesc *) dirbuff;
    dirlen = (FILE_ID_DESC_CONSTANT_LEN +
              direntry->L_FI +
              direntry->L_IU +
              ot_dirpad( direntry->L_FI, direntry->L_IU));

    uioInit( &uio, &iovec, (caddr_t) dirbuff, (INT32) offset , (INT32) dirlen);
    
    /* The transfer may be one of three kinds: */
    /* 1) (unlikely) The new entry will start on  a logical block */
    /* boundary, so we just go get the needed logical block(s) and */
    /* write it. */
    /* 2) (more likely) The new entry starts somewhere in the middle */
    /* of an existing block and ends in that block */
    /* 3) Ends in a new block. Hence, we start do/while loop to get */
    /* the entire job done */

    do {
        
        /* what will be the starting and ending blocks in the file? */
        /* ISSUE -Requires changes for proper handling of logical blocks */
        
        startBlk  = uio.uio_offset/fsp->SectorSize;
        endBlk    = (uio.uio_offset + uio.uio_resid - 1) / fsp->SectorSize;
        XtentSize = (endBlk-startBlk+1) * fsp->SectorSize;
        lboffset  = uio.uio_offset % fsp->SectorSize;
        numbytes  = uio.uio_resid;
        
        /* now, is this a partial or full update of an already written */
        /* LB? (Cases 2) or 3) above) Look into the ADL to find out */
        /* if  the targeted sectors are already part of the file */

        found   = ot_searchADL(directory,startBlk,&begExt,&ADIndex);
        if (found)
            ADp  = directory->ADL.beginADs + ADIndex;
        
        /* if we found some overlap, we have to read the overlapped */
        /* logical block and stick all or part of the directory onto */
        /* it and then rewrite it. We may also have to run through the */
        /* loop one more time to allocate a block and finish writing */
        /* the directory data */

        /* Here we see if the length of the directory to write will */
        /* fall off the end of the extent.  If it does, we trim the */
        /* directory to fit, and we'll get the rest on the next pass */
        
        if ((found) &&
            (endBlk > begExt + (ADp->InfoLength/fsp->SectorSize)-1) ) {
            endBlk = begExt + (ADp->InfoLength/fsp->SectorSize) -1;
            XtentSize = (endBlk-startBlk+1) * fsp->SectorSize;
            numbytes = XtentSize - lboffset;
        }
        
        /* read the extent to update if it is written */
        
        bzero( fsp->iobuf, IOBUFSIZE);
        if( found && ADp->ExtentLength.bf.Type==E_RECORDED) {
            physec = startBlk - begExt + ADp->Location_LBN;
            amtRead = XtentSize;
            if( error = ot_read(fsp, physec, fsp->iobuf, &amtRead)) {
                return error;  
            }
        }
        
        /* now do we update in place or get some space to do the write? */
        
        if( found && (fsp->MediaType > (int) fsWORM ||
                      ADp->ExtentLength.bf.Type==E_ALLOCATED)) {
            
            /* update in place */
            
            physec = startBlk - begExt + ADp->Location_LBN;
            if (FirstWrite) {
                FillTag(TAGID_FILE_ID, 1, 1, NOCRC, 0, physec,
                        (struct Tag *) dirbuff);
#if !IBM
                ot_xlate(TAGID_FILE_ID, TO_OP, (caddr_t) dirbuff,dirlen,0);
#endif
                FirstWrite = FALSE;
            }

            /* move what we can into the iobuffer */
        
            uiomove( fsp->iobuf + lboffset, numbytes, UIO_WRITE, &uio);
            writeptr = fsp->iobuf;
            
            if( fsp->MediaType == (int) fsMO) {
                error = ot_erase(fsp, 1, (UINT32) physec,
                                 (UINT32) XtentSize/fsp->SectorSize);
            }

            if (!error) {
                error = ot_write(fsp, physec, writeptr, XtentSize);
            }

            ADp->ExtentLength.bf.Type = E_RECORDED;
            
        } else {
            
            /* get space for the write */
            
            addrList = ot_getSpace(fsp, (UINT32)(endBlk-startBlk+1),
                                   nonContiguous);
            
            if( addrList == NULL) {
                error = ENOSPC;
            } else {
                
                /* now that we know where the data is going, assemble */
                /* the tag - what a pain! */

                if (FirstWrite) {
                    FillTag(TAGID_FILE_ID,1,1,NOCRC,0,addrList->start,
                            (struct Tag *) dirbuff);
#if !IBM
                    ot_xlate(TAGID_FILE_ID,TO_OP,(caddr_t) direntry,dirlen,0);
#endif
                    FirstWrite = FALSE;
                }

                /* now we can do the write.... */
                /* move what we can into the iobuffer */
        
                uiomove( fsp->iobuf + lboffset, numbytes, UIO_WRITE, &uio);
                writeptr = fsp->iobuf;
                
                for (nextSct = addrList; nextSct && !error;
                     nextSct = nextSct->next) 
                {
                    dirsize = dirlen;
                    if(nextSct->length * fsp->SectorSize < dirsize) {
                        xferlen = nextSct->length * fsp->SectorSize;
                    } else {
                        xferlen = dirsize;
                    }
                    error = ot_write(fsp, nextSct->start,
                                     (caddr_t)writeptr,
                                     xferlen < fsp->SectorSize ? fsp->SectorSize
                                                     : fsp->SectorSize,xferlen);
                    writeptr += nextSct->length * fsp->SectorSize;
                    dirsize -= xferlen;
                }
                if( !error) {
                    error = ot_addToADL(directory,startBlk,E_RECORDED,addrList);
                }
                killSpaceList( &addrList);
            } /* addrList != NULL */
        } /* not an update in place */
        
        /* adjust the filesize and the modified flag */
        /* ISSUE - How about modification dates, etc.? */
        if(offset + dirlen  > directory->icbe.fe.InfoLengthL && !error) {
            /* ISSUE - dirlen should not be used here - eg, if we just wrote
            * part of the direntry.
            */
            directory->icbe.fe.InfoLengthL = offset + dirlen;
            /* ISSUE - need to also update LogBlocks field */
            directory->flags |= OT_MOD;
        }
    } while (uio.uio_resid > 0 && numbytes != 0 && !error);
    return( error);
}

#pragma page
/********************************************************************/
/*  ot_addDirentry                                                  */
/*      routine to add a directory entry                            */
/*                                                                  */
/*  input parameters:                                               */
/*      fsp:        the otfs structure for the volume               */
/*      otp:        the otnode for the file (data fork)             */
/*      directory:  the otnode for the directory in which the file  */
/*                      is to be created                            */
/*      rfotp:      the otnode for the resource fork for the file   */
/*                                                                  */
/*  returns:                                                        */
/*      0           success - This might mean nothing happened if   */
/*                      the entry filename already exists           */
/*                  ISSUE - do we validate character sets here      */
/*                  ISSUE - out of memory ENOMEM                    */
/*                                                                  */
/*                                                                  */
/*  notes:                                                          */
/*      rfotp may be zero if associated files are not used          */
/*          by the implementation                                   */
/*                                                                  */
/*      Does not support sorted directories. New entries are just   */
/*          appended to the end of the directory                    */
/*      Does not support logical blocks, e.g. log blk size = sector */
/*          size.                                                   */
/*      Does not support implementation use area <> 0               */
/*      Even though directories are files, we get none of the ad-   */
/*          vantage here because we can't properly fill the tags    */
/*          at the beginning of ever entry without knowing the      */
/*          block address of ever entry,hence we have to do all of  */
/*          the work of writing and updating them.  %&@&^#$!!!      */
/*                                                                  */
/*  Modification history:                                           */
/*      27.Dec.89   DRaM    sorted directories                      */
/*      16.Sep.92   ECB     ECMA167 directory format                */
/********************************************************************/
INT32 ot_addDirentry(register struct otfs *fsp,
                     struct otnode *otp,
                     struct otnode *rfotp,
                     struct otnode *directory)
     
{
    UINT32            sectorBegins,sectorIndex;
    struct space      *addrList;     /* space list directory append */
    struct space      *nextSct;      /* intermediate variable in write loop */
    register INT32    error;
    char              *dirbuff;      /* entire directory entry contents */
    struct FileIDDesc *direntry;     /* constant len part of the directory */
    int               charset=5;     /* character set for filename work */
    UCHAR8            namelen;       /* length of filename in entry */
    UINT16            impuselen=4;   /* length of implementation use field */
    int               dirlen;        /* length of directory entry */
    UINT32            offset;        /* bytes into first LB of extent to write */
    caddr_t           writeptr;      /* intermediate buffer pointer */
    
    /* This call to ot_getDirentry presumes the file will NOT be */
    /* found, we are just using it to get sectorIndex, the byte offset */
    /* of the end of the file so we know where to append. Deleted */
    /* files WILL NOT be found. (see getDirentry). */
    error = ot_getDirentry( fsp, directory, &direntry, &sectorBegins,
                           &sectorIndex, otp->name, (UINT32) 0,
                           findByName);
    if (error == SUCCESS){
        kmem_free((caddr_t) direntry, (FILE_ID_DESC_CONSTANT_LEN
                                       + direntry->L_FI
                                       + direntry->L_IU), BUFRAM);
    }
    if (error == ENOENT) {
        error = SUCCESS;
        /* assemble the entry */

        /* start by figuring out how long it is (ECMA 167 directories */
        /* are variable length namelen returns the file id length. */
        /* Charset identifies how to interpret the bytes */
        /* ISSUE - charset is CS5 - ISO 8859-1 for now */
        
        error = ot_fileName((UCHAR8 *)otp->name, charset, &namelen);
        dirlen = (FILE_ID_DESC_CONSTANT_LEN + namelen
                  + impuselen + ot_dirpad(namelen, impuselen));
        
        /* go ahead and allocate some space for the directory entry */
        dirbuff = kmem_alloc( (UINT32) dirlen, BUFRAM);
        if (dirbuff == NULL) {
            error = ENOMEM;
        }
        bzero((caddr_t) dirbuff, dirlen);

        /* load up the directory entry */
        direntry = (struct FileIDDesc *) dirbuff;
        
        direntry->VersionNum = 1;
        if( otp->vnode.v_type == VDIR) {
            direntry->Characteristics |= DIR_ATTR;
        }
        direntry->L_FI = namelen;
        direntry->ICB.ExtentLength.bf.Length = fsp->ICBsize*fsp->SectorSize;
        direntry->ICB.ExtentLength.bf.Type = E_RECORDED;
        direntry->ICB.Location_LBN = otp->icb_addr_LBN;
        direntry->ICB.Location_PartNo = otp->icb_addr_PartNo;
        direntry->L_IU = impuselen;
        
        /* copy the name field */
        memcpy(dirbuff + FILE_ID_DESC_CONSTANT_LEN + direntry->L_IU,
               otp->name, namelen); 
        
        /* if a resource fork otnode is present, copy that into the */
        /* Implementation use area of the ICB address. */
        /* ISSUE - The resource fork address will go into the icb entry  */
        
        if( rfotp) {
            *( UINT32 *) direntry->ICB.aImpUse[2] = rfotp->icb_addr_LBN;
        }
        error = ot_writeDirentry(fsp, directory, dirbuff,
                                 sectorBegins*fsp->SectorSize
                                 + sectorIndex);
        /* get rid of directory buffer */
        kmem_free( (caddr_t) dirbuff, dirlen, BUFRAM);
    } /* end ot_getDirentry == ENOENT */
    return error;
} /* end ot_addDirentry */

/********************************************************************/
/*  ot_deleteDirentry                                               */
/*      routine to delete a directory entry                         */
/*                                                                  */
/*  input parameters:                                               */
/*      fsp:        the otfs structure for the volume               */
/*      directory:  the otnode for the directory                    */
/*      name:       the name of the entry to delete                 */
/*                                                                  */
/*  returns:                                                        */
/*      0           success                                         */
/*                                                                  */
/*  notes:                                                          */
/*      Just marks an entry as deleted.  Run ot_reorgDirectory      */
/*         to eliminate "voids" and make directories contiguous.    */
/*                                                                  */
/*  Modification history:                                           */
/*      27.Dec.89   DRaM    sorted directories                      */
/*      17.Aug.89   ECB     ECMA 167 compatibility                  */
/********************************************************************/
INT32 ot_deleteDirentry(register struct otfs *fsp,
                        struct otnode *directory,
                        caddr_t name)
{
    UINT32                          beginOffset,sectorOffset;
    struct FileIDDesc               *theEntry;
    register INT32                  error;
    UINT32                          dirlen;

    /* get the desired entry pointed to by theEntry */
    error = ot_getDirentry(fsp, directory, &theEntry, &beginOffset,
                           &sectorOffset, name, (UINT32) 0,
                           findByName);
    if (!error) {
        dirlen = (FILE_ID_DESC_CONSTANT_LEN
                  + theEntry->L_FI
                  + theEntry->L_IU
                  + ot_dirpad(theEntry->L_FI + theEntry->L_IU));

        /* mark the entry as deleted */
        theEntry->Characteristics |= DELETE_ATTR;
        
        /* now write it back */
        error = ot_writeDirentry(fsp, directory, (caddr_t) theEntry,
                                 sectorOffset+beginOffset); 

        kmem_free((caddr_t)theEntry, dirlen, 0);

        /* mark the otnode as modified */
        directory->flags |= OT_MOD;
    }
    return error;
} /* end ot_deleteDirentry */

/********************************************************************/
/*  ot_dirempty                                                     */
/*      routine to check if a directory is empty                    */
/*                                                                  */
/*  input parameters:                                               */
/*      fsp:        the otfs structure for the volume               */
/*      directory:  the otnode for the directory to check           */
/*                                                                  */
/*  returns:                                                        */
/*      0           success, the directory is empty                 */
/*      ENOENT:     directory is not the vnode for a directory      */
/*      ENOTEMPTY:  the directory is not empty                      */
/*                                                                  */
/*  notes:                                                          */
/*      a directory is empty if it only contains the ".." entry     */
/********************************************************************/
INT32 ot_dirempty(struct otfs *fsp,
                  struct otnode *directory)

{
    UINT32          NumFiles = 0;
    register INT32  error;
    UINT16 vacancy;

    if (directory->vnode.v_type != VDIR)
        return ENOENT;
    error = ot_getNumFiles( fsp, directory->icb_addr_LBN, sumTotal,
                           &NumFiles, &vacancy);
    if (!error && NumFiles)
        return ENOTEMPTY;
    else if (!error && !NumFiles)
        return 0;
    else
        return error;
} /* end ot_dirempty */

/********************************************************************/
/*  ot_direntry                                                     */
/*      routine to find or create a file or directory entry         */
/*                                                                  */
/*  input parameters:                                               */
/*      fsp:    the otfs structure for the volume                   */
/*      otp:    the otnode for the directory in which to look       */
/*      name:   the name to look for                                */
/*      vap:    the attributes for the new file/directory, if being */
/*                  created                                         */
/*      otpp:   for the file (data fork)                            */
/*      type:   VREG or VFORKS                                      */
/*      rfotpp: for the associated file (resource fork)             */
/*                                                                  */
/*  output parameters:                                              */
/*      otpp:   contains the otnode for the data fork whether       */
/*                  found or created                                */
/*      rfotpp: contains the otnode for the resource fork if found  */
/*                  or created (and requested)                      */
/*  returns:                                                        */
/*      0       success                                             */
/*                                                                  */
/*  notes:                                                          */
/*      if type == VFORKS, both the data and resource forks will    */
/*          be found/created. if the data fork exists but the       */
/*          resource doesn't, then otpp will be filled, but rfotpp  */
/*          will be 0.                                              */
/*      if type == VREG, then just the data fork will be            */
/*          found/created. if the resource fork does exist, it will */
/*          not be returned.                                        */
/********************************************************************/
INT32 ot_direntry(register struct otfs *fsp,
                  struct otnode *otp,
                  caddr_t name,
                  struct vattr *vap,
                  register struct otnode **otpp,
                  enum vtype type,
                  struct otnode **rfotpp)
{
    register INT32  error;
    *otpp = *rfotpp = (struct otnode *) 0;

    /* check for existence */
    error = ot_getAndFillOTNode( fsp, otp, otpp, rfotpp, name, type,
            (struct FileIDDesc *) 0,LAST);
    if (error == ENOENT) {
        DMLPRINTF(oDirLev)("ot_direntry: ");
        DMLPRINTF(oDirLev)("Doesn't already exist, make a new one\n");

        /* does not exist, get new one */
        error = ot_setupDirentryOTNode( fsp, otpp, vap, vap->va_type,
                                       name, otp->icb_addr_LBN);
        /* create resource fork? */
        if (error == 0 && type == VFORKS) {
            error = ot_setupDirentryOTNode( fsp, rfotpp, vap, VASSOC,
                                           name, otp->icb_addr_LBN); 
            if (error) {
                returnSectors( fsp, (struct space *) 0,
                              (*otpp)->icb_addr_LBN, (UINT32)
                              fsp->ICBsize, bmapErased); 
                /* force release of new otnode */
                (*otpp)->flags = 0;
                (*otpp)->vnode.v_count = 0;
                otvn_inactive(&(*otpp)->vnode,(struct ucred *) 0);
                *otpp = (struct otnode *)0;

                /* was the resource fork otnode allocated? if so, */
                /* force its release */ 

                if (*rfotpp) {
                    (*rfotpp)->flags = 0;
                    (*rfotpp)->vnode.v_count = 0;
                    otvn_inactive( &(*rfotpp)->vnode,(struct ucred *) 0);
                    *rfotpp = (struct otnode *)0;
                }
            }
        }
    }
    DMLPRINTF(oDirLev)("ot_direntry returning error %x\n",error);
    return error;
} /* end ot_direntry */

/* ISSUE - DO we need to clean up the language of versions and */
/* revisions as established by the committee? */

/********************************************************************/
/*  ot_dirlook                                                      */
/*      routine to look in a directory for an entry and return      */
/*      an otnode for it                                            */
/*                                                                  */
/*  input parameters:                                               */
/*      fsp:        the otfs structure for the file                 */
/*      directory:  the otnode for the directory in which to look   */
/*      nm:         the name to look for                            */
/*      type:       VASSOC or VREG                                  */
/*      version:    the version of the file's icb entry to find     */
/*                                                                  */
/*  output parameters:                                              */
/*      otpp:       contains the otnode for the entry, if found     */
/*                                                                  */
/*  returns:                                                        */
/*      EEXIST:     file exists (success)                           */
/*      ENOENT:     directory vnode's type is not VDIR              */
/*                                                                  */
/*  notes:                                                          */
/*      type == VREG means find the data fork icb entry,            */
/*      type == VASSOC  means resource fork icb                     */
/*                                                                  */
/*  Modification history:                                           */
/********************************************************************/
INT32 ot_dirlook(struct otfs      *fsp,
                 struct otnode    *directory,
                 caddr_t          nm,
                 struct otnode    **otpp,
                 enum vtype       type,
                 UINT32           version)

{
    int                     isSpecial;
    struct otnode           *df,
                            *rf;
    register INT32          error;
    register struct otnode  *active,
                            *dfotp,
                            *rfotp;

    DMLPRINTF(oDirLev)("ot_dirlook: top\n");
    *otpp = df = rf = (struct otnode *) NULL;

    /* ISSUE - String handling for file names not set up for multibyte */
    /* char sets*/

    /* you can only look inside a directory */
    if (directory->vnode.v_type != VDIR) {
        DMLPRINTF(oDirLev)("ot_dirlook: return ENOENT 1\n");
        return ENOENT;
    }
    /* at root and looking for parent? if so, return error since the */
    /* root has no parent */
    if ((directory->icb_addr_LBN == fsp->rootAddr) &&
        (strcmp( nm,"..") == 0)) {
        DMLPRINTF(oDirLev)("ot_dirlook: return ENOENT 2\n");
        return ENOENT;
    }

    /* Looking for info on current directory? */
    if (strcmp(nm,".") == 0) {
        directory->vnode.v_count++;
        *otpp = directory;
        DMLPRINTF(oDirLev)("ot_dirlook: return EEXIST 1\n");
        return EEXIST;
    }
    isSpecial = strcmp(nm,"..") == 0;

    /* check active otnode list */
    for (active = fsp->active; active; active = active->next) {

/* ISSUE - references to LBN are only good for single volume references */

        if (ot_active(active) &&
            (isSpecial ? directory->parent_addr_LBN == active->icb_addr_LBN :
             active->parent_addr_LBN == directory->icb_addr_LBN) &&
            (isSpecial ? 1 : strcmp(active->name,nm) == 0) &&
            ((version == LAST && active->isLast) || version ==
             active->version) &&
            (type == VASSOC) == (active->vnode.v_type == VASSOC)) {
                active->vnode.v_count++;
                *otpp = active;
                DMLPRINTF(oDirLev)("ot_dirlook: return EEXIST 2\n");
                return EEXIST;
        }
    }

    /* check inactive cache; might still be hanging around */
    for (active = fsp->inactive; active; active = active->next) {

        if ((isSpecial ? 1 : strcmp( active->name,nm) == NULL) &&
            (isSpecial ?
             directory->parent_addr_LBN == active->icb_addr_LBN :
             active->parent_addr_LBN == directory->icb_addr_LBN) &&
            ((version == LAST && active->isLast)
             || version == active->version)
            && (type == VASSOC) == (active->vnode.v_type == VASSOC)) {
            ot_dequeueOTNode(&fsp->inactive,active);
            ot_enqueueOTNode(&fsp->active,active);
            active->vnode.v_count = 1;
            *otpp = active;
            DMLPRINTF(oDirLev)("ot_dirlook: return EEXIST 3\n");
            return EEXIST;
        }
    }

    DMLPRINTF(oDirLev)("ot_dirlook: before call to ot_getAndFillOTNode\n");
    error = ot_getAndFillOTNode( fsp, directory, &df, &rf, nm,
                                (type == VASSOC) ? type : VNON,
                                (struct FileIDDesc *) NULL, version);
    dfotp = df;
    rfotp = rf;

    if (error == EEXIST) {
        if (type == VASSOC)
            *otpp = rfotp;
        else
            *otpp = dfotp;
    }

    if (error && error != EEXIST) {
        if (dfotp) {        
            dfotp->vnode.v_count--;
            otvn_inactive(&dfotp->vnode,(struct ucred *) 0);
        }
        if (rfotp) {
            rfotp->vnode.v_count--;
            otvn_inactive(&rfotp->vnode,(struct ucred *) 0);
        }
    }
    DMLPRINTF(oDirLev)("ot_dirlook: bottom error = %x\n", error);
    return error;
} /* end ot_dirlook */

/********************************************************************/
/*  ot_getDirentry                                                  */
/*      routine to get directory information for a given named file */
/*                                                                  */
/*  input parameters:                                               */
/*      fsp:        the otfs structure for the volume               */
/*      directory:  the otnode for the directory in which to look   */
/*      name:       name of entry for which to look                 */
/*      icbaddr:    address of icb entry for which to look          */
/*      searchType: findByName, findByAddress                       */
/*                                                                  */
/*  output parameters:                                              */
/*      *DEntrypp:  actual directory entry, if found                */
/*                  **needs to be deallocated by calling routine**  */
/*      *sectorBegins:  byte offset of sector that holds directory  */
/*      *sectorIndex:   byte offset of start of directory entry     */
/*                      from beginning of sector.  sectorBegins +   */
/*                      sectorIndex = offset from start of file for */
/*                      found entry.  If entry not found, then      */
/*                      sectorBegins and sectorIndex point to byte  */
/*                      (currently at end of file) where directory  */
/*                      would go.                                   */
/*                                                                  */
/*  returns:                                                        */
/*      0           success                                         */
/*      ENOENT:     entry not found                                 */
/*      ENOMEM:     memory allocation failure                       */
/*                                                                  */
/*  notes:                                                          */
/*      if searchType is findByAddress, a linear search for an      */
/*          entry with icb address "icbaddr" is performed, and      */
/*          "name" is ignored.                                      */
/*      if searchType is findByName, a linear search for an entry   */
/*          with name "name" is performed, and "icbaddr" is ignored.*/
/*      if entry was not found, return value is ENOENT.             */
/*                                                                  */
/*  Modification history:                                           */
/*      27.Dec.89   DRaM    sorted directories                      */
/*      18.Sep.92   ECB     Major mods for ECMA 167 nonsorted dirs  */
/********************************************************************/
INT32 ot_getDirentry(struct otfs *fsp,
                     struct otnode *directory,
                     struct FileIDDesc **DEntrypp,
                     UINT32 *sectorBegins,
                     UINT32 *sectorIndex,
                     caddr_t name,
                     UINT32 icbaddr,
                     enum searchType searchType)

{
    int                        found;
    register INT32             error;
    register struct FileIDDesc *tempEntry;
    UINT32                     bytesread;   /* bytes processed so far reading directory */
    UINT32                     bytestoread; /* remaining bytes in directory file */
                                            /* to process */
    UINT32                     filelen;     /* just what it sez... */
    UINT32                     offset;      /* bytes from beginning of iobuff to start */
                                            /* of directory entry under consideration */
    struct uio                 uio;         /* for I/O */
    struct iovec               iovec;
    caddr_t                    iobuff;      /* buffer for directory read/writes */
    UINT32                     iobuffsize;
    UINT32                     xferlen;     /* transfer length */
    struct FileIDDesc          *direntry;
    UINT32                     dirlen;
    
    *DEntrypp = NULL;
    found = FALSE;

    filelen = directory->icbe.fe.InfoLengthL;
    bytestoread = filelen;
    bytesread = 0;
    
    /* allocate a temporary I/O buffer */

    iobuffsize = (UINT32) fsp->SectorSize;
    iobuff = kmem_alloc( iobuffsize, BUFRAM);
    if( iobuff == NULL) {
        return ENOMEM;
    }

    /* pass through the entire directory file */
    
    do {
        
        /* how many bytes to transfer? */
        
        if( bytestoread < iobuffsize) {
            xferlen = bytestoread;
        } else {
            xferlen = iobuffsize;
        }
        
        /* read the data */
        
        /* ISSUE not sure the fourth parameter is correct */

        uioInit( &uio, &iovec, iobuff, bytesread, (INT32) xferlen);
        error = ot_readData( fsp, directory, &uio);
        
        if (error == EEOF || error == SUCCESS) {
            
            offset = 0;
            
            /* while the entire amount of the directory entry appears */
            /* to be in buffer - process */
            
            while( ( FILE_ID_DESC_CONSTANT_LEN + offset) <= xferlen) {
                
                direntry = ( struct FileIDDesc *) (iobuff + offset);

                /* now flip the representation */
#if !IBM
                ot_xlate(TAGID_FILE_ID, FROM_OP, (caddr_t) direntry, iobuffsize-offset,0);
#endif

                dirlen = (FILE_ID_DESC_CONSTANT_LEN
                          + direntry->L_FI + direntry->L_IU
                          + ot_dirpad(direntry->L_FI + direntry->L_IU)); 
                
                /* now we can make the final judgement if directory is */
                /* entirely enclosed in current buffer */
                
                if( dirlen + offset <= xferlen) {
                    if (!( direntry->Characteristics & DELETE_ATTR)) {
                        if( searchType == findByAddress) {
                            if( direntry->ICB.Location_LBN == icbaddr) {
                                found = TRUE;
                            }
                        } else { /* find by name */
                            if( strlen( name) == direntry->L_FI) {
                                if( strncmp((char *)ot_getFileID(direntry),
                                            name,direntry->L_FI) == 0)  {
                                    found = TRUE;
                                }
                            }
                        }
                    }
                    if( found) {
                        *sectorBegins = (bytesread+offset) / fsp->SectorSize;
                        *sectorIndex = (bytesread+offset) % fsp->SectorSize; 
                        break;
                    } else {
                        offset += dirlen;
                    }
                } else {
                    break;
                }
            } /* while */
            /* not all in buffer, adjust pointers and read some more */
            bytestoread -= offset;
            bytesread += offset;
        } /* not error */
    } while ((!error) && (bytesread < filelen) && !(found));
    if (found) {
        tempEntry = (struct FileIDDesc *) kmem_alloc((UINT32) dirlen,
                                                     BUFRAM);
        error = (tempEntry == NULL ? ENOMEM : SUCCESS);
    } else {
        *sectorBegins = filelen / fsp->SectorSize;
        *sectorIndex = filelen % fsp->SectorSize;
        error = ENOENT;
    }
    if (!error) {
        ot_copy( (char *) direntry, (char *)tempEntry, dirlen);
        *DEntrypp = tempEntry;
    }
    if( iobuff) {
        kmem_free( iobuff, iobuffsize, 0);
    }
    return error;
} /* end ot_getDirentry */


/********************************************************************/
/* ot_dirpad                                                        */
/*       calculate the length of the pad field in a directory entry */
/*                                                                  */
/* input parameters                                                 */
/*       filenamelen:  length of fileID in directory                */
/*       impuselen:    length of implementation use area in dir     */
/*                                                                  */
/* output parameters                                                */
/*       none                                                       */
/*                                                                  */
/* returns                                                          */
/*       integer value of pad.  Value can be 0-3                    */
/*                                                                  */
/* notes                                                            */
/*       padlength returned insured directory is aligned on 4 byte  */
/*       boundaries                                                 */
/*                                                                  */
/* Modification History:                                            */
/*       18.Sep.92   ECB   First Entered                            */
/*                                                                  */
/********************************************************************/

int ot_dirpad( int filenamelen, int impuselen)

{
    return(4 * ((filenamelen + impuselen
                 + FILE_ID_DESC_CONSTANT_LEN + 3) /4)
           - (filenamelen
              + impuselen
              + FILE_ID_DESC_CONSTANT_LEN)); 
}


UINT32 ot_reorgDirectory()
{
    return(0);
}

/********************************************************************/
/*   ot_getFileID                                                   */
/*      return a pointer to the start of a file identifier in a     */
/*      supplied directory entry                                    */
/*                                                                  */
/*   input parameters:                                              */
/*      Direntry   pointer to a struct FileIDDesc that includes     */
/*                 a valid file idemtifier length field, a imp-     */
/*                 lementation use length field and the file        */
/*                 identifier                                       */
/*                                                                  */
/*   output parameters                                              */
/*      none                                                        */
/*                                                                  */
/*   returns                                                        */
/*      an unsigned char * to the start of the file identifier      */
/*                                                                  */
/*   Revision History                                               */
/*      18.Sep.92   ECB    Initial entry                            */
/*                                                                  */
/********************************************************************/

UCHAR8 *ot_getFileID( struct FileIDDesc *Direntry)
{

    /* the pointer is just the start address + fixed structure + */
    /* Implementation use length */

    return ( (UCHAR8 *)Direntry + FILE_ID_DESC_CONSTANT_LEN
            + Direntry->L_IU); 
}



UINT32 ot_fileName(UCHAR8 *name , int charset, UCHAR8 *len)
{
    /* just return the length of the string pointed to by name */
    /* as if the string contained 8-bit characters */

    *len = strlen((char *)name);
    return( SUCCESS);
}
