/********************************************************************/
/*  o_icb.c - NSR support routines for icb operations               */
/*                                                                  */
/*      Unpublished Confidential Information of Hewlett-Packard     */
/*      Company. Do not disclose.  Copyright (c) Hewlett-Packard    */
/*      Company 1996. All rights reserved                           */
/*                                                                  */
/* The contents of this software are proprietary and confidential   */
/* the Hewlett- Packard Company, and are limited in distribution to */
/* those with a direct need to know.  Individuals having access to  */
/* this software are responsible for maintaining the                */
/* confidentiality of the content and for keeping the software      */
/* secure when not in use.  Transfer to any party is strictly       */
/* forbidden other than as expressly permitted in writing by        */
/* Hewlett-Packard Company.  Unauthorized transfer to or possession */
/* by any unauthorized party may be a criminal offense.             */
/*                                                                  */
/*                   RESTRICTED RIGHTS LEGEND                       */
/*                                                                  */
/*      Use,  duplication,  or disclosure by the                    */
/*      U. S. Government is subject to restrictions as set forth    */
/*      in subparagraph (c) (1) (ii) in the Rights in Technical     */
/*      Data and Computer Software clause in DFARS 252.227-7013     */
/*      or any other successor clause.                              */
/*                                                                  */
/*                       HEWLETT-PACKARD COMPANY                    */
/*                       3000 Hanover St.                           */
/*                       Palo Alto, CA  94304 U.S.A.                */
/*                                                                  */
/*      Rights for non-DOD U.S. Government Departments and Agencies */
/*      are as set forth in FAR 52.227-19 (c) (1,2) or any other    */
/*      successor clause.                                           */
/*                                                                  */
/*  Module name:    o_icb.c                                         */
/*                                                                  */
/*  Description:                                                    */
/*      Routines to read, write and search icb's                    */
/*                                                                  */
/*  Functions in this module:                                       */
/*      ot_indICBWr                 ot_ICBRdWr                      */
/*      ot_lastInBlock              ot_ICBWt                        */
/*      ot_makeICBVsnBlock          ot_searchLast                   */
/*      ot_ICBDelete                ot_searchMO                     */
/*      ot_searchVersion                                            */
/*                                                                  */
/*  Modification history:                                           */
/*      16.Mar.92   ECB     Added ANSI function declarations        */
/********************************************************************/
/*
 $Source: /projects/HMS/nsr/udfOld/src/nsrCore/o_icb.c,v $
 $Revision: 1.27 $      $Author: hmgr $
 $State: Exp $          $Locker:  $
 $Date: 96/05/30 14:52:57 $
 ********************************************************************/
#include "Debug.h"
#include "../nsrHdrs/def.h"
#include "../nsrHdrs/o_vndops.h"
#include "../nsrHdrs/nsr.h"
#include "../nsrHdrs/o_node.h"
#include "../nsrHdrs/o_proto.h"

#if SUN             /* to enable debuggging */
#include "o_fs.h"
#endif

/********************************************************************/
/*  ot_indICBWr                                                     */
/*      write a indirect icb entry                                  */
/*                                                                  */
/*  input parameters:                                               */
/*      fs:         the otfs structure for the volume               */
/*      icb:        indirect icb to write/buffer to use             */
/*      addr:       where to write the icb                          */
/*      addrList:   space entry, indicating icb's icb block         */
/*                      address and length                          */
/*      level:      the number of indirects between the primary     */
/*                      icb block and the direct icb block in       */
/*                      the chain being created.                    */
/*  notes:                                                          */
/*      icb parameter MUST be at least 1 sector long                */
/*                                                                  */
/*  Modification History:                                           */
/*                                                                  */
/********************************************************************/
INT32 ot_indICBWr(register struct otfs *fs,
                  register struct IndirectEntry *icbe,
                  UINT32 addr,
                  struct space *addrList,
                  UINT16 level)
{
    INT32   error;

    bzero((caddr_t) icbe,(INT32) fs->SectorSize);
    FillTag(TAGID_INDIRECT,1,1,NOCRC,0,addr,(struct Tag *)icbe);
    
    /* Fill in the ICB Tag portion of the Indirect icb entry */

    icbe->sICBTag.PriorDirects = pow( fs->ICBsize, level-1) +
      (fs->ICBsize - fs->IndirectSize);
    if( level == 1)
      icbe->sICBTag.PriorDirects--;

    icbe->sICBTag.StrategyType = 1;
    *((UINT16 *) icbe->sICBTag.StrategyParm) = fs->ICBsize - fs->IndirectSize; 
    icbe->sICBTag.NumberEntries = fs->ICBsize;
    icbe->sICBTag.Reserved = 0;
    icbe->sICBTag.FileType = FILE_TYPE_INDIRECT_ENTRY;
    /* ISSUE - This may need specification, but is not for now */
    icbe->sICBTag.sParentICB_LBN = 0;
    icbe->sICBTag.sParentICB_PartNo = 0;
    icbe->sICBTag.Flags = 0;

    /* fill in address of icbe block (direct or indirect) pointed to */

    icbe->sIndirectICB.ExtentLength.bf.Length = addrList->length;
    icbe->sIndirectICB.ExtentLength.bf.Type = E_RECORDED;
    icbe->sIndirectICB.Location_LBN = addrList->start;
    icbe->sIndirectICB.Location_PartNo = 0;

    if (error = ot_erase(fs, 1, addrList->start, addrList->length))
        return error;
#if !IBM
    ot_xlate(TAGID_INDIRECT, TO_OP,(caddr_t) icbe, fs->SectorSize, 0);
#endif
    error = ot_write(fs, addr, (caddr_t) icbe, fs->SectorSize);

    return( error);

} /* end ot_indICBWr */

/********************************************************************/
/*  ot_lastInBlock                                                  */
/*      read last recorded sector in an icb                         */
/*                                                                  */
/*  input parameters:                                               */
/*      fs:             the otfs structure for the volume           */
/*      blockAddr:      beginning of the block                      */
/*      blockOffset:    for address of last written sector in block */  
/*                                                                  */
/*  output parameters:                                              */
/*      fs->iobuf:      contains the untranslated icbe              */
/*      blockOffset:    address of last written sector in block     */
/*                                                                  */
/*  returns:                                                        */
/*      0               success                                     */
/*      ENOENT:         never found any icb's                       */
/********************************************************************/
INT32 ot_lastInBlock(register struct otfs *fs,
                     UINT32 blockAddr,
                     UINT16 *blockOffset)
{
    register UINT16     offset;             /* from blockAddr to icbe being considered */
    register INT32      error;
    UINT16              atFirst,            /* bufptr at beginning of buffer? */
                        bufSectCount = IOBUFSIZE / fs->SectorSize;
    UINT32              readLength,         /* # of bytes to read */
                        redLength,
                        blockLength;        /* # of bytes remaining in icbe block */
    UINT16              nodeTag;            /* for icbe verification */
    register caddr_t    bufptr;             /* for traversing sectors in buffer */
    UINT32              sector;

    /* scan backwards from the end for the first nonempty sector in the block */
    /* note: is this slower than going forward and (possibly) rereading
        when the first empty one is found? */
    blockLength = fs->SectorSize * fs->ICBsize;
    offset = fs->ICBsize;
    do {
        offset = offset > bufSectCount ? offset-bufSectCount : 0;
        readLength = blockLength > IOBUFSIZE ? IOBUFSIZE : blockLength;
        blockLength -= readLength;
        redLength = readLength;
        error = ot_read(fs,blockAddr + offset,fs->iobuf,&redLength);
        if (!error)
            atFirst = 0;
            sector = blockAddr + offset + readLength/fs->SectorSize - 1;
            for (bufptr = fs->iobuf+(readLength-fs->SectorSize);
                 !atFirst && !ot_validICB(TAGID_NONE, (struct EntryHdr *)bufptr, sector);
                sector--)
            {
                atFirst = bufptr == fs->iobuf;
                bufptr -= fs->SectorSize;
            }
    } while (!error && blockLength > 0 && atFirst);

    /* if one was found, put it at the beginning of fs->iobuf, and return its sector */
    if (!error && !atFirst) {
        if (bufptr > fs->iobuf) {
            offset += (bufptr - fs->iobuf)/fs->SectorSize;
            ot_copy(bufptr,fs->iobuf,fs->SectorSize);
        }
        *blockOffset = offset;
    } else if (!error)  /* no error, but never found any icbs at all */
        error = ENOENT;
    return error;
} /* end ot_lastInBlock */

/********************************************************************/
/*  ot_makeICBVsnBlock                                              */
/*      create a new partial icb (for versioning)                   */
/*                                                                  */
/*  input parameters:                                               */
/*      fs:         the otfs structure for the volume               */
/*      mainAddr:   address of last primary icb entry recorded      */
/*      newAddr:    address within block of where additions begin   */  
/*      newLevel:   number of indirects to create                   */
/*      nodeAddr:   for the address of the new icb entry            */
/*                                                                  */
/*  output parameters:                                              */
/*      nodeAddr:   address where new direct icb entry may be       */
/*                  written                                         */
/*                                                                  */
/*  notes:                                                          */
/*      mainAddr may be ignored, unless newAddr is 0. this means we */
/*          must create an entirely new primary icb in              */
/*          mainAddr's successor area.                              */
/*      if newAddr is nonzero, this is where the first indirect     */
/*          icb entry in the chain we will be creating must go.     */
/*      newLevel is the number of indirects (1-based) which must be */
/*          chained from the newAddr icb entry. newLevel >= 1.      */
/*      uses fs->iobuf.  ot_search_last(), which is our only caller,*/
/*          also uses fs->iobuf, but doesn't compete.               */
/*      fs->iobuf must be at least fs->SectorSize bytes long.       */
/*                                                                  */
/*  modification history:                                           */
/*      03.DEC.92  ECB  Removed creation of successors              */
/*                                                                  */
/*                                                                  */
/*                                                                  */
/*                                                                  */
/********************************************************************/
INT32 ot_makeICBVsnBlock(struct otfs *fs,
                         UINT32 mainAddr,
                         register UINT32 newAddr,
                         UINT16 newLevel,
                         UINT32 *nodeAddr)
{
    register UINT16         level;
    register struct space   *addrList,  /* base of disk space taken list */
                            *addrPtr;   /* traverses addrList */
    INT32                   error;
    struct IndirectEntry    *icbe = (struct IndirectEntry *) fs->iobuf;
    struct space            *head;

    /* see if enough space for the entire chain can be allocated before writing anything */

    addrList = ot_getSpace( fs, (UINT32) fs->ICBsize, contiguous);
    addrPtr = addrList;

    for (level = newLevel-1; addrPtr && level > 0; level--) {
        addrPtr->next = ot_getSpace( fs, (UINT32) fs->ICBsize, contiguous);
        addrPtr = addrPtr->next;
    }

    error = 0;

    if (addrPtr) {  /* enough space existed */

        addrPtr = addrList;
        level = newLevel;

        /* make a chain of indirect icb blocks */
        /* DRaM - do this backwards, so if a write error occurs, there won't
            be an indirect icbe pointing off to nirvana? */

        while (level > 0 && !error) {
            error = ot_indICBWr( fs, icbe, newAddr, addrPtr, newLevel);
            newAddr = addrPtr->start;
            level--;
            addrPtr = addrPtr->next;
        }
        *nodeAddr = newAddr;

    } else {
        /* ISSUE - do we want to return the space we did get? */
        error = ENOSPC;
    }

    if (addrList) {
        head = addrList;
        killSpaceList(&head);
    }
    return error;
} /* end ot_makeICBVsnBlock */

/********************************************************************/
/*  ot_ICBDelete                                                    */
/*      routine to delete an icb entry                              */
/*                                                                  */
/*  input parameters:                                               */
/*      fs:     the otfs structure for the volume                   */
/*      otp:    the otnode for the file whose icb is to be deleted  */
/*                                                                  */
/*  returns:                                                        */
/*      0       success                                             */
/*                                                                  */
/*  notes:                                                          */
/*      the otnode is now invalid                                   */
/********************************************************************/
INT32 ot_ICBDelete(register struct otfs *fs, struct otnode *otp)
{
    INT32           error;

    otp->flags |= OT_INVAL;

    /* erase icb sectors, put on free list */
    if (error = ot_erase( fs, 1, otp->icb_addr_LBN, (UINT32) fs->ICBsize))
        return error;
    error = returnSectors( fs, (struct space *) 0, otp->icb_addr_LBN,
                        (UINT32) fs->ICBsize, bmapErased);
    return error;
} /* end ot_ICBDelete */

/********************************************************************/
/*  ot_ICBRdWr                                                      */
/*      routine to read or write an icb entry                       */
/*                                                                  */
/*  input parameters:                                               */
/*      fsp:        the otfs structure for the volume               */
/*      otp:        the otnode; contains the icb to read/write      */
/*      rw:         UIO_READ or UIO_WRITE                           */
/*      version:    the version of the icb to read                  */
/*                                                                  */
/*  returns:                                                        */
/*      0           success                                         */
/*      EINVAL:     an invalid icb was found                        */
/*      ENOENT:     the otnode is invalid                           */
/*                                                                  */
/*  notes:                                                          */
/*      version is valid only on a read; a write ALWAYS creates the */
/*          last version                                            */
/*      if the icb has already been read and modified, the read     */
/*          will not be performed, saving the modified icb          */
/*      if the icb has not been modified, a write will not be       */
/*          performed, saving disk space                            */
/*                                                                  */
/********************************************************************/
INT32 ot_ICBRdWr(register struct otfs *fsp,
                 register struct otnode *otp,
                 enum uio_rw rw,
                 UINT32 version)
{
    int             index;
    INT16           wasLast = version == LAST;
    UINT32          addr = 0L;
    struct otnode   *active;
    register INT32  error;

    /* is otnode invalid? */
    if (otp->flags & OT_INVAL)
        return ENOENT;

    /* don't trash an already read and modified otnode  */
    /* it may not have been flushed to the disk yet     */
    /* and re-reading will get the previous one         */
    /* also, don't flush if not modified                */
    if (((otp->flags & OT_MOD) && (rw == UIO_READ)) ||
        (!(otp->flags & OT_MOD) && (rw == UIO_WRITE)))
        return 0;

    /* clear I/O buffer */
    bzero(fsp->iobuf,(INT32) IOBUFSIZE);

    if (fsp->MediaType <= (int) fsWORM) {   /* WORM or WORM Emulation */
        if (version == LAST)
            error = ot_searchLast( fsp, otp->icb_addr_LBN, &addr, fsp->iobuf, rw, &version);
        else
            error = ot_searchVersion( fsp, otp->icb_addr_LBN, &addr, fsp->iobuf, version);
        if (error != 0)
            return error;                            
    } else {
        error = ot_searchMO( fsp, otp->icb_addr_LBN, &addr, fsp->iobuf, rw);
        if (error)
            return error;
        version = 1;
    }

    if (rw == UIO_WRITE) {
        /* validate icb */
        /* if (otp->icb_addr_LBN != otp->icbe.ge.sICBTag.sParentICB_LBN)
            return EINVAL; */   /* Removed for MDI - DVM 6-1-93 */
        if (ot_ICBHasADs(otp)) {
            DMLPRINTF(oADLev)("ot_ICBRdWr() - calling ot_writeADL.\n");
            error = ot_writeADL(otp);
            if (error) {
                DMLPRINTF(oADLev)("ot_ICBRdWr() error = %d\n",error);
            }
        }
            
        /* update tag location, checksum */
        FillTag(otp->icbe.ge.sTag.uTagID,
                otp->icbe.ge.sTag.uDescriptorVersion,
                otp->icbe.ge.sTag.uTagSerialNum,
                0,0,addr,(struct Tag *)&otp->icbe.ge.sTag);

        bzero(fsp->iobuf,(INT32) IOBUFSIZE);
        ot_copy( (caddr_t) &otp->icbe, fsp->iobuf,fsp->SectorSize);

        /* translate to disk structure and write */
#if !IBM
        ot_xlate(TAGID_NONE,TO_OP,fsp->iobuf,fsp->SectorSize,0);
#endif
        error = ot_write(fsp,addr,fsp->iobuf,fsp->SectorSize);

        if (!error) {
            if (fsp->MediaType <= (int) fsWORM) {
                ot_clearContext( ctextOne, fsp, otp);
                otp->version = version++;
            }
            otp->isLast = TRUE;
            otp->flags &= ~OT_MOD;
        }   

    } else {

        ot_copy( fsp->iobuf, (caddr_t) &otp->icbe, fsp->SectorSize);
#if !IBM
        ot_xlate(TAGID_NONE, FROM_OP, (caddr_t) &otp->icbe,fsp->SectorSize,0);
#endif
        error = ot_goodTag(TAGID_NONE, &otp->icbe, addr);
        if (!error)
            error = ot_isICB(&otp->icbe);
        if (!error && otp->icbe.ge.sTag.uTagID == TAGID_FILE_ENTRY)
            ot_setSystemArea(otp);
        
        otp->flags |= OT_INITIALIZED; 
        if (!error) {
            otp->version = version;
            otp->isLast = wasLast;
        }
    }

    /* reset version numbers of other otnodes for same file; the one just written is
        now the latest version */

    if (!error && fsp->MediaType <= (int) fsWORM && otp->isLast) {
        for (index = 0; index < 2; index ++)
            for (   active = index == 0 ? fsp->active : fsp->inactive;
                    active;
                    active = active->next) {
                if (!(active->flags & OT_INVAL) &&
                    active->icb_addr_LBN == otp->icb_addr_LBN &&
                    active != otp &&
                    otp->vnode.v_type == active->vnode.v_type)
                        active->isLast = 0;
            }
    }
    return error;
} /* end ot_ICBRdWr */

/********************************************************************/
/*  ot_searchLast                                                   */
/*      find the last version of an icb entry                       */
/*                                                                  */
/*  input parameters:                                               */
/*      fs:         the otfs structure for the volume               */
/*      blockAddr:  address of icb                                  */
/*      nodeAddr:   for the icb address of the icb entry found      */
/*      buffer:     for I/O                                         */
/*      rw:         UIO_READ or UIO_WRITE                           */
/*      version:    for the version number of the icb entry found   */
/*                                                                  */
/*  output parameters:                                              */
/*      nodeAddr:   contains actual disk icb address of icb entry   */
/*      version:    actual version number of icb entry              */
/*                                                                  */
/*  returns:                                                        */
/*      0           success                                         */
/*                                                                  */
/*  notes:                                                          */
/*      if rw == UIO_READ, returns untranslated icb entry in buffer */
/*      if rw == UIO_WRITE, makes space for new icb entry in buffer */
/*          contents undefined                                      */
/*      uses fs->iobuf (though fs->iobuf can be used as "buffer"    */
/*                      parameter)                                  */
/*      if rw==UIO_WRITE, "buffer" parameter is unused and may be 0 */
/*      "buffer",if used, must be at least fs->SectorSize bytes long*/
/*                                                                  */
/*  modification list:                                              */
/*      05.Dec.92  ECB  Removed successor block handling            */
/*                                                                  */
/*                                                                  */
/*                                                                  */
/*                                                                  */
/********************************************************************/
INT32 ot_searchLast(register struct otfs *fs,
                    register UINT32 blockAddr,
                    UINT32 *nodeAddr,
                    caddr_t buffer,
                    enum uio_rw rw,
                    UINT32 *version)
{

    INT32                   error = 0;
    union icbe              *icb = (union icbe *) fs->iobuf;
    UINT16                  level;          /* of indirection */
    UINT16                  offset,         /* into block for latest icb */
                            levelCtr,
                            firstNewLevel;  /* in indirect chain when writing */
    UINT32                  lastMainBlock,  /* last primary icb block in successor chain */
                            firstNewAddr,   /* first new icb in indirect chain when writing */
                            lastVersion,    /* what version # IS last, anyway? */
                            vsnCount,       /* total versions below a level of indirection */
                            secSize = fs->SectorSize,
                            levelCount;     /* # of versions at a level of indirection */

    *nodeAddr = 0;
    lastVersion = 1;

    lastMainBlock = blockAddr;

    /* find the last icb in the main block */
    error = ot_lastInBlock( fs, (UINT32) blockAddr, &offset);

    /* find last icb, wherever it may lie */
    if (error == ENOENT && rw == UIO_WRITE) {   /* empty primary icb block, and writing... */
        error = 0;                                  /* it's OK */
        offset = 0;                                 /* write first icb here */
    } else                                      /* otherwise... */
        offset++;                                   /* write next node here */

    if (!error) {
        level = fs->ICBsize - fs->IndirectSize;

        /* keep lastVersion at # corresponding to last of second to last icb */

        levelCount = 1;
        if (offset == 0)
            lastVersion--;  /* it was 1-based, so this won't go negative */
        else if (offset <= level)
            lastVersion += offset-1;
        else {
            for (vsnCount = level, levelCtr = level+1;
                levelCtr < offset;
                levelCtr++, levelCount *= fs->ICBsize, vsnCount += levelCount);
            lastVersion += vsnCount;
        }

        /* in case we're writing, save best guess about where new icb must go */

        firstNewAddr = blockAddr+offset;
        if (offset < level)
            firstNewLevel = 0;
        else if (offset < fs->ICBsize)
            firstNewLevel = offset - level + 1;
        else    /* new successor may be necessary */
            firstNewAddr = 0;

        /*
        * now traverse the indirects, saving best guess on new icb as we go
        */
        if (offset > level) {
            level = offset-level;
            do {
                level--;
#if !IBM
                ot_xlate(TAGID_INDIRECT, FROM_OP, (caddr_t) icb, fs->SectorSize, 0);
#endif
                blockAddr = icb->ie.sIndirectICB.Location_LBN;
                if (!error)
                    error = ot_lastInBlock(fs,(UINT32) blockAddr,&offset);

                /* we'll accept an empty block at trail's end if writing */
                if (error == ENOENT && rw == UIO_WRITE && level == 0) {
                    error = 0;
                    offset = 0;
                } else
                    offset++;
                if (!error) {
                    if (offset == 0)
                        lastVersion--;
                    else if (offset > 1)
                        lastVersion += levelCount*(offset-1);
                    levelCount /= fs->ICBsize;
                    if (offset < fs->ICBsize) {  /* new best guess */
                        firstNewAddr = blockAddr+offset;
                        firstNewLevel = level;
                    }
                }
            } while (!error && level > 0);
        }
    }

    if (!error) {
        if (rw == UIO_READ) {
            if (buffer != fs->iobuf)
                ot_copy(fs->iobuf,buffer,fs->SectorSize);
            *nodeAddr = blockAddr+(offset-1);
            *version = lastVersion;
        } else {    /* UIO_WRITE */
            if (firstNewAddr == 0 || firstNewLevel > 0) /* punt */
                error = ot_makeICBVsnBlock(fs,lastMainBlock,firstNewAddr,firstNewLevel,
                            (UINT32 *) nodeAddr);
            else
                *nodeAddr = firstNewAddr;
            *version = lastVersion+1;
        }
    }

    return error;
} /* end ot_searchLast */

/********************************************************************/
/*  ot_searchMO                                                     */
/*      routine to search for an icb on an MO volume                */
/*                                                                  */
/*  input parameters:                                               */
/*      fsp:    the otfs structure for the volume                   */
/*      icbaddr: the icb block address                              */
/*      addr:   for the actual one read/to write                    */
/*      buffer: for I/O                                             */
/*      rw:     UIO_READ or UIO_WRITE                               */
/*                                                                  */
/*  output parameters:                                              */
/*      addr:   contains actual address of one read or the address  */
/*              to write                                            */
/*  returns:                                                        */
/*      0       success                                             */
/*                                                                  */
/*                                                                  */
/*  NOTE: The original version of the toolkit wrote onodes in a     */
/*     swing-buffer strategy. Each onode would take 2 sectors. This */
/*     technique is no longer in use.                               */
/********************************************************************/
INT32 ot_searchMO(register struct otfs *fsp,
                  register UINT32 icbaddr,
                  UINT32 *addr,
                  caddr_t buffer,
                  enum uio_rw rw)
{
    INT32   error;
    UINT32  secSize = fsp->SectorSize;

    if ( rw == UIO_READ )
        error = ot_read(fsp,icbaddr,buffer,&secSize);
    *addr = icbaddr;
    return 0;
} /* end ot_searchMO */

/********************************************************************/
/*  ot_searchVersion                                                */
/*      find any extant version of an icb                           */
/*                                                                  */
/*  input parameters:                                               */
/*      fs:         the otfs structure for the volume               */
/*      blockAddr:  address of the icb                              */
/*      nodeAddr:   for the actual address                          */
/*      buffer:     for I/O                                         */
/*      version:    version of icb entry desired                    */
/*                                                                  */
/*  output parameters:                                              */
/*      nodeAddr:   contains the icb entry actual sector address    */
/*      buffer:     if reading, contains the untranslated icb entry */
/*                                                                  */
/*  returns:                                                        */
/*      0           success                                         */
/*      ENOENT:     an invalid icb entry was found, or the requested*/
/*                      version does not exist                      */
/********************************************************************/
INT32 ot_searchVersion(register struct otfs *fs,
                       UINT32 blockAddr,
                       UINT32 *nodeAddr,
                       caddr_t buffer,
                       register UINT32 version)

{
    INT32             error;
    UINT32            vsnCount;   /* version counter while looking */
    union icbe        *icb = (union icbe *)buffer;
    UINT16            level;      /* # of indirects counter */
    UINT16            offset;     /* to icb from block base address */
    UINT32            secSize = fs->SectorSize,
                      levelCount; /* # versions at a level of indirection */

    error = 0;          /* optimistically speaking */
    *nodeAddr = 0;
    version--;          /* 0-based for ease of use */

    /* find out what level of indirection icb will be at */

    for (levelCount = 1, vsnCount = fs->ICBsize - fs->IndirectSize, level = 0;
         vsnCount <= version && level < fs->IndirectSize;
         level++, levelCount *= fs->ICBsize, vsnCount += levelCount);

    if (vsnCount > version) {

        /* travel indirect chain until we reach the right direct icb block */

        if (level == 0)     /* it's a direct icb */
            offset = version;
        else {              /* it's indirect */
            offset = fs->ICBsize - fs->IndirectSize + level - 1;
            version -= vsnCount - levelCount;
            while (level > 0 && !error) {
                error = ot_read( fs, blockAddr + offset, (caddr_t) icb, &secSize);
                if (!error) {
#if !IBM
                    ot_xlate(TAGID_INDIRECT,FROM_OP,
                             (caddr_t) icb,fs->SectorSize,0);
#endif
                    error = ot_goodTag(TAGID_INDIRECT,icb,blockAddr+offset);
                }
                if (error == EINVAL)
                    error = ENOENT;
                blockAddr = icb->ie.sIndirectICB.Location_LBN;
                levelCount /= fs->ICBsize;
                offset = version/levelCount;
                version %= levelCount;
                level--;
            }
        }
        if (!error)
            error = ot_read( fs, blockAddr + offset, buffer, &secSize);
        if (!error && isTag((struct Tag *) buffer) != TAGID_FILE_ENTRY)
            error = ENOENT;
        if (!error)
            *nodeAddr = blockAddr + offset;
    } else
        error = ENOENT;

    return error;

} /* end ot_searchVersion */

/********************************************************************/
/*  ot_validICB                                                     */
/*      return TRUE if icb entry has a valid tag, icb tag, valid    */
/*      location and valid checksum                                 */
/*                                                                  */
/*  input parameters:                                               */
/*      tagid:      what kind of icb entry to check for             */
/*      hdr:        pointer to tag, icb tag                         */
/*      location:   location where this icb entry came from         */
/*                                                                  */
/*  returns:                                                        */
/*      1           if valid                                        */
/*      0           not valid                                       */
/********************************************************************/
INT32 ot_validICB(UINT16 tagid, struct EntryHdr *hdr, UINT32 location)
{
    struct Tag    *tag = (struct Tag *)hdr;
    struct ICBTag *icbtag = (struct ICBTag *)&hdr->sICBTag;
    UCHAR8 CheckSum;
    UCHAR8 *BytePtr;
    UINT16 i;
    struct Tag xtag;
    struct ICBTag xicbtag;

    /* 
    * xlate the tags 
    */
    xtag = *tag;
    xlate_tag(&xtag);
    xicbtag = *icbtag;
    xlate_ICBTag(&xicbtag);
    
    /* 
    * First, if checking for a specific tag, compare
    * tagid with the data.
    */
    if ( tagid != TAGID_NONE && xtag.uTagID != tagid )
        return 0; 

    /*
    * Check for valid location.
    */
    if ( xtag.uTagLoc != location ) {
        return 0;
    }

    /* 
     * Check for valid tagID in tag and that the icbtag
     * file type makes sense with the tagID.
     */
    switch ( xtag.uTagID ) {
    case TAGID_FSD:
    case TAGID_FILE_ID:
    case TAGID_ALIAS:
    case TAGID_FILE_ENTRY:
    case TAGID_EXT_ATTR:
    case TAGID_SPACE_BMAP:
    case TAGID_ALLOC_EXTENT: 
        break;
    case TAGID_INDIRECT: if ( xicbtag.FileType != FILE_TYPE_INDIRECT_ENTRY ) return 0;
        break;
    case TAGID_TERM_ENTRY: if ( xicbtag.FileType != FILE_TYPE_TERMINAL_ENTRY ) return 0;
        break;
    case TAGID_UNALLOC_SP_ENTRY: if ( xicbtag.FileType != FILE_TYPE_UNALLOC_SP_ENTRY ) return 0;
        break;
    case TAGID_PART_INTEGRITY: if ( xicbtag.FileType != FILE_TYPE_PARTITION_INTEGRITY ) return 0;
        break;
    default: return 0;
        break;
    }

    /*
    * Test the checksum.
    */
    BytePtr = (UCHAR8 *) &xtag;
    CheckSum = 0;
      
    for (i=1; i<5; i++, BytePtr++)
        CheckSum+= *BytePtr;
    *BytePtr++;                            /* skip the 8 bit checksum field */
    for (i=1; i<12; i++, BytePtr++)
        CheckSum+= *BytePtr;
            
    if( xtag.uTagChecksum != CheckSum)
        return 0;

    return 1;
}

INT32 ot_isICB(void *ptr)
{
    struct Tag *tagp = ptr;
    
    return((tagp->uTagID == TAGID_INDIRECT ||
            tagp->uTagID == TAGID_FILE_ENTRY ||
            tagp->uTagID == TAGID_PART_INTEGRITY ||
            tagp->uTagID == TAGID_UNALLOC_SP_ENTRY ||
            tagp->uTagID == TAGID_TERM_ENTRY) ? SUCCESS : EINVAL);
}
