/********************************************************************/
/*  o_vndops.c - NSR routines for vnode/otnode operations           */
/*                                                                  */
/*      Unpublished Confidential Information of Hewlett-Packard     */
/*      Company. Do not disclose.  Copyright (c) Hewlett-Packard    */
/*      Company 1996. All rights reserved                           */
/*                                                                  */
/* The contents of this software are proprietary and confidential   */
/* the Hewlett- Packard Company, and are limited in distribution to */
/* those with a direct need to know.  Individuals having access to  */
/* this software are responsible for maintaining the                */
/* confidentiality of the content and for keeping the software      */
/* secure when not in use.  Transfer to any party is strictly       */
/* forbidden other than as expressly permitted in writing by        */
/* Hewlett-Packard Company.  Unauthorized transfer to or possession */
/* by any unauthorized party may be a criminal offense.             */
/*                                                                  */
/*                   RESTRICTED RIGHTS LEGEND                       */
/*                                                                  */
/*      Use,  duplication,  or disclosure by the                    */
/*      U. S. Government is subject to restrictions as set forth    */
/*      in subparagraph (c) (1) (ii) in the Rights in Technical     */
/*      Data and Computer Software clause in DFARS 252.227-7013     */
/*      or any other successor clause.                              */
/*                                                                  */
/*                       HEWLETT-PACKARD COMPANY                    */
/*                       3000 Hanover St.                           */
/*                       Palo Alto, CA  94304 U.S.A.                */
/*                                                                  */
/*      Rights for non-DOD U.S. Government Departments and Agencies */
/*      are as set forth in FAR 52.227-19 (c) (1,2) or any other    */
/*      successor clause.                                           */
/*                                                                  */
/*  Module name:    o_vndops.c                                      */
/*                                                                  */
/*  Description:    VFS routines for the management of files and    */
/*                  directories; routines needed but not supported  */
/*                  by the VFS interface related to files and       */
/*                  directories; routines to manage otnodes: the    */
/*                  the in-memory vnode/icb entry combination       */
/*                                                                  */
/*  Functions in this module:                                       */
/*      otvn_access             otvn_close                          */
/*      otvn_create             otvn_create_assoc                   */
/*      otvn_fid                otvn_fsync                          */
/*      otvn_getattr            otvn_inactive                       */
/*      otvn_invalop            otvn_lookup                         */
/*      otvn_lookup_assoc       otvn_mkdir                          */
/*      otvn_open               otvn_rdwr                           */
/*      otvn_readdir            otvn_remove                         */
/*      otvn_rename             otvn_rmdir                          */
/*      otvn_setattr                                                */
/*                                                                  */
/*      ot_delete               ot_deleteExtAttrICBs                */
/*      ot_deleteData           ot_getAndFillOTNode                 */
/*      ot_setFileSize          ot_getName                          */
/*      ot_getNumFiles          ot_getOTNode                        */
/*      ot_readData             ot_readVerify                       */
/*      ot_release              ot_setupDirentryOTNode              */
/*      ot_writeData            ot_fillOtNode                       */
/*                                                                  */
/*  Modification history:                                           */
/*                                                                  */
/********************************************************************/
/*
 $Source: /projects/HMS/nsr/udfOld/src/nsrCore/o_vndops.c,v $
 $Revision: 1.42 $      $Author: hmgr $
 $State: Exp $          $Locker:  $
 $Date: 96/05/30 14:53:04 $
 ********************************************************************/

#include "../nsrHdrs/def.h"
#include <string.h>
#include "Debug.h"
#include "o_vndops.h"
#include "nsr.h"
#include "o_node.h"
#include "o_proto.h"
#include "nsrerrno.h"
#include "vnode.h"

#if SUN             /* to enable debuggging */
#include "o_fs.h"
#endif

#if (MACINTOSH || IBM || HP)
struct vnodeops otfs_fvnodeops;
struct vnodeops otfs_dvnodeops;
#endif

#if HP
struct vnodeops otfs_fvnodeops = {
    (int (*)()) otvn_open,       /* vn_open      */
    (int (*)()) otvn_close,      /* vn_close     */
    (int (*)()) otvn_rdwr,       /* vn_rdwr      */
    (int (*)()) otvn_invalop,    /* vn_ioctl     */
    (int (*)()) otvn_invalop,    /* vn_select    */
    (int (*)()) otvn_getattr,    /* vn_getattr   */
    (int (*)()) otvn_setattr,    /* vn_setattr   */
    (int (*)()) otvn_access,     /* vn_access    */
    (int (*)()) otvn_invalop,    /* vn_lookup    */
    (int (*)()) otvn_invalop,    /* vn_create    */
    (int (*)()) otvn_invalop,    /* vn_remove    */
    (int (*)()) otvn_link,       /* vn_link      */
    (int (*)()) otvn_invalop,    /* vn_rename    */
    (int (*)()) otvn_invalop,    /* vn_mkdir     */
    (int (*)()) otvn_invalop,    /* vn_rmdir     */
    (int (*)()) otvn_invalop,    /* vn_readdir   */
    (int (*)()) otvn_invalop,    /* vn_symlink   */
    (int (*)()) otvn_readlink,   /* vn_readlink  */
    (int (*)()) otvn_fsync,      /* vn_fsync     */
    (int (*)()) otvn_inactive,   /* vn_inactive  */
    (int (*)()) otvn_bmap,       /* vn_bmap      */
    (int (*)()) otvn_strategy,   /* vn_strategy  */
    (int (*)()) otvn_invalop,    /* vn_bread     */
    (int (*)()) otvn_invalop,    /* vn_brelse    */
    (int (*)()) otvn_invalop,    /* vn_lockctl   */
    (int (*)()) otvn_fid         /* vn_fid       */
};

struct vnodeops otfs_dvnodeops = {
    (int (*)()) otvn_open,       /* vn_open      */
    (int (*)()) otvn_close,      /* vn_close     */
    (int (*)()) otvn_rdwr,       /* vn_rdwr      */
    (int (*)()) otvn_invalop,    /* vn_ioctl     */
    (int (*)()) otvn_invalop,    /* vn_select    */
    (int (*)()) otvn_getattr,    /* vn_getattr   */
    (int (*)()) otvn_setattr,    /* vn_setattr   */
    (int (*)()) otvn_access,     /* vn_access    */
    (int (*)()) otvn_lookup_dcp, /* vn_lookup    */
    (int (*)()) otvn_create_dcp, /* vn_create    */
    (int (*)()) otvn_remove,     /* vn_remove    */
    (int (*)()) otvn_link,       /* vn_link      */
    (int (*)()) otvn_rename,     /* vn_rename    */
    (int (*)()) otvn_mkdir,      /* vn_mkdir     */
    (int (*)()) otvn_rmdir,      /* vn_rmdir     */
    (int (*)()) otvn_readdir,    /* vn_readdir   */
    (int (*)()) otvn_symlink,    /* vn_symlink   */
    (int (*)()) otvn_readlink,   /* vn_readlink  */
    (int (*)()) otvn_invalop,    /* vn_fsync     */
    (int (*)()) otvn_inactive,   /* vn_inactive  */
    (int (*)()) otvn_invalop,    /* vn_bmap      */
    (int (*)()) otvn_invalop,    /* vn_strategy  */
    (int (*)()) otvn_invalop,    /* vn_bread     */
    (int (*)()) otvn_invalop,    /* vn_brelse    */
    (int (*)()) otvn_invalop,    /* vn_lockctl   */
    (int (*)()) otvn_fid         /* vn_fid       */
};
#endif /* HP */

#if SUN
/* vnode operation vectors for files. */
struct vnodeops otfs_fvnodeops = {
    otvn_open,          /* vn_open      */
    otvn_close_lck,     /* vn_close     */
    otvn_rdwr_lck,      /* vn_rdwr      */
    otvn_invalop,       /* vn_ioctl     */
    otvn_invalop,       /* vn_select    */
    otvn_getattr_lck,   /* vn_getattr   */
    otvn_setattr_lck,   /* vn_setattr   */
    otvn_access_lck,    /* vn_access    */
    otvn_invalop,       /* vn_lookup    */
    otvn_invalop,       /* vn_create    */
    otvn_invalop,       /* vn_remove    */
    otvn_link_lck,      /* vn_link      */
    otvn_invalop,       /* vn_rename    */
    otvn_invalop,       /* vn_mkdir     */
    otvn_invalop,       /* vn_rmdir     */
    otvn_invalop,       /* vn_readdir   */
    otvn_invalop,       /* vn_symlink   */
    otvn_readlink_lck,  /* vn_readlink  */
    otvn_fsync_lck,     /* vn_fsync     */
    otvn_inactive_lck,  /* vn_inactive  */
    otvn_bmap_lck,      /* vn_bmap      */
    otvn_strategy_lck,  /* vn_strategy  */
    otvn_invalop,       /* vn_bread     */
    otvn_invalop,       /* vn_brelse    */
    otvn_invalop,       /* vn_lockctl   */
    otvn_fid_lck        /* vn_fid       */
};

/* vnode operation vectors for directories. */
struct vnodeops otfs_dvnodeops = {
    otvn_open,          /* vn_open      */
    otvn_close_lck,     /* vn_close     */
    otvn_rdwr_lck,      /* vn_rdwr      */
    otvn_invalop,       /* vn_ioctl     */
    otvn_invalop,       /* vn_select    */
    otvn_getattr_lck,   /* vn_getattr   */
    otvn_setattr_lck,   /* vn_setattr   */
    otvn_access_lck,    /* vn_access    */
    otvn_lookup_lck,    /* vn_lookup    */
    otvn_create_lck,    /* vn_create    */
    otvn_remove_lck,    /* vn_remove    */
    otvn_link_lck,      /* vn_link      */
    otvn_rename_lck,    /* vn_rename    */
    otvn_mkdir_lck,     /* vn_mkdir     */
    otvn_rmdir_lck,     /* vn_rmdir     */
    otvn_readdir_lck,   /* vn_readdir   */
    otvn_symlink_lck,   /* vn_symlink   */
    otvn_readlink_lck,  /* vn_readlink  */
    otvn_invalop,       /* vn_fsync     */
    otvn_inactive_lck,  /* vn_inactive  */
    otvn_invalop,       /* vn_bmap      */
    otvn_invalop,       /* vn_strategy  */
    otvn_invalop,       /* vn_bread     */
    otvn_invalop,       /* vn_brelse    */
    otvn_invalop,       /* vn_lockctl   */
    otvn_fid_lck        /* vn_fid       */
};
#endif

INT32   otfs_debug = 1;

    
/********************************************************************/
/*  otvn_access:                                                    */
/*      VFS routine to check access permissions for a file or       */
/*      directory                                                   */
/*                                                                  */
/*  input parameters:                                               */
/*      vp:     the vnode for the file or directory                 */
/*      mode:   VREAD, VWRITE or VEXEC                              */
/*                                                                  */
/*  returns:                                                        */
/*      0 if successful                                             */
/*      EACESS if permission is denied                              */
/*                                                                  */
/*  notes:                                                          */
/*      only denied access is the attempt to write a read-only file */
/********************************************************************/
INT32 otvn_access(struct vnode    *vp,
                  INT32           mode,
                  struct ucred    *cred)
{
    INT32 error;
    
    error = ot_check_permissions(vp,cred);
#if CHECK_CRED
    if (!error) {
        /* now we need to check the file's mode.  permission is
           granted unless an attempt is made to write a read-only
           file.  should we be checking more than this??? */
        if ((otp->attr & OTA_RDONLY) && mode == VWRITE)
            error = EPERM;
    }
#endif
    return error ? EACCES : error;
} /* end otvn_access */

/********************************************************************/
/*  otvn_close:                                                     */
/*      VFS routine to close a file or directory                    */
/*                                                                  */
/*  input parameters:                                               */
/*      vp:     the vnode for the file or directory to be closed    */
/*      flag:   indicates how the file was opened                   */
/*                                                                  */
/*  returns:                                                        */
/*      0                                                           */
/*                                                                  */
/*  Modification History                                            */
/*                                                                  */
/********************************************************************/
INT32 otvn_close(struct vnode    *vp,
                 INT32           flag,
                 struct ucred    *cred)
{
    return 0;
} /* end otvn_close */

/********************************************************************/
/*  otvn_create:                                                    */
/*      VFS routine to create a file in a given directory           */
/*                                                                  */
/*  input parameters:                                               */
/*      dvp       vnode for the directory in which to create the    */
/*                file                                              */
/*      nm        the name of the file to create                    */
/*      vap       attributes for the new file                       */
/*      exclusive exclusive/non-exclusive tag                       */
/*      mode      the open mode                                     */
/*      cred      credential structure for dvp?                     */
/*      setSize                                                     */
/*                                                                  */
/*  output parameters                                               */
/*      vpp       for the newly created files vnode                 */
/*      assocVpp  for the new created associated file vnode         */
/*                                                                  */
/*  returns:                                                        */
/*      SUCCESS   plus vnode(s) for newly created file(s)           */
/*      EEXIST    file(s) already exist                             */
/*      EISDIR    requested to make a directory, use otvn_mkdir     */
/*                                                                  */
/*  notes:                                                          */
/*      vpp will always contain the vnode for the data file, never  */
/*           the associated file                                    */
/*                                                                  */
/*  Modification History                                            */
/*      22.Sep.92  ECB  Added changes for ECMA 167 directory strc.  */
/*                                                                  */
/********************************************************************/
INT32 otvn_create(register struct vnode   *dvp,
                  caddr_t                 nm,
                  register struct vattr   *vap,
                  enum vcexcl             exclusive,
                  INT32                   mode,
                  struct vnode            **vpp,
                  struct vnode            **assocVpp,
                  struct ucred            *cred,
                  int                     setSize)
{
    UINT16                  numWritten;
    UINT32                  sectorBegins,entry;
    enum vtype              reqType;
    struct otnode           *df, *rf;
    struct otfs             *fsp = (struct otfs *) dvp->v_vfsp->vfs_data;
    struct FileIDDesc       *dirEntry;
    register INT32          error;
    
    DMLPRINTF(oVndLev)("otvn_create: create %s in %s at %lx\n",
                     nm,((struct otnode *) dvp->v_data)->name,
                     ((struct otnode *) dvp->v_data)->icb_addr_LBN);
    
    DMLPRINTF(oVndLev)("otvn_create: type to create is 0x%x\n",vap->va_type);
    DMLPRINTF(oVndLev)("otvn_create: size is 0x%x\n",vap->va_size);
    
    if (vap->va_type == VDIR)   /* use otvn_mkdir for directories */
        return EISDIR;
    
    reqType = vap->va_type;     /* save it, so we know what to return */
    if (reqType == VFORKS)      /* make both forks */
        vap->va_type = VREG;    /* so the icb entry contains the proper attribute */
    
    if (reqType == VASSOC) {    /* just create the resource fork */
        DMLPRINTF(oVndLev)("otvn_create: it thinks it should create the resource fork\n");
        
        error = ot_getDirentry( fsp, (struct otnode *) dvp->v_data,
                               &dirEntry, &sectorBegins, &entry, nm,
                               (UINT32) 0, findByName);
        if (!error)
            /* ISSUE - this changes with fix to Resource Fork Location */
            if (* ((UINT32 *) dirEntry->ICB.aImpUse[2]) > 0)
                error = ot_getAndFillOTNode(fsp, (struct otnode *)
                                            dvp->v_data, &df, &rf, nm,
                                            VASSOC, dirEntry, LAST);
            else {
                error = ot_setupDirentryOTNode( fsp, &rf, vap, VASSOC,
                                               nm, ((struct otnode *)
                                                    dvp->v_data)->icb_addr_LBN);
                /* update directory data */
                if (!error) {
                    /* ISSUE - this changes when we perm locate Resource Fork Addr */
                    *(UINT32 *)(dirEntry->ICB.aImpUse[2])
                        = rf->icb_addr_LBN; 
                    error = ot_writeDirentry( fsp, (struct otnode *)
                                             dvp->v_data, (caddr_t)
                                             dirEntry,sectorBegins+entry);
                    /* ISSUE - This is not the correct response for all errors */
                    if (error) {
                        rf->flags &= ~OT_MOD;
                    }
                }
            }
        if (!error || (error == EEXIST && setSize))
            ot_setFileSize( fsp, rf, vap->va_size, nonContiguous);
    } else {
        DMLPRINTF(oVndLev)("otvn_create: just create the data fork\n");
        
        /* get or create otnode(s) */
        DMLPRINTF(oVndLev)("otvn_create: get/create the otnode for the file\n");
        error = ot_direntry( fsp, (struct otnode *)dvp->v_data, nm,
                            vap, &df, reqType, &rf);
        
        /* add new entry to parent directory and set the file size for the new file */
        if (!error) {
            DMLPRINTF(oVndLev)("otvn_create: add new name to directory\n");
            error = ot_addDirentry( fsp, df, rf, (struct otnode *)
                                   dvp->v_data);
        }
        if (!error || (setSize && error == EEXIST)) {
            DMLPRINTF(oVndLev)("otvn_create: set the file's size\n");
            error = ot_setFileSize( fsp, df, vap->va_size,
                                   nonContiguous);
        }
    }
    if ((error == 0 || error == EEXIST) && vpp) {
        DMLPRINTF(oVndLev)("otvn_create: successfully created file\n");
        *vpp = &df->vnode;
    } else {
        if (df) {
            DMLPRINTF(oVndLev)("otvn_create: something bad happened, release otnode\n");
            df->flags = 0;
            df->vnode.v_count--;
            otvn_inactive(&df->vnode,cred);
        }
    }
    
    if ((error == 0 || error == EEXIST) && assocVpp)
        *assocVpp = &rf->vnode;
    else {
        if (rf) {
            DMLPRINTF(oVndLev)("otvn_create: it thinks it should");
            DMLPRINTF(oVndLev)("release an otnode for the resource");
            DMLPRINTF(oVndLev)("fork\n"); 
            rf->flags = 0;
            rf->vnode.v_count--;
            otvn_inactive(&rf->vnode,cred);
        }
    }
    
    DMLPRINTF(oVndLev)("otvn_create: returning error 0x%x\n",error);
    return error;
} /* end otvn_create */

/********************************************************************/
/*  otvn_fid:                                                       */
/*      VFS routine to get a unique file identifier for a vnode     */
/*                                                                  */
/*  input parameters:                                               */
/*      vfsp:       the vfs structure for the volume                */
/*                  NOTE: This has been removed, as NO ONE uses it  */
/*                  including NSR, SUN, HP, ...                     */
/*      vp:         the vnode for which a unique identifier is      */
/*                  desired                                         */
/*      fidpp:      for a structure which can hold the identifier   */
/*                                                                  */
/*  output parameters:                                              */
/*      fidpp:      newly allocated and contains the identifier:    */
/*                  the file/directory icb entry addr and version   */
/*  returns:                                                        */
/*      0:          if successful                                   */
/*      ENOMEM:     a memory allocation failure occurred            */
/********************************************************************/
INT32 otvn_fid(struct vnode    *vp,
               struct fid      **fidpp)
{
    
    register struct fidData *fidData;
    
    if (!(*fidpp))
        *fidpp = (struct fid *) kmem_alloc((UINT32) sizeof(struct fid) +
                                           (sizeof(struct fidData) - 1),0);
    if (*fidpp) {
        (*fidpp)->fid_len = sizeof(struct fidData);
        fidData = (struct fidData *) (*fidpp)->fid_data;
        fidData->icbAddr = ((struct otnode *) vp->v_data)->icb_addr_LBN;
        fidData->version = ((struct otnode *) vp->v_data)->version;
        fidData->attribute = vp->v_type;
        ot_copy(((struct otnode *) vp->v_data)->name, fidData->name,
                (INT32) FILENAMESIZE);
    }   
    return *fidpp ? 0 : ENOMEM;
} /* end otvn_fid */

/********************************************************************/
/*  otvn_fsync:                                                     */
/*      VFS routine to flush cached information for a given file    */
/*                                                                  */
/*  input parameters:                                               */
/*      vp: the vnode for the file to be flushed                    */
/*                                                                  */
/*  returns:                                                        */
/*      0:  success                                                 */
/*                                                                  */
/*  notes:                                                          */
/*      once flushed, an attempt will be made to inactivate the     */
/*          vnode for the file. note that the vnode will be released*/
/*          iff it has been flushed and is no longer referenced     */
/********************************************************************/
INT32 otvn_fsync(struct vnode     *vp,
                 struct ucred    *cred)
{
    INT32          error;
    struct otnode  *otp = (struct otnode *) vp->v_data;
    struct otfs    *fsp = (struct otfs *) vp->v_vfsp->vfs_data;
    
    /* check that we have permission to flush the file */
    error = otvn_access(vp,VWRITE,cred);
    if (error)
        return error;
#ifdef USECONSOLIDATE
    error = ot_consolidateADL(otp);
#endif
    if (otp->flags & OT_MOD && !(otp->flags & OT_INVAL) && !error) {
        error = ot_ICBRdWr(fsp,otp,UIO_WRITE,LAST);
        if (!error) {
            otp->flags &= ~OT_MOD;
        }
    }
    if (!error)
        otvn_inactive(&otp->vnode,cred);
    return error;
} /* end otvn_fsync */

/********************************************************************/
/*  otvn_getattr:                                                   */
/*      VFS routine to get attributes for a file or directory       */
/*                                                                  */
/*  input parameters:                                               */
/*      vp:     the vnode for the file or directory                 */
/*      vap:    the attribute structure to be filled                */
/*                                                                  */
/*  output paremeters:                                              */
/*      vap:    contains the file/directory's attributes            */
/*                                                                  */
/*  returns:                                                        */
/*      0:      success                                             */
/********************************************************************/
INT32 otvn_getattr(register struct vnode   *vp,
                   struct vattr            *vap,
                   struct ucred            *cred)
{
    struct otnode   *otp = (struct otnode *) vp->v_data;
    struct otfs     *fsp = (struct otfs *) vp->v_vfsp->vfs_data;
    INT32  error;
    
    /* check that we have permission to look at the file */
    error = ot_check_permissions(vp,cred);
    
    /* provide fs attr */
    if (!error) {
        vap->va_nodeid = otp->icb_addr_LBN;
        vap->va_size = otp->icbe.fe.InfoLengthL;
        vap->va_type = vp->v_type;
        vap->va_fsid = fsp->VolSection[0].devvp->v_rdev;
        vap->va_blocksize = (INT32) fsp->SectorSize;
        vap->va_rdev = fsp->VolSection[0].devvp->v_rdev;
        vap->va_blocks = (INT32) (vap->va_size / fsp->SectorSize) + 1;
        
        /* get host specific attr */
        ot_getAttributes(otp,vap);
    }
    return error;
} /* end otvn_getattr */

/********************************************************************/
/*  otvn_inactive:                                                  */
/*      VFS routine to inactivate a vnode                           */
/*                                                                  */
/*  input parameters:                                               */
/*      vp: the vnode to be inactivated                             */
/*                                                                  */
/*  returns:                                                        */
/*      0   success                                                 */
/********************************************************************/
INT32 otvn_inactive(struct vnode    *vp,
                    struct ucred    *cred)
{
    INT32 error;
    register struct otnode *otp = (struct otnode *) vp->v_data;
    
    /* check that we have permission to do anything with the file */
    error = ot_check_permissions(vp,cred);
    
    /* release only if there isn't a release already pending */
    if (!(otp->flags & OT_RELE_PEND) && !error)
        ot_release(otp);
    return error;
} /* end otvn_inactive */

/********************************************************************/
/*  otvn_invalop:                                                   */
/*      routine to indicate an invalid request for this file system */
/*      was made                                                    */
/*                                                                  */
/*  returns:                                                        */
/*      EINVAL: inavlid request                                     */  
/********************************************************************/
INT32 otvn_invalop(void)
{
    return EINVAL;
} /* end otvn_invalop */

/********************************************************************/
/*  otvn_lookup:                                                    */
/*      VFS routine to look up a name in a directory                */
/*                                                                  */
/*  input parameters:                                               */
/*      dvp:        the vnode for the directory                     */
/*      nm:         the name of the file for which to look          */
/*      vpp:        for the file desired                            */
/*      version:    the version of the file to be located           */
/*                                                                  */
/*  output parameters:                                              */
/*      vpp:        contains the vnode for the file if it exists    */
/*  returns:                                                        */
/*      0           success                                         */
/*      ENOENT:     the file does not exist                         */
/*                                                                  */  
/*  notes:                                                          */
/*      requesting ".." at the root will result in ENOENT           */
/********************************************************************/
INT32 otvn_lookup(struct vnode    *dvp,
                  caddr_t         nm,
                  struct vnode    **vpp,
                  struct ucred    *cred,
                  UINT32          version)
{
    register INT32  error;
    struct otnode   *otp;
    struct otfs     *fsp = (struct otfs *) dvp->v_vfsp->vfs_data;
    
    /* check that we have permission to access the file */
    error = ot_check_permissions(dvp,cred);
    
    /* look in dir for name - VNON indicates not RFork */
    if (!error)
        error = ot_dirlook( fsp, (struct otnode *) dvp->v_data, nm,
                           &otp, VNON, version);
    if (error == EEXIST)
        *vpp = &otp->vnode;
    return error == EEXIST ? 0 : error;
} /* end otvn_lookup */

/********************************************************************/
/*  otvn_lookup_assoc:                                              */
/*      routine to look up an associated file in a directory        */
/*                                                                  */
/*  input parameters:                                               */
/*      dvp:        the vnode for the directory                     */
/*      nm:         the name of the file for which the associated   */
/*                      file is desired                             */
/*      vpp:        for the file desired                            */
/*      version:    the version of the associated file to be located*/
/*                                                                  */
/*  output parameters:                                              */
/*      vpp:        contains the vnode for the file if found        */
/*                                                                  */
/*  returns:                                                        */
/*      0           success                                         */
/*      ENOENT:     the file does not exist                         */
/*                                                                  */  
/*  notes:                                                          */
/*      requesting ".." at the root will result in ENOENT           */
/********************************************************************/
INT32 otvn_lookup_assoc(struct vnode    *dvp,
                        caddr_t         nm,
                        struct vnode    **vpp,
                        struct ucred    *cred,
                        UINT32          version)
{
    struct otnode   *otp;
    struct otfs     *fsp = (struct otfs *) dvp->v_vfsp->vfs_data;
    register INT32  error;
    
    /* look in dir for name - VASSOC indicates associated file */
    error = ot_dirlook( fsp, (struct otnode *) dvp->v_data, nm, &otp,
                       VASSOC, version);
    if (error == EEXIST)
        *vpp = &otp->vnode;
    return error == EEXIST ? 0 : error;
} /* end otvn_lookup_assoc */

/********************************************************************/
/*  otvn_mkdir                                                      */
/*      VFS routine to make a directory                             */
/*                                                                  */
/*  input parameters:                                               */
/*      dvp:    the vnode for the directory in which the new        */
/*                  directory is to be added                        */
/*      nm:     the name of the new directory                       */
/*      vap:    the attributes for the new directory                */
/*      vpp:    for the new directory                               */
/*                                                                  */
/*  output parameters:                                              */
/*      vpp:    contains the vnode for the newly created directory  */
/*                  if the creation was successful                  */
/*  returns:                                                        */
/*      0       success                                             */
/*      EEXIST  the directory already existed                       */
/*      ENOTDIR a request was made to make something other than a   */
/*              directory, or the parent vnode is not a directory   */
/*                                                                  */
/*  notes:                                                          */
/*      if vap->va_type == VFORKS, then both the data and resource  */
/*          (i.e., associated file) forks should be made.           */
/*      if vap->va_type == VDIR, then just the data fork will be    */
/*          created.                                                */
/*      vpp will always contain the data fork vnode, never the      */
/*          resource fork vnode.                                    */
/*      if error==EEXIST, then the directory already existed and    */
/*          vpp contains the vnode for it                           */
/*      a newly created directory contains only the ".." entry      */
/*                                                                  */
/*  Modification history:                                           */
/*      27.Dec.89   DRaM    sorted directories                      */
/*      22.Sep.92   ECB     Changes for ECMA 167 directory struct.  */
/********************************************************************/
INT32 otvn_mkdir(struct vnode    *dvp,
                 caddr_t         nm,
                 struct vattr    *vap,
                 struct vnode    **vpp,
                 struct ucred    *cred)
{
    UINT16                          numWritten;
    struct otnode                   *newdir,
    *rfotp,
    *parent = (struct otnode *) dvp->v_data;
    register INT32                  error;
    register INT16                  BothForks = 0;
    register struct FileIDDesc      *DirEntry;
    register struct otfs            *fsp = (struct otfs *) dvp->v_vfsp->vfs_data;
    UINT32                          dirlen; /* length of created entry */
    
    if (dvp->v_type != VDIR ||
        (vap->va_type != VFORKS && vap->va_type != VDIR))
        return ENOTDIR;
    
    /* check that we have permission to access the directory */
    error = ot_check_permissions(dvp,cred);
    if (error)
        return error;
    
    if (vap->va_type == VFORKS) {   /* make both forks */
        vap->va_type = VDIR;        /* so the icb entry contains the proper attribute */
        BothForks = 1;
    }
    
    /* get or create otnode(s) */

    error = ot_direntry( fsp, (struct otnode *) dvp->v_data, nm, vap,
                        &newdir, BothForks ? VFORKS : VREG, &rfotp);
    
    /* create ".." entry and write to disk */

    if (!error) {

        bzero( fsp->dirbuf, (INT32) fsp->SectorSize);
        DirEntry = (struct FileIDDesc *) fsp->dirbuf;

        /* ISSUE - handling of .. changes and character sets */
       
        DirEntry->L_FI = 2;
        DirEntry->L_IU = 4;
        DirEntry->VersionNum = 1;
        DirEntry->ICB.Location_LBN = newdir->parent_addr_LBN;
        DirEntry->ICB.Location_PartNo = newdir->parent_addr_PartNo;
        DirEntry->Characteristics = DIR_ATTR;

        /* ISSUE - must set resource fork pointer to zero */

        ot_copy( "..", ((caddr_t) DirEntry
                        + FILE_ID_DESC_CONSTANT_LEN
                        + DirEntry->L_IU), (INT32) 2); 
        
        /* write the entry at first byte in the directory file */

        error = ot_writeDirentry( fsp,newdir,(caddr_t)fsp->dirbuf,0);
    }

    /* add new directory's name to parent directory data */
    if (!error)
        error = ot_addDirentry( fsp, newdir, rfotp, parent);
    
    if (BothForks && rfotp) {
        rfotp->vnode.v_count--;
        otvn_inactive(&rfotp->vnode,cred);
    }
    
    if (error == 0 || error == EEXIST)
        *vpp = &newdir->vnode;
    else {
        /* error other than EEXIST means both otnodes are brand new, */
        /* and must really be taken off the active list */ 
        if (newdir) {
            newdir->flags = 0;
            newdir->vnode.v_count--;
            otvn_inactive(&newdir->vnode,cred);
        }
        if (rfotp) {
            rfotp->flags = 0;
            otvn_inactive(&rfotp->vnode,cred);
        }
    }
    return error;
} /* end otvn_mkdir */

/********************************************************************/
/*  otvn_open                                                       */
/*      VFS routine to open a file or directory                     */
/*                                                                  */
/*  input parameters:                                               */
/*      vpp:    for the opened file or directory                    */
/*      flags:  the open flags                                      */
/*      cred:   pointer to the credentials structure                */
/*                                                                  */
/*  returns:                                                        */
/*      0       success                                             */
/*                                                                  */
/*  notes:                                                          */
/*      within NSR, otvn_lookup does all the work; and, the concept */
/*          of opening an unnamed file really doesn't exist.        */ 
/********************************************************************/
INT32 otvn_open(struct vnode    **vpp,
                INT32           flag,
                struct ucred    *cred)
{
    return 0;
} /* end otvn_open */


/********************************************************************/
/*  otvn_rdwr                                                       */
/*      VFS routine to read or write file data                      */
/*                                                                  */
/*  input parameters:                                               */
/*      vp:     the vnode for the file to be read/written           */
/*      uiop:   the uio structure for the data:                     */
/*              uiop->uio_offset indicates where in the file,       */
/*                  in terms of bytes, the read/write should begin  */
/*              uiop->uio_resid indicates how many bytes should be  */
/*                  read/written                                    */
/*              uiop->uio_iov->iov_base is the data to be           */
/*                  written/where to put the data read              */
/*      rw:     UIO_READ or UIO_WRITE                               */
/*      ioflag: indicates whether or not to append, zero otherwise  */
/*                                                                  */
/*  output parameters:                                              */
/*      uiop:   uiop->uio_offset has been incremented by the number */
/*                  of bytes read or written                        */
/*              uiop->resid has been decremented by the number of   */
/*                  bytes read or written                           */
/*              uiop->uio_iov->iov_base contains new data           */
/*                  (if UIO_READ) and has been incremented by the   */
/*                  number of bytes read/written                    */
/*  returns:                                                        */
/*      0       success                                             */
/*      ENOENT: the vnode/otnode has been invalidated               */
/********************************************************************/
INT32 otvn_rdwr(struct vnode    *vp,
                struct uio      *uiop,
                enum uio_rw     rw,
                INT32           ioflag,
                struct ucred    *cred)
{
    struct otfs             *fsp = (struct otfs *) vp->v_vfsp->vfs_data;
    register INT32          error;
    register struct otnode  *otp = (struct otnode *) vp->v_data;
    
    /* is otnode invalid? */
    if (otp->flags & OT_INVAL)
        return ENOENT;
    
    /* check that we have permhission to touch the file */
    error = ot_check_permissions(vp,cred);
    
    if (!error) {
        if ((ioflag & IO_APPEND) && rw == UIO_WRITE)
            uiop->uio_offset = otp->icbe.fe.InfoLengthL;
        if (rw == UIO_READ)
            error = ot_readData(fsp,otp,uiop);
        else
            error = ot_writeData(fsp,otp,uiop);
    }
    return error;        
} /* end otvn_rdwr */

/********************************************************************/
/*  otvn_readdir                                                    */
/*      VFS routine to read directory entries                       */
/*                                                                  */
/*  input parameters:                                               */
/*      dvp:    the vnode for the directory                         */
/*      uiop:   for the data read:                                  */
/*              uiop->uio_offset indicates where, in terms of bytes,*/
/*                  the read should begin                           */
/*              uiop->uio_resid indicates how many bytes should be  */
/*                  read. uiop->uio_resid == 0 on return iff all    */
/*                  entries desired fit in the buffer passed        */
/*                                                                  */
/*  output parameters:                                              */
/*      uiop:   uiop->uio_iov->iov_base contains the entries        */
/*                  formatted according to the definition of the    */
/*                  direct structure. in addition, the actual       */
/*                  directory entry immediately follows the filename*/
/*  returns:                                                        */
/*      0       success                                             */
/*      EINVAL  an invalid read request was made                    */
/*      ENOENT  dvp is not the vnode for a directory                */
/********************************************************************/
INT32 otvn_readdir(struct vnode        *dvp,
                   register struct uio *uiop,
                   struct ucred        *cred)
{
    UINT32                 ReturnBufIdx; /* next free spot in return buffer */
    UINT32                 maxIdx;       /* set to initial uio_resid */
    UINT16                 len;          /* directory entry length in bytes */
    UINT32                 avail_len;    /* Number of bytes left in dirbuf */
    UINT32                 retlen;       /* return length of dir in bytes */
    struct direct          *d;
    struct otnode          *otp = (struct otnode *) dvp->v_data;
    struct otfs            *fsp = (struct otfs *) otp->vnode.v_vfsp->vfs_data;
    struct FileIDDesc      *direntry;
    INT32                  error = 0;
    struct uio             dir_uio;        /* uio struct for actual io */
    struct iovec           dir_iovec;      /* iovec struct for actual io */
    INT32                  bytesToInterpret = 0;
    INT32                  partialDirentry;
    INT32                  byteOffset;
    
    if (otp->vnode.v_type != VDIR)
        return ENOENT;
    
    /* check that we have permission to look at the directory */
    if ( error = ot_check_permissions(dvp,cred) )
        return error;
    
    /* validate request */
    if (uiop->uio_iovcnt != 1)
        return EINVAL;
  
    /*
    * In an attempt to speed up this routine, otp->lastEntry indicates the 
    * number of readdir() bytes returned to the user. otp->lastOffset holds
    * the offset into the actual directory data on the disk.
    */
    if (uiop->uio_offset == otp->lastEntry ) {
        /* start from where we left off last time we read in this directory */
        byteOffset = otp->lastOffset;
    } else {
        /* start from beginning of directory data */
        byteOffset = 0;
        otp->lastOffset = 0;
        otp->lastEntry = 0;
    }
    
    ReturnBufIdx = 0;
    maxIdx = uiop->uio_resid;
    
    do 
    {
        partialDirentry = FALSE;
        /* read a buffer */
        uioInit(&dir_uio,&dir_iovec,fsp->dirbuf,byteOffset,DIRBUFSIZE);
        error = ot_readData(fsp,otp,&dir_uio);
        bytesToInterpret = DIRBUFSIZE - dir_uio.uio_resid;
        if( !error || error == EEOF) {
            if( bytesToInterpret ) {
                direntry = (struct FileIDDesc *) fsp->dirbuf;
                do 
                {
                    /*
                    * go through each file entry in buffer just read
                    */
                    avail_len = (UINT32)(fsp->dirbuf+DIRBUFSIZE-(char *)direntry);
                    partialDirentry = (avail_len < FILE_ID_DESC_CONSTANT_LEN);
                    if (!partialDirentry) {
#if !IBM
                        ot_xlate(TAGID_FILE_ID, FROM_OP,
                                 (caddr_t) direntry, avail_len,0);
#endif
                        len = (FILE_ID_DESC_CONSTANT_LEN
                               + direntry->L_FI
                               + direntry->L_IU
                               + ot_dirpad(direntry->L_FI,direntry->L_IU));
                        partialDirentry = (len > avail_len);
                    }
                    if (!partialDirentry) {
                        retlen = sizeof(d->d_fileno)
                               + sizeof(d->d_reclen)
                               + sizeof(d->d_namlen)
                               + direntry->L_FI + 1;
                        
                        if ( retlen % 4 )
                            retlen += (4 - (retlen % 4)); /* double word alignment */
    
                        if(!( direntry->Characteristics & DELETE_ATTR)) {
                            /* 
                            * set the start location in the return buffer 
                            * if there is room
                            */
                            if ( ReturnBufIdx + retlen < maxIdx ) {
                                if ( ReturnBufIdx >= uiop->uio_offset ) {
                                    d = (struct direct *)(uiop->uio_iov->iov_base + ReturnBufIdx); 
                                    memset((void *)d,0,retlen);
                                    d->d_fileno = direntry->ICB.Location_LBN;
                                    d->d_reclen = retlen;
                                    d->d_namlen = direntry->L_FI;
                                    memmove(d->d_name,
                                            (UCHAR8 *)direntry
                                            + FILE_ID_DESC_CONSTANT_LEN
                                            + direntry->L_IU,
                                            direntry->L_FI);
                                    uiop->uio_resid -= retlen; 
                                }
                                ReturnBufIdx += retlen;
                            } else {
                                /* would overflow- cause loop to end */
                                ReturnBufIdx = maxIdx;
                            }
                        } /* not deleted */
                        bytesToInterpret -= len;
                        byteOffset += len;
                        direntry = (struct FileIDDesc *)((UCHAR8 *)direntry + len);
                    }
                } while ( uiop->uio_resid > 0
                       && bytesToInterpret > 0
                       && ReturnBufIdx < maxIdx
                       && !partialDirentry );
            } /* bytesToInterpret */
        } /* not error */
    } while( ! error && uiop->uio_resid > 0 && dir_uio.uio_resid == 0 );

    /* do the equivalent of uiomove() */
    uiop->uio_offset += maxIdx-uiop->uio_resid; 
    uiop->uio_iov->iov_base += maxIdx-uiop->uio_resid; 

    otp->lastOffset = byteOffset;
    otp->lastEntry =  uiop->uio_offset;

    if (!error || error == EEOF)  {
      return( SUCCESS);
    } else {
      return( error);
    }
} /* end otvn_readdir */

/********************************************************************/
/*  otvn_remove                                                     */
/*      VFS routine to remove a file from a directory               */
/*                                                                  */
/*  input parameters:                                               */
/*      dvp:    the vnode for the directory                         */
/*      nm:     the name of the file to remove                      */
/********************************************************************/
INT32 otvn_remove(struct vnode    *dvp,
                  caddr_t         nm,
                  struct ucred    *cred)
{
    struct otnode   *dotp = (struct otnode *) dvp->v_data;
    struct otfs     *fsp = (struct otfs *) dvp->v_vfsp->vfs_data;
    register INT32  error;
    
    /* check access permission on dir - delete if ok */
    if (error = otvn_access(dvp,(INT32) VWRITE,cred))
        return error;
    error = ot_delete(fsp,dotp,nm);
    return error;
} /* end otvn_remove */

/********************************************************************/
/*  otvn_rename                                                     */
/*      VFS routine to rename a file or directory                   */
/*                                                                  */
/*  input parameters:                                               */
/*      sdvp:       the vnode for the source directory              */
/*      OldName:    the name of the file within sdvp to be renamed  */
/*      ddvp:       the vnode for the destination directory         */
/*      NewName:    the new name for the file                       */
/*                                                                  */
/*  returns:                                                        */
/*      0           success                                         */
/*      EEXIST:     the file already exists in destination directory*/
/*                                                                  */
/*  notes:                                                          */
/*      if renaming within a directory, sdvp == ddvp                */
/*                                                                  */
/********************************************************************/
INT32 otvn_rename(struct vnode   *sdvp,
                  caddr_t         OldName,
                  struct vnode   *ddvp,
                  caddr_t         NewName,
                  struct ucred   *cred)
{
    int                    index;
    struct otnode          *SrcDir = (struct otnode *) sdvp->v_data;
    struct otnode          *TgtDir = (struct otnode *) ddvp->v_data;
    struct otnode          *fotStack;
    struct otnode          *rfotStack;
    struct otfs            *fsp = (struct otfs *) sdvp->v_vfsp->vfs_data;
    register struct otnode *fotp;
    register struct otnode *rfotp;
    register struct otnode *active;
    register INT32         error;
    
    /* check access permissions on both directories */
    if (error = otvn_access(sdvp,(INT32) VWRITE,cred))
        return error;
    if (error = otvn_access(ddvp,(INT32) VWRITE,cred))
        return error;
    
    error =ot_getAndFillOTNode(fsp,SrcDir,&fotStack,&rfotStack,OldName,
                               VFORKS,(struct FileIDDesc *) 0,LAST);
    fotp = fotStack;
    rfotp = rfotStack;
    
    if (error == EEXIST) {
        error = 0;
        
        /* update otnodes' names and ParentAddrs */
        fotp->parent_addr_LBN = TgtDir->icb_addr_LBN;
        fotp->parent_addr_PartNo = TgtDir->icb_addr_PartNo;
        strcpy(fotp->name,NewName);
        if (rfotp) {
            rfotp->parent_addr_LBN = TgtDir->icb_addr_LBN;
            rfotp->parent_addr_PartNo = TgtDir->icb_addr_PartNo;
            strcpy(rfotp->name,NewName);
        }
        
        /* use updated otnodes to add a new entry to the target */
        /* directory */
        error = ot_addDirentry(fsp,fotp,rfotp,TgtDir);
        
        if (error) {    /* (could even be EEXIST) */
            /* put the otnodes back the way you got them */
            fotp->parent_addr_LBN = SrcDir->icb_addr_LBN;
            fotp->parent_addr_PartNo = SrcDir->icb_addr_PartNo;
            strcpy(fotp->name,OldName);
            if (rfotp) {
                rfotp->parent_addr_LBN = SrcDir->icb_addr_LBN;
                rfotp->parent_addr_PartNo = SrcDir->icb_addr_PartNo;
                strcpy(rfotp->name,OldName);
            }
        } else {
            fotp->flags |= OT_MOD;
            if (rfotp)
                rfotp->flags |= OT_MOD;
            error = ot_deleteDirentry(fsp,SrcDir,OldName);
            for (index = 0; index < 2; index++)
                for (active = index == 0 ? fsp->active : fsp->inactive;
                     active; active = active->next)
                    if ((strcmp(active->name,OldName) == 0) &&
                        (active->icb_addr_LBN == fotp->icb_addr_LBN ||
                         (rfotp && active->icb_addr_LBN == rfotp->icb_addr_LBN))) {
                        bzero(active->name,(INT32) FILENAMESIZE);
                        strcpy(active->name,NewName);
                    }
        }
        
        fotp->vnode.v_count--;
        otvn_inactive(&fotp->vnode,cred);
        if (rfotp) {
            rfotp->vnode.v_count--;
            otvn_inactive(&rfotp->vnode,cred);
        }   
    } /* end ot_getAndFillOTNode() == EEXIST */
    
    return error;
} /* end otvn_rename */

/********************************************************************/
/*  otvn_rmdir                                                      */
/*      VFS routine to remove a directory                           */
/*                                                                  */
/*  input parameters:                                               */
/*      Parentvp:       the vnode for the directory in which the    */
/*                          directory to be removed is contained    */
/*      DirToDelete:    the name of the directory to delete         */
/********************************************************************/
INT32 otvn_rmdir(struct vnode    *Parentvp,
                 caddr_t         DirToDelete,
                 struct ucred    *cred)
{
    struct otfs     *fsp = (struct otfs *) Parentvp->v_vfsp->vfs_data;
    struct otnode   *potp = (struct otnode *) Parentvp->v_data;
    register INT32  error;
    
    /* check access permissions for parent dir */
    error = otvn_access(Parentvp,(INT32) VWRITE,cred);
    if (!error)
        error = ot_delete( fsp, potp, DirToDelete);
    return error;
} /* end otvn_rmdir */

/********************************************************************/
/*  otvn_setattr                                                    */
/*      VFS routine to set attributes for a file or directory       */
/*                                                                  */
/*  input parameters:                                               */
/*      vp:     the vnode for the file or directory                 */
/*      vap:    the attributes                                      */
/*                                                                  */
/*  returns:                                                        */
/*      0       success                                             */
/*      EINVAL: an attempt was made to set some attributes which    */
/*                  may not be changed                              */
/*                                                                  */
/********************************************************************/
INT32 otvn_setattr(struct vnode            *vp,
                   register struct vattr   *vap,
                   struct ucred            *cred)
{
    INT32           error;
    
    error = 0;
    
    DMLPRINTF(oVndLev)("otvn_setattr: top\n");
    
    /* check to see if we can touch the file */
    error = ot_check_permissions(vp,cred);
    if (error)
        return error;
    
    /* can only set some */
    if ((vap->va_nlink != -1) || (vap->va_blocksize != -1) ||
        (vap->va_rdev  != -1) || (vap->va_blocks    != -1) ||
        (vap->va_fsid  != -1) || (vap->va_nodeid    != -1)) {
        error = EINVAL;
    }
    
    /* set file size to that specified in vattr */
    if (!error && vap->va_size != (UINT32) -1) {
        DMLPRINTF(oVndLev)("otvn_setattr: before ot_setFileSize\n");
        error = ot_setFileSize((struct otfs *) vp->v_vfsp->vfs_data,
                               (struct otnode *) vp->v_data,
                               vap->va_size,nonContiguous);
    }
    DMLPRINTF(oVndLev)("otvn_setattr: after ot_setFileSize\n");
    
    /* set host specific attributes */
    if (!error) {
        DMLPRINTF(oVndLev)("otvn_setattr: before ot_setAttributes\n");
        error = ot_setAttributes((struct otnode *) vp->v_data,vap);
    }
    
    DMLPRINTF(oVndLev)("otvn_setattr: returning\n");
    return error;
} /* end otvn_setattr */

/********************************************************************/
/*  ot_check_permissions:                                           */
/*      Support routine to check access permissions for a file or   */
/*      directory                                                   */
/*                                                                  */
/*  input parameters:                                               */
/*      vp:     the vnode for the file or directory                 */
/*      cred:   the credentials structure                           */
/*                                                                  */
/*  returns:                                                        */
/*      0 if successful                                             */
/*      EPERM if permission is denied                               */
/*                                                                  */
/*  Modification History                                            */
/*      06JAN93  Permissions not initially supported                */
/*                                                                  */
/********************************************************************/
INT32 ot_check_permissions(struct vnode    *vp,
                           struct ucred    *cred)
{
    INT32 error = 0;
    return error;
} /* end ot_check_permissions */

/********************************************************************/
/*  ot_delete                                                       */
/*      routine to delete a file from a directory                   */
/*                                                                  */
/*  input parameters:                                               */
/*      fsp:            the otfs structure for the volume           */
/*      directory:      the otnode for the directory                */
/*      NameToDelete:   the name of the file to delete              */
/*                                                                  */
/*  returns:                                                        */
/*      0               success                                     */
/*                                                                  */
/*  notes:                                                          */
/*      if a resource fork exists, it too will be deleted           */
/********************************************************************/
INT32 ot_delete(struct otfs     *fsp,
                struct otnode   *directory,
                caddr_t         NameToDelete)
{
    struct otnode           *df,
                            *rf;
    register struct otnode  *fotp,  /* data fork otnode */
                            *rfotp; /* resource fork otnode */
    register INT32          error;
    
    /* get the otnode(s) */
    error = ot_getAndFillOTNode( fsp, directory, &df, &rf,
                                NameToDelete, VFORKS,
                                (struct FileIDDesc *) 0,LAST);
    if (error != EEXIST)
        return error;
    
    fotp = df;
    rfotp = rf;
    
    /* delete extended attribute icb entries and data */
    error = ot_deleteExtAttrICBs( fsp, fotp);
    if (!error && rfotp)
        error = ot_deleteExtAttrICBs( fsp, rfotp);
    
    /* delete data and icb for data fork */
    if (!error)
        error = ot_deleteData( fsp, fotp);
    if (!error)
        error = ot_ICBDelete( fsp, fotp);
    
    /* force data fork otnode to be inactivated */
    fotp->vnode.v_count--;
    otvn_inactive( &fotp->vnode,(struct ucred *) 0);
    
    if (rfotp) {    /* delete resource fork data and icb entry */
        if (!error)
            error = ot_deleteData( fsp, rfotp);
        if (!error)
            error = ot_ICBDelete( fsp, rfotp);
        
        /* force resource fork otnode to be inactivated */
        rfotp->vnode.v_count--;
        otvn_inactive( &rfotp->vnode, (struct ucred *) 0);
    }
    
    if (!error)
        error = ot_deleteDirentry( fsp, directory ,NameToDelete);
    return error;
} /* end ot_delete */

/********************************************************************/
/*  ot_deleteExtAttrICBs                                            */
/*      routine to delete a chain of extended attribute icb entries */
/*      and their data                                              */
/*                                                                  */
/*  input parameters:                                               */
/*      fsp:    the otfs structure for the volume                   */
/*      otp:    the otnode whose extended attribute icb entries are */
/*                  to be deleted                                   */
/*  returns:                                                        */
/*      0       success                                             */
/********************************************************************/
INT32 ot_deleteExtAttrICBs(register struct otfs *fsp,struct otnode *otp)
{
    struct otnode   *ExtAttrOtp;
    register INT32  error;
    register UINT32 ExtAttrICBAddr;   /* addr of first in chain */
    
    error = 0;
    ExtAttrICBAddr = otp->icbe.fe.sExtAttrICB.Location_LBN;
    
    while (ExtAttrICBAddr > 0 && !error) {

        error = ot_getOTNode( 0, fsp, "", ExtAttrICBAddr, LAST, VNON,
                             otp, &ExtAttrOtp);

        if (!error) {
            if (!(ExtAttrOtp->flags & OT_INITIALIZED)) {
                ExtAttrOtp->icb_addr_LBN = ExtAttrICBAddr;
                error = ot_ICBRdWr( fsp, ExtAttrOtp, UIO_READ, LAST);
            }
        }
        
        if (!error)
            ExtAttrICBAddr = ExtAttrOtp->icbe.fe.sExtAttrICB.Location_LBN;
        if (!error)
            error = ot_deleteData(fsp, ExtAttrOtp);
        if (!error)
            error = ot_ICBDelete(fsp, ExtAttrOtp);
        
        /* force ExtAttrOtp to be inactivated */

        ExtAttrOtp->vnode.v_count--;
        otvn_inactive(&ExtAttrOtp->vnode,(struct ucred *) 0);
    }
    return error;
} /* end ot_deleteExtAttrICBs */

/********************************************************************/
/*  ot_deleteData                                                   */
/*      routine to delete file data                                 */
/*                                                                  */
/*  input parameters:                                               */
/*      fsp:    the otfs structure for the volume                   */
/*      otp:    the otnode for the file to delete                   */
/*                                                                  */
/*  returns:                                                        */
/*      0       success                                             */
/********************************************************************/
INT32 ot_deleteData(register struct otfs    *fsp,
                    struct otnode           *otp)
{
    register INT32          error;
    register struct gen_AD  *ADp;
    error = 0;
    
    /* only deletes flat file data or empty dir data */
    if (otp->vnode.v_type == VDIR)
        if (error = ot_dirempty(fsp,otp))
            return error;
    
    /* if no data, nothing to do */
    if (otp->icbe.fe.InfoLengthL == 0)
        return 0;

    /* Now get rid of the entire ADL */
    error = ot_truncateADL(otp,0);

    otp->icbe.fe.InfoLengthL = 0L;
    otp->icbe.fe.LogBlocksL = 0L;
    otp->flags |= OT_MOD;
    return error;
} /* end ot_deleteData */

/********************************************************************/
/*  ot_getAndFillOTNode                                             */
/*      routine to get/allocate read/fill an otnode for a given file*/
/*                                                                  */
/*  input parameters:                                               */
/*      fsp:        the otfs structure for the volume               */
/*      directory:  the otnode for the directory in which to locate */
/*                      the file                                    */
/*      name:       the name of the file whose otnode is desired    */
/*      type:       VREG, VASSOC or VFORKS                          */
/*      DirEntry:   the actual directory entry                      */
/*      version:    the version of the file's icb entry desired     */
/*                                                                  */
/*  output parameters:                                              */
/*      otpp:       contains the icb entry for the file (data fork),*/
/*                      if found                                    */
/*      rfotpp:     contains the icb entry for the resource fork, if*/
/*                      if found and requested                      */
/*  returns:                                                        */
/*      EEXIST:     the file existed (success)                      */
/*                                                                  */
/*  notes:                                                          */
/*      type == VREG means find just the data fork; type == VASSOC  */
/*          means find just the resource fork; type == VFORKS means */
/*          find both.                                              */
/*      DirEntry may or may not be supplied. if not, it             */
/*          directory entry will be read from disk.                 */
/*                                                                  */
/*  Modification History:                                           */
/*      ??        ??      Original Entry                            */
/*      18.Sep.92 ECB     Modified to use ECMA 167 directory str.   */
/*                                                                  */
/*                                                                  */
/*  ISSUE - Numerous string usage not compatible w/ multibyte sets  */
/********************************************************************/
INT32 ot_getAndFillOTNode(struct otfs             *fsp,
                          struct otnode           *directory,
                          register struct otnode  **otpp,
                          register struct otnode  **rfotpp,
                          caddr_t                 name,
                          enum vtype              type,
                          struct FileIDDesc      *DirEntry,
                          UINT32                  version)
{
    INT16                   MadeDirEntry;
    UINT32                  junk1, junk2;  
    INT32                   error;
    UCHAR8                  *FID;                 /* name of the file */
    UINT32                  icbAddr; 
    UINT32                  AssocICBAddr;
    UINT16                  Version;
    enum vtype              vtype;
    UINT32                  dirlen;
    
    error = 0;
    *otpp = *rfotpp = (struct otnode *) NULL;
    MadeDirEntry = (DirEntry == NULL);
    
    /* if not supplied, get the directory entry */
    if (!DirEntry) {
        error = ot_getDirentry( fsp, directory, &DirEntry, &junk1, &junk2, name, (UINT32) 0,findByName);
        if (error)
            return error;
    }
    
    /* tease out needed values from directory entry */

    FID = ot_getFileID( DirEntry);
    icbAddr = DirEntry->ICB.Location_LBN;
    AssocICBAddr = *((UINT32 *) DirEntry->ICB.aImpUse[2]);
    Version = DirEntry->VersionNum;
    dirlen = FILE_ID_DESC_CONSTANT_LEN + DirEntry->L_FI + DirEntry->L_IU +
        ot_dirpad( DirEntry->L_FI, DirEntry->L_IU);
    if( DirEntry->Characteristics & DIR_ATTR) {
        vtype = VDIR;
    } else {
        vtype = VREG;
    }
    
    if (type != VASSOC) {
        error = ot_getOTNode( 0, fsp,(caddr_t)FID, icbAddr, version,
                             vtype, directory, otpp);
    }
    
    if ((type == VFORKS || type == VASSOC) && AssocICBAddr > 0) {
        error = ot_getOTNode( 0, fsp,(caddr_t)FID, AssocICBAddr,
                             version, VASSOC, directory, rfotpp);
    }
    
    /* did one (or both) allocations fail? */
    if ((type != VASSOC && !(*otpp)) || 
        ((type==VFORKS || type==VASSOC) &&
         AssocICBAddr > 0 && !(*rfotpp))) {
        error = ENOMEM;
    }
    
    if (strcmp( name, "..") == 0) {
        if (!error)
            error = ot_getName(fsp, *otpp, directory->version);
        if (!error && *rfotpp)
            ot_copy((*otpp)->name,(*rfotpp)->name,(INT32)
                    strlen((*otpp)->name));
    }
    
    if (MadeDirEntry)
        kmem_free((caddr_t) DirEntry,(UINT32) dirlen, BUFRAM);

    if (!error) {
        error = EEXIST;
    } else {
        if (*otpp) {
            (*otpp)->vnode.v_count--;
            otvn_inactive(&(*otpp)->vnode,(struct ucred *) 0);
        }
        if (*rfotpp) {
            (*rfotpp)->vnode.v_count--;
            otvn_inactive(&(*rfotpp)->vnode,(struct ucred *) 0);
        }
        *otpp = *rfotpp = (struct otnode *) 0;
    }
    return error;
} /* end ot_getAndFillOTNode */
    
/********************************************************************/
/*  ot_getName                                                      */
/*      fill out a file's otnode given absolute minimal information */
/*      on it, by reading it and adding information from its entry  */
/*      in its parent directory.                                    */
/*                                                                  */
/*  input parameters:                                               */
/*      fsp:            the otfs structure for the volume           */
/*      otp:            the otnode to be filled                     */  
/*      pDirVersion:    which version of the parent directory       */
/*                          should be searched                      */
/*                                                                  */
/*  output parameters:                                              */
/*      otp:        contains the appropriate icb entry for the file */
/*                                                                  */
/*  returns:                                                        */
/*      0               success                                     */
/*      ENOMEM:         memory allocation failure                   */
/*                                                                  */
/*  notes:                                                          */
/*      "Absolute minimal" information is the parent directory's    */
/*          version, and in the otnode itself, the icb address,     */
/*          the desired version, and the address of the parent vnode*/
/*      PS - neither OT_MOD nor OT_INVAL can be set in otp->flags   */
/*                                                                  */
/*  Modification History                                            */
/*      ??        ??    Original entry                              */
/*      18.Sep.92 ECB   Chabges for ECMA 167 directory structure    */
/*                                                                  */
/********************************************************************/
INT32 ot_getName(struct otfs             *fsp,
                 register struct otnode  *otp,
                 UINT32                  pDirVersion)
{
    UINT32                  trashMe1,trashMe2;
    struct otnode           *parent;
    struct FileIDDesc       *Entry;
    register INT32          error;
    UINT32                  dirlen;            /* length of allocated directory entry */
    UINT32                  junk1, junk2;
    
    parent = NULL;
    Entry = NULL;

    /* read the icb entry  */

    error = ot_ICBRdWr( fsp, otp, UIO_READ, otp->version);

    /* see if otnode represents the root directory (no parent) */
    
    if (otp->parent_addr_LBN == 0 && otp->parent_addr_PartNo == 0) {

        /* Synthesize a directory entry with the important fields for */
        /* the root no implementation use */

        dirlen = FILE_ID_DESC_CONSTANT_LEN + strlen( fsp->VolumeName) + 
            ot_dirpad( strlen( fsp->VolumeName), 0);
        Entry = (struct FileIDDesc *) kmem_alloc((UINT32) dirlen, BUFRAM);
        bzero( (caddr_t) Entry, dirlen);

        /* ISSUE - Improper string handling for multibyte character sets */

        ot_copy(fsp->VolumeName,(char *)ot_getFileID( Entry), (INT32)
                strlen(fsp->VolumeName));
        Entry->Characteristics |= DIR_ATTR;
        otp->vnode.v_flag = VROOT;

    } else {
        /* otp doesn't represent the root, so just get an Entry for */
        /* the file in question from its parent directory */

        otp->vnode.v_flag = 0;
        error = ot_getOTNode( 0, fsp, "", otp->parent_addr_LBN, pDirVersion, VNON,
                             (struct otnode *) NULL, &parent);
        if (!error)
            error = ot_getDirentry(fsp, parent, &Entry, &junk1,&junk2,
                                   (caddr_t) NULL, otp->icb_addr_LBN,
                                   findByAddress);
    }
    
    if (!error) {
        ot_copy((char *)ot_getFileID(Entry),(char *)otp->name,
                (INT32)strlen((char *)ot_getFileID( Entry))); 
        otp->attr = Entry->Characteristics;
        otp->vnode.v_type = (otp->attr == DIR_ATTR) ? VDIR : VREG;
        otp->vnode.v_op = ((otp->attr == DIR_ATTR) ?
                           &otfs_dvnodeops : &otfs_fvnodeops) ;
    }
    
    if (parent) {
        parent->vnode.v_count--;
        otvn_inactive(&parent->vnode,(struct ucred *) 0);
    }

    if (Entry)
        kmem_free((caddr_t) Entry,(UINT32) dirlen, BUFRAM);

    return error;

} /* end ot_getName */

/*
 * *****************************************************************
 *  ot_getNumFiles
 *      routine to count the number of entries in a given directory
 *
 *  input parameters:
 *      fsp:        the otfs structure for the volume
 *      icbaddr:    the icb address of the directory to search
 *      sumType:    sumDirectories, sumFiles or sumTotal
 *      numFiles    for the entry count
 *
 *  output parameters:
 *      numFiles:   contains the # of enries found of type requested
 *      vacancy:    Integer percentage of directory occupied by
 *                  deleted entries.  Call ot_reorgDirectory to
 *                  eliminate the voids and defragment the file
 *                  if vacancy gets too high
 *  returns:
 *      0           success
 *      ENOMEM:     memory allocation failure
 *                  I/O errors from read except end of file which
 *                  is filtered out.
 *
 *  notes:
 *      if sumType == sumDirectories, only the directories will be
 *      counted; if sumType == sumFiles, only the files found
 *      will be counted; if sumType == sumTotal, files and
 *      directories will be counted.  None of the totals re-
 *      turned include number of deleted entries
 *      the ".." entry does not count; so, a directory which has
 *      only the ".." entry is considered empty and numFiles=0
 *
 *  Modification History
 *      ??               Original entry
 *      17.Sep.92  ECB   Changes to support ECMA 167 Directories
 *                       and added return of vacancy
 *
 * *****************************************************************
 */
INT32 ot_getNumFiles(struct otfs         *fsp,
                     UINT32              icbaddr,
                     enum sumEntryType   sumType,
                     UINT32              *numFiles,
                     UINT16              *vacancy)
{
    UINT16            numEntries;
    INT32             totalDirs = 0L;
    INT32             totalFiles = 0L;
    INT32             totalDeleted = 0L;
    INT32             total = 0L;         /* totalDirs + totalFiles */
    caddr_t           iobuff;
    UINT32            iobuffsize;
    struct otnode     *directory;
    register INT32    error;
    register UINT16   entry;
    struct uio        uio;                /* for I/O */
    struct iovec      iovec;
    UINT32            bytesread;          /* bytes processed so far reading directory */
    UINT32            bytestoread;        /* remaining bytes in directory file */
                                          /* to process */
    UINT32            xferlen;            /* transfer length */
    UINT32            filelen;            /* just what it sez... */
    UINT32            offset;             /* bytes from beginning of iobuff to start */
                                          /* of directory entry under consideration */
    struct FileIDDesc *direntry;          /* pointer to constant part of directory */
    UINT16            dirlen;             /* size of entire directory entry */
    
    
    error = ot_getOTNode(1, fsp, "", icbaddr, LAST, VDIR,
                         (struct otnode *) 0, &directory);
    if (!directory)
        return ENOMEM;
    
    /* allocate a temporary I/O buffer */
    iobuffsize = (UINT32) fsp->SectorSize;
    iobuff = kmem_alloc( iobuffsize, BUFRAM);
    if (!iobuff) {
        directory->vnode.v_count--;
        otvn_inactive( &directory->vnode, (struct ucred *) 0);
        return ENOMEM;  
    }
    
    error = 0;
    *numFiles = 0L;
    
    if (!error) {
        filelen = directory->icbe.fe.InfoLengthL;
        bytesread = 0;
        bytestoread = filelen;

        /* pass through the entire directory file */

        do {

            /* how many bytes to transfer? */

            if( bytestoread < iobuffsize) {
                xferlen = bytestoread;
            } else {
                xferlen = iobuffsize;
            }

            /* read the data */

            uioInit( &uio, &iovec, iobuff, bytesread, (INT32) xferlen);
            error = ot_readData( fsp, directory, &uio);
#if !IBM
            ot_xlate(TAGID_FILE_ID, FROM_OP, (caddr_t) iobuff,xferlen,0);
#endif

            if (error == EEOF || error == SUCCESS) {

                offset = 0;

                /* while the entire amount of the directory entry */
                /* appears to be in buffer - process */

                while( ( FILE_ID_DESC_CONSTANT_LEN + offset) <= xferlen) {

                    direntry = ( struct FileIDDesc *) (iobuff + offset);
                    dirlen = (FILE_ID_DESC_CONSTANT_LEN
                              + direntry->L_FI + direntry->L_IU
                              + ot_dirpad(direntry->L_FI +
                                       direntry->L_IU));

                    /* now we can make the final judgement if directory is entirely enclosed */
                    /* in current buffer */

                    if( dirlen + offset <= xferlen) {

                        /* do the counting */

                        total++;
                        if( direntry->Characteristics & DIR_ATTR) {
                            totalDirs++;
                        }
                        if( direntry->Characteristics & FILE_ATTR) {
                            totalFiles++;
                        }
                        if( direntry->Characteristics & DELETE_ATTR) {
                            totalDeleted++;
                            total--;
                        }
                        
                        /* correct for '..' if seen */
                        /* ISSUE - not multibyte aware */

                        if( direntry->L_FI == 2 ) {
                            if(!strncmp((char *)ot_getFileID( direntry),"..",2)){
                                total--;
                                totalDirs--;
                            }
                        }

                        /* point to start of next directory entry */

                        offset += dirlen;


                    } else {

                        break;

                    }

                } /* while */

                /* not all in buffer, adjust pointers and read some more */

                bytesread += offset;
                bytestoread -= offset;
                break;  /* out of while */

            } /* not error */

        } while ((!error) && (bytesread < filelen));

    }
    
    directory->vnode.v_count--;
    otvn_inactive( &directory->vnode, (struct ucred *) 0);
    kmem_free( iobuff, iobuffsize, 0);
    
    if (sumType == sumDirectories)
        *numFiles = totalDirs;  
    else if (sumType == sumFiles)
        *numFiles = totalFiles;
    else if (sumType == sumTotal)
        *numFiles = total;

    /* calculate vacancy rate */

    if(total != 0) {
        *vacancy = (totalDeleted * 100) / total;
    } else {
        *vacancy = 0;
    }

    return error == EEOF ? 0 : error;
} /* end ot_getNumFiles */

/********************************************************************/
/*  ot_getOTNode                                                    */
/*      do whatever it takes to get an otnode from disk or memory   */
/*                                                                  */
/*  input parameters:                                               */
/*      getname             whether to completely fill out otnode   */
/*      fs                  file system                             */
/*      name                name of otnode's file (for fillout)     */
/*      icbaddr             sector of beginning of icb              */
/*      version             which version of icb entry is desired?  */
/*      vtype               which type of icb entry is desired?     */
/*      parent              &(otnode's parent)                      */
/*                                                                  */
/*  output parameters:                                              */
/*      theNode             the otnode it got                       */
/*                                                                  */
/*  returns:                                                        */
/*      0           success                                         */
/*      ENOMEM      out of memory                                   */
/*      ENOSPC      out of disk space                               */
/*      see also:   ot_erase, ot_getName, ot_ICBRdWr                */
/*                                                                  */
/*  notes:                                                          */
/*      if icbaddr < 0, make one from scratch. in this case, type,  */
/*      name and parent are important.  if icbaddr >= 0, get icb    */
/*      entry from disk, if necessary.  here, version is important. */
/********************************************************************/
INT32 ot_getOTNode(int                  getname,
                   register struct otfs *fs,
                   caddr_t              name,
                   UINT32               icbaddr,
                   UINT32               version,
                   enum vtype           vtype,
                   struct otnode        *parent,
                   struct otnode        **theNode)
{
    INT32          error;
    UINT32         parentLBN;
    UINT16         parentPartNo;
    struct space   *addr_list;
    register int   index;
    struct otnode  *ot = 0;
    struct otnode  *active;
    
    *theNode = 0;
    if( parent) {
        parentLBN = parent->icb_addr_LBN;
        parentPartNo = parent->icb_addr_PartNo;
    } else {
        parentLBN = 0;
        parentPartNo = 0;
    }
    
    DMLPRINTF(oVnuLev)("getOTNode: top. icbaddr= %lx\n",icbaddr);
    
    /* print_fs_ICBlists(fs); */
    
    /* check active and inactive otnode lists */

    for (index = 0; index < 2; index++) {

        for (active = index == 0 ? fs->active : fs->inactive;
             active;
             active = active->next) {
            
            if ((index == 1 || ot_active(active)) &&
                icbaddr == active->icb_addr_LBN &&
                ((version == LAST && active->isLast) || version == active->version) &&
                (vtype == VASSOC) == (active->vnode.v_type == VASSOC)) {
                
                if (index == 1) {
                    ot_dequeueOTNode(&fs->inactive,active);
                    ot_enqueueOTNode(&fs->active,active);
                }
                active->vnode.v_count++;    /* was 0 if inactive so increment is ok */
                *theNode = active;
                DMLPRINTF(oVnuLev)("getOTNode: Found on active list.\n");
                return SUCCESS;
            }
        }
    }
    
    /****************************************************************/
    /* otnode is not active. Need to allocate one for use if        */
    /* resources permit.  Note that we allocate room for the otnode */
    /* specific stuff plus a sector for the icb entry               */
    /****************************************************************/
    ot = (struct otnode *) kmem_alloc((UINT32)sizeof(struct otnode)
                                      - sizeof(union icbe)
                                      + fs->SectorSize, 0);
    if (!ot) {
        /* sync and try again */
        DMLPRINTF(oVnuLev)("getOTNode: Syncing fs.\n");
        otfs_sync(fs->vfs);
        ot = (struct otnode *)kmem_alloc((UINT32) sizeof(struct otnode)
                                         - sizeof(union icbe) 
                                         + fs->SectorSize, 0);
        if (!ot) {
            DMLPRINTF(oVnuLev)("getOTNode: ENOMEM returned.\n");
            return ENOMEM;
        }
    }
    
    bzero((caddr_t) ot, (INT32) (sizeof(struct otnode)
                                 - sizeof(union icbe)
                                 + fs->SectorSize));
    ot->vnode.v_data = (caddr_t) ot;
    ot->vnode.v_vfsp = fs->vfs;
    ot->vnode.v_rdev = fs->VolSection[0].devvp->v_rdev;
    ot->vnode.v_count = 1;

/* ISSUE - Should maybe make a subroutine that fills in a regid.
 * Search to make sure we are setting all three fields of the regid
 * elsewhere in the code.
 * ISSUE - Should we remove this altogether? NOTE that it contains
 * references to the fe field, so it should at least be put inside an
 * if (uTagID == TAGID_FILE_ENTRY) statement. - DVM
 */
    strcpy((char *)ot->icbe.fe.sImpID.aRegisteredID, SFS_REGID);
    strcpy((char *)ot->icbe.fe.sImpID.aIDSuffix, SFS_FE_REGID);
    ot->icbe.fe.sImpID.uFlags = PROTECTREGID;

    DMLPRINTF(oVnuLev)("getOTNode: Making new otnode active.\n");
    ot_enqueueOTNode( &fs->active,ot);
    DMLPRINTF(oVnuLev)("getOTNode: Made new otnode active.\n");
    
    if (icbaddr == (UINT32) -1) {
        DMLPRINTF(oVnuLev)("getOTNode: Making otnode from scratch. Getting space.\n");
        addr_list = ot_getSpace( fs, (UINT32) (fs->ICBsize),
                                contiguous);
        if (addr_list) {
            DMLPRINTF(oVnuLev)("getOTNode: Successfully got space.\n");
            icbaddr = addr_list->start;
            DMLPRINTF(oVnuLev)("getOTNode: Erasing space.\n");
            error = ot_erase( fs, 1, addr_list->start, addr_list->length);
            kmem_free((caddr_t) addr_list, (UINT32) sizeof(struct space), 0);
            if (!error) {
                DMLPRINTF(oVnuLev)("getOTNode: Filling otnode.\n");
                ot_fillOtNode( fs, ot, name, icbaddr, vtype, parentLBN,
                              version);
            }
        } else {
            error = ENOSPC;
            ot->vnode.v_count--;
            otvn_inactive( &ot->vnode, (struct ucred *) 0);
        }
    } else {
        DMLPRINTF(oVnuLev)("getOTNode: Getting otnode from disk.\n");
        error = 0;
        if (!(ot->flags & OT_INITIALIZED)) {
            DMLPRINTF(oVnuLev)("getOTNode: Filling in otnode.\n");
            ot_fillOtNode( fs, ot, name, icbaddr, vtype, parentLBN, version);
            if (getname) {
                DMLPRINTF(oVnuLev)("getOTNode: Calling getName.\n");
                error = ot_getName( fs, ot, LAST);
            } else {
                DMLPRINTF(oVnuLev)("getOTNode: Reading otnode.\n");
                error = ot_ICBRdWr( fs, ot, UIO_READ, version);
            }
        }
    }
    if (!error && ot_ICBHasADs(ot)){
        error = ot_readADL(ot);
    }
    if (error) {
        if (ot) {
            ot->vnode.v_count--;
            otvn_inactive(&ot->vnode,(struct ucred *) 0);
        }
    } else
        *theNode = ot;
    DMLPRINTF(oVnuLev)("getOTNode: bottom\n");
    return error;
} /* end ot_getOTNode */

/********************************************************************/
/*  ot_readData                                                     */
/*      routine to read file data                                   */
/*                                                                  */
/*  input parameters:                                               */
/*      fsp:    the otfs structure for the volume                   */
/*      otp:    the otnode for the file                             */
/*      uiop:   uiop->uio_offset is the byte offset into the file   */
/*                  where the read should begin                     */
/*              uiop->uio_resid indicates how many bytes should be  */
/*                  read                                            */
/*              uiop->uio_iov->iov_base: where to put data read     */
/*                                                                  */
/*  output parameters:                                              */
/*      uiop:   uiop->uio_offset and uiop->uio_iov->iov_base get    */
/*                  incremented by the number of bytes read         */
/*              uiop->resid gets decremented by the number of       */
/*                  bytes read                                      */
/*  returns:                                                        */
/*      0       success                                             */
/*      EEOF    end of file encountered while locating data to read */
/*      EINVAL  invalid request made                                */
/********************************************************************/
INT32 ot_readData(register struct otfs    *fsp,
                  struct otnode           *otp,
                  register struct uio     *uiop)
{
    int             readDirect;
    INT32           error;
    UINT32          begExt;
    UINT32          offset;
    UINT32          bsize;
    UINT32          amtRead;
    UINT32          skip;
    UINT32          blockSize;
    INT32           diff;
    register UINT32 startBlk;
    register UINT32 endBlk;
    register UINT32 numBytes;
    struct gen_AD   *ADp;
    UINT32          ADIndex;
    
    error = 0;
    blockSize = IOBUFSIZE / fsp->SectorSize;
    
    /* check for valid request */
    if (uiop->uio_offset < 0 || uiop->uio_offset + uiop->uio_resid < 0)
        return EINVAL;
    
    /* check if done or no data */
    if (uiop->uio_resid == 0 || otp->icbe.fe.InfoLengthL == 0L)
        return 0;
    
    do {
        diff = otp->icbe.fe.InfoLengthL - uiop->uio_offset;
        if (diff <= 0)
            return 0;
        offset = uiop->uio_offset % fsp->SectorSize;
        numBytes = uiop->uio_resid < diff ? uiop->uio_resid : diff;
        
        /* are the conditions ALMOST right for readDirect? if so, make them right */
        diff = numBytes % fsp->SectorSize;
        if (offset == 0 && diff != 0 && numBytes > IOBUFSIZE)
            numBytes -= diff;
        
        readDirect = offset == 0 && numBytes % fsp->SectorSize == 0;
        startBlk = uiop->uio_offset / fsp->SectorSize;
        endBlk = (uiop->uio_offset + numBytes - 1) / fsp->SectorSize;
        bsize = (endBlk - startBlk + 1) * fsp->SectorSize;
        
        if (!readDirect && bsize > IOBUFSIZE) {
            endBlk = startBlk + blockSize - 1;
            bsize = (endBlk - startBlk + 1) * fsp->SectorSize;
            numBytes = bsize - offset;
        }
        
        if (ot_searchADL(otp,startBlk,&begExt,&ADIndex)) {
            ADp = otp->ADL.beginADs + ADIndex;
        } else {
            return EEOF;
        }
        
        if (endBlk > begExt + ADp->InfoLength/fsp->SectorSize - 1) {
            endBlk = begExt + ADp->InfoLength/fsp->SectorSize - 1;
            bsize = (endBlk - startBlk + 1) * fsp->SectorSize;
            numBytes = bsize - offset;
        }
        
        skip = startBlk - begExt;
        
        if (readDirect) {
            /*
             * NOTE: Reads typically go through fsp->iobuf and data gets
             * moved via uiomove(). For speed, data will be read directly
             * into user buffer, assuming of course, that the extent is
             * written. If the extent is not written, just zero the
             * buffer.
             */
            
            if (ADp->ExtentLength.bf.Type==E_RECORDED) {
                amtRead = bsize;
                error = ot_read(fsp,
                                ADp->Location_LBN+skip,
                                uiop->uio_iov->iov_base,
                                &amtRead);
            } else {    /* either unallocated or allocated */
                bzero(uiop->uio_iov->iov_base,bsize);
                amtRead = bsize;
            }
            if (error)
                return error;
            
            /* This "fakes" the uiomove call */
            uiop->uio_offset += amtRead;
            uiop->uio_resid -= amtRead;
            uiop->uio_iov->iov_base += amtRead;
        } else {
            /* not a direct read, but still check if written */
            if (ADp->ExtentLength.bf.Type==E_RECORDED) {
                amtRead = bsize;
                error = ot_read(fsp,
                                ADp->Location_LBN+skip,
                                fsp->iobuf,
                                &amtRead);
            } else {            /* allocated or unallocated */
                bzero(fsp->iobuf,bsize);
            }
            if (error)
                return error;
            uiomove(fsp->iobuf+offset,numBytes,UIO_READ,uiop);
        }
    } while (uiop->uio_resid > 0 && numBytes != 0);
    return error;
} /* end ot_readData */

/********************************************************************/
/*  ot_readVerify                                                   */
/*      routine to read and verify data                             */
/*                                                                  */
/*  input parameters:                                               */
/*      vp:     the vnode for the file                              */
/*      uiop:   uiop->uio_offset is the byte offset into the file   */
/*                  where the verify should begin                   */
/*              uiop->uio_resid is the number of bytes to verify    */
/*              uiop->uio_iov->iov_base is the data to compare it to*/  
/*                                                                  */
/*  output parameters:                                              */
/*      uiop:   uiop->uio_offset has been incremented by the number */
/*                  of bytes read and compared                      */
/*              uiop->uio_resid has been decremented by the number  */
/*                  of bytes read and compared                      */
/*              uiop->uio_iov->iov_base has been incremented by     */
/*                  (approximately) the number of bytes verified    */
/*  returns:                                                        */
/*      0       success                                             */
/*      EIO:    verify failed or ran out of data                    */
/********************************************************************/
INT32 ot_readVerify(struct vnode        *vp,
                    register struct uio *uiop)
{
    struct uio           uio;
    struct iovec         iovec;
    register INT32       error;
    register INT32       sectorOffset; /* offset into sector, for alignment */
    register INT32       ioLength;        /* # bytes in each read */
    register struct otfs *fsp = (struct otfs *) vp->v_vfsp->vfs_data;
    
    error = 0;
    do {
        /* until error, or end of verify request, or incomplete read (EOF?) */
        /* read as many bytes as remain in the request, truncated to buffer */
        /* size */

        ioLength = (uiop->uio_resid >= IOBUFSIZE ?
                    IOBUFSIZE : uiop->uio_resid);

        /* but adjust to align subsequent reads on sector boundaries */

        sectorOffset = uiop->uio_offset % fsp->SectorSize;

        if (sectorOffset != 0) {
            sectorOffset = IOBUFSIZE - sectorOffset;
            if (ioLength > sectorOffset)
                ioLength = sectorOffset;    
        }
        
        uioInit( &uio, &iovec, fsp->iobuf, uiop->uio_offset, ioLength);
        error = otvn_rdwr(vp, &uio, UIO_READ, IO_FLAG,
                          (struct ucred *)0);
        ioLength -= uio.uio_resid;
        if (!error && ot_memCompare(fsp->iobuf,
                                    uiop->uio_iov->iov_base,
                                    ioLength) != 0)
            error = EIO;
        uiop->uio_offset += ioLength;
        uiop->uio_resid -= ioLength;
        uiop->uio_iov->iov_base += ioLength;

    } while (!error && ioLength > 0 && uio.uio_resid == 0);
    DMLPRINTF(oVndLev)("ot_readVerify: ioLen = 0x%x, error = 0x%x\n",
                       ioLength,error);
    return error == 0 && ioLength > 0 ? EIO : error;
} /* end ot_readVerify */

/********************************************************************/
/*  ot_release                                                      */
/*      routine to free memory used by an inactive otnode           */
/*                                                                  */
/*  input parameters:                                               */
/*      otp:    the otnode to be released                           */
/*                                                                  */
/*  returns:                                                        */
/*      0       success                                             */
/*                                                                  */
/********************************************************************/
void ot_release(register struct otnode *otp)
{
    register INT32          numInactive;
    register struct otnode  *inactive,
    *previous;
    
    struct otfs *fsp = (struct otfs *) otp->vnode.v_vfsp->vfs_data;
    
    /* if active or modified, can't release it */
    if ((otp->vnode.v_count != 0) || 
        ((otp->flags & OT_MOD) && !(otp->flags & OT_INVAL)) || 
        (otp->ADL.modified))
      return;
    
    DMLPRINTF(oVnuLev)("release: Before dequeueOTNode(active).\n");
    /* take it off the active list */
    ot_dequeueOTNode(&fsp->active,otp);
    DMLPRINTF(oVnuLev)("release: After dequeueOTNode(active).\n");
    
    if (INACTIVE_CACHE_SIZE == 0 || otp->flags & OT_INVAL) {
        if (otp->ADL.bufLen != 0) {
            kmem_free(otp->ADL.beginADs,otp->ADL.bufLen,CPURAM);
        }
        kmem_free((caddr_t) otp, (UINT32) sizeof(struct otnode) 
                  - sizeof( union icbe) + fsp->SectorSize, 0);
        return;
    }
    
    /* count number already in inactive otnode cache */
    for (previous = inactive = fsp->inactive, numInactive = 0;
         inactive;
         previous = inactive, inactive = inactive->next, numInactive++);
    
    /* free least recently used otnode to make room for this one... */
    if (numInactive + 1 > INACTIVE_CACHE_SIZE) {
        if (previous) {
            DMLPRINTF(oVnuLev)("release: Before dequeueOTNode(inactive).\n");
            ot_dequeueOTNode(&fsp->inactive,previous);
            DMLPRINTF(oVnuLev)("release: After  dequeueOTNode(inactive).\n");
            kmem_free((caddr_t) previous,
                      (UINT32) sizeof(struct otnode)
                      - sizeof( union icbe)
                      + fsp->SectorSize, 0); 
        }
    }
    DMLPRINTF(oVnuLev)("release: Before enqueueOTNode(inactive).\n");
    ot_enqueueOTNode(&fsp->inactive,otp);   /* add new one to cache */
    DMLPRINTF(oVnuLev)("release: After  enqueueOTNode(inactive).\n");
    
    return;
} /* end ot_release */

/*
 *  ot_setupDirentryOTNode
 *      does the redundant part of ot_direntry();making a new otnode.
 *
 *  input parameters:
 *      fsp:            the otfs structure for the volume
 *      otpp:           for the new one
 *      vap:            the attributes for the new one
 *      vtype:          VREG, VDIR or VASSOC
 *      name:           this one's name
 *      parentAddress:  address of this one's parent
 *
 *  output parameters:
 *      otpp:           contains the filled out otnode
 *
 *  returns:
 *      0               success
 *      ENOMEM:         memory allocation failure
 *      ENOSPC:         couldn't get space on disk for the icb entry
 */
INT32 ot_setupDirentryOTNode(struct otfs   *fsp,
                             struct otnode **otpp,
                             struct vattr  *vap,
                             enum vtype    vtype,
                             caddr_t       name,
                             UINT32        parentAddress)
{
    INT32         error;
    struct otnode *new;
    struct otnode *parent = NULL;
    
    error = 0;
    
    error = ot_getOTNode(0,fsp,name,(UINT32) -1,LAST,vtype,
                         parent,&new);
    if (new) {
        new->flags |= OT_MOD;
        new->parent_addr_LBN = parentAddress;
        new->parent_addr_PartNo = 0;
        otvn_setattr(&new->vnode,vap,(struct ucred *) 0);
    }
    
    *otpp = error == 0 ? new : 0;
    return error;
} /* end ot_setupDirentryOTNode */

/*
 *  ot_setFileSize
 *      truncate or extend a file
 *
 *  input parameters:
 *      fsp:            the otfs structure for the volume
 *      otp:            otnode for the file
 *      NewFileSize:    the desired file size
 *      allocationType: contiguous or nonContiguous
 *
 *  returns:
 *      0               success
 *      ENOSPC:         no space left on device
 *
 *  notes:
 *      if an extension is desired, "allocationType" indicates
 *          whether or not the new space obtained must be contiguous
 */
INT32 ot_setFileSize(struct otfs             *fsp,
                     register struct otnode  *otp,
                     register UINT32         NewFileSize,
                     enum spRequestType      allocationType)
{
    INT32         error;
    UINT32        numSec;
    UINT32        ADIndex;
    UINT32        begLastExt;
    UINT32        endLastExt;
    UINT32        lastBlk;
    UINT32        logSector;
    UINT32        WillFitInLastSct;
    UINT32        retLen;
    UINT32        retLoc;
    struct space  *addrList;
    struct gen_AD *ADp;

    const UINT32  MaxFileSectors = 0x7fffffff;
    
    DMLPRINTF(oVndLev)("ot_setFileSize: top\n");
    
    if (NewFileSize == 0L) {                            /* effectively delete */
        DMLPRINTF( oVndLev)("ot_setFileSize: size == 0\n");
        error = ot_deleteData(fsp,otp);
        return error;
    }
    
    if (NewFileSize == otp->icbe.fe.InfoLengthL)     /* nothing to do */
        return 0;
    
    if (NewFileSize > otp->icbe.fe.InfoLengthL) {    /* Extension */

        DMLPRINTF(oVndLev)("ot_setFileSize: Extension\n");

        /* count # sectors used in ADL */
        /* Use searchADL to count logSectors by searching for an */
        /* arbitrarily high sector address (to get to the end). */
        ot_searchADL(otp,MaxFileSectors,&logSector,&ADIndex);

        /* calc # bytes which will fit in last sector used */
        if (otp->icbe.fe.InfoLengthL >= fsp->SectorSize)
            WillFitInLastSct = (logSector*fsp->SectorSize) -
                otp->icbe.fe.InfoLengthL;
        else if (otp->icbe.fe.InfoLengthL == 0L)
            WillFitInLastSct = 0;
        else
            WillFitInLastSct = (fsp->SectorSize
                                - otp->icbe.fe.InfoLengthL);
        
        /* calc # sectors still needed */
        if (NewFileSize-otp->icbe.fe.InfoLengthL <= WillFitInLastSct) {
            otp->icbe.fe.InfoLengthL = NewFileSize;  /* fits */
            otp->flags |= OT_MOD;
        } else {
            numSec = ((NewFileSize
                       - otp->icbe.fe.InfoLengthL
                       - WillFitInLastSct-1) / fsp->SectorSize) + 1L;
            
            addrList = ot_getSpace( fsp, numSec, allocationType);
            if (!addrList)
                error = ENOSPC;
            
            if (!error) {   /* create AD */
                ot_addToADL(otp,logSector,E_ALLOCATED,addrList);
                killSpaceList( &addrList);
                otp->icbe.fe.InfoLengthL = NewFileSize;
                otp->flags |= OT_MOD;
            }
        }
    } else {    /* Truncate - error recovery!? */

        DMLPRINTF(oVndLev)("ot_setFileSize: truncate\n");
        lastBlk = (NewFileSize-1) / fsp->SectorSize;  /* new last logical sector */
        
        /* check ADL */
        ot_searchADL(otp,lastBlk,&begLastExt,&ADIndex);
        ADp = otp->ADL.beginADs + ADIndex;
        endLastExt = begLastExt + ADp->InfoLength/fsp->SectorSize - 1;

        /* endLastExt is last sector in last extent which will remain */
        /* modify last AD which remains valid and recover (MO) */
        /* sectors */

        if (lastBlk < endLastExt) {
            retLoc = ADp->Location_LBN + lastBlk - begLastExt + 1;
            retLen = endLastExt-lastBlk;
            if (ADp->ExtentLength.bf.Type == E_RECORDED) {
                returnSectors( fsp, (struct space *) 0,
                              retLoc,retLen,bmapUnerased);
            } else if (ADp->ExtentLength.bf.Type == E_ALLOCATED) {
                returnSectors( fsp, (struct space *) 0,
                              retLoc,retLen,bmapErased);
            }
            ADp->InfoLength =
                ADp->RecordedLength =
                    ADp->ExtentLength.bf.Length -= retLen;
        }
        
        DMLPRINTF(oVndLev)("ot_setFileSize: returning unused extents\n");
        ot_truncateADL(otp,ADIndex+1);

        otp->icbe.fe.InfoLengthL = NewFileSize;
        otp->flags |= OT_MOD;
        return 0;
    } /* truncate */
    DMLPRINTF(oVndLev)("ot_setFileSize: returning\n");
    return error;
} /* end ot_setFileSize */

/********************************************************************/
/*  ot_writeData                                                    */
/*      routine to write file data                                  */
/*                                                                  */
/*  input parameters:                                               */
/*      fsp:    the otfs structure for the volume                   */
/*      otp:    the otnode for the file                             */
/*      uiop:   uiop->uio_offset is the byte offset into the file   */
/*                  where the write should begin                    */
/*              uiop->uio_resid indicates how many bytes should be  */
/*                  written                                         */
/*              uiop->uio_iov->iov_base is the data to be written   */
/*                                                                  */
/*  output parameters:                                              */
/*      uiop:   uiop->uio_offset and uiop->uio_iov->iov_base get    */
/*                  incremented by the number of bytes written      */
/*              uiop->resid gets decremented by the number of       */
/*                  bytes written                                   */
/*  returns:                                                        */
/*      0       success                                             */
/*      EINVAL  invalid request                                     */
/*      ENOSPC  no space left on volume                             */
/********************************************************************/
INT32 ot_writeData(register struct otfs    *fsp,
                   struct otnode           *otp,
                   register struct uio     *uiop)
{
    int            writeDirect;
    int            anUpdate;
    struct gen_AD  *ADp;
    UINT32         ADIndex;
    UINT32         begExt;
    UINT32         offset;      /* offset into a sector */
    UINT32         phySec;
    UINT32         amtRead;
    INT32          diff;
    INT32          error;
    caddr_t        writePtr;
    struct space   *spcList;
    struct space   *spcPtr;
    UINT32         startBlk;
    UINT32         endBlk;
    UINT32         bsize;
    UINT32         numBytes;

    /* check for valid request */
    if (uiop->uio_offset < 0 || uiop->uio_offset + uiop->uio_resid < 0)
        return EINVAL;
    
    /* check if done */
    if (uiop->uio_resid == 0)
        return 0;
    
    error = 0;
    ADp = otp->ADL.beginADs;
    do {
        offset = (UINT32) uiop->uio_offset % fsp->SectorSize;
        numBytes = (UINT32) uiop->uio_resid;
        
        /*
         * Are the conditions ALMOST right for writeDirect? if so, make
         * them right:
         *
         *   If it starts on a sector boundary (offset ==0) AND
         *   if multiple iterations are required anyway
         *   (numBytes > IOBUFSIZE), and the request does not end on a
         *   sector boundary (diff != 0), then round down the
         *   requested length to the nearest sector boundary
         *   (numBytes -= diff)
         */

        diff = numBytes % fsp->SectorSize;
        if (offset == 0 && diff != 0 && numBytes > IOBUFSIZE)
            numBytes -= diff;

        /*
         * writeDirect : if start and stop on a sector boundary
         * startBlk    : block address of beginning
         * endBlk      : block address of end
         * bsize       : number of bytes to be written, rounded up to
         *               the nearest sector. 
         */

        writeDirect = offset == 0 && numBytes % fsp->SectorSize == 0;
        startBlk = (UINT32) uiop->uio_offset / fsp->SectorSize;
        endBlk = (UINT32) (uiop->uio_offset + numBytes - 1) / fsp->SectorSize;
        bsize = (endBlk - startBlk + 1) * fsp->SectorSize;

        /*
         * If entire request will not fit inside the buffer,
         *   set endBlk, bsize, and numBytes appropriately.
         */

        if (!writeDirect && bsize > IOBUFSIZE) {
            endBlk = startBlk + IOBUFSIZE / fsp->SectorSize - 1;
            bsize = (endBlk - startBlk + 1) * fsp->SectorSize;
            numBytes = bsize - offset;
        }

        /*
         * Check ADL for extent at beginning of write:
         *   begExt   : Updated during search. Equals sector offset into
         *              file matching beginning of extent.
         *   ADIndex  : Equals extent's index ADL.
         *   ADp      : Pointer to found AD.
         *   anUpdate : TRUE if found AD.
         *      
         */

        anUpdate = ot_searchADL(otp,startBlk,&begExt,&ADIndex);
        ADp      = otp->ADL.beginADs + ADIndex;

        /*
         * If request is an update (i.e., not an append), and request
         * spans into another extent (endBlk is not in same extent), then
         * set endBlk, bsize, and numBytes to modify only that extent
         * (all of it).
         */

        if (anUpdate && endBlk > begExt+ADp->InfoLength/fsp->SectorSize-1) {
            endBlk = begExt + ADp->InfoLength/fsp->SectorSize - 1;
            bsize = (endBlk-startBlk+1) * fsp->SectorSize;
            numBytes = bsize - offset;
        }
        /*
         * If writeDirect, change state of uio structure to reflect data
         * being taken out of it.
         * If NOT, zero out the IOBUF, read the extent, and overlay the
         * data as required.
         */

        if (writeDirect) {
            writePtr = uiop->uio_iov->iov_base;
            uiop->uio_offset += bsize;
            uiop->uio_resid -= bsize;
            uiop->uio_iov->iov_base += bsize;
        } else {
            /* read extent to update only if written */
            bzero(fsp->iobuf,IOBUFSIZE);
            if (anUpdate && ADp->ExtentLength.bf.Type==E_RECORDED) {
                phySec = startBlk - begExt + ADp->Location_LBN;
                amtRead = bsize;
                if (error = ot_read(fsp,phySec,fsp->iobuf,&amtRead))
                    return error;
            } 
            
            /* Update data */
            uiomove(fsp->iobuf + offset,numBytes,UIO_WRITE,uiop);
            writePtr = fsp->iobuf;
        }
        
        if (anUpdate && (fsp->MediaType > (int) fsWORM ||
                         !ADp->ExtentLength.bf.Type==E_RECORDED)) {
            /*
             * Find physical address (phySec) of start of write, erase
             * the region (if MO), write the data, and then mark the
             * extent as written.
             */

            phySec = ADp->Location_LBN + startBlk - begExt;
            /* ISSUE: Shouldn't this NOT erase if Type==E_ALLOCATED? */
            if (fsp->MediaType == (int) fsMO)   /* erase only if MO */
                error = ot_erase(fsp,1,(UINT32) phySec,
                                 (UINT32) bsize/fsp->SectorSize); 
            if (!error)
                error = ot_write(fsp,phySec,writePtr,bsize);
            
            /*
             * ISSUE: needs to consider an update which uses only a portion
             * of an AD -- It is marking the entire extent as
             * recorded. This is not correct. - DVM
             */

            ADp->ExtentLength.bf.Type = E_RECORDED;
        } else {
            /* get space */
            spcList = ot_getSpace(fsp,(UINT32) endBlk-startBlk+1,
                                  nonContiguous);
            if (!spcList)
                return ENOSPC;
            
            /* write data */
            for (spcPtr = spcList;
                 spcPtr && !error;
                 spcPtr = spcPtr->next) {
                error = ot_write(fsp,spcPtr->start,writePtr,
                                 spcPtr->length * fsp->SectorSize);
                writePtr += spcPtr->length * fsp->SectorSize;
            }
            if (!error)
                error = ot_addToADL(otp,startBlk,E_RECORDED,spcList);
            if (!error)
                otp->flags |= OT_MOD;
            killSpaceList(&spcList);
        } /* end not anUpdate */

        /*
         * Change the file size and set the MOD flag in the otnode if
         * we've modified it.
         */

        if (uiop->uio_offset > otp->icbe.fe.InfoLengthL && !error) {
            otp->icbe.fe.InfoLengthL = uiop->uio_offset;
            otp->flags |= OT_MOD;
        }
    } while (uiop->uio_resid > 0 && numBytes != 0 && !error);
    
    return error;
} /* end ot_writeData */


/* ISSUE - Clean up the return values here after done fixing up icb entry */
/* changes */

/********************************************************************/
/*  ot_fillOtNode                                                   */
/*      fill out a brand new otnode                                 */
/*                                                                  */
/*  input parameters:                                               */
/*      ot                  the otnode to fill out                  */
/*      name                name of file controlled by otnode       */
/*      icbaddr             otnode's sector address                 */
/*      vtype               vnode's vtype                           */
/*      paddr               otnode's parent's sector address        */
/*      version             otnode's version                        */
/*                                                                  */
/*  modification history                                            */
/*      05.Dec.92    Start passing otfs structure                   */
/*                                                                  */
/********************************************************************/
void ot_fillOtNode(struct otfs             *fs,
                   register struct otnode  *ot,
                   caddr_t                 name,
                   UINT32                  icbaddr,
                   enum vtype              vtype,
                   UINT32                  parentAddr,
                   UINT32                  version)
{
    ot->vnode.v_type = vtype;
    ot->vnode.v_op = vtype == VDIR ? &otfs_dvnodeops : &otfs_fvnodeops;
    ot->parent_addr_LBN = parentAddr;
    ot->parent_addr_PartNo = 0;
    ot->icb_addr_LBN = icbaddr;
    ot->icb_addr_PartNo = 0;
    ot->attr = vtype == VDIR ? DIR_ATTR : FILE_ATTR;
    ot->version = version;
    ot->isLast = version == LAST;
    if (name) strcpy(ot->name,name);
    FillTag(TAGID_FILE_ENTRY,1,1,NOCRC,0,0,&ot->icbe.ge.sTag);
    ot->ADL.modified  = 0;
    ot->ADL.beginADs  = ot->ADL.endADs = NULL;
    ot->ADL.diskSpace = NULL;
    ot->ADL.bufLen    = 0;
    ot->icbe.fe.InfoLengthH = 0L;
    ot->icbe.fe.InfoLengthL = 0L;
    ot->icbe.fe.LogBlocksH = 0L;
    ot->icbe.fe.LogBlocksL = 0L;
    ot->icbe.fe.L_EA = 0;
    ot->icbe.fe.L_AD = 0;
    ot->icbe.ge.sICBTag.NumberEntries = 0;
    ot->icbe.ge.sICBTag.StrategyType
        = (fs->MediaType < fsWORM ? 4 : 1);
    *((UINT16 *)ot->icbe.fe.sICBTag.StrategyParm)
        = (fs->MediaType < fsWORM ? fs->ICBsize - fs->IndirectSize : 1);

    /* note that vtype must be supplied by the calling routine */
    switch (vtype) {
    VDIR:  ot->icbe.ge.sICBTag.FileType = FILE_TYPE_DIRECTORY ; break;
    VREG:  ot->icbe.ge.sICBTag.FileType = FILE_TYPE_RAW       ; break;
    VBLK:  ot->icbe.ge.sICBTag.FileType = FILE_TYPE_BDEV      ; break;
    VCHR:  ot->icbe.ge.sICBTag.FileType = FILE_TYPE_CDEV      ; break;
    VFIFO: ot->icbe.ge.sICBTag.FileType = FILE_TYPE_FIFO      ; break;
    VSOCK: ot->icbe.ge.sICBTag.FileType = FILE_TYPE_SOCKET    ; break;
    VLNK:  ot->icbe.ge.sICBTag.FileType = FILE_TYPE_SYMLINK   ; break;
    default: ot->icbe.ge.sICBTag.FileType = FILE_TYPE_UNSPECIFIED;
    }

    /* Not needed. Filled in later... */
    ot->icbe.ge.sICBTag.Flags = (vtype == VDIR ? ADLONG : fs->ADType);
    /* set other ICBTag Flags values here by OR'ing them in */
    ot->icbe.ge.sICBTag.sParentICB_LBN = icbaddr;
    ot->icbe.ge.sICBTag.sParentICB_PartNo = 0;
    /* ISSUE: This should go away! (it assumes file entry...!) */
    ot->flags |= OT_INITIALIZED;
} /* end ot_fillOtNode */

