#ifndef DEF_H
#define DEF_H

/********************************************************************/
/*  def.h - System definition header for NSR                        */
/*                                                                  */
/*      Unpublished Confidential Information of Hewlett-Packard     */
/*      Company. Do not disclose.  Copyright (c) Hewlett-Packard    */
/*      Company 1996. All rights reserved                           */
/*                                                                  */
/* The contents of this software are proprietary and confidential   */
/* the Hewlett- Packard Company, and are limited in distribution to */
/* those with a direct need to know.  Individuals having access to  */
/* this software are responsible for maintaining the                */
/* confidentiality of the content and for keeping the software      */
/* secure when not in use.  Transfer to any party is strictly       */
/* forbidden other than as expressly permitted in writing by        */
/* Hewlett-Packard Company.  Unauthorized transfer to or possession */
/* by any unauthorized party may be a criminal offense.             */
/*                                                                  */
/*                   RESTRICTED RIGHTS LEGEND                       */
/*                                                                  */
/*      Use,  duplication,  or disclosure by the                    */
/*      U. S. Government is subject to restrictions as set forth    */
/*      in subparagraph (c) (1) (ii) in the Rights in Technical     */
/*      Data and Computer Software clause in DFARS 252.227-7013     */
/*      or any other successor clause.                              */
/*                                                                  */
/*                       HEWLETT-PACKARD COMPANY                    */
/*                       3000 Hanover St.                           */
/*                       Palo Alto, CA  94304 U.S.A.                */
/*                                                                  */
/*      Rights for non-DOD U.S. Government Departments and Agencies */
/*      are as set forth in FAR 52.227-19 (c) (1,2) or any other    */
/*      successor clause.                                           */
/*                                                                  */
/********************************************************************/
/*
 $Source: /projects/HMS/nsr/udfOld/src/nsrHdrs/def.h,v $
 $Revision: 1.11 $      $Author: hmgr $
 $State: Exp $          $Locker:  $
 $Date: 96/05/30 14:53:07 $
 ********************************************************************/

#define HP              1
#define SUN             0                   /* machine type         */
#define IBM             0   
#define MACINTOSH       0

#define CHECK_CRED      0
#define SUPP_LINK       0

/* Used by kmem_alloc and kmem_free as the ctl argument. */

                                /* '0' is default        */
#define BUFRAM          0       /* alloc/free buffer ram */
#define CPURAM          1       /* alloc/free cpu ram    */

#define TRUE            1
#define FALSE           0


                        /* machine orginator ids         */
#define MAC_ORIG_ID     (UINT32) 0x204d6163    /* ' Mac' */
#define SUN_ORIG_ID     (UINT32) 0x00000053
#define IBM_ORIG_ID     (UINT32) 0x00000049
#define HP_ORIG_ID      (UINT32) 0x5f48505f   /* '_HP_' */

#if (HP)

/* Globals used for debugging inside kernel */
extern int (*debPrintf)(const char *, ...);
extern int debMinLev;
extern int debMaxLev;

/* Main debug macro */

/* Debug level definitions */
#define oAllLev debMaxLev
#define oNevLev 101
#define oStrLev 61
#define oSysLev 55
#define oVfsLev 51
#define oADLev  45
#define oXfrLev 39
#define oCmpLev 31
#define oCrtLev 30
#define oVnuLev 29
#define oVndLev 21
#define oDirLev 11
#define oF2vLev 7
#define oMntLev 3
#define oFoxLev 2
#define oPrmLev 1

#define UIOSEG_KERNEL   1   /* This value just here as a place holder */
#endif /* NITRO || lynx */

#if (SUN || HP)
#define ORIG_ID         SUN_ORIG_ID 
#endif /* SUN || HP || lynx || NITRO */

#if MACINTOSH
#define ORIG_ID         MAC_ORIG_ID 
#endif

#if IBM
#define ORIG_ID         IBM_ORIG_ID 
#endif

/************************ generic NSR defines ***********************/

#define IO_FLAG         (INT32) 0           /* I/O flag for r/w     */
#define FROM_OP         (INT32) 0x00        /* xlate from optical   */
#define TO_OP           (INT32) 0x01        /* xlate to optical     */

/********************** UNIX configuration **************************/
#if (SUN || HP)

#define INACTIVE_CACHE_SIZE (INT32) 0x02    /* number of otnodes allowed to cache */

#define IOBUFSIZE       (INT32) 0x2000      /* fs buffer sizes      */
#define DIRBUFSIZE      (INT32) 0x400

typedef char            CHAR8;              /* portability          */
typedef unsigned char   UCHAR8;
typedef short           INT16;
typedef unsigned short  UINT16;
typedef long            INT32;              /* int and long are same*/
typedef unsigned long   UINT32;

#if (HP)
#define _INCLUDE_HPUX_SOURCE
#define _INCLUDE_XOPEN_SOURCE
#define _INCLUDE_POSIX_SOURCE
#define _INCLUDE_AES_SOURCE
#define _HPUX_SOURCE
typedef struct {        /* used for device file descriptor on HP    */
    INT32   devfd;
} otfsVendorField;
#endif /* HP */



#if SUN
typedef struct {        /* Sun has to have something...             */
    CHAR8   empty[4];
} otfsVendorField;
#endif /* SUN */

#endif  /* (SUN || HP || lynx || NITRO) */

/********************* Macintosh configuration *********************/
#if MACINTOSH

#include <HFS.h>

typedef char            CHAR8;
typedef unsigned char   UCHAR8;
typedef int             INT16;
typedef unsigned int    UINT16;
typedef long            INT32;
typedef unsigned long   UINT32;

#define INACTIVE_CACHE_SIZE (INT32) 0x00    /* number of otnodes allowed to cache */

#define KERNEL
#define DIRBUFFER
#define IOBUFSIZE       (INT32) 0x400
#define DIRBUFSIZE      (INT32) 0x400

typedef struct {
    HVCB            *hvcb;
#ifdef DIRBUFFER
    unsigned int    dbufActive; /* bit 0 means initialized, bit 1 means active */
    unsigned long   dbufSector;
    unsigned char   dbuf[1024]; /* maximum sector size */
#endif /* DIRBUFFER */
} otfsVendorField;  
        
#endif /* MACINTOSH */

/************************ IBM configuration ************************/
#if IBM

typedef char            CHAR8;
typedef unsigned char   UCHAR8;
typedef int             INT16;
typedef unsigned int    UINT16;
typedef long            INT32;
typedef unsigned long   UINT32;

#define INACTIVE_CACHE_SIZE (INT32) 0x00    /* number of otnodes allowed to cache */

#define KERNEL
#define DIRBUFFER
#define IOBUFSIZE       (INT32) 0x2000
#define DIRBUFSIZE      (INT32) 0x400

typedef struct {        /* PC has to have something... */
    CHAR8   empty[2];
} otfsVendorField;

#endif /* IBM */

#endif /* DEF_H */
