#ifndef NSR_H
#define NSR_H
/********************************************************************/
/*  nsr.h - NSR structures, tags and definitions                    */
/*                                                                  */
/*      Unpublished Confidential Information of Hewlett-Packard     */
/*      Company. Do not disclose.  Copyright (c) Hewlett-Packard    */
/*      Company 1996. All rights reserved                           */
/*                                                                  */
/* The contents of this software are proprietary and confidential   */
/* the Hewlett- Packard Company, and are limited in distribution to */
/* those with a direct need to know.  Individuals having access to  */
/* this software are responsible for maintaining the                */
/* confidentiality of the content and for keeping the software      */
/* secure when not in use.  Transfer to any party is strictly       */
/* forbidden other than as expressly permitted in writing by        */
/* Hewlett-Packard Company.  Unauthorized transfer to or possession */
/* by any unauthorized party may be a criminal offense.             */
/*                                                                  */
/*                   RESTRICTED RIGHTS LEGEND                       */
/*                                                                  */
/*      Use,  duplication,  or disclosure by the                    */
/*      U. S. Government is subject to restrictions as set forth    */
/*      in subparagraph (c) (1) (ii) in the Rights in Technical     */
/*      Data and Computer Software clause in DFARS 252.227-7013     */
/*      or any other successor clause.                              */
/*                                                                  */
/*                       HEWLETT-PACKARD COMPANY                    */
/*                       3000 Hanover St.                           */
/*                       Palo Alto, CA  94304 U.S.A.                */
/*                                                                  */
/*      Rights for non-DOD U.S. Government Departments and Agencies */
/*      are as set forth in FAR 52.227-19 (c) (1,2) or any other    */
/*      successor clause.                                           */
/*                                                                  */
/********************************************************************/
/*
 $Source: /projects/HMS/nsr/udfOld/src/nsrHdrs/nsr.h,v $
 $Revision: 1.47 $      $Author: hmgr $
 $State: Exp $          $Locker:  $
 $Date: 96/05/30 14:53:19 $
 ********************************************************************/

#define byteClear   0x00

#define BITZERO     0x01
#define BITONE      0x02
#define BITTWO      0x04
#define BITTHREE    0x08
#define BITFOUR     0x10
#define BITFIVE     0x20
#define BITSIX      0x40
#define BITSEVEN    0x80
#define BITEIGHT    0x100
#define BITNINE     0x200
#define BITTEN      0x400
#define BITELEVEN   0x800
#define BITTWELVE   0x1000
#define BITTHIRTEEN 0x2000
#define BITFOURTEEN 0x4000
#define BITFIFTEEN  0x8000


#define bitsPerByte 0x08

#define FORWARD             (CHAR8) 0x00
#define BACKWARD            (CHAR8) 0x01

#define FILE_ATTR           (UCHAR8) BITZERO   /* attribute for a file     */
#define HIDDEN_ATTR         (UCHAR8) BITZERO   /* ... existence bit        */
#define DIR_ATTR            (UCHAR8) BITONE    /* ... for a directory      */
#define DELETE_ATTR         (UCHAR8) BITTWO    /* ... for a deleted file   */
#define PARENT_ATTR         (UCHAR8) BITTHREE  /* ... for a parent         */

#define LAST                (UINT32) 0      /* Request for last version */

/* tag id's for ISO/IEC 13346 structures */

#define TAGID_NONE             (UINT16) 0   /* no tag */
#define TAGID_PVD              (UINT16) 1   /* primary volume desc */
#define TAGID_ANCHOR           (UINT16) 2   /* anchor desc */
#define TAGID_POINTER          (UINT16) 3   /* pointer desc */
#define TAGID_IUD              (UINT16) 4   /* implementation use desc */
#define TAGID_PD               (UINT16) 5   /* volume partition desc */
#define TAGID_LVD              (UINT16) 6   /* logical volume desc */
#define TAGID_USD              (UINT16) 7   /* unallocated volume space desc */
#define TAGID_TERM_DESC        (UINT16) 8   /* terminator desc */
#define TAGID_LVID             (UINT16) 9   /* logical volume integrity desc */
#define TAGID_FSD              (UINT16) 256  /* file set desc */
#define TAGID_FILE_ID          (UINT16) 257  /* file identifier desc */
#define TAGID_ALLOC_EXTENT     (UINT16) 258  /* Allocation extent desc */
#define TAGID_INDIRECT         (UINT16) 259  /* Indirect entry */
#define TAGID_TERM_ENTRY       (UINT16) 260  /* Terminal entry */
#define TAGID_FILE_ENTRY       (UINT16) 261  /* File entry */
#define TAGID_EXT_ATTR         (UINT16) 262  /* Extended attribute desc */
#define TAGID_UNALLOC_SP_ENTRY (UINT16) 263  /* Unallocated space entry (WORM)*/
#define TAGID_SPACE_BMAP       (UINT16) 264  /* Space bitmap desc */
#define TAGID_PART_INTEGRITY   (UINT16) 265  /* Partition integrity desc */


/* controls for data structure functions */

/* FillTag function */

#define NOCRC 0
#define DOCRC 1

/* regids for ECMA 167 structures on Nitro */

#define SFS_REGID       "-HP CXXXXX NSR " /* MUST translate to ISO 646 */
#define SFS_PVD_REGID   "PVD XX.XX"       /* MUST translate to ISO 646 */
#define SFS_PD_REGID    "PD XX.XX"        /* MUST translate to ISO 646 */
#define SFS_LVD_REGID   "LVD XX.XX"       /* MUST translate to ISO 646 */
#define SFS_FE_REGID    "FE XX.XX"        /* MUST translate to ISO 646 */
#define NSRPARTCONTENTS "+NSR02"          /* MUST translate to ISO 646 */

#define FILENAMESIZE        0x20   /* size of a file name          */
#define SYS_USE_SIZE        8      /* ISSUE - Temporary for transition */


enum sumEntryType {     /* the ways the items in a directory can be summed */
    sumDirectories = 0, /* get the number of subdirectory entries */
    sumFiles,           /* get the number of flat files */
    sumTotal            /* get the total number of entries in the directory */
};

enum searchType {       /* the ways a directory item can be located */
    findByName = 0,     /* find an entry by its name */
    findByAddress       /* find an entry by its icb   address */
};

enum bitmapType {       /* the types of bitmap we use */
    bmapErased = 0,     /* Unallocated Bit Map: sectors ready to use */
    bmapUnerased        /* Freed Bit Map: sectors ready to be erased */
};

enum spRequestType {    /* list of ways to get disk space */
    nonContiguous = 0,  /* non-contiguous disk space request */
    contiguous          /* contiguous disk space request */
};

/* File System/Medium Type list: note, keep all WORM-types together at low end */

enum fstype {
    fsWORMonMO = 0,     /* WORM file system on MO media */
    fsWORMonMagnetic,   /* WORM file system on magnetic media */
    fsWORM = 31,        /* WORM file system on WORM media */
    fsMO,               /* erasable file system on MO media */
    fsMagnetic          /* erasable file system on magnetic media */
};

/* Access Interpretation field in PD */
#define ACCESS_UNSPECIFIED      0
#define ACCESS_READ_ONLY        1
#define ACCESS_WORM             2
#define ACCESS_REWRITABLE       3
#define ACCESS_OVERWRITABLE     4

/* context clear parameters: clear one, all of a file system, or all */
enum ctextType {
    ctextOne,
    ctextFS,
    ctextAll
};

#define ICB_BLOCK_SIZE_WORM       5
#define ICB_INDIRECT_SIZE_WORM    2
#define ICB_BLOCK_SIZE_MO         1
#define ICB_INDIRECT_SIZE_MO      0


struct space {                          /* space list in memory        */
    UINT32              modified;       /* used for high water mark    */
    UINT32              start;          /* starting physical sector    */
    UINT32              length;         /* number of sectors           */
    struct space        *next;          /* pointer to next link        */
    struct space        *previous;      /* pointer to previous link    */
};

struct bitmap {
    UINT32  logSector;  /* logical sector of bitmap file it comes from */
    UINT32  lastSector; /* last sector of bitmap file it comes from    */
    UINT32  modified;   /* dirty flag                                  */
    char    *bitmapData;  /* ptr to start of 1's & 0's bitmap data     */
    char    *bitmapHdr;   /* ptr to start of sector of bitmap data     */
                         /* NOTE: bitmapData==bitmapHdr except for 1st sector*/
    char    *bitmapEnd;   /* pointer to end of bitmap data for this sector */
    UINT32  lastByte;     /* calculated last byte in last sector */
};

struct otfsSection {                /* Volume section definition    */
    struct vnode    *devvp;         /* vnode for the device         */
    INT32           cdev;           /* character device number      */
    INT16           lun;            /* logical unit number          */
    UINT32          baseaddr;       /* address of base record icb   */
    UINT32          baseVersion;    /* version of base record using */
    UINT32          volsiz;         /* size of vol sect, in sectors */
};


struct sfHeaderId {                 /* System File Identification */
    UINT32  Tag;
    UINT32  OnodeAddr;
    UINT32  Reserved;
};

struct fidData {                    /* overlay for unique file id */
    UINT32      icbAddr;            /* the vnode's icb address    */
    UINT32      version;            /* the vnode's version number */
    CHAR8       name[FILENAMESIZE]; /* the vnode's name           */
    enum vtype  attribute;          /* the vnode's type           */
};

struct bitmapInfo {             /* overlay for bimap system use area */
    UINT32  sectorsFree;        /* number of sectors free on volume  */
    UINT32  reserved;           /* for 64-bit addressing             */
};

/* new definitions to support the ECMA 167 Standard */

/* some basic type definitions */

typedef CHAR8 dstring;

struct extent_ad {
    UINT32 Length;
    UINT32 Location;
};

/*
 *           Allocation Descriptors - See ECMA-167 P4-14
 *
 * struct short_AD: This is used for a single partition
 * struct long_AD: This is used for multiple partitions
 * struct ext_AD: This is used for compression
 */

/*
 *                             WARNING
 *                          struct lb_addr
 *                             WARNING
 *
 * struct lb_addr is NOT recommended for use. The size of this structure
 * will be rounded up to 8 bytes on some machines (e.g. HP-PA)!
 *
 */

struct lb_addr {
    UINT32          LBN;              /* partition relative block # */
    UINT16          PartNo;
};

/*
 *                           **** WARNING ****
 *
 * The ELENGTH type corresponds to the ExtentLength field. Note that the
 * two most-significant bits correspond to the type of extent. This is
 * compiler-dependent, so the order of the Type and Length fields may
 * need to be reversed for some compilers.
 *
 *                           **** WARNING ****
 */
typedef union {
    UINT32          Length32;
    struct {
        UINT32          Type:2;       /* See Extent Type Constants */
        UINT32          Length:30;    /* length in bytes */
    } bf;                             /* "bitfield" */
} ELENGTH;

/*
 * Extent Type Constants 
 */

#define E_RECORDED    0 /* extent allocated and recorded */
#define E_ALLOCATED   1 /* extent allocated but unrecorded */
#define E_UNALLOCATED 2 /* extent unallocated and unrecorded */
#define E_ALLOCEXTENT 3 /* extent is next extent of ADs */

/* udf flag definitions */
#define EXT_ERASED ((UINT16)BITZERO) /* Only valid for ALLOCATED ADs */

/* short allocation descriptor (8 bytes) */
struct short_AD {
    ELENGTH         ExtentLength;
    UINT32          Location;         /* logical block address */
};

/* long allocation descriptor (16 bytes) */
struct long_AD {
    ELENGTH         ExtentLength;
    UINT32          Location_LBN;     /* partition relative block # */
    UINT16          Location_PartNo;  /* partition number */
    UINT16          uUdfFlags;        /* See UDF1.01 2.3.10.1 */
    UCHAR8          aImpUse[4];
};

/* extended allocation descriptor (20 bytes) */
struct ext_AD {
    ELENGTH         ExtentLength;
    UINT32          RecordedLength;   /* in bytes too... */
    UINT32          InfoLength;       /* bytes */
    UINT32          Location_LBN;     /* partition relative block # */
    UINT16          Location_PartNo;  /* partition number */
    UCHAR8          aImpUse[2];
};
/*
 * generic allocation descriptor (24 bytes)
 *
 * This is the struct used in the ADL -- It is NOT part of NSR, but
 * rather allows all 4 bytes of implementation use (needed for long_ADs)
 * as well as RecordedLength and InfoLength fields. It holds enough
 * information for all AD types.
 */
struct gen_AD {
    ELENGTH         ExtentLength;
    UINT32          RecordedLength;   /* in bytes too... */
    UINT32          InfoLength;       /* bytes */
    UINT32          Location_LBN;     /* partition relative block # */
    UINT16          Location_PartNo;  /* partition number */
    UINT16          uUdfFlags;        /* See UDF1.01 2.3.10.1 */
    UCHAR8          aImpUse[4];
};


#define CS0  0
#define CS1  1
#define CS2  2
#define CS3  3
#define CS4  4
#define CS5  5
#define CS6  6
#define CS7  7
#define CS8  8

struct charspec {
    UCHAR8 uCharSetType;
    UCHAR8 aCharSetInfo[63];
};

/* definitions for the domain flags field */
#define WRPROTECT_HARD BITZERO
#define WRPROTECT_SOFT BITONE

/* definitions for the flags field of a regid */
#define DIRTYREGID   BITZERO
#define PROTECTREGID BITONE

/* Pre-defined strings for ID field. */
#define OSTA_LVINFO_ID ("*UDF LV Info")
#define OSTA_DOMAIN_ID ("*OSTA UDF Compliant")

/* Pre-defined OS Class values */
#define OSCLASS_UNDEF 0
#define OSCLASS_DOS   1
#define OSCLASS_OS2   2
#define OSCLASS_MAC   3 
#define OSCLASS_UNIX  4

/* Pre-defined OS Identifier values */
#define OSID_GENERIC     0
#define OSID_IBM_AIX     1
#define OSID_SUN_SOLARIS 2
#define OSID_HPUX        3 
#define OSID_SGI_IRIX    4

/* ISO-definition of regid */
struct regid {
    UCHAR8 uFlags;
    UCHAR8 aRegisteredID[23];
    UCHAR8 aIDSuffix[8];
};

/* UDF1.01 / 2.1.4.3 */
struct domainEntityId {
    UCHAR8 uFlags;
    UCHAR8 aID[23];
    UINT16 uUDFRevision;
    UCHAR8 uDomainFlags;
    UCHAR8 aReserved[5];
};

/* UDF1.01 / 2.1.4.3 */
struct udfEntityId {
    UCHAR8 uFlags;
    UCHAR8 aID[23];
    UINT16 uUDFRevision;
    UCHAR8 uOSClass;
    UCHAR8 uOSIdentifier;
    UCHAR8 aReserved[4];
};

/* UDF1.01 / 2.1.4.3 */
struct implEntityId {
    UCHAR8 uFlags        ;
    UCHAR8 aID[23]       ;
    UCHAR8 uOSClass      ;
    UCHAR8 uOSIdentifier ;
    UCHAR8 uDescVersion  ;
    UCHAR8 uLibrVersion  ;
    UINT16 uImplRegNo    ;
    UCHAR8 uImplVersion  ;
    UCHAR8 uImplRevision ;
};

struct timestamp {
    UINT16 uTypeAndTimeZone;
    INT16  iYear;
    UCHAR8 uMonth;
    UCHAR8 uDay;
    UCHAR8 uHour;
    UCHAR8 uMinute;
    UCHAR8 uSecond;
    UCHAR8 uCentiseconds;
    UCHAR8 uHundredMicroseconds;
    UCHAR8 uMicroseconds;
};
/* Macros for dissecting type and time zone. */
/* masks and shift constants */

#define TPMask  0xf000
#define TPShift 12

#define TZMask  0x0fff
#define TZSignBit 0x0800
#define TZSignExt 0xf000
#define TZSignExtend(tz)((INT16)((TZSignBit&tz)?(tz|TZSignExt):(tz)))

/* getters */
#define GetTSTP(ttz)(ttz>>TPShift)
#define GetTSTZ(ttz)(TZSignExtend(ttz&TZMask))

/* setters */
#define SetTSTP(ttz,tp)ttz=((tp<<TPShift)|(ttz&TZMask))
#define SetTSTZ(ttz,tz)ttz=(GetTSTP(ttz)|(tz&TZMask))

/* Partition Header Descriptor */

struct PartHeaderDesc {
    struct short_AD UST;            /* unallocated space table */
    struct short_AD USB;            /* unallocated space bitmap */
    struct short_AD PIT;            /* partition integrity table */
    struct short_AD FST;            /* freed space table */
    struct short_AD FSB;            /* freed space bitmap */
    UCHAR8 aReserved[88];
};
    
struct Tag {
    UINT16 uTagID;
    UINT16 uDescriptorVersion;
    UCHAR8 uTagChecksum;
    UCHAR8 uReserved;
    UINT16 uTagSerialNum;
    UINT16 uDescriptorCRC;
    UINT16 uCRCLen;
    UINT32 uTagLoc;
};

/* Volume Structure Descriptor (ECMA 3/9.1) */
struct VolStructDesc {
    UCHAR8 structType;
    UCHAR8 stdIdentifier[5];
    UCHAR8 structVersion;
    UCHAR8 reserved;
    UCHAR8 structData[2040];
};

/* Primary Volume Descriptor */
struct PrimaryVolDes {
    struct Tag sTag;       /* uTagID = 1 */
    UINT32 uVolDescSeqNum;
    UINT32 uPrimVolDesNum;
    dstring aVolID[32];
    UINT16 uVSN;
    UINT16 uMaxVSN;
    UINT16 uInterchangeLev;
    UINT16 uMaxInterchangeLev;
    UINT32 uCharSetList;
    UINT32 uMaxCharSetList;
    dstring aVolSetID[128];
    struct charspec sDesCharSet;
    struct charspec sExplanatoryCharSet;
    struct extent_ad sVolAbstract;
    struct extent_ad sVolCopyrightNotice;
    struct regid sApplicationID;
    struct timestamp sRecordingTime;
    struct implEntityId sImplementationID;
    UCHAR8 aImplementationUse[64];
    UINT32 uPredecessorVDSLoc;
    UINT16 uFlags;
    UCHAR8 aReserved[22];
};

/* PrimaryVolDesc Flags Constants */

/* Bit 0 : Only one defined so far. */
#define COMMON_VOLSET_IDENTIFICATION ((UINT16)BITZERO)

/* Anchor Volume Descriptor Pointer */
struct AnchorVolDesPtr {
    struct Tag sTag;       /* uTagID = 2 */
    struct extent_ad sMainVDSAdr;
    struct extent_ad sReserveVDSAdr;
    UCHAR8 Reserved[480];
};

/* Volume Descriptor Pointer */
struct VolDescPtr {
    struct Tag sTag;      /* uTagID = 3 */
    UINT32 uVolDescSeqNum;
    struct extent_ad sNextVDS;
    UCHAR8 aReserved[484];
};

/* Implementation Use Descriptor */
struct ImpUseDesc {
    struct Tag sTag;      /* uTagID = 4 */
    UINT32 uVolDescSeqNum;
    struct udfEntityId sImplementationIdentifier;
    UCHAR8 aReserved[460];
};

/* OSTA Volume Descriptor (fits inside of implUseDesc.aReserved). */
struct LVInformation {
    struct charspec sLVICharset;
    dstring aLogicalVolumeIdentifier[128];
    dstring aLVInfo1[36];
    dstring aLVInfo2[36];
    dstring aLVInfo3[36];
    struct implEntityId sImplementationID;
    UCHAR8 aImplementationUse[128];
};

/* Partition Descriptor */
struct PartDesc {
    struct Tag sTag;        /* uTagID = 5 */
    UINT32 uVolDescSeqNum;
    UINT16 uPartFlags;
    UINT16 uPartNumber;
    struct regid sPartContents;
    UCHAR8 aPartContentsUse[128];
    UINT32 uAccessType; 
    UINT32 uPartStartingLoc;
    UINT32 uPartLength;
    struct implEntityId sImplementationID;
    UCHAR8 aImplementationUse[128];
    UCHAR8 aReserved[156];
};

/* Partition Flags definition */
/* (Only Bit 0 defined) */
#define PARTITION_ALLOCATED ((UINT16) BITZERO)

/* Type 1 Partition Map */
struct PartMap1 {
    UCHAR8 uPartMapType;
    UCHAR8 uPartMapLen;
    UINT16 uVSN;
    UINT16 uPartNum;
};

/* Type 2 Partition Map */
struct PartMap2 {
    UCHAR8 uPartMapType;
    UCHAR8 uPartMapLen;
    UCHAR8 uPartID[62];
};

/* Logical Volume Descriptor */
struct LogVolDesc {
    struct Tag sTag;       /* uTagID = 6 */
    UINT32 uVolDescSeqNum;
    struct charspec sDesCharSet;
    dstring uLogVolID[128];
    UINT32 uLogBlkSize;
    struct domainEntityId sDomainID;
    UCHAR8 uLogVolUse[16];
    UINT32 uMapTabLen;
    UINT32 uNumPartMaps;
    struct implEntityId sImplementationID;
    UCHAR8 aImplementationUse[128];
    struct extent_ad integritySeqExtent;
/* Variable-length stuff follows: (see Macros below) */
/*    struct PartMap1 sPartMaps[uNumPartMaps];       */
/* Or:                                               */
/*    struct PartMap2 sPartMaps[uNumPartMaps];       */
};

/* Macros for LVDs */
#define LVD_HdrLen      (sizeof(struct LogVolDesc))
#define LVD_PM1Start(x) (struct PartMap1 *)((char *)(x)+LVD_HdrLen)
#define LVD_PM2Start(x) (struct PartMap2 *)((char *)(x)+LVD_HdrLen)

/* Unallocated Space (volume) Entry */
struct UnallocSpDesHead {
    struct Tag sTag;          /* uTagID = 7 */
    UINT32 uVolDescSeqNum;     
    UINT32 uNumAllocationDes;
};

/* Terminator Descriptor */
struct Terminator {
    struct Tag sTag;          /* uTagID = 8 */
    UCHAR8 aReserved[496];
};

/* Logical Volume Integrity Descriptor */
struct LogicalVolumeIntegrityDesc {
    struct Tag          sTag;          /* uTagID = 9 */
    struct timestamp    sRecordingTime;
    UINT32              integrityType;
    struct extent_ad    nextIntegrityExtent;
    UINT32              UniqueIdH;
    UINT32              UniqueIdL;
    UCHAR8              reserved[24];
    UINT32              N_P;    /* num Partitions */
    UINT32              L_IU;   /* Len implement use */
/* Variable-length stuff follows: (see Macros below) */
/*    UINT32              FreeSpaceTable[N_P];       */
/*    UINT32              SizeTable[N_P];            */
/*    struct LVIDImplUse  ImplementUse;              */
};

/* Macros for LVIDs */
/* NOTE that the IUStart() macro provides the proper address for the
   LVIDImplUse structure, rather than the implementation use field of
   it. */
#define LVID_HdrLen       (sizeof(struct LogicalVolumeIntegrityDesc))
#define LVID_FSTStart(x)  ((UINT32 *)((char *)(x)+LVID_HdrLen))
#define LVID_SZTStart(x)  (LVID_FSTStart(x)+(x)->N_P)
#define LVID_IUStart(x)   (UCHAR8 *)(LVID_SZTStart(x)+(x)->N_P)

/* Logical Volume Integrity Descriptor Implementation Use area */
struct LVIDImplUse {
    struct implEntityId implementationID;
    UINT32              numFiles;
    UINT32              numDirectories;
/* Variable-length stuff follows: (see Macros below) */
/*    UCHAR8             ImplementUse[L_IU-40]       */
};

/* Macros for LVIDImplUse */
/* NOTE that LVIDIU_IUStart() needs the LVID_IUStart() as its
   argument. Thus, the supported use model is for the two structs to
   be instantiated separately, and the LVIDIU to be cast to the
   implementation use pointer supplied by the LVID. */
#define LVIDIU_HdrLen     (sizeof(struct LVIDImplUse))
#define LVIDIU_IUStart(x) (UCHAR8 *)((char *)(x)+LVIDIU_HdrLen)

/* File Set Descriptor */
struct FileSetDesc {
    struct Tag            sTag;          /* uTagID = 256 */
    struct timestamp      sRecordingTime;
    UINT16                uInterchangeLev;
    UINT16                uMaxInterchangeLev;
    UINT32                uCharSetList;
    UINT32                uMaxCharSetList;
    UINT32                uFileSetNum;
    UINT32                uFileSetDescNum;
    struct charspec       sLogVolIDCharSet;
    dstring               aLogVolID[128];
    struct charspec       sFileSetCharSet;
    dstring               aFileSetID[32];
    dstring               aCopyrightFileID[32];
    dstring               aAbstractFileID[32];
    struct long_AD        sRootDirICB;
    struct domainEntityId DomainID;
    struct long_AD        sNextExtent;
    UCHAR8                aReserved[48];
};

/*
 * constant length part of a ECMA 167 File Identifier Desc 
 *
 * define a constant here to represent the constant size of the
 * file identifier descriptor since sizeof(FileIDDesc) may be
 * erroneous on risc machines that pad to a multiple of 4 bytes.
 * To obtain length of entire FID, add L_FI, L_IU, and the following
 * constant and then round up to a multiple of four (for padding).
 */
#define FILE_ID_DESC_CONSTANT_LEN       38

/* WARNING - Do not use sizeof(FileIDDesc) !!!
 *  This is placing a struct at a non-multiple-of-four location!!!
 *  There is not much we can do about this, since UDF1.01 mandates it.
 *  There are no apparent problems on g++/hp-ux, since the structure
 *  itself does not contain any word-long pieces.
 */
struct FileIDDesc {
    struct Tag          sTag;   /* uTagID = 257 */
    UINT16              VersionNum;
    UCHAR8              Characteristics;
    UCHAR8              L_FI;        
    struct long_AD      ICB;
    UINT16              L_IU;
    struct implEntityId sImplementationID;
    /* Implementation Use Falls here */
    /* File ID falls here */
    /* padding falls here */
};

/* - ISSUE -
 * define a constant here to specify the info length field for the 
 * root directory. Used in makefs.c
 */
#define ROOTDIR_INFO_LENGTH       6

/* alias descriptor format */
struct AliasDesc {
    struct Tag sTag;   /* uTagID = 258 */
    UINT16 VersionNum;
    UCHAR8 Characteristics;
    UCHAR8 L_FI;
    UINT16 L_AV;
    UINT16 L_IU;
    /* Implementation Use Falls here */
    /* File ID falls here */
    /* alias value falls here */
    /* padding falls here */
};

/* ECMA 167 Information Control Block Details */
struct ICBTag {
    UINT32 PriorDirects;
    UINT16 StrategyType;
    UCHAR8 StrategyParm[2];
    UINT16 NumberEntries;
    UCHAR8 Reserved;
    UCHAR8 FileType;
    /* lb_addr   sParentICB; */
    UINT32 sParentICB_LBN;       /* partition relative block # */
    UINT16 sParentICB_PartNo;    /* partition number */
    UINT16 Flags;
};

/* ICB file types */
#define FILE_TYPE_UNSPECIFIED          0
#define FILE_TYPE_UNALLOC_SP_ENTRY     1
#define FILE_TYPE_PARTITION_INTEGRITY  2
#define FILE_TYPE_INDIRECT_ENTRY       3
#define FILE_TYPE_DIRECTORY            4
#define FILE_TYPE_RAW                  5
#define FILE_TYPE_BDEV                 6
#define FILE_TYPE_CDEV                 7
#define FILE_TYPE_EXT_ATTR             8
#define FILE_TYPE_FIFO                 9
#define FILE_TYPE_SOCKET              10
#define FILE_TYPE_TERMINAL_ENTRY      11
#define FILE_TYPE_SYMLINK             12

/* ICB Flags Constants */

/* Bits 0,1,2 - ADType */
#define ADTYPEMASK     (UINT16) 0x0007    /* allocation descriptor type */
                                          /* mask for ICBTag flags field */
#define ADGENERIC      (UINT16) 0x2000    /* Internal use ONLY */
#define ADSHORT        (UINT16) 0x00      /* Short ADs */
#define ADLONG         (UINT16) 0x01      /* Long ADs */
#define ADEXTENDED     (UINT16) 0x02      /* Extended ADs */
#define ADNONE         (UINT16) 0x03      /* Data replaces ADs */

#define ValidADType(x) ((x)==ADSHORT||(x)==ADLONG||(x)==ADEXTENDED)

/* Bits 3-9 : Miscellaneous */
#define SORTED_DIRECTORY   0x0008 /* This should be cleared. */
#define NON_RELOCATABLE    0x0010 /* We don't set this. */
#define ARCHIVE            0x0020 /* Should always be set when written or modified. */
#define ICBF_S_ISUID       0x0040
#define ICBF_S_ISGID       0x0080
#define ICBF_C_ISVTX       0x0100
#define CONTIGUOUS         0x0200
#define FLAG_SYSTEM        0x0400
#define FLAG_TANSFORMED    0x0800
#define FLAG_MULTIVERS     0x1000

struct EntryHdr {
    struct Tag sTag;
    struct ICBTag sICBTag;
};

/* ECMA 167 Indirect Entry */
struct IndirectEntry {
    struct Tag sTag;       /* uTagID = 259 */
    struct ICBTag sICBTag; /* FileType = 3 */ 
    struct long_AD sIndirectICB;
};

/* ECMA 167 Terminal Entry */
struct TerminalEntry {
    struct Tag sTag;        /* uTagID = 260 */
    struct ICBTag sICBTag;  /* FileType = 11 */
};

#define MIN_ARRAY_SIZE    1  /* some cc's allow 0 */

/* ECMA 167 File Entry */
struct FileEntry {
    struct Tag sTag;        /* uTagID = 261 */
    struct ICBTag sICBTag;  /* FileType = [4-10] */
    UINT32 UID;
    UINT32 GID;
    UINT32 Permissions;
    UINT16 LinkCount;
    UCHAR8 RecFormat;
    UCHAR8 RecDisplayAttr;
    UINT32 RecLength;
    UINT32 InfoLengthH;
    UINT32 InfoLengthL;
    UINT32 LogBlocksH;
    UINT32 LogBlocksL;
    struct timestamp sAccessTime;   /* each timestamp is 12 bytes long */
    struct timestamp sModifyTime;
    struct timestamp sAttrTime;
    UINT32 Checkpoint;
    struct long_AD sExtAttrICB;     /* 16 bytes */
    struct implEntityId sImpID;
    UINT32 UniqueIdH;           /* for 13346 */
    UINT32 UniqueIdL;
    UINT32 L_EA;
    UINT32 L_AD;
    /* allocation descriptors go here: ADMacros at end of file */
};
    
/* ECMA 167 Extended Attribute Defines */

/* Extended Attribute Header Descriptor */
struct ExtAttrHeaderDesc {
    struct Tag sTag;   /* uTagID = 262 */
    UINT32 ImpAttrLoc;
    UINT32 AppAttrLoc;
};


/* Unallocated Space Entry */
struct UnallocSpEntry {
    struct Tag sTag;   /* uTagID = 263 */
    struct ICBTag sICBTag;  /* FileType = 1 */
    UINT32 L_AD;
    /* allocation descriptors go here: ADMacros at end of file */
};

/* Space Bitmap Entry */
struct SpaceBitmapHdr {
    struct Tag sTag;   /* uTagID = 264 */
    UINT32 N_Bits;
    UINT32 N_Bytes;
};
struct SpaceBitmapEntry {
    struct SpaceBitmapHdr hdr;
    UCHAR8 bitmap[MIN_ARRAY_SIZE]; /* bitmap */
};

/* Partition Integrity Entry */
struct PartIntegrityEntry {
    struct Tag sTag;         /* uTagID = 265 */
    struct ICBTag sICBTag;   /* FileType = 2 */
    struct timestamp sRecordingTime;
    UCHAR8 IntegrityType;
    UCHAR8 reserved[175];
    struct regid sImpID;
    UCHAR8 aImpUse[256];
};

/* INTEGRITY CONSTANTS
   The first two are valid for LVIDs as well as PIEs.
   The final one, INTEGRITY_STABLE, is ONLY valid for PIE.
   */

#define INTEGRITY_OPEN     0
#define INTEGRITY_CLOSE    1
#define INTEGRITY_STABLE   2

/* Allocation Extent Descriptor */
struct AllocationExtentDesc {
    struct Tag sTag;         /* uTagID = 266 */
    UINT32 prevAllocExtLoc;
    UINT32 L_AD;
    /* allocation descriptors go here: ADMacros at end of file */
};

struct ExtAttrPrologue {
    UINT32 AttrType;
    UCHAR8 AttrSubType;
    UCHAR8 Reserved[3];
    UINT32 A_L;
    /* extended attribute specific data follows */
};

/*
 *                       The icb entry structure.
 *
 * Note that only one AD is declared in an fe icb. When an otnode is
 * allocated, a larger AD space will be allocated and subsequently
 * referenced. This is necessary to accommodate variable length sector
 * sizes. 
 */

union icbe {
    struct IndirectEntry      ie;
    struct FileEntry          fe;      
    struct PartIntegrityEntry pie;
    struct UnallocSpEntry     uspe;
    struct TerminalEntry      te; 
    struct EntryHdr           ge;   /* generic entry */
};

struct ADList {
    struct gen_AD *beginADs;    /* Beginning of array */
    struct gen_AD *endADs;      /* After last array ent */
    struct space  *diskSpace;   /* allocation extents */
    UINT32        modified;     /* disk <> RAM */
    UINT32        bufLen;       /* Size of buffer allocated */
};

#define ADBufLim(x) ((struct gen_AD *)((char *)(x)->beginADs + (x)->bufLen))

/*
 *                      The otnode structure.
 *
 */

struct otnode {                         /* Otnode structure definition  */
    struct vnode    vnode;              /* vnode for the file/directory */
    INT32           flags;
    UINT32          lastOffset;
    UINT32          lastEntry;
    struct ADList   ADL;
    UINT32          version;            /* file version number          */
    CHAR8           isLast;             /* is file the last version?    */  
    UCHAR8          attr;
    CHAR8           name[FILENAMESIZE];
    UINT32          parent_addr_LBN;    /* partition relative block # */
    UINT16          parent_addr_PartNo; /* partition number */
    UINT32          icb_addr_LBN;       /* partition relative block # */
    UINT16          icb_addr_PartNo;    /* partition number */
    struct otnode   *previous;          /* ptr to prev otnode in pool   */
    struct otnode   *next;              /* ptr to next otnode in pool   */
    union icbe      icbe;               /* icb entry for the file/directory */
};

struct otfs {                           /* Otfs structure definition    */
    struct vfs          *vfs;           /* ptr to corresponding vfs     */
    INT32               flags;          /* flags                        */
    UINT32              SectorSize;     /* sector size, in bytes        */
    UCHAR8              MediaType;      /* media type                   */
    UCHAR8              FileSystemType; /* type of file system          */
    UCHAR8              ADType;         /* default allocation rec type  */
    UINT16              ICBsize;        /* in sectors                   */
    UINT16              IndirectSize;   /* in sectors                   */
    UINT32              DiskSectors;    /* size of surface, in sectors  */
    UINT32              SectorsInBMap;  /* # sectors in bitmaps         */
    UINT32              sectorsFree;    /* # sectors free on volume     */
    UINT32              rootAddr;       /* address of the root dir icb  */
    UINT32              erasedAddr;     /* address of unalloced bmap icb*/
    UINT32              unerasedAddr;   /* address of freed bmap icb    */
    UINT32              pitAddr;        /* addr of PIT icb              */
    UINT32              hwaterVersion;  /* v # of highwater mark file   */
    char *              iobuf;          /* allocated at mount time      */
    char *              dirbuf;         /* allocated at mount time      */
    struct otnode       *active;        /* ptr to active otnode list    */
    struct otnode       *inactive;      /* ptr to inactive otnode list  */
    struct space        *highWater;     /* ptr to WORM highwater mark   */
    struct bitmap       *erasedBitmap;  /* ptr to unalloced bitmap      */
    struct bitmap       *unerasedBitmap;/* ptr to freed bitmap          */
    otfsVendorField     vf;             /* implementation specific stuff*/
    INT16               sections;       /* number of sections in volume */
    CHAR8               VolumeName[FILENAMESIZE];
    struct otfsSection  VolSection[1];  /* volume section information   */
};

/*
 * ADMacros
 *
 * These provide access to various points in the FE, USE, and AE
 * structures.
 */

#define FE_HdrLen(x)   (sizeof(struct FileEntry)+(x)->L_EA)
#define USE_HdrLen     (sizeof(struct UnallocSpEntry))
#define AE_HdrLen      (sizeof(struct AllocationExtentDesc))

#define FE_ADStart(x)  (void *)((char *)(x)+FE_HdrLen(x))
#define USE_ADStart(x) (void *)((char *)(x)+USE_HdrLen)
#define AE_ADStart(x)  (void *)((char *)(x)+AE_HdrLen)

#define FE_EAStart(x)  (void *)((char *)(x)+(sizeof(struct FileEntry)))

#endif  /* NSR_H */
