#ifndef _sys_vfs_h
#define _sys_vfs_h
/********************************************************************/
/*  vfs.h - lots of file system defines.                            */
/*                                                                  */
/*      Unpublished Confidential Information of Hewlett-Packard     */
/*      Company. Do not disclose.  Copyright (c) Hewlett-Packard    */
/*      Company 1996. All rights reserved                           */
/*                                                                  */
/* The contents of this software are proprietary and confidential   */
/* the Hewlett- Packard Company, and are limited in distribution to */
/* those with a direct need to know.  Individuals having access to  */
/* this software are responsible for maintaining the                */
/* confidentiality of the content and for keeping the software      */
/* secure when not in use.  Transfer to any party is strictly       */
/* forbidden other than as expressly permitted in writing by        */
/* Hewlett-Packard Company.  Unauthorized transfer to or possession */
/* by any unauthorized party may be a criminal offense.             */
/*                                                                  */
/*                   RESTRICTED RIGHTS LEGEND                       */
/*                                                                  */
/*      Use,  duplication,  or disclosure by the                    */
/*      U. S. Government is subject to restrictions as set forth    */
/*      in subparagraph (c) (1) (ii) in the Rights in Technical     */
/*      Data and Computer Software clause in DFARS 252.227-7013     */
/*      or any other successor clause.                              */
/*                                                                  */
/*                       HEWLETT-PACKARD COMPANY                    */
/*                       3000 Hanover St.                           */
/*                       Palo Alto, CA  94304 U.S.A.                */
/*                                                                  */
/*      Rights for non-DOD U.S. Government Departments and Agencies */
/*      are as set forth in FAR 52.227-19 (c) (1,2) or any other    */
/*      successor clause.                                           */
/*                                                                  */
/********************************************************************/
/*
 $Source: /projects/HMS/nsr/udfOld/src/nsrHdrs/vfs.h,v $
 $Revision: 1.6 $       $Author: hmgr $
 $State: Exp $          $Locker:  $
 $Date: 96/05/30 14:53:01 $
 ********************************************************************/


/*
 * File system identifier. Should be unique (at least per machine).
 */

typedef struct {
        long val[2];                    /* file system id type */
} fsid_t;

/*
 * File identifier. Should be unique per filesystem on a single machine.
 */
#define MAXFIDSZ        16
#define freefid(fidp) \
    kmem_free((char *)(fidp), sizeof (struct fid) - MAXFIDSZ + (fidp)->fid_len,0)

struct fid {
        UINT32          fid_len;                /* length of data in bytes */
        char            fid_data[MAXFIDSZ];     /* data (variable length) */
};

/*
 * Structure per mounted file system.
 * Each mounted file system has an array of
 * operations and an instance record.
 * The file systems are put on a singly linked list.
 * If vfs_stats is non-NULL statistics are gathered, see vfs_stat.h
 */
struct vfs {
        struct vfs      *vfs_next;              /* next vfs in vfs list */
        struct vfsops   *vfs_op;                /* operations on vfs */
        struct vnode    *vfs_vnodecovered;      /* vnode we mounted on */
        int             vfs_flag;               /* flags */
        int             vfs_bsize;              /* native block size */
        fsid_t          vfs_fsid;               /* file system id */
        char *  vfs_stats;              /* filesystem statistics */
        char *          vfs_data;               /* private data */
};

/*
 * vfs flags.
 * VFS_MLOCK lock the vfs so that name lookup cannot proceed past the vfs.
 * This keeps the subtree stable during mounts and unmounts.
 */
#define VFS_RDONLY      0x01            /* read only vfs */
#define VFS_MLOCK       0x02            /* lock vfs so that subtree is stable */
#define VFS_MWAIT       0x04            /* someone is waiting for lock */
#define VFS_NOSUID      0x08            /* turn off set-uid on exec */
#define VFS_GRPID       0x10            /* Old BSD group-id on create */
#define VFS_NOSUB       0x20            /* No mounts allowed beneath this fs */
#define VFS_REMOUNT     0x40            /* modify mount otions only */
#define VFS_MULTI       0x80            /* Do multi-component lookup on files */

/*
 * Operations supported on virtual file system.
 */
struct vfsops {
        int     (*vfs_mount)();         /* mount file system */
        int     (*vfs_unmount)();       /* unmount file system */
        int     (*vfs_root)();          /* get root vnode */
        int     (*vfs_statfs)();        /* get fs statistics */
        int     (*vfs_sync)();          /* flush fs buffers */
        int     (*vfs_vget)();          /* get vnode from fid */
        int     (*vfs_mountroot)();     /* mount the root filesystem */
        int     (*vfs_swapvp)();        /* return vnode for swap */
};

#define VFS_MOUNT(VFSP, PATH, DATA) \
                                 (*(VFSP)->vfs_op->vfs_mount)(VFSP, PATH, DATA)
#define VFS_UNMOUNT(VFSP)        (*(VFSP)->vfs_op->vfs_unmount)(VFSP)
#define VFS_ROOT(VFSP,VPP,NAME)  (*(VFSP)->vfs_op->vfs_root)(VFSP,VPP,NAME)
#define VFS_STATFS(VFSP, SBP)    (*(VFSP)->vfs_op->vfs_statfs)(VFSP,SBP)
#define VFS_SYNC(VFSP)           (*(VFSP)->vfs_op->vfs_sync)(VFSP)
#define VFS_VGET(VFSP, VPP, FIDP) (*(VFSP)->vfs_op->vfs_vget)(VFSP, VPP, FIDP)
#define VFS_MOUNTROOT(VFSP, VPP, NM) \
                                 (*(VFSP)->vfs_op->vfs_mountroot)(VFSP, VPP, NM)
#define VFS_SWAPVP(VFSP, VPP, NM) (*(VFSP)->vfs_op->vfs_swapvp)(VFSP, VPP, NM)

/*
 * file system statistics
 */
struct statfs {
        long f_type;                    /* type of info, zero for now */
        long f_bsize;                   /* fundamental file system block size */
        long f_blocks;                  /* total blocks in file system */
        long f_bfree;                   /* free block in fs */
        long f_bavail;                  /* free blocks avail to non-superuser */
        long f_files;                   /* total file nodes in file system */
        long f_ffree;                   /* free file nodes in fs */
        fsid_t f_fsid;                  /* file system id */
        long f_spare[7];                /* spare for later */
};

#ifdef KERNEL
/*
 * Filesystem type switch table
 */
struct vfssw {
        char            *vsw_name;      /* type name string */
        struct vfsops   *vsw_ops;       /* filesystem operations vector */
};

/*
 * public operations
 */
extern void     vfs_mountroot();        /* mount the root */
extern int      vfs_add();              /* add a new vfs to mounted vfs list */
extern void     vfs_remove();           /* remove a vfs from mounted vfs list */
extern int      vfs_lock();             /* lock a vfs */
extern void     vfs_unlock();           /* unlock a vfs */
extern struct vfs *getvfs();            /* return vfs given fsid */
/* extern struct vfssw *getfstype();       find default filesystem type */
extern int vfs_getmajor();              /* get major device # for an fs type */
extern void vfs_putmajor();             /* free major device # for an fs type */
extern int vfs_getnum();                /* get device # for an fs type */
extern void vfs_putnum();               /* release device # for an fs type */

#define VFS_INIT(VFSP, OP, DATA)        { \
        (VFSP)->vfs_next = (struct vfs *)0; \
        (VFSP)->vfs_op = (OP); \
        (VFSP)->vfs_flag = 0; \
        (VFSP)->vfs_stats = NULL; \
        (VFSP)->vfs_data = (DATA); \
}

/*
 * globals
 */
extern struct vfs *rootvfs;             /* ptr to root vfs structure */
extern struct vfssw vfssw[];            /* table of filesystem types */
extern struct vfssw *vfsNVFS;           /* vfs switch table end marker */
#endif

#endif /* _sys_vfs_h_ */
