/********************************************************************/
/* osys.c  - Routines that allow low level I/O.                     */
/*                                                                  */
/*      Unpublished Confidential Information of Hewlett-Packard     */
/*      Company. Do not disclose.  Copyright (c) Hewlett-Packard    */
/*      Company 1996. All rights reserved                           */
/*                                                                  */
/* The contents of this software are proprietary and confidential   */
/* the Hewlett- Packard Company, and are limited in distribution to */
/* those with a direct need to know.  Individuals having access to  */
/* this software are responsible for maintaining the                */
/* confidentiality of the content and for keeping the software      */
/* secure when not in use.  Transfer to any party is strictly       */
/* forbidden other than as expressly permitted in writing by        */
/* Hewlett-Packard Company.  Unauthorized transfer to or possession */
/* by any unauthorized party may be a criminal offense.             */
/*                                                                  */
/*                   RESTRICTED RIGHTS LEGEND                       */
/*                                                                  */
/*      Use,  duplication,  or disclosure by the                    */
/*      U. S. Government is subject to restrictions as set forth    */
/*      in subparagraph (c) (1) (ii) in the Rights in Technical     */
/*      Data and Computer Software clause in DFARS 252.227-7013     */
/*      or any other successor clause.                              */
/*                                                                  */
/*                       HEWLETT-PACKARD COMPANY                    */
/*                       3000 Hanover St.                           */
/*                       Palo Alto, CA  94304 U.S.A.                */
/*                                                                  */
/*      Rights for non-DOD U.S. Government Departments and Agencies */
/*      are as set forth in FAR 52.227-19 (c) (1,2) or any other    */
/*      successor clause.                                           */
/*                                                                  */
/*                                                                  */
/*           ot_mopread, ot_mopwrite, ot_erase, ot_lock_cart,       */
/*           ot_lockfs, ot_read, ot_readCapacity, ot_testready      */
/*           ot_unlock_cart, ot_unlockfs, ot_write, kmem_alloc      */
/*           kmem_free, bzero                                       */
/*                                                                  */
/********************************************************************/
/*
 $Source: /projects/HMS/nsr/udfOld/src/nsrSys/o_sys.c,v $
 $Revision: 1.40 $      $Author: hmgr $
 $State: Exp $          $Locker:  $
 $Date: 96/05/30 14:53:06 $
 ********************************************************************/
#include "def.h"

#if HP
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "o_vfsops.h"
#include "vnode.h"
#include "nsr.h"
#include "o_proto.h"
#include "nsrerrno.h"
#include "Debug.h"
#include "o_node.h"
#include "makefs.h"
#include "../nsrTools/magvol/magvol.h"
#endif

#ifdef _HPUX_SOURCE 
#include <sys/ioctl.h>
#include <sys/scsi.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif

static INT32 mon[] = {0,31,28,31,30,31,30,31,31,30,31,30,31};


#if HP
INT32 hpux_open(void);
INT32 hpux_close(void);
INT32 hpux_bfd;
INT32 hpux_cfd;
INT32 hpux_read(int offset,UCHAR8 *iobuf,int bsize);
INT32 hpux_write(int offset,UCHAR8 *iobuf,int bsize);
INT32 hpux_readCapacity(char *file,UINT32 *sectors,UINT32 *sectorsize);
#endif


/* make the volume data global to these routines */
/* yes, yes, I know about the hazards of global data!! */
/********** (ed beshore did this)  ***********/
struct    ArgDat VolData;  /* holds first 512 bytes of volfile */


/********************************************************************/
/*    ot_erase                                                      */
/*        routine to erase one or more sectors on the disk          */
/*                                                                  */
/*    input parameters:                                             */
/*        fsp:        the otfs structure for the volume             */
/*        sector:        where to start erasing                     */
/*        numsec:        the number of sectors to erase             */
/*                                                                  */
/*    notes:                                                        */
/*        this is (currently) a stub for the magnetic drives used   */
/*            in file system development                            */
/********************************************************************/
INT32 ot_erase( struct otfs *fsp, int override, UINT32 sector, UINT32 numsec)
{
    INT32            error=0L;
    UINT32            i;
    register CHAR8    *buffer;

    if (fsp->MediaType == (int) fsWORM || !override && fsp->MediaType != (int) fsMO)
        return 0;

    buffer = kmem_alloc(fsp->SectorSize, 0);
    if (buffer) {
        bzero(buffer,fsp->SectorSize);
        *((long *) buffer) = 0x6C6C6C6C;

        for (i = 0, error = 0; i < numsec && !error; i++) 
        {
            error = ot_write(fsp,sector,buffer,fsp->SectorSize);    
            sector++;
        }
        kmem_free(buffer,fsp->SectorSize, 0);
    } else
        error = ENOMEM;

    return error;
} /* end ot_erase */

/********************************************************************/
/*    ot_lock_cart                                                  */
/*        routine to lock the cartridge in the drive                */
/********************************************************************/
INT32 ot_lock_cart( struct otfs *fsp, INT32 partition)

{
    return 0;
} /* end ot_lock_cart */

/********************************************************************/
/* ot_lockfs                                                        */
/*        routine to lock the file system                           */
/********************************************************************/
void ot_lockfs(register struct otfs *fsp)
{
    return;
} /* end ot_lockfs */


/********************************************************************/
/*    ot_mopread                                                    */
/*        routine to read sectors from the optical disk             */
/*                                                                  */
/*    input parameters:                                             */
/*        fsp:        the otfs structure for the volume             */
/*        vnode:      ????????????                                  */
/*        iobuf:      where the data goes                           */
/*        sector:     what sector to read                           */
/*        bsize:      ptr to how many bytes to read                 */
/*                                                                  */
/*    output parameters:                                            */
/*        bsize:      how many bytes were read                      */
/*                                                                  */
/*    notes:                                                        */
/*        On encountering EBLKCHK, the data will be zero'd out      */
/*        and the bsize returned will be made less by the amount    */
/*        of bytes for the sectors where EBLKCHK occured.           */
/*        Note that error = 0 on EBLKCHK.                           */
/********************************************************************/
INT32 ot_mopread(struct otfs *fsp,
                 struct vnode *vn,
                 caddr_t iobuf,
                 INT32 sector,
                 INT32 *bsize)
{
    INT32 Error=0;
    char *devfile = getenv("VOLUME");

    if ( strncmp(devfile,"/dev",4) == 0 ) {
        Error = hpux_read(sector*fsp->SectorSize,(UCHAR8 *)iobuf,*bsize);
    }
    return (Error == EBLKCHK ? 0 : Error);
} /* end ot_mopread */


/********************************************************************/
/*    ot_mopwrite                                                   */
/*        routine to write sectors to the optical disk              */
/********************************************************************/
INT32 ot_mopwrite(struct otfs *fsp, struct vnode *vn, caddr_t iobuf, UINT32 sector, UINT32 bsize)
 {
    INT32 Error=0;
    char *devfile = getenv("VOLUME");

    if ( strncmp(devfile,"/dev",4) == 0 ) {
        Error = hpux_write(sector*fsp->SectorSize,(UCHAR8 *)iobuf,bsize);
    }

    return Error;
} /* end ot_mopwrite */

/********************************************************************/
/*    ot_read                                                       */
/*        routine to read sectors from the optical disk             */
/*                                                                  */
/*    notes:                                                        */
/*        translates volume relative addresses to physical addresse */
/*            on a given section of the volume                      */
/********************************************************************/
INT32 ot_read(struct otfs *fsp,
              UINT32 sector,
              caddr_t iobuf,
              UINT32 *bsize)
{ 
    INT32    error;

    error = ot_mopread(fsp, (struct vnode *) NULL, iobuf,
                       sector + fsp->VolSection[0].baseaddr,
                       (INT32 *)bsize);
    
    return error;
} /* end ot_read */

/********************************************************************/
/*    ot_readCapacity                                               */
/*        routine to get size of disk                               */
/********************************************************************/
INT32 ot_readCapacity( struct otfs *fsp)
{
    INT32 Error=0;
    FILE *handle;
    char *cdevfile = getenv("RVOLUME");
   
    if ( strncmp(cdevfile,"/dev/",5) == 0 ) {
        hpux_readCapacity(cdevfile,&fsp->DiskSectors,&fsp->SectorSize);
    }
    return Error;
} /* end ot_readCapacity */

/********************************************************************/
/*    ot_testready                                                  */
/*        routine to issue a TEST UNIT READY command                */
/********************************************************************/
INT32 ot_testready( struct otfs *fsp, INT32 ScsiId)

{
    INT32 Error=0;
    
    return Error;
} /* end ot_testready */

/********************************************************************/
/*    ot_unlock_cart                                                */
/*        routine to allow cartridge removal                        */
/********************************************************************/
INT32 ot_unlock_cart(struct otfs *fsp, INT32 partition)

{
    return 0;
} /* end ot_unlock_cart */

/********************************************************************/
/*    ot_unlockfs                                                   */
/*        routine to unlock the file system                         */
/********************************************************************/
void ot_unlockfs( register struct otfs *fsp)
{
    return;
} /* end ot_unlockfs */

/********************************************************************/
/*    ot_write                                                      */
/*        routine to write sectors to the optical disk              */
/*                                                                  */
/*    notes:                                                        */
/*        translates volume relative addresses to physical addresses*/
/*        on a given section of the volume                          */
/********************************************************************/
INT32 ot_write(struct otfs *fsp, UINT32 sector, caddr_t iobuf, UINT32 bsize)

{
    INT32    error;

    error = ot_mopwrite(fsp, ( struct vnode *) NULL, iobuf,
                        sector + fsp->VolSection[0].baseaddr,bsize);
    return error;
} /* end ot_write */

/********************************************************************/
/* SetUpOtfs                                                        */
/*                                                                  */
/* routine to create a otfs structure for the first time            */
/* When mop_read,write or test unit ready or read capacity is       */
/* called.    Return a null if unable to open the file.             */
/* The only location where a non-null handle is returned is the     */
/* last statement in the file.                                      */
/*                                                                  */
/********************************************************************/

/********************************************************************/
/* SectorWritten                                                    */
/*                                                                  */
/*        Examines the requested sector and compares against        */
/*        the initialization pattern.  Returns >1 if the pattern    */
/*        is not seen (e.g., something else has been written there) */
/*        Returns -1 error if can't process request.  CAUTION!      */
/*        DO NOT JUST TEST FOR TRUE, it could mean an ERROR!        */
/*                                                                  */
/********************************************************************/

/********************************************************************/
/*    kmem_alloc, kmem_free, and kmem_realloc                       */
/*                                                                  */
/*        Versions for application level use.                       */
/*                                                                  */
/*                                                                  */
/********************************************************************/
void *kmem_alloc(UINT32 size, UINT16 control)
{
    return(malloc((size_t)size));
}

void kmem_free(void *ptr, UINT32 oldSize, UINT16 control)
{
    free(ptr);
}

void *kmem_realloc(void *ptr, UINT32 newSize,
                   UINT32 oldSize, UINT16 control)
{
    /* This may be easily rewritten for systems not providing a */
    /* realloc() */
    return(realloc(ptr,(size_t)newSize));
}

INT16 bzero( char * buff, INT32 bufflen)
{
    INT32 i;
    char *buffpt;

    buffpt = buff;

    for(i=0; i<bufflen; *buffpt++ = 0, i++);

    return(0);
}


/* convert a string to a dstring type */
void str2dstr( dstring * outstring, UCHAR8 dlen, char * instring)
{
    UCHAR8 StringLen, CopyLen;

    bzero( (char *) outstring, dlen);
    
    /* get the length of the input char string */
    StringLen = (UCHAR8) strlen( instring);

    if (StringLen + (UCHAR8) 1    > dlen) {
        CopyLen = dlen - (UCHAR8) 1;
    } else {
        CopyLen = StringLen;
    }
    memcpy((void *) outstring, (void *) instring, CopyLen);
    outstring[dlen-1] = CopyLen;
}



/********************************************************************/
/*  ot_setOtTime                                                    */
/*      routine to get local time and date and return a structure   */
/*      amenable to the disk format                                 */
/********************************************************************/
INT32 ot_setOtTime(struct timestamp *tsp)
{
    
    SetTSTP(tsp->uTypeAndTimeZone,1);
    SetTSTZ(tsp->uTypeAndTimeZone,-7*60); /* 7 hours from GMT */
    tsp->iYear = 1954;
    tsp->uMonth = 10;
    tsp->uDay = 4;
    tsp->uHour = 6;
    tsp->uMinute = 4;
    tsp->uSecond = 0;
    tsp->uCentiseconds = 0;
    tsp->uHundredMicroseconds = 0;
    tsp->uMicroseconds = 0;
    
    return SUCCESS;

} /* end ot_setOtTime */


/********************************************************************/
/*  setFsId                                                         */
/*      set the fsId to some unique tag value                       */
/*                                                                  */
/*  input parameters:                                               */
/*      id:     create a unique volume tag                          */
/*                                                                  */
/*  output parameters:                                              */
/*      id:     uniquely created volume tag                         */
/*                                                                  */
/*  returns:                                                        */
/*      0       success                                             */
/********************************************************************/
INT32 setFsId(UINT32 *id)
{
    /* Of course, it would be more useful here to set the tag to some
     * meaningful value other than zero.  We are simply not using
     * the tag yet.
     */
    *id = (UINT32) 0;

    return 0;
}

/********************************************************************/
/*  makeLocalName                                                   */
/*      routine to transform a directory entry name into one the    */
/*      native system can handle                                    */
/*                                                                  */
/*  input parameters:                                               */
/*      name    the entry name, as it lives on disk                 */
/*      type    the type of structure                               */
/*                                                                  */
/*  output parameters:                                              */
/*      name    contains the transformed name                       */
/*                                                                  */
/*  notes:                                                          */
/*      the name is translated in situ                              */
/********************************************************************/
void makeLocalName(caddr_t name, int     type)
{
} /* end makeLocalName */

/********************************************************************/
/*  otfs_mount                                                      */
/*      VFS routine to mount the file system                        */
/*                                                                  */
/*  input parameters:                                               */
/*      vfsp:       the vfs for the volume                          */
/*      VolName:    the name of the volume to mount                 */
/*      VolTag:     the volume tag (if known)                       */
/*      NumSects:   number of sections in the volume                */
/*      dev:        device file of character device                 */
/*                                                                  */
/*  output parameters:                                              */
/*      vfsp:       vfsp->vfs_data points to private data (the otfs)*/
/*                      for the volume                              */
/*  returns:                                                        */
/*      0           success                                         */
/*      ENOMEM:     memory allocation failure                       */
/********************************************************************/
INT32 otfs_mount(struct vfs  *vfsp,
                 caddr_t     VolName,
                 UINT32      VolTag,
                 INT32       NumSects,
                 char        *dev)
{
    INT32                   err;
    INT32                   error;
    struct otfs             *fs;
    UINT32                  bsize;
    
    err = error = 0;
    
    DMLPRINTF(oMntLev)("otfs_mount: top\n");
    
    /* alloc and clear otfs structure */
    fs = (struct otfs *) kmem_alloc((UINT32) sizeof(struct otfs),0);
    if (!fs)
        return ENOMEM;
    bzero((caddr_t) fs,(INT32) sizeof(struct otfs));
    
    DMLPRINTF(oMntLev)("otfs_mount: after bzero\n");
    
    /* establish otfs and vfs links */
    fs->vfs = vfsp;
    vfsp->vfs_data = (caddr_t) fs;

    if (!error) {
        /* get size of disk */
        DMLPRINTF(oMntLev)("otfs_mount: calling ot_readCapacity()\n");
        error = ot_readCapacity(fs);
        DMLPRINTF(oMntLev)("otfs_mount: ot_readCapacity() returned %d\n",error);
    }

#if HP
    hpux_open();
#endif

    if (!error) {
    
        /* default allocation record type - Extended Allocation Descriptor */
        fs->ADType = ADEXTENDED;

        /* alloc I/O buffers */

        fs->iobuf = kmem_alloc((UINT32) (IOBUFSIZE * sizeof(char)),0);
        fs->dirbuf = kmem_alloc((UINT32) (DIRBUFSIZE * sizeof(char)),0);
    }
    if (!fs->iobuf || !fs->dirbuf)
        error = ENOMEM;
    
    if( !error) {

        /* should be for each section */
        fs->sections = NumSects;

        ot_testready(fs,err);
        DMLPRINTF(oMntLev)("otfs_mount: after ot_testready\n");

    }
    
    if (!error) {
        error = ot_getVolInfo(fs,VolName,VolTag,(INT32) 0);
        DMLPRINTF(oMntLev)("otfs_mount: after ot_getVolInfo\n");
    }


    if (!error && fs->MediaType > (int) fsWORM) {

        DMLPRINTF(oMntLev)("otfs_mount: allocate bitmaps...\n");

        fs->erasedBitmap = (struct bitmap *) kmem_alloc((UINT32) sizeof(struct bitmap),0);
        fs->unerasedBitmap = (struct bitmap *) kmem_alloc((UINT32) sizeof(struct bitmap),0);

        if (!fs->erasedBitmap || !fs->unerasedBitmap)
            error = ENOMEM;

        if (!error) {
            bzero((caddr_t) fs->erasedBitmap,(INT32) sizeof(struct bitmap));
            bzero((caddr_t) fs->unerasedBitmap,(INT32) sizeof(struct bitmap));
            fs->erasedBitmap->bitmapHdr = kmem_alloc(fs->SectorSize,0);
            fs->unerasedBitmap->bitmapHdr = kmem_alloc(fs->SectorSize,0);
            if (!fs->erasedBitmap->bitmapHdr || !fs->unerasedBitmap->bitmapHdr )
                error = ENOMEM;
        }
    }

    DMLPRINTF(oMntLev)("otfs_mount: point 1\n");

    if (!error && fs->MediaType > (int) fsWORM)
        error = ot_countFreeSectors(fs);

    if (!error) {

        if (fs->MediaType <= (int) fsWORM)  /* get highwater mark */

            error = ot_rdHighwaterMark(fs);

        else {                              /* read bitmaps */

            DMLPRINTF(oMntLev)("otfs_mount: read bitmaps\n");

            error = ot_readBitmap( fs, (UINT32) 0, bmapErased, 1);

            if (error == 0 || error == EEOF)
                error = ot_readBitmap( fs, (UINT32) 0, bmapUnerased, 1);

            if (error == EEOF)
                error = 0;
        }
    }
    
    if (!error) {   /* set flag in base record to indicate that it's mounted */
        /* ISSUE - what if mounted read only ?? */
        error = pit_write(fs,INTEGRITY_OPEN);
    }

    if (error) {

        killBitmap(fs,bmapErased);
        killBitmap(fs,bmapUnerased);

        if (fs->iobuf)
            kmem_free(fs->iobuf,(UINT32) IOBUFSIZE,0);
        if (fs->dirbuf)
            kmem_free(fs->dirbuf,(UINT32) DIRBUFSIZE,0);

        binval(fs->VolSection[0].devvp);

        if (fs) {
            kmem_free((caddr_t) fs,(UINT32) sizeof(struct otfs),0);
            vfsp->vfs_data = 0;
        }

    }
    
    if(error) {

        DMLPRINTF(oMntLev)("otfs_mount: fs->VolumeName %s, fs->rootAddr %lx\n",
                         fs->VolumeName, fs->rootAddr);
        DMLPRINTF(oMntLev)("otfs_mount: fs->ICBsize 0x%x\n", fs->ICBsize);

    }
    return error;
} /* end otfs_mount */

/********************************************************************/
/*  otfs_unmount                                                    */
/*      VFS routine to unmount the file system                      */
/*                                                                  */
/*  input parameters:                                               */
/*      vfsp:   the vfs structure for the volume to unmount         */
/*                                                                  */
/*  output parameters:                                              */
/*      vfsp:   vfsp->vfs_data (the otfs) has been flushed & freed  */
/*                  and vfsp->vfs_data == 0                         */
/*  returns:                                                        */
/*      0       success                                             */
/********************************************************************/
INT32 otfs_unmount(struct vfs *vfsp)
{
    INT32                   error = 0;
    register struct otfs    *fsp = (struct otfs *) vfsp->vfs_data;
   
    /* ISSUE - what if mounted read only ?? */
    error = pit_write(fsp,INTEGRITY_CLOSE);
    
    /* flush and prepare for complete unmount */

    if (!error)
        error = ot_takeOffline(fsp);
    
    if (!error) {

        /* invalidate buffers and free structures */

        binval(fsp->VolSection[0].devvp);
        kmem_free((caddr_t) fsp,(UINT32) sizeof(struct otfs),0);
        vfsp->vfs_data = 0;
    }

#if HP
    hpux_close();
#endif

    return error;
} /* end otfs_unmount */

/********************************************************************/
/*  otvn_bmap                                                       */
/*      VFS routine to premap blocks for the paging system          */
/********************************************************************/
INT32 otvn_bmap(struct vnode    *vp,
                INT32           lbn,
                struct vnode    **vpp,
                INT32           *bnp)
{
    return EINVAL;
} /* end otvn_bmap */

/********************************************************************/
/*  otvn_link                                                       */
/*      VFS routine to link a file                                  */
/********************************************************************/
INT32 otvn_link(struct vnode *vp,
                struct vnode *dvp,
                caddr_t       name,
                struct ucred *cred)
{
    return EMLINK;  /* HP implementation doesn't support links yet */
} /* end otvn_link */

/********************************************************************/
/*  otvn_readlink                                                   */
/*      VFS routine to read a symbolic link                         */
/********************************************************************/
INT32 otvn_readlink(struct vnode *vp,
                    struct uio   *uiop,
                    struct ucred *cred)
{
    return EINVAL;   /* HP implementation doesn't support links yet */
} /* end otvn_readlink */

/********************************************************************/
/*  otvn_strategy                                                   */
/*      VFS routine to read blocks into memory for the paging system*/
/********************************************************************/
INT32 otvn_strategy(register struct buf *bp)
{
    return EINVAL;   /* HP implementation doesn't support executability yet */
} /* end otvn_strategy */

/********************************************************************/
/*  otvn_symlink                                                    */
/*      VFS routine to symbolically link a file                     */
/********************************************************************/
INT32 otvn_symlink(struct vnode *vp,
                   caddr_t       lnm,
                   struct vattr *vap,
                   caddr_t       tnm,
                   struct ucred *cred)
{
    return EINVAL;   /* HP implementation doesn't support links yet */
} /* end otvn_symlink */


/********************************************************************/
/*  ot_getAttributes                                                */
/*      routine to get operating system specific attributes         */
/********************************************************************/
INT32 ot_getAttributes(register struct otnode  *otp,
                       register struct vattr   *vap)
{
    struct vnode            *vp = &otp->vnode;
    register INT32          perm;
    
    vap->va_uid = 0;    /* default */
    vap->va_gid = 0;
    perm = (otp->attr & OTA_RDONLY) ? 0555 : 0777;
    
    if (vp->v_type == VDIR)
        vap->va_mode = S_IFDIR | perm;
    else if (vp->v_type == VLNK)
        vap->va_mode = S_IFLNK | perm;
    else if (vp->v_type == VREG)
        vap->va_mode = S_IFREG | perm;
    
    vap->va_nlink = otp->icbe.fe.LinkCount;

    ot_getTimeAndDate(otp,vap);

    return 0;
} /* end ot_getAttributes */

/********************************************************************/
/*  ot_getVolInfo()                                                 */
/*      routine to get volume information                           */
/********************************************************************/
INT32 ot_getVolInfo(register struct otfs    *fsp,
                    caddr_t                 VolumeName,
                    UINT32                  VolumeTag,
                    UINT32                  Section)
{
    INT16               found = 0;
    struct FindVDSFlags sFindVDSFlags;
    struct ReadVDSFlags sReadVDSFlags;
    struct extent_ad    sPrimaryAdr, sReserveAdr;
    struct extent_ad    sPVDAdr, sPDAdr, sLVDAdr, sUSDAdr, sIUDAdr;
    struct LogVolDesc   lvd;
    struct FileSetDesc  fsd;
    struct PartDesc     pd; 
    struct PartHeaderDesc *PHDp;
    struct FileEntry    *root;
    UINT32              fsd_loc;
    INT32               error = 0;
    UINT32              WrittenVDSLen;
    UINT32              bsize = fsp->SectorSize;
  
    /* 
    * Walk through the volume descriptors so can get info for fsp.
    */
    if ( FindVDS(fsp,&sFindVDSFlags,&sPrimaryAdr,&sReserveAdr) == SUCCESS ) {
        if (ReadVDS(fsp,sPrimaryAdr,&sReadVDSFlags,
                         &sPVDAdr, &sPDAdr, &sLVDAdr, &sUSDAdr, &sIUDAdr,
                         &WrittenVDSLen) == SUCCESS ) {
            if ( sReadVDSFlags.uPDValid ) {
                if ( ! error ) 
                    error = ot_read(fsp,sPDAdr.Location,(caddr_t)fsp->iobuf, &bsize);
                if ( ! error )  {
                    memmove(&pd,fsp->iobuf,sizeof(struct PartDesc));
#if !IBM
                    ot_xlate(TAGID_PD, FROM_OP, (caddr_t) &pd,0,0);
#endif
                    error = ot_goodTag(TAGID_PD,&pd,sPDAdr.Location);
                }
                if ( ! error ) {
                    PHDp = (struct PartHeaderDesc *) &pd.aPartContentsUse;
                }
            }
            if ( !error && sReadVDSFlags.uLVDValid ) {
                /* read the lvd to get location of fsd */
                error = ot_read(fsp,sLVDAdr.Location,(caddr_t)fsp->iobuf, &bsize);
                if ( ! error ) {
                    memmove(&lvd,fsp->iobuf,sizeof(struct LogVolDesc));
#if !IBM
                    ot_xlate(TAGID_LVD, FROM_OP, (caddr_t) &lvd,0,0);
#endif
                    error = ot_goodTag(TAGID_LVD, &lvd, sLVDAdr.Location);
                }
                if ( ! error ) 
                    fsd_loc = ((struct long_AD *) (lvd.uLogVolUse))->Location_LBN;
            }
        }
    }
    if ( error ) {
        return error;
    }
    
    /*
    * Now that all the reads are done, set the baseaddr. This
    * will affect all subsequenct ot_read(),ot_write() addresses.
    */
    fsp->VolSection[Section].baseaddr = pd.uPartStartingLoc; 
    
    /*
    * read the fsd to get location of root dir ICB
    */
    if ( ! error ) 
        error = ot_read(fsp,fsd_loc,(caddr_t) fsp->iobuf, &bsize);
    if ( ! error ) {
        memmove(&fsd,fsp->iobuf,sizeof(struct FileSetDesc));
#if !IBM
        ot_xlate(TAGID_FSD, FROM_OP, (caddr_t) &fsd,0,0);
#endif
        error = ot_goodTag(TAGID_FSD, &fsd, fsd_loc);
    }
    if ( ! error ) {
        fsp->rootAddr = fsd.sRootDirICB.Location_LBN;
        /*
        * The ICBsize is taken from the Strategy Parameter field in
        * the root dir icb. We assume that all icbs on the volume 
        * will be of this same size.
        *
        * A more robust implementation would dynamically adjust the
        * size on the fly, so each entry would have its own size.
        */
        error = ot_read(fsp,fsp->rootAddr,(caddr_t) fsp->iobuf, &bsize);
        if ( !error ) {
#if !IBM
            ot_xlate(TAGID_FILE_ENTRY, FROM_OP,
                             (caddr_t)fsp->iobuf,fsp->SectorSize,0);
#endif
            error = ot_goodTag(TAGID_FILE_ENTRY,fsp->iobuf,fsp->rootAddr);
        }
        if ( ! error ) {
            root = (struct FileEntry *)fsp->iobuf;
            fsp->ICBsize =  root->sICBTag.NumberEntries;
            fsp->IndirectSize = root->sICBTag.NumberEntries
                               - *(UINT16 *)&(root->sICBTag.StrategyParm);
        }
        found = 1;
    }
    
    if ( ! error ) {
        fsp->pitAddr = PHDp->PIT.Location;
        switch (pd.uAccessType) {
        case 1: return EROFS; 
        case 2: fsp->MediaType = fsWORM;
            fsp->erasedAddr = PHDp->UST.Location;
            fsp->unerasedAddr = PHDp->FST.Location;
            break;
        case 3: fsp->MediaType = fsMO;
            fsp->erasedAddr = PHDp->USB.Location;
            fsp->unerasedAddr = PHDp->FSB.Location;
            break;
        default: return ERANGE;
        }
        
        /* 
        * Calculate volume size. This is used in statfs to indicate the size
        * of the logical volume. To calculate correctly, add up the partition
        * spaces of each volume in the volume set.
        */
        fsp->VolSection[Section].volsiz = pd.uPartLength;
        
        /*
        * Copy volume name from lvd.
        */
        memmove((void *)fsp->VolumeName,(void *)lvd.uLogVolID,FILENAMESIZE);
    }

    DMLPRINTF(oSysLev)("ot_getVolInfo: returning error %d\n",error==0 && !found ? ENOENT : error);

    return error == 0 && !found ? ENOENT : error;

} /* end ot_getVolInfo */


/********************************************************************/
/*  ot_getTimeval                                                   */
/*      routine to translate NSR timestamp to the VFS time struct-  */
/*      ure, timeval.                                               */
/*                                                                  */
/*  input:                                                          */
/*      tsp - a pointer to a struct timestamp that contains the     */
/*            NSR structure to be converted.                        */
/*                                                                  */
/*      tvp - a pointer to a vfs struct timeval that will receive   */
/*            the converted data.                                   */
/*                                                                  */
/*  output:                                                         */
/*      SUCCESS (always)                                            */
/*                                                                  */
/*  notes:                                                          */
/*      if the incoming tsp structure has a month value of zero,    */
/*      ot_getTimeval assumes that the value in tsp is not speci-   */
/*      fied, and zero is returned in tvp.                          */
/*                                                                  */
/*      ot_getTimeval returns the time in the same zone as the      */
/*      YYMMDDHHMMSS specified in the timestamp.  The Type and      */
/*      TimeZone fields are NOT used to supply a correction to      */
/*      to local time.  This could be easily done with a third      */
/*      parameter called convert that when TRUE, would add or       */
/*      subtract from the final value depending on the value of     */
/*      the iTimeZone field in the timestamp.                       */
/*                                                                  */
/********************************************************************/
INT32 ot_getTimeval(struct timestamp *tsp, struct timeval *tvp)
{
    register INT32      year;
    UINT32              second;
    UINT32              month;
    UINT32              NumDays = 0;
    
    
    /* if time not specified, then set to ZERO */
    /* presumes that month must be a non-zero value */
    
    if( tsp->uMonth == 0) {

        tvp->tv_sec = 0;
        tvp->tv_usec = 0;
        
    } else {
        
        /* turn of century is not a leap year */
        
        for (year=1970 ; year < (INT32)tsp->iYear; year++)  {
            NumDays += ((year % 4) || (year % 100) == 0) ? 365 : 366;
        }
        
        mon[2] = (((INT32) tsp->iYear % 4) || ((INT32) tsp->iYear % 100) == 0) ? 28 : 29;

        for (month=1; month < (INT32)tsp->uMonth; month++) {
            NumDays += mon[month];
        }
        
        NumDays += (INT32) tsp->uDay - 1;
        second = tsp->uSecond;
        second+=(NumDays * 24 * 60 * 60) +
          ((INT32) tsp->uHour * 60 * 60) +
            ((INT32) tsp->uMinute * 60); 

        tvp->tv_sec  = second;
        tvp->tv_usec = tsp->uCentiseconds * 10000 + tsp->uHundredMicroseconds * 100 +
          tsp->uMicroseconds;

        /* could do a conversion to local time here if desired */
    }
    return SUCCESS;
}    

/********************************************************************/
/*  ot_getTimeAndDate                                               */
/*      routine to translate the time and date values in an otnode  */
/*      (which are Access, Modify and Attribute times) and place    */
/*      them in a Vnode attribute structure (struct vattr).         */
/*                                                                  */
/*  input                                                           */
/*      otp - pointer to an otnode with the values to convert       */
/*      vap - pointer to a struct vattr that receive the convert-   */
/*            ed values.                                            */
/*                                                                  */
/*  output                                                          */
/*      returns SUCCESS always                                      */
/*                                                                  */
/*  notes                                                           */
/*      vattr value   gets the following  otnode.icbe               */
/*       va_atime                         sAccessTime               */
/*       va_mtime                         sModifyTime               */
/*       va_ctime                         sModifyTime               */
/*                                                                  */
/*                                                                  */
/********************************************************************/
INT32 ot_getTimeAndDate(register struct otnode  *otp,
                        register struct vattr   *vap)
{

    ot_getTimeval(&(otp->icbe.fe.sAccessTime), &(vap->va_atime));
    ot_getTimeval(&(otp->icbe.fe.sModifyTime), &(vap->va_mtime));
    ot_getTimeval(&(otp->icbe.fe.sModifyTime), &(vap->va_ctime));

    return SUCCESS;

} /* end ot_getTimeAndDate */

/********************************************************************/
/*  ot_memCompare                                                   */
/*      routine to compare two buffers                              */
/********************************************************************/
INT32 ot_memCompare(
    register caddr_t buffer1,
    register caddr_t buffer2,
    INT32            length)
{
    caddr_t end = buffer1 + length;
    
    while (buffer1 < end && *buffer1 == *buffer2) {
        buffer1++;
        buffer2++;
    }
    return buffer1 >= end;
} /* end ot_memCompare */

/********************************************************************/
/*  ot_setAttributes                                                */
/*      routine to set operating system specific attributes         */
/********************************************************************/
INT32 ot_setAttributes(register struct otnode *otp,
                       register struct vattr  *vap)
{
    
    if (vap->va_mode != (UINT16) -1) {
        if ((vap->va_mode & 0222) == 0) {
            otp->attr |= OTA_RDONLY;
            otp->icbe.fe.Permissions = 0555;
        } else {
            otp->attr &= ~OTA_RDONLY;
            otp->icbe.fe.Permissions = 0777;
        }
    }
    
    if (vap->va_uid == -1) {
        vap->va_uid = 0;    /* default to current user, what is that */
        vap->va_gid = 0;    /* under HP-UX? */
    } else {
        otp->icbe.fe.UID = (UINT32) vap->va_uid;
        otp->icbe.fe.GID = (UINT32) vap->va_gid;
    }
    
    ot_setTimeAndDate( otp);

    otp->flags |= OT_MOD;

    return 0; 
} /* end ot_setAttributes */



/********************************************************************/
/*  ot_setSystemArea                                                */
/*      fill the system area in an icb with implementation ud       */
/*      specific data, regardless of creator                        */
/*                                                                  */
/*  notes:                                                          */
/*      note that the otnode is not marked as modified by this      */
/*      routine. this guarantees that the system area will belong   */
/*      to the implementation as long as the icb is in memory.      */
/*      if the file is modified, then other routines will set the   */
/*      modified flag, forcing the icb to be written to disk.       */
/********************************************************************/
void ot_setSystemArea(struct otnode *ot)
{
    
    /* just set default permissions.  An implementation may have
       to make this specific for the permission and user identification
       of a specific system */

    ot->icbe.fe.Permissions = 0555;
    ot->icbe.fe.UID = (UINT32) 0;
    ot->icbe.fe.GID = (UINT32) 0;
    strcpy((char *)ot->icbe.fe.sImpID.aRegisteredID, SFS_FE_REGID);
    /* set flag to not allow a change to the regid */
    ot->icbe.fe.sImpID.uFlags = BITONE;
    
} /* end ot_setSystemArea */


/********************************************************************/
/*  ot_setTimeAndDate                                               */
/*      routine to set the time and date fields in an otnode.       */
/*      NOTE that this replaces both ot_setTime() and ot_setDate()  */
/*      NOTE ALSO that this does NOT allow the host computer to set */
/*      the time and date from his vattr structure, though this is  */
/*      allowed in the vnode interface...             -DVM          */
/********************************************************************/
INT32 ot_setTimeAndDate(register struct otnode  *otp)
{
    /* ever been set? */

    if (otp->icbe.fe.sModifyTime.uMonth == 0) {
        ot_setOtTime(&(otp->icbe.fe.sModifyTime));
    }
    ot_setOtTime(&(otp->icbe.fe.sAttrTime));
    ot_setOtTime(&(otp->icbe.fe.sAccessTime));

    return 0;
} /* end ot_setTimeAndDate */

/********************************************************************/
/*  ot_takeOffline                                                  */
/*      routine to do fundamentals of unmount without freeing the   */
/*      fs structures                                               */
/*                                                                  */
/********************************************************************/
INT32 ot_takeOffline(register struct otfs *fs)
{
    struct vfs              *vfsp = fs->vfs;
    register INT32          error;
    register struct otnode  *inActive,
    *next;
    
    /* check for free fs */

    if (error = ot_busy(fs))
        return EBUSY;
    
    /* sync the file system */

    if (error = otfs_sync(vfsp))
        return EBUSY;
    
    /* free I/O buffers */

    kmem_free(fs->iobuf,(UINT32) IOBUFSIZE,0);
    kmem_free(fs->dirbuf,(UINT32) DIRBUFSIZE,0);
    
    if (fs->MediaType <= (int) fsWORM) {

        kmem_free((caddr_t) fs->highWater,(UINT32) sizeof(struct space),0);
        fs->highWater = (struct space *) 0;

    } else {

        killBitmap(fs,bmapErased);
        killBitmap(fs,bmapUnerased);

    }
    
    /* free the inactive otnode cache */

    for (inActive = fs->inactive; inActive; inActive = next) {
        next = inActive->next;
        ot_dequeueOTNode(&fs->inactive,inActive);
        kmem_free((caddr_t) inActive, (UINT32) sizeof(struct otnode) - 
                  sizeof(union icbe) + fs->SectorSize, 0);
    }
    
    return 0;
} /* end ot_takeOffline */

/********************************************************************/
/*  binval - invalidate Unix buffers                                */
/********************************************************************/
void binval(struct vnode *devvp)
{
} /* end binval */

INT32 ot_clearContext(enum ctextType  clearMode,
                      struct otfs    *fsp,
                      struct otnode  *otp)
{
    return 0;
} /* end clearContext */

INT32 ot_bflush(struct otfs *fsp)
{
    return (0);
}


/********************************************************************/
/* uiomove - move data from/to user input/output structure          */
/*                                                                  */
/* input parameters                                                 */
/*     buf        where the data is, or where it is to go           */
/*     len        the amount of data to transfer                    */
/*     rw         UIO_READ or UIO_WRITE                             */
/*     uiop       a pointer to the uio structure                    */
/*                                                                  */
/* output parameters                                                */
/*     uiop->uio_iov->iov_base and uiop->uio_offset have been       */
/*        incremented by the number of bytes moved                  */
/*     uiop->uio_resid have been decremented by the number of       */
/*        bytes moved.                                              */
/*                                                                  */
/* notes                                                            */
/*     uiomove is dumb.  No range checking or overflow checking     */
/*        is done.                                                  */
/*                                                                  */
/* Modifcation History                                              */
/*     04Oct92  ECB  Re-entered after mysterious loss               */
/*                                                                  */
/*                                                                  */
/*                                                                  */
/*                                                                  */
/********************************************************************/
void uiomove(caddr_t buf, INT32 len, enum uio_rw rw, struct uio *uiop)
{
    if( rw == UIO_READ) {
        memcpy( uiop->uio_iov->iov_base, buf, len);
    } else {
        memcpy( buf, uiop->uio_iov->iov_base, len);
    }

    uiop->uio_offset += len;
    uiop->uio_resid -= len;
    uiop->uio_iov->iov_base += len;
}


/********************************************************************/
/* pow - integer exponentiation calc x^^y                           */
/*                                                                  */
/* input parameters                                                 */
/*     x     base to be raised to power y                           */
/*     y     power y                                                */
/*                                                                  */
/* returns                                                          */
/*     UIN32 value, 1 if y is 0                                     */
/*                                                                  */
/* notes                                                            */
/*     both x and y must be >=0                                     */
/*     behavior on overflow is undefined (arguments are expected    */
/*         to be small.)                                            */
/*     lifted from Kernigan and Ritchie Second edition page 25      */
/*                                                                  */
/*                                                                  */
/********************************************************************/
UINT32 pow( int x, int y)
{
    int i, p;

    p=1;

    for( i=1; i<= y; ++i)
       p = p*x;

    return p;
}

INT32 pit_write(struct otfs *fs,UCHAR8 newtype)
{
    UINT32 error;
    struct PartIntegrityEntry *pit;
    struct otnode *pit_otnode;
    
    /*
    * Read the old one just to get the tags, etc.
    */
    error = ot_getOTNode(0,fs, NULL,fs->pitAddr,LAST,VREG,
                         NULL,&pit_otnode);
    if ( !error )
        pit = &pit_otnode->icbe.pie;
    
    /*
    * Set the IntegrityType and update the recording time.
    */
    if ( ! error ) {
        pit->IntegrityType = newtype;
        ot_setOtTime(&pit->sRecordingTime);
    }
    if ( ! error ) {
        pit_otnode->flags |= OT_MOD;
        error = ot_ICBRdWr(fs,pit_otnode,UIO_WRITE,LAST);
    }
    if ( pit_otnode ) {
        pit_otnode->vnode.v_count--;
        otvn_inactive(&pit_otnode->vnode,NULL);
    }
    return error;
}



#if HP
INT32 hpux_read(int offset,UCHAR8 *iobuf,int bsize)
{
    int count;
    int length;
    
    if (lseek(hpux_bfd, (off_t)offset, SEEK_SET) < 0) {
        perror("hpux_read,lseek");
        exit(0);
    }
    count = read(hpux_bfd, (void *)iobuf, bsize);
    if (count == -1) {
        /* try the character device - may be at end of disk */
        if (lseek(hpux_cfd, (off_t)offset, SEEK_SET) < 0) {
            perror("hpux_read,lseek");
            exit(0);
        }
        count = read(hpux_cfd, (void *)iobuf, bsize);
        if (count == -1) {
            perror("hpux_read,read");
            exit(0);
        }
    }
    else if (count < bsize) {
        printf("WARNING / only read %d bytes, asked for %d\n", count, bsize);
    }
    return 0;
}

INT32 hpux_write(int offset,UCHAR8 *iobuf,int bsize)
{
    int count;
    int length;
    
    if (lseek(hpux_bfd, (off_t)offset, SEEK_SET) < 0) {
        perror("hpux_write,lseek");
        exit(0);
    }
    count = write(hpux_bfd, (void *)iobuf, bsize);
    if (count == -1) {
        perror("hpux_write,write");
        exit(0);
    }
    else if (count < bsize) {
        printf("WARNING / only read %d bytes, asked for %d\n", count, bsize);
    }
    return 0;
}

INT32 hpux_readCapacity(char *devfile, UINT32 *DiskSectors, UINT32 *sectorSize)
{
    int error;
    int fd;
    int flag = 1;
    int ret;
    unsigned char buf[8];
    static const struct scsi_cmd_parms read_cap = {
      10,      /* 10 bytes in cdb                       */
      1,       /* 1 -> select with ATN                  */
      100,     /* clock ticks (50 ticks = 1 sec)        */
      0x25,    /* cdb[0] read capacity                  */
      0x00,    /* cdb[1] lun = 0; reserved = 0          */
      0x00,    /* cdb[2] lba = 0                        */
      0x00,    /* cdb[3] lba = 0                        */
      0x00,    /* cdb[4] lba = 0                        */
      0x00,    /* cdb[5] lba = 0                        */
      0x00,    /* cdb[6] reserved = 0                   */
      0x00,    /* cdb[7] reserved = 0                   */
      0x00,    /* cdb[8] reserved = 0                   */
      0x00     /* cdb[9] reserved,flag,link = 0         */
    };
        
    /* open and setup device */
    if ((fd=open(devfile, 2))<0) {
        perror("hpux_readCapacity,open");
        exit(1);
    }
   
    /* Sets device up in command mode: must be super-user */
    if ( (ret = ioctl(fd,SIOC_CMD_MODE,&flag) ) <0 ) {
        perror("hpux_readCapacity,ioctl");
        exit(1);
    }
    /* Send command */
    if ( ioctl(fd,SIOC_SET_CMD,&read_cap) < 0 ) {
        perror("hpux_readCapacity,SET_CMD");
        exit(1);
    }
    if ( read(fd,buf,8) != 8 ) {
        perror("hpux_readCapacity");
        exit(1);
    }

    /* Take driver out of command mode. */
    if ((error=ioctl(fd,SIOC_CMD_MODE,0)) != 0)  {
        perror("hpux_readCapacity,SIOC_CMD");
        exit(1);
    }

    /* Close device */
    close(fd);

    *DiskSectors = buf[3]+(buf[2]<<8)+(buf[1]<<16)+(buf[0]<<24);
    *sectorSize = buf[7]+(buf[6]<<8)+(buf[5]<<16)+(buf[4]<<24);
    return 0;
}

INT32 hpux_open()
{
    char *bdev = getenv("VOLUME");
    char *cdev = getenv("RVOLUME");

    hpux_bfd = open(bdev,O_RDWR);
    if ( hpux_bfd < 0 ) {
        perror("bdev open");
        exit(0);
    }
    hpux_cfd = open(cdev,O_RDWR);
    if ( hpux_cfd < 0 ) {
        perror("cdev open");
        exit(0);
    }
    return 0;
}

INT32 hpux_close()
{
    close(hpux_bfd);
    close(hpux_cfd);
    return 0;
}
#endif /* HP */
