/********************************************************************/
/*  adTest.c - Tests the Allocation Descriptor stuff                */
/*                                                                  */
/*      Unpublished Confidential Information of Hewlett-Packard     */
/*      Company. Do not disclose.  Copyright (c) Hewlett-Packard    */
/*      Company 1996. All rights reserved                           */
/*                                                                  */
/* The contents of this software are proprietary and confidential   */
/* the Hewlett- Packard Company, and are limited in distribution to */
/* those with a direct need to know.  Individuals having access to  */
/* this software are responsible for maintaining the                */
/* confidentiality of the content and for keeping the software      */
/* secure when not in use.  Transfer to any party is strictly       */
/* forbidden other than as expressly permitted in writing by        */
/* Hewlett-Packard Company.  Unauthorized transfer to or possession */
/* by any unauthorized party may be a criminal offense.             */
/*                                                                  */
/*                   RESTRICTED RIGHTS LEGEND                       */
/*                                                                  */
/*      Use,  duplication,  or disclosure by the                    */
/*      U. S. Government is subject to restrictions as set forth    */
/*      in subparagraph (c) (1) (ii) in the Rights in Technical     */
/*      Data and Computer Software clause in DFARS 252.227-7013     */
/*      or any other successor clause.                              */
/*                                                                  */
/*                       HEWLETT-PACKARD COMPANY                    */
/*                       3000 Hanover St.                           */
/*                       Palo Alto, CA  94304 U.S.A.                */
/*                                                                  */
/*      Rights for non-DOD U.S. Government Departments and Agencies */
/*      are as set forth in FAR 52.227-19 (c) (1,2) or any other    */
/*      successor clause.                                           */
/*                                                                  */
/********************************************************************/
/*
 $Source: /projects/HMS/nsr/udfOld/src/nsrTools/adtest.c,v $
 $Revision: 1.7 $       $Author: hmgr $
 $State: Exp $          $Locker:  $
 $Date: 96/05/30 14:53:10 $
 ********************************************************************/

#include "def.h"
#include "o_vfsops.h"
#include "vnode.h"
#include "nsr.h"
#include "o_proto.h"
/* #include "magvol.h" */
#include "nsrerrno.h"
#include "Debug.h"
#include "o_node.h"
#include "makefs.h"

#define TIOBUFSIZE       32768
#define EXTSIZE          8192
#define MAXSIZE          (1024*1024*2)
#define MAXEXTS          12
#define MANYEXTS         200
#define LONGSIZE         (64*1024)
#define FILELEN          (EXTSIZE*MAXEXTS)
#define LONGCOUNT        (FILELEN / LONGSIZE + (FILELEN % LONGSIZE ? 1 : 0))
#define SMALLLEN         (501)

main()
{
    char *bufp;                 /* for misc use */
    char *fbufp;                /* for writing into a file */
    char *vbufp;                /* for verifying file contents */
    char *lbufp;                /* for long reads */
    
    struct vfs *vfsp;
    struct vnode *vnp;
    struct vnode *rvnp;   /* root vnode ptr */
    struct vnode *dvnp;
    struct vnode *fvp;
    struct vnode *favp;
    
    struct vattr rvas;
    struct vattr dvas;
    struct vattr vas;
    struct ucred ucreds;
    struct iovec iovecs;
    struct uio uios;
    
    INT32 error;
    INT32 i;                    /* temp */
    
    /* Set the debug levels */

    /* debugLevel = -45;                 ADs only */
    /* debugMask  = -1;          Any and all areas */

    /* Zero out the attributes. */
    memset(&vas,0,sizeof(struct vattr));
    
    vas.va_fsid          = -1;  /* not settable */
    vas.va_nodeid        = -1;  /* not settable */
    vas.va_nlink         = -1;  /* not settable */
    vas.va_blocksize     = -1;  /* not settable */
    vas.va_rdev          = -1;  /* not settable */
    vas.va_blocks        = -1;  /* not settable */
    vas.va_type          = VREG;
    
    dvas = vas;
    dvas.va_size          = -1; /* Let's not set it... */
    dvas.va_type=VDIR;
    
    /* allocate a g.p. buffer */
    bufp= (char *) kmem_alloc( TIOBUFSIZE, 0);
    bzero( bufp, TIOBUFSIZE);
    
    /* try doing a mount */
    
    /* allocate a vfs and otfs structure */
    vfsp = (struct vfs *) kmem_alloc( sizeof( struct vfs), 0);
    
    error = otfs_mount( vfsp, "TEST", 1, 1, "DUMMY");
    
    /* get the root */
    if (!error) 
        error = otfs_root( vfsp, &rvnp, 0);
    
    if ( !error ) 
        error = otvn_getattr(rvnp,&rvas,&ucreds);
    
    vnp = rvnp;
    
    /* read the root directory */
    
    printf("Reading / directory.\n");
    
    bzero( (caddr_t) &uios, (INT32) sizeof( struct uio));
    uios.uio_iovcnt = 1;
    uios.uio_offset = 0;
    uios.uio_resid = TIOBUFSIZE;
    uios.uio_iov = &iovecs;
    iovecs.iov_base = bufp;
    iovecs.iov_len = TIOBUFSIZE;
    
    uioInit( &uios, &iovecs, bufp, 0, TIOBUFSIZE);
    if(!error) 
        error = otvn_readdir( vnp, &uios, &ucreds);
    
 testWrite:
    if (!error ) {
        printf("Now beginning file Write test\n");
        
        /* allocate a file buffer */
        printf("  - Allocating buffers (size = %d)...",MAXSIZE);
        fbufp= (char *) kmem_alloc( MAXSIZE, 0);
        vbufp= (char *) kmem_alloc( MAXSIZE, 0);
        if (!fbufp) {
            error = ENOMEM;
            printf("done. error = %d.\n",error);
        } else {
            printf("done. SUCCESS!\n");
        }
        
        if (!error) {
            /* Filling buffer with pattern */
            printf("  - Filling buffer with pattern...");
            for (i=0 ; i <= MAXSIZE; i++) {
                fbufp[i] = i & 0xff;
            }
            printf("done.\n");
        }
        
        if (!error) {
            /* Creating file "TESTDATA" in Root directory. */
            printf("  - Creating file /TESTDATA ...");
            error = otvn_create(rvnp, "TESTDATA", &vas, NONEXCL, VWRITE, &fvp, &favp, &ucreds, 0);
            printf("done. (error = %d)\n",error);
        }
        
        if (!error) {
            /* Writing pieces to file */
            printf("  - Writing Extents to /TESTDATA ");
            for (i=0; !error && i <= MAXEXTS; i++) {
                uioInit( &uios, &iovecs, fbufp, i*EXTSIZE, EXTSIZE);
                printf (".");
                error = otvn_rdwr( fvp, &uios, UIO_WRITE, IO_APPEND, &ucreds);
            }
            if (error) {
                printf("done. error = %d\n");
            } else {
                printf("done. SUCCESS!\n");
            }
        }
        
        if (!error) {
            /* syncing, inactivating */
            printf("  - Inactivating vnode ... ");
            fvp->v_count--;
            otvn_inactive(fvp, &ucreds);
            if (error) {
                printf("done. error = %d\n");
            } else {
                printf("done. SUCCESS!\n");
            }
        }


        if (!error) {
            /* Creating file "TESTSMALL" in Root directory. */
            printf("  - Creating file /TESTSMALL ...");
            error = otvn_create(rvnp, "TESTSMALL", &vas, NONEXCL, VWRITE, &fvp, &favp, &ucreds, 0);
            printf("done. (error = %d)\n",error);
        }
        
        if (!error) {
            /* Writing to file */
            printf("  - Writing partial extent to /TESTSMALL .");
            uioInit( &uios, &iovecs, fbufp, 0, SMALLLEN);
            error = otvn_rdwr( fvp, &uios, UIO_WRITE, IO_APPEND, &ucreds);
            if (error) {
                printf("done. error = %d\n");
            } else {
                printf("done. SUCCESS!\n");
            }
        }
        
        if (!error) {
            /* syncing, inactivating */
            printf("  - Inactivating vnode ... ");
            fvp->v_count--;
            otvn_inactive(fvp, &ucreds);
            if (error) {
                printf("done. error = %d\n");
            } else {
                printf("done. SUCCESS!\n");
            }
        }


        if (!error) {
            /* Creating file "TESTMANY" in Root directory. */
            printf("  - Creating file /TESTMANY ...");
            error = otvn_create(rvnp, "TESTMANY", &vas, NONEXCL, VWRITE, &fvp, &favp, &ucreds, 0);
            printf("done. (error = %d)\n",error);
        }
        
        if (!error) {
            /* Writing pieces to file */
            printf("  - Writing Extents to /TESTMANY ");
            for (i=0; !error && i <= MANYEXTS; i++) {
                uioInit( &uios, &iovecs, fbufp, i*EXTSIZE, EXTSIZE);
                printf (".");
                error = otvn_rdwr( fvp, &uios, UIO_WRITE, IO_APPEND, &ucreds);
            }
            if (error) {
                printf("done. error = %d\n");
            } else {
                printf("done. SUCCESS!\n");
            }
        }
        
        if (!error) {
            /* syncing, inactivating */
            printf("  - Inactivating vnode ... ");
            fvp->v_count--;
            otvn_inactive(fvp, &ucreds);
            if (error) {
                printf("done. error = %d\n");
            } else {
                printf("done. SUCCESS!\n");
            }
        }

        if (!error) {
            /* Creating file "TESTBIG" in Root directory. */
            printf("  - Creating file /TESTBIG ...");
            error = otvn_create(rvnp, "TESTBIG", &vas, NONEXCL, VWRITE, &fvp, &favp, &ucreds, 0);
            printf("done. (error = %d)\n",error);
        }
        if (!error) {
            /* Writing pieces to file */
            printf("  - Writing /TESTBIG (all at once) ");
            uioInit( &uios, &iovecs, fbufp, 0, MAXSIZE);
            error = otvn_rdwr( fvp, &uios, UIO_WRITE, IO_APPEND, &ucreds);
            if (error) {
                printf("done. error = %d\n");
            } else {
                printf("done. SUCCESS!\n");
            }
        }
        if (!error) {
            /* syncing, inactivating */
            printf("  - Inactivating vnode ... ");
            fvp->v_count--;
            otvn_inactive(fvp, &ucreds);
            if (error) {
                printf("done. error = %d\n");
            } else {
                printf("done. SUCCESS!\n");
            }
        }
    }
    
 testMkdir:
    
    if (!error) {
        
        /*create a directory in the root */
        
        printf("Creating DIRECTORY1 through DIRECTORYPAD4.\n");
        
        if( !error)
            error = otvn_mkdir( rvnp, "DIRECTORY1", &dvas, &dvnp, &ucreds);
        if( !error)
            error = otvn_mkdir( rvnp, "DIRECTORYP2", &dvas, &dvnp, &ucreds);
        if( !error)
            error = otvn_mkdir( rvnp, "DIRECTORYPA3", &dvas, &dvnp, &ucreds);
        if( !error)
            error = otvn_mkdir( rvnp, "DIRECTORYPAD4", &dvas, &dvnp, &ucreds);
        
        printf("Reading root directory.\n");
        
        uioInit( &uios, &iovecs, bufp, 0, TIOBUFSIZE);
        if ( !error )
            error = otvn_readdir( rvnp, &uios, &ucreds);
        if ( !error ) {
            /* read again to see if nothing gets returned */
            uios.uio_resid = TIOBUFSIZE;
            error = otvn_readdir( rvnp, &uios, &ucreds);
        }
        
        printf("Creating DIRECTORY5 through DIRECTORY10.\n");
        
        if( !error)
            error = otvn_mkdir( rvnp, "DIRECTORY5", &dvas, &dvnp, &ucreds);
        if( !error)
            error = otvn_mkdir( rvnp, "DIRECTORYP6", &dvas, &dvnp, &ucreds);
        if( !error)
            error = otvn_mkdir( rvnp, "DIRECTORYPA7", &dvas, &dvnp, &ucreds);
        if( !error)
            error = otvn_mkdir( rvnp, "DIRECTORYPAD8", &dvas, &dvnp, &ucreds);
        if( !error)
            error = otvn_mkdir( rvnp, "DIRECTORY9", &dvas, &dvnp, &ucreds);
        if( !error)
            error = otvn_mkdir( rvnp, "DIRECTORY10", &dvas, &dvnp, &ucreds);
        if( !error)
            error = otvn_mkdir( rvnp, "NDIRECTORY1", &dvas, &dvnp, &ucreds);
        if( !error)
            error = otvn_mkdir( rvnp, "NDIRECTORYP2", &dvas, &dvnp, &ucreds);
        if( !error)
            error = otvn_mkdir( rvnp, "NDIRECTORYPA3", &dvas, &dvnp, &ucreds);
        if( !error)
            error = otvn_mkdir( rvnp, "NDIRECTORYPAD4", &dvas, &dvnp, &ucreds);
        if( !error)
            error = otvn_mkdir( rvnp, "NDIRECTORY5", &dvas, &dvnp, &ucreds);
        if( !error)
            error = otvn_mkdir( rvnp, "NDIRECTORYP6", &dvas, &dvnp, &ucreds);
        if( !error)
            error = otvn_mkdir( rvnp, "NDIRECTORYPA7", &dvas, &dvnp, &ucreds);
        if( !error)
            error = otvn_mkdir( rvnp, "NDIRECTORYPAD8", &dvas, &dvnp, &ucreds);
        if( !error)
            error = otvn_mkdir( rvnp, "NDIRECTORY9", &dvas, &dvnp, &ucreds);
        if( !error)
            error = otvn_mkdir( rvnp, "NDIRECTORY10", &dvas, &dvnp, &ucreds);
        
        uioInit( &uios, &iovecs, bufp, 0, TIOBUFSIZE);
        if ( ! error ) {
            do {
                uios.uio_resid = 100;
                error = otvn_readdir( rvnp, &uios, &ucreds);
            }
            while ( !error && uios.uio_resid != 100 );
        }
        
        printf("Removing DIRECTORY5.\n");
        
        if( !error)
            error = otvn_rmdir( rvnp, "DIRECTORY5", &ucreds);
        
        printf("Creating eight levels of directories \n");
        
        /* now create a chain of directories that go eight levels deep */
        
        if( !error) {
            error = otvn_mkdir( rvnp, "LEVEL1", &dvas, &dvnp, &ucreds);
            vnp = dvnp;
        }
        if( !error) {
            error = otvn_mkdir( vnp, "LEVEL2", &dvas, &dvnp, &ucreds);
            vnp = dvnp;
        }
        if( !error) {
            error = otvn_mkdir( vnp, "LEVEL3", &dvas, &dvnp, &ucreds);
            vnp = dvnp;
        }
        if( !error) {
            error = otvn_mkdir( vnp, "LEVEL4", &dvas, &dvnp, &ucreds);
            vnp = dvnp;
        }
        if( !error) {
            error = otvn_mkdir( vnp, "LEVEL5", &dvas, &dvnp, &ucreds);
            vnp = dvnp;
        }
        if( !error) {
            error = otvn_mkdir( vnp, "LEVEL6", &dvas, &dvnp, &ucreds);
            vnp = dvnp;
        }
        if( !error) {
            error = otvn_mkdir( vnp, "LEVEL7", &dvas, &dvnp, &ucreds);
            vnp = dvnp;
        }
        if( !error) {
            error = otvn_mkdir( vnp, "LEVEL8", &dvas, &dvnp, &ucreds);
        }
        
        if ( !error ) {
            /* now create a file */
            printf("Creating file LEVEL8/FILE1.\n");
            error = otvn_create(vnp, "FILE1", &vas, NONEXCL, VWRITE, &fvp, &favp, &ucreds, 0);
            /* fill a buffer and write it */
            if (!error) {
                memcpy( bufp, "TESTSTRING", 10);
                uioInit( &uios, &iovecs, bufp, 0, 10);
                printf("Writing 'TESTSTRING' to LEVEL8/FILE1.\n");
                error = otvn_rdwr( fvp, &uios, UIO_WRITE, IO_APPEND, &ucreds);
            }
        }
        
        printf("Reading root directory.\n");
        uioInit( &uios, &iovecs, bufp, 0, 1024);
        if ( ! error ) {
            do {
                uios.uio_resid = 1024;
                error = otvn_readdir( rvnp, &uios, &ucreds);
            }
            while ( !error && uios.uio_resid != 1024);
        }
    }
    
 testRead:
    if (!error) {
        printf("  - unmounting fs ... ");
        error = otfs_unmount(vfsp);
        if (error) {
            printf("done. error = %d\n");
        } else {
            printf("done. SUCCESS!\n");
        }
    }
    
    if (!error) {
        printf("  - Re-mounting fs ... ");
        error = otfs_mount( vfsp, "TEST", 1, 1, "DUMMY");
        if (error) {
            printf("done. error = %d\n");
        } else {
            printf("done. SUCCESS!\n");
        }
    }
    
    if (!error) {
        printf("  - Getting ROOT ... ");
        error = otfs_root( vfsp, &rvnp, 0);
        if (error) {
            printf("done. error = %d\n");
        } else {
            printf("done. SUCCESS!\n");
        }
    }
    
    if ( !error ) {
        printf("  - Getting Root's attributes ...");
        error = otvn_getattr(rvnp,&rvas,&ucreds);
        if (error) {
            printf("done. error = %d\n");
        } else {
            printf("done. SUCCESS!\n");
        }
    }
    
    if (!error) {
        printf("  - Looking up /TESTDATA ... ");
        error = otvn_lookup(rvnp, "TESTDATA", &fvp, &ucreds, LAST);
        if (error) {
            printf("done. error = %d\n");
        } else {
            printf("done. SUCCESS!\n");
        }
    }
    
    if (!error) {
        /* Reading from file */
        printf("  - Reading and verifying extents in file /TESTDATA\n");
        for (i=0; !error && i <= MAXEXTS; i++) {
            uioInit( &uios, &iovecs, vbufp, i*EXTSIZE, EXTSIZE);
            printf (".");
            error = otvn_rdwr( fvp, &uios, UIO_READ, 0, &ucreds);
            if (!error) {
                if (error = memcmp(vbufp,fbufp,EXTSIZE))
                    printf("\n   memcmp(w,r) failure: returned %d.\n",error);
            }
        }
        if (error) {
            printf("done. error = %d\n");
        } else {
            printf("done. SUCCESS!\n");
        }
    }
    
    if (!error) {
        /* Reading from file */
        printf("  - Reading and verifying 256 byte pieces in /TESTDATA\n");
        for (i=0; !error && i <= (MAXEXTS * EXTSIZE / 256); i++) {
            uioInit( &uios, &iovecs, vbufp, i*256, 256);
            printf (".");
            error = otvn_rdwr( fvp, &uios, UIO_READ, 0, &ucreds);
            if (!error) {
                if (error = memcmp(vbufp,fbufp,256))
                    printf("\n   memcmp(w,r) failure: returned %d.\n",error);
            }
        }
        if (error) {
            printf("done. error = %d\n");
        } else {
            printf("done. SUCCESS!\n");
        }
    }
    
 testLongRead:
    if (!error) {
        lbufp = (char *) kmem_alloc( LONGSIZE, 0);
        /* Reading from file */
        printf("  - Reading %d pieces from file /TESTDATA\n",LONGSIZE);
        for (i=0; !error && i < LONGCOUNT; i++) {
            uioInit( &uios, &iovecs, lbufp, i*LONGSIZE, LONGSIZE);
            printf (".");
            error = otvn_rdwr( fvp, &uios, UIO_READ, 0, &ucreds);
        }
        if (error) {
            printf("done. error = %d\n");
        } else {
            printf("done. SUCCESS!\n");
        }
    }
    
    if (!error) {
        /* inactivating */
        printf("  - Inactivating vnode ... ");
        fvp->v_count--;
        otvn_inactive(fvp, &ucreds);
        if (error) {
            printf("done. error = %d\n");
        } else {
            printf("done. SUCCESS!\n");
        }
    }

 testSmallRead:
    if (!error) {
        printf("  - Looking up /TESTSMALL ... ");
        error = otvn_lookup(rvnp, "TESTSMALL", &fvp, &ucreds, LAST);
        if (error) {
            printf("done. error = %d\n");
        } else {
            printf("done. SUCCESS!\n");
        }
    }
    
    if (!error) {
        /* Reading from file */
        printf("  - Reading and verifying single extent in /TESTSMALL\n");
        uioInit( &uios, &iovecs, vbufp, 0, SMALLLEN);
        error = otvn_rdwr( fvp, &uios, UIO_READ, 0, &ucreds);
        if (!error) {
            if (error = memcmp(vbufp,fbufp,SMALLLEN))
                printf("\n   memcmp(w,r) failure: returned %d.\n",error);
        }
        if (error) {
            printf("done. error = %d\n");
        } else {
            printf("done. SUCCESS!\n");
        }
    }

    if (!error) {
        /* inactivating */
        printf("  - Inactivating vnode ... ");
        fvp->v_count--;
        otvn_inactive(fvp, &ucreds);
        if (error) {
            printf("done. error = %d\n");
        } else {
            printf("done. SUCCESS!\n");
        }
    }

 testManyRead:
    if (!error) {
        printf("  - Looking up /TESTMANY ... ");
        error = otvn_lookup(rvnp, "TESTMANY", &fvp, &ucreds, LAST);
        if (error) {
            printf("done. error = %d\n");
        } else {
            printf("done. SUCCESS!\n");
        }
    }
    
    if (!error) {
        /* Reading from file */
        printf("  - Reading and verifying extents in file /TESTMANY\n");
        for (i=0; !error && i <= MANYEXTS; i++) {
            uioInit( &uios, &iovecs, vbufp, i*EXTSIZE, EXTSIZE);
            printf (".");
            error = otvn_rdwr( fvp, &uios, UIO_READ, 0, &ucreds);
            if (!error) {
                if (error = memcmp(vbufp,fbufp,EXTSIZE))
                    printf("\n   memcmp(w,r) failure: returned %d.\n",error);
            }
        }
        if (error) {
            printf("done. error = %d\n");
        } else {
            printf("done. SUCCESS!\n");
        }
    }
    
 unmount:    
    if ( !error ) 
        error = otfs_unmount(vfsp);

    if ( error )
        printf("adtest: error 0x%x (%d)\n",error,error);
    printf("Done!\n");
}
