/********************************************************************/
/*  dclean.c - Program to erase the anchors on an mo disk.          */
/*                                                                  */
/*      Unpublished Confidential Information of Hewlett-Packard     */
/*      Company. Do not disclose.  Copyright (c) Hewlett-Packard    */
/*      Company 1996. All rights reserved                           */
/*                                                                  */
/* The contents of this software are proprietary and confidential   */
/* the Hewlett- Packard Company, and are limited in distribution to */
/* those with a direct need to know.  Individuals having access to  */
/* this software are responsible for maintaining the                */
/* confidentiality of the content and for keeping the software      */
/* secure when not in use.  Transfer to any party is strictly       */
/* forbidden other than as expressly permitted in writing by        */
/* Hewlett-Packard Company.  Unauthorized transfer to or possession */
/* by any unauthorized party may be a criminal offense.             */
/*                                                                  */
/*                   RESTRICTED RIGHTS LEGEND                       */
/*                                                                  */
/*      Use,  duplication,  or disclosure by the                    */
/*      U. S. Government is subject to restrictions as set forth    */
/*      in subparagraph (c) (1) (ii) in the Rights in Technical     */
/*      Data and Computer Software clause in DFARS 252.227-7013     */
/*      or any other successor clause.                              */
/*                                                                  */
/*                       HEWLETT-PACKARD COMPANY                    */
/*                       3000 Hanover St.                           */
/*                       Palo Alto, CA  94304 U.S.A.                */
/*                                                                  */
/*      Rights for non-DOD U.S. Government Departments and Agencies */
/*      are as set forth in FAR 52.227-19 (c) (1,2) or any other    */
/*      successor clause.                                           */
/*                                                                  */
/********************************************************************/
/*
 $Source: /projects/HMS/nsr/udfOld/src/nsrTools/dclean.c,v $
 $Revision: 1.5 $       $Author: hmgr $
 $State: Exp $          $Locker:  $
 $Date: 96/05/30 14:53:12 $
 ********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/scsi.h>

typedef int INT32;
typedef unsigned int UINT32;
typedef unsigned char UCHAR8;

INT32 hpux_readCapacity(char *devfile, UINT32 *DiskSectors, UINT32 *sectorSize)
{
    int error;
    int fd;
    int flag = 1;
    int ret;
    unsigned char buf[8];
    static const struct scsi_cmd_parms read_cap = {
      10,      /* 10 bytes in cdb                       */
      1,       /* 1 -> select with ATN                  */
      100,     /* clock ticks (50 ticks = 1 sec)        */
      0x25,    /* cdb[0] read capacity                  */
      0x00,    /* cdb[1] lun = 0; reserved = 0          */
      0x00,    /* cdb[2] lba = 0                        */
      0x00,    /* cdb[3] lba = 0                        */
      0x00,    /* cdb[4] lba = 0                        */
      0x00,    /* cdb[5] lba = 0                        */
      0x00,    /* cdb[6] reserved = 0                   */
      0x00,    /* cdb[7] reserved = 0                   */
      0x00,    /* cdb[8] reserved = 0                   */
      0x00     /* cdb[9] reserved,flag,link = 0         */
    };
        
    /* open and setup device */
    if ((fd=open(devfile, 2))<0) {
        perror("hpux_readCapacity,open");
        exit(1);
    }
   
    /* Sets device up in command mode: must be super-user */
    if ( (ret = ioctl(fd,SIOC_CMD_MODE,&flag) ) <0 ) {
        perror("hpux_readCapacity,ioctl");
        exit(1);
    }
    /* Send command */
    if ( ioctl(fd,SIOC_SET_CMD,&read_cap) < 0 ) {
        perror("hpux_readCapacity,SET_CMD");
        exit(1);
    }
    if ( read(fd,buf,8) != 8 ) {
        perror("hpux_readCapacity");
        exit(1);
    }

    /* Take driver out of command mode. */
    if ((error=ioctl(fd,SIOC_CMD_MODE,0)) != 0)  {
        perror("hpux_readCapacity,SIOC_CMD");
        exit(1);
    }

    /* Close device */
    close(fd);

    *DiskSectors = buf[3]+(buf[2]<<8)+(buf[1]<<16)+(buf[0]<<24);
    *sectorSize = buf[7]+(buf[6]<<8)+(buf[5]<<16)+(buf[4]<<24);
    return 0;
}

main(argc, argv)
int argc;
char *argv[];
{
    int fd;
    int ret;
    unsigned char iobuf[1024];
    int idx;
    char *devfile = getenv("RVOLUME");
    UINT32 DiskSectors;
    UINT32 sectorSize;
    UINT32 loc;

    if ( strncmp(devfile,"/dev/",5) != 0 ) {
        printf("RVOLUME must be /dev/ file\n");
        exit(1);
    }
    
    hpux_readCapacity(devfile,&DiskSectors,&sectorSize);

    memset(iobuf,0,1024);
    strcpy(iobuf,"hello world");
    
    fd = open(devfile,O_RDWR);
    if ( fd == -1 )  {
        perror("open");
        exit(1);
    }
   
    for ( idx=0; idx <= 2; idx++ )
    {
        switch (idx) {
        case 0: loc = 256; break;
        case 1: loc = DiskSectors-1; break;
        case 2: loc = DiskSectors-1-256; break;
        default: loc = ((idx-3)*(DiskSectors-1))/59;
        }
        printf("erasing anchor at %d\n",loc);
        ret = lseek(fd,loc*sectorSize,0);
        if ( ret == -1 )  {
            perror("seek");
            exit(1);
        }
        ret = write(fd,iobuf,sectorSize);
        if ( ret != sectorSize )  {
            perror("write");
            exit(1);
        }
    } 
    close(fd);
}
