/********************************************************************/
/*  mkvolio.c - Main for making a file system?                      */
/*                                                                  */
/*      Unpublished Confidential Information of Hewlett-Packard     */
/*      Company. Do not disclose.  Copyright (c) Hewlett-Packard    */
/*      Company 1996. All rights reserved                           */
/*                                                                  */
/* The contents of this software are proprietary and confidential   */
/* the Hewlett- Packard Company, and are limited in distribution to */
/* those with a direct need to know.  Individuals having access to  */
/* this software are responsible for maintaining the                */
/* confidentiality of the content and for keeping the software      */
/* secure when not in use.  Transfer to any party is strictly       */
/* forbidden other than as expressly permitted in writing by        */
/* Hewlett-Packard Company.  Unauthorized transfer to or possession */
/* by any unauthorized party may be a criminal offense.             */
/*                                                                  */
/*                   RESTRICTED RIGHTS LEGEND                       */
/*                                                                  */
/*      Use,  duplication,  or disclosure by the                    */
/*      U. S. Government is subject to restrictions as set forth    */
/*      in subparagraph (c) (1) (ii) in the Rights in Technical     */
/*      Data and Computer Software clause in DFARS 252.227-7013     */
/*      or any other successor clause.                              */
/*                                                                  */
/*                       HEWLETT-PACKARD COMPANY                    */
/*                       3000 Hanover St.                           */
/*                       Palo Alto, CA  94304 U.S.A.                */
/*                                                                  */
/*      Rights for non-DOD U.S. Government Departments and Agencies */
/*      are as set forth in FAR 52.227-19 (c) (1,2) or any other    */
/*      successor clause.                                           */
/*                                                                  */
/* modification history:                                            */
/*      3.Dec.92  Made default for WORM to not allow successors     */
/********************************************************************/
/*
 $Source: /projects/HMS/nsr/udfOld/src/nsrTools/mkvol/mkvolio.c,v $
 $Revision: 1.16 $      $Author: hmgr $
 $State: Exp $          $Locker:  $
 $Date: 96/05/30 14:53:18 $
 ********************************************************************/
#include "def.h"

/* includes */

/* IBM version */
#if IBM
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <errno.h>
#include <conio.h>
#include <process.h>
#include "d:\pc-nsr\include\times.h"
#include "d:\corecode\src\nsrhdrs\vnode.h"
#include "d:\corecode\src\nsrhdrs\o_proto.h"
#include "d:\nitro\include\nsr.h"
#include "d:\nitro\include\newnsrer.h"
#include "d:\nitro\makefs\makefs.h"
#include "mkvol.h"
#endif

#if HP
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "def.h"
#include "o_vfsops.h"
#include "nsr.h"
#include "makefs.h"
#include "mkvol.h"
#include "o_proto.h"
#include "nsrerrno.h"
#include "Debug.h"
#endif


/* prototype definitions */

INT32 makefs( struct fs_info *x, struct PartDesc *pd, struct LogVolDesc *lvd);
UINT32 mkvol( struct vol_info *x, struct fs_info *fsi,struct PartDesc *pd, struct LogVolDesc *lvd);
INT16 ParseCommandLine( INT16 a, CHAR8 *b[], struct vol_info *c, struct fs_info *d);
void PrintMessage( INT16 a);

#if HP
extern INT32 hpux_open(void);
extern INT32 hpux_close(void);
#endif

/* globals */

INT16 Messages = TRUE;

const char *usage = 
"mkvol -[vdw]\n"
"\n"
"-d <device spec>\n"
"-v <volume name>\n"
"-w make a worm volume\n";

#pragma page()
void main( INT16 argc, CHAR8 *argv[]) 
{
    struct vol_info vol_data;
    struct fs_info file_data;
    struct PartDesc pd;
    struct LogVolDesc lvd;
    struct otfs fs;
    
    UINT32 error;

    /* set up defaults */
    bzero( (char *)&vol_data, sizeof( struct vol_info));

    strcpy( vol_data.VolID, "TESTVOLUME");
    strcpy( vol_data.LogID, "TESTLOGVOLUME");
    strcpy( vol_data.DevID, "NOTUSED");
    vol_data.PartLen = 1000;
    vol_data.BlkSize = 512;
    vol_data.MediaType = fsMO;
    vol_data.CharSet[0] = 5;     /* portable posix 191 chars from 8859-1 */
    vol_data.SectorSize = 512;
    vol_data.DiskSectors = 2048;
    
    /*
    * Do a read capacity here since it was taken out of mkvol().
    * It was taken out of mkvol() because a read capacity is already
    * done in makefs_dcp(). If arguments were passed in they 
    * will override the sector size, number of sectors found in
    * this read capacity.
    */
    ot_readCapacity(&fs);
    vol_data.SectorSize = fs.SectorSize;
    vol_data.DiskSectors = fs.DiskSectors;
    /* 
    * Make partition length a function of disk sectors. 
    * Must subtract off 0 to 256,N-256 to N.
    */
    vol_data.PartLen = fs.DiskSectors-257-257;
    file_data.ADType = ADLONG;
    
    /* Parse the command line into the volume and file definition structures */
    error = ParseCommandLine( argc, argv, &vol_data, &file_data);

#if HP
    hpux_open();
#endif
    
    /* make the volume structure */
    if( !error)
        error = mkvol(&vol_data, &file_data, &pd, &lvd);

    /* set up defaults for makefs call */
    strcpy(file_data.VolumeName, vol_data.VolID);
    file_data.Type = vol_data.MediaType;
    file_data.SectorSize = vol_data.SectorSize;
    file_data.DiskSectors   = vol_data.DiskSectors;
    /* Set onode block size, indirect size, and successors according
     * to the type of volume requested.
     */
    if (file_data.Type == fsMO) {
        file_data.ICBsize = ICB_BLOCK_SIZE_MO;
        file_data.IndirectSize = ICB_INDIRECT_SIZE_MO;  
    } else {
        /* some kind of write-once volume */
        file_data.ICBsize = ICB_BLOCK_SIZE_WORM;
        file_data.IndirectSize = ICB_INDIRECT_SIZE_WORM;  
    }

    /* At this release, we are imposing some limitations. */
    file_data.StripeSize = 0;
    file_data.FileSystemType = 0;
    file_data.VolSeqNumber = 0;
    file_data.VolumeSize = 1;

    if( !error) error = makefs( &file_data, &pd, &lvd);

#if HP
    hpux_close();
#endif

    exit(0);
}


#pragma page()
INT16 ParseCommandLine(INT16 argc, CHAR8 *argv[],
                       struct vol_info *vol_data,
                       struct fs_info *fs_data)
{
    INT16 tokensread;
    INT16 status = SUCCESS;
        
    
    tokensread = 1;                 /* don't want the program name */
        
    while ((tokensread < argc) && (status == SUCCESS)) {

        /* should see a switch character followed by a switchval */
        if ((*++argv)[0] == SWITCHCHAR) {

            /* saw a switchchar....which one? */
            switch( (*argv)[1]) {
                case 'v':
                case 'V':

                    /* next token will be the volumeID
                        validation other than length is not done here */
                    if( strlen( *++argv) <= MAXVOLIDLEN) {
                        strcpy( vol_data->VolID, *argv);
                        tokensread += 2;
                    } else {
                        status = BADVOLNAME;
                        PrintMessage( status);
                    }
                    break;

                case 'w':
                case 'W':
                   
                    vol_data->MediaType = fsWORM;
                    tokensread++;
                    break;

                case 'a':
                case 'A':
                   
                    /* next character specifies ADType */
                    switch( (*argv)[2]) {
                    case 's':
                    case 'S':
                        fs_data->ADType  = ADSHORT;
                        break;
                    case 'l':
                    case 'L':
                        fs_data->ADType  = ADLONG;
                        break;
                    case 'x':
                    case 'X':
                        fs_data->ADType  = ADEXTENDED;
                        break;
                    default:
                        status = BADADTYPE;
                        PrintMessage( status);
                    }
                    tokensread++;
                    break;

                case 'D':
                case 'd':

                /* next token will be a device specifier.  This will
                   be very implementation and device specific.  Should
                   probably end up some where in the otfs structure */

                    if( strlen( *++argv) <= MAXDEVIDLEN) {
                        strcpy( vol_data->DevID, *argv);
                        tokensread += 2;
                    } else {
                        status = BADDEVID;
                        PrintMessage( status);
                    }
                    break;

                case 'm':
                case 'M':

                    /* toggle messages */
                    Messages = !Messages;
                    tokensread++;
                    break;
                
                case 'u':
                    status = HELPLINE;
                    PrintMessage(status);
                    break;

                default:
                    /* couldn't make anything out */
                    status = BADSWITCHVAL;
                    printf("(%s) ", *argv);
                    PrintMessage( status);
                    break;
            }
        } else {
            /* didn't see the switch char */
            status = BADSWITCHCHAR;
            printf("(%s) ", *argv);
            PrintMessage( status);
        }
    }
    return ( status);
}

#pragma page()
void PrintMessage( INT16 MessageNum)
{
    const CHAR8 *messagedata[8] = {
        {""}, /* First Entry always blank */
        {"Unrecognized token, bad switch value.\n"},    /* BADSWITCHVAL */
        {"Unrecognized token, no switch character seen.\n"}, /* BADSWITCHCHAR */
        {"%TYPE D:\\nitro\\mkvol\\MKVOL.MAN | MORE"}, /* HELPLINE */
        {"Volume name too long.\n"}, /* BADVOLNAME */
        {"Device name too long.\n"}, /* BADDEVID */
        {"ADType not [Ss], [Ll], or [Xx].\n"}, /* BADADTYPE */
        {""} /* last entry always blank */
    };

    if(Messages) {
        if( messagedata[MessageNum][0] == '%')
            /* edb: system( &(messagedata[MessageNum][1])); */
            printf("%s",usage);
        else   
            printf("%s", messagedata[MessageNum]);
    }
    return;
}
