/*
  ----------------------------------------------------------------------
  rs.c
    Reads sectors and/or blocks

    To read block addresses, link executable to 'rb', and environment
    variable BIAS will be added into supplied argument, making it a
    block number instead of a sector number.
  
        Unpublished Confidential Information of Hewlett-Packard
        Company. Do not disclose.  Copyright (c) Hewlett-Packard
        Company 1996. All rights reserved
   
  The contents of this software are proprietary and confidential to
  the Hewlett- Packard Company, and are limited in distribution to
  those with a direct need to know.  Individuals having access to
  this software are responsible for maintaining the confidentiality
  of the content and for keeping the software secure when not in use.
  Transfer to any party is strictly forbidden other than as expressly
  permitted in writing by Hewlett-Packard Company.  Unauthorized
  transfer to or possession by any unauthorized party may be a
  criminal offense.
  
                     RESTRICTED RIGHTS LEGEND
  
       Use,  duplication,  or disclosure by the U. S. Government is
       subject to restrictions as set forth in subparagraph (c) (1)
       (ii) in the Rights in Technical Data and Computer Software
       clause in DFARS 252.227-7013 or any other successor clause.
  
                         HEWLETT-PACKARD COMPANY
                         3000 Hanover St.
                         Palo Alto, CA  94304 U.S.A.
  
       Rights for non-DOD U.S. Government Departments and Agencies
       are as set forth in FAR 52.227-19 (c) (1,2) or any other
       successor clause.
  
  ---------------------------------------------------------------------

  $Source: /projects/HMS/nsr/udfOld/src/nsrTools/sniff/rs.c,v $
  $Revision: 1.12 $     $Author: hmgr $
  $State: Exp $         $Locker:  $
  $Date: 96/05/30 14:53:22 $

  */

#include "def.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <sys/scsi.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/unistd.h>
#include <fcntl.h>

#define DEBUG       1
#define ONE_SECOND 50

#define WAIT   ONE_SECOND*2

/* prototype definitions */
void testunit(char *special);
void read_capacity(char *special,int *sectorsize);
void read_sector(char *special,long offset,unsigned char *iobuf,int bsize);


#define SECSIZE 512
#define MAXSECS 10
#define SPACECHAR "."

/* prototype definitions */

int main( int argc, char *argv[])
{
    FILE *handle, *chandle;
    int result;
    int secsize = SECSIZE;
    unsigned char *iobuff;
    char *bdevfile = NULL;
    char *cdevfile = NULL;
    char tmpstring[32];
    char config[64];
    int i,j;
    long int secnum = -1;
    long int seclist[MAXSECS];
    long int bias = 0;
    char *stopstring;
    int sectorsread = 0;
    int isDev;

    iobuff = (unsigned char *) malloc( SECSIZE);

    if( iobuff == NULL) {
        printf("rs: couldn't allocate enough memory.\n");
        exit(-1);
    }

    for(i=0; i<MAXSECS; seclist[i++] = 0);

    /* assemble a string that defines the config file */

    strcpy((char *)config, getenv("HOME"));
    if( config != NULL) {
        strcat((char *)config, "/.rs");
    }
    
    /* get the name of the file to open from the environment variables
       VOLUME and RVOLUME */ 
    
    bdevfile = getenv("VOLUME");
    cdevfile = getenv("RVOLUME");
    isDev = (strncmp(bdevfile,"/dev/",5) == 0 ? 1 : 0);

    if( argv[1] == NULL) {

        /* find the last sector number written in the .rs file at
           local root */ 

        if(( chandle = fopen(config, "r+")) != NULL) {
            result = fread( (void *) &secnum, sizeof( int), 1, chandle);
            fclose( chandle);
        }

        seclist[0] = ++secnum;
        sectorsread++;

    } else {

        for(i=1; argv[i] != NULL; i++ ) {

            if( *(argv[i]) == '-') {

                switch (*(argv[i]++)) {

                  /* volume name */
    
                  case 'v':
                    bdevfile = (argv[i]++);
                    break;

                  default:
                    printf("rs: bad argument value %s\n", argv[i]);
                    break;
                }

            } else {
                
                /* must be a sector */
                
                if( sectorsread < MAXSECS) {
                    secnum = strtoul( argv[i], &stopstring, 10);
                    seclist[sectorsread++] = secnum;
                }
            }
        }
    }
           

    /* open the config file and write the value of the last sector read */

    if(( chandle = fopen( config, "w")) != NULL) {
        result = fseek( chandle, (long int) 0, SEEK_CUR);
        result = fwrite( (void *) &secnum, sizeof( int), 1, chandle);
        fclose( chandle);
    }

    /* If invoked as 'rb', grab the BIAS environment variable */

    if (strcmp(argv[0],"rb") == 0) {
        strcpy( tmpstring, getenv("BIAS"));
        if( tmpstring != NULL) {
            sscanf( tmpstring, "%d", &bias);
        }
    }

    if ( isDev ) {
        /*
        * Do a read capacity.
        */
        testunit(cdevfile);
        read_capacity(cdevfile,&secsize);
    }
    else {
        /* now open the file */
        if( ( handle = fopen( bdevfile, "r")) == NULL) {
           sprintf(tmpstring,"cat %s/src/nsrTools/sniff/rs.man | more",
                               getenv("NSR"));
           system(tmpstring);
           fprintf(stderr, "rs: can't open volume.\n");
           exit( -1);
        }
    }

    /* emit the number of bytes in per sector */
    /* max is 65536 */

    putw( secsize, stdout);
    putw( sectorsread, stdout);

    for(i=0; i<sectorsread; i++) 
    {
        if ( isDev ) {
            read_sector(cdevfile,(seclist[i]+bias)*secsize,iobuff,secsize);
        }
        else {
            result = fseek(handle,(seclist[i]+bias)*secsize, SEEK_SET);
            if (result != 0) {
                printf("rs: can't seek.\n");
                exit(-2);
            }
            
            result = fread( (void *) (iobuff), (size_t) secsize, 1, handle);
            if (result != 1) {
                printf("rs: can't read.\n");
                exit(-3);
            }
        }
        putw( seclist[i], stdout);
        putw( seclist[i]+bias, stdout);

        for(j=0; j<secsize; putchar( iobuff[j++]));
    }
    exit(0);
}

void testunit(char *special)
{
    static const struct scsi_cmd_parms testunit = {
       6,       /* 6 bytes in cdb                       */
       1,       /* 1 -> select with ATN                  */
       WAIT,    /* clock ticks (50 ticks = 1 sec)        */
       0x00,    /* cdb[0] testunit opcode                */
       0x00,    /* cdb[1] */
       0x00,    /* cdb[2] */
       0x00,    /* cdb[3] */
       0x00,    /* cdb[4] */
       0x00,    /* cdb[5] */
    };
    int fd;
    int flag = 1;
    int ret;
    char dummy[10];

    /* open and setup device */
    if ((fd=open(special, 2))<0) {
        perror("can't open device");
        exit(1);
    }
   
    /* Sets device up in command mode: must be super-user */
    if ( (ret = ioctl(fd,SIOC_CMD_MODE,&flag) ) <0 ) {
        perror("can't set CMD_MODE, not super-user");
        exit(1);
    }

    /* Do a testunit to clear power up check condition */
    if ( ioctl(fd,SIOC_SET_CMD,&testunit) < 0 ) {
        perror("can't set up SCSI command");
        exit(1);
    }
    /* Must do a r/w to cause the above ioctl to execute */
    if (( ret = read(fd,dummy,4)) < 0 ) {
        perror("can't execute testunit");
        exit(1);
    }
    close(fd);
}

void read_capacity(char *special,int *sectorsize)
{
    static const struct scsi_cmd_parms read_cap = {
      10,      /* 10 bytes in cdb                       */
      1,       /* 1 -> select with ATN                  */
      WAIT,    /* clock ticks (50 ticks = 1 sec)        */
      0x25,    /* cdb[0] read capacity                  */
      0x00,    /* cdb[1] lun = 0; reserved = 0          */
      0x00,    /* cdb[2] lba = 0                        */
      0x00,    /* cdb[3] lba = 0                        */
      0x00,    /* cdb[4] lba = 0                        */
      0x00,    /* cdb[5] lba = 0                        */
      0x00,    /* cdb[6] reserved = 0                   */
      0x00,    /* cdb[7] reserved = 0                   */
      0x00,    /* cdb[8] reserved = 0                   */
      0x00     /* cdb[9] reserved,flag,link = 0         */
    };
    int fd;
    int flag = 1;
    int ret;
    unsigned char buf[8];
    int cap;
    int error;

    /* open and setup device */
    if ((fd=open(special, 2))<0) {
        perror("can't open device");
        exit(1);
    }
   
    /* Sets device up in command mode: must be super-user */
    if ( (ret = ioctl(fd,SIOC_CMD_MODE,&flag) ) <0 ) {
        perror("can't set CMD_MODE");
        exit(1);
    }
    /* Send command */
    if ( ioctl(fd,SIOC_SET_CMD,&read_cap) < 0 ) {
        perror("can't set up SCSI command");
        exit(1);
    }
    if ( read(fd,buf,8) != 8 ) {
        perror("didn't read 8 bytes");
        exit(1);
    }

    /* Take driver out of command mode. */
    if ((error=ioctl(fd,SIOC_CMD_MODE,0)) != 0)  {
        perror("ioctl: could not clear SIOC_CMD_MODE");
        exit(error);
    }

    /* Close device */
    close(fd);

    /* Print out read capacity information. */
    cap = buf[3] + (buf[2] << 8) + (buf[1] << 16) + (buf[0] << 24);
    *sectorsize  = buf[7] + (buf[6] << 8) + (buf[5] << 16) + (buf[4] << 24);
}

void read_sector(char *special,long offset,unsigned char *iobuf,int bsize)
{
    int fd;
    int count;
    int length;
    
    fd = open(special, O_RDONLY);

    if (fd < 0) {
        perror(" open ");
        exit(0);
    }

    if (lseek(fd, (long)offset, SEEK_SET) < 0) {
        perror(" lseek ");
        exit(0);
    }
    count = read(fd, (void *)iobuf, bsize);
    if (count != bsize ) {
        perror(" read ");
        exit(0);
    }
    else if (count < bsize) {
        printf("WARNING / only read %d bytes, asked for %d\n", count, bsize);
    }

    close(fd);
}
