/********************************************************************/
/*  st                                                              */
/*    Interpret standard input stream as a NSR structure otherwise  */
/*    do a hex dump.  Typically used in conjunction with rs (read   */
/*    sector) as the source of the standard input via a pipe. See   */
/*    notes.                                                        */
/*                                                                  */
/*      Unpublished Confidential Information of Hewlett-Packard     */
/*      Company. Do not disclose.  Copyright (c) Hewlett-Packard    */
/*      Company 1996. All rights reserved                           */
/*                                                                  */
/* The contents of this software are proprietary and confidential   */
/* the Hewlett- Packard Company, and are limited in distribution to */
/* those with a direct need to know.  Individuals having access to  */
/* this software are responsible for maintaining the                */
/* confidentiality of the content and for keeping the software      */
/* secure when not in use.  Transfer to any party is strictly       */
/* forbidden other than as expressly permitted in writing by        */
/* Hewlett-Packard Company.  Unauthorized transfer to or possession */
/* by any unauthorized party may be a criminal offense.             */
/*                                                                  */
/*                   RESTRICTED RIGHTS LEGEND                       */
/*                                                                  */
/*      Use,  duplication,  or disclosure by the                    */
/*      U. S. Government is subject to restrictions as set forth    */
/*      in subparagraph (c) (1) (ii) in the Rights in Technical     */
/*      Data and Computer Software clause in DFARS 252.227-7013     */
/*      or any other successor clause.                              */
/*                                                                  */
/*                       HEWLETT-PACKARD COMPANY                    */
/*                       3000 Hanover St.                           */
/*                       Palo Alto, CA  94304 U.S.A.                */
/*                                                                  */
/*      Rights for non-DOD U.S. Government Departments and Agencies */
/*      are as set forth in FAR 52.227-19 (c) (1,2) or any other    */
/*      successor clause.                                           */
/*                                                                  */
/*                                                                  */
/*  notes:                                                          */
/*    st expects no command line arguments.  If one is seen, help   */
/*       message is printed.                                        */
/*    st expects to see a 16 bit sector size as the first 2 bytes   */
/*       of standard input followed by a 2 byte sector number.      */
/*       st prints these as confirmation.  It then expects          */
/*       sectorsize bytes.  The first word of these bytes is        */
/*       examined.  If it corresponds to a recognized ISO/IEC 13346 */
/*       tag then the entire buffer is cast to the suspected        */
/*       structure and printed in a useful form.                    */
/*    every attempt has been made to use the same terminology as    */
/*       that found in the draft of ISO/IEC 13346 dated Nov 1993    */
/*                                                                  */
/*  Modification history:                                           */
/*       12.Sep.92  ECB    Initial Entry                            */
/*       18.Nov.92  ECB    Accepts multi sector stream from rs      */
/*                         attempts to interpret entire stream con- */
/*                         tiguously                                */
/*                                                                  */
/********************************************************************/
/*
  $Source: /projects/HMS/nsr/udfOld/src/nsrTools/sniff/st.c,v $
  $Revision: 1.42 $     $Author: hmgr $
  $State: Exp $         $Locker:  $
  $Date: 96/05/30 14:53:24 $
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

/*
 * Get definitions for ISO/IEC 13346 structures
 *
 * but first, fake out the definition of "vnode" and "vtype" so we don't
 * have to bring in a raft of other superfluous include files
 */

struct vnode {
    int  dummy;
};

enum vtype { DUM1, DUM2 };

/* Also, instead of including sys/param.h, simply define the following: */

#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))

#include "../../nsrHdrs/def.h"
#include "../../nsrHdrs/nsr.h"
#include "../../nsrHdrs/o_proto.h"

/* some default values */

#define SECSIZE 1024
#define SPACECHAR "."
#define TITLELEN 99
#define MAXCOLUMNS 16

/* forward declaration of procedures in st */

void DisplayPVD( char *x);
void DisplayFSD( char *x);
void DisplayTag( char *x);
void DisplayICBTag( char *x);
void DisplayAnchor( char *x);
void DisplayPointer( char *x);
void DisplayImpUse( char *x);
void DisplayPD( char *x);
void DisplayLVD( char *x, UINT32 y);
void DisplayUSD( char *x);
void DisplayTerm( char *x);
void DisplayLVID( char *x, UINT32 y);
void DisplayFID( char *x, UINT32 length);
void DisplayPIT( char *x);
void DisplayBMAP(char *x,int nbytes);
void DisplayUST( char *x,UINT32 length);
void DisplayFE(char *x,UINT32 length);
void DisplayAE(void *x,UINT32 length,UINT32 ADType);
void DisplayIE(char *x);
void TitleLine(char *x);
void Displayextent_ad( struct extent_ad x);
void DisplayLb_addr(UINT32 lbn,UINT16 partition_num);
void DisplayShort_AD(struct short_AD *address);
void DisplayLong_AD(struct long_AD *x);
void DisplayByteStream( char *x, int y, int z, int a);
void DisplayLongString( char* x, int y);
void DisplayCharSet( struct charspec x);
void DisplayRegID( struct regid * x);
void DisplayDomainID( struct domainEntityId * deip);
void DisplayUdfID( struct udfEntityId * ueip);
void DisplayImplID( struct implEntityId * ieip);
void DisplayTimestamp( struct timestamp x, int y);
void printspace( int x);
void printDstring( char *x, UCHAR8 l, UCHAR8 indent);
void printDchars( char *x, UCHAR8 l, UCHAR8 indent);

char *usage = "rs [n] | st";

/* For extent type displaying */
char ETypeStr[4] = {'R','A','U','X'};

int
allzeroes( int nbytes, char *iobuf )
{
    int i;
    for (i=0;i<nbytes;i++){
        if (iobuf[i] != 0)
            return FALSE;
    }
    return TRUE;
}

void
hexdump( int nbytes, char *iobuf )
{
    int inner, outer;
    
    if( allzeroes(nbytes,iobuf) ) {
        TitleLine("ALL ZEROES");
    } else {
        /* can't see a tag that make sense, do a hex dump */
        TitleLine("HEX DUMP");
        for( outer=0; outer < (nbytes)/16; outer++) {
            printf("%3.3X  ", outer*16);
            for( inner = 0; inner < 16; inner++) {
                if( inner == 8)
                  printf("  ");
                printf("%2.2X ", (UCHAR8)iobuf[outer*16+inner]);
            }
            printf("  ");
            for( inner = 0; inner < 16; inner++) {
                if( inner == 8)
                  printf(" ");
                if( isgraph((int) iobuf[outer*16+inner])) {
                    printf("%c", iobuf[outer*16+inner]);
                } else {
                    printf(SPACECHAR);
                }
            }
            printf("\n");
        }
    }
}

void
main( int argc, char *argv[])
{
    int secsize, sectorsread;
    char *iobuf;
    int i,j;
    long int blockNum, sectorNum;
    UINT16 TagID;
    UINT16 error = 0;
    UINT32 AEADType = ADLONG;   /* ISSUE Should be parm! */

    /* help if we need it */

    if( argv[1] != NULL) {
      printf("%s\n,usage");
      exit(-1);
    }

    /* get the sector size */

    secsize = getw( stdin);
    sectorsread = getw( stdin);
    
    /* allocate the buffer */

    iobuf = malloc( secsize*sectorsread);
    if( iobuf == NULL ) {
        printf("st: can't allocate i/o buffer.\n");
        exit(-1);
    }
    
    /* fill the buffer */

    for(j=0; j<sectorsread; j++) {
        blockNum = getw( stdin);
        sectorNum = getw( stdin);
        if (blockNum == sectorNum) /* BIAS not added in => is sector address */
            printf("--Sector %ld--\n", sectorNum);
        else
            printf("--Block %ld (Sector %ld)--\n", blockNum, sectorNum);
        for(i=0; i<secsize; iobuf[j*secsize+i++] = getchar());
    }

    /* determine what kind of structure this is */
    /* first, swap byte sex if required */

    /* now display the structure if there is a match */
    /* by looking at the Tag ID at the beginning of the sector */

    TagID = *( (UINT16 *) iobuf );
    
    /* translate the tag id */

    swapUINT16( &TagID);

    switch (TagID) {

    case TAGID_PVD:
        DisplayPVD( iobuf );
        break;

    case TAGID_ANCHOR:
        DisplayAnchor( iobuf );
        break;

    case TAGID_POINTER:
        DisplayPointer( iobuf );
        break;

    case TAGID_IUD:
        DisplayImpUse( iobuf );
        break;

    case TAGID_PD:
        DisplayPD( iobuf );
        break;

    case TAGID_LVD:
        DisplayLVD( iobuf,secsize*sectorsread);
        break;

    case TAGID_USD:
        DisplayUSD( iobuf );
        break;

    case TAGID_TERM_DESC:
        DisplayTerm( iobuf );
        break;

    case TAGID_LVID:
        DisplayLVID( iobuf,secsize*sectorsread);
        break;
        
    case TAGID_FSD:
        DisplayFSD( iobuf );
        break;

    case TAGID_FILE_ID:
        DisplayFID(iobuf,secsize*sectorsread );
        break;

    case TAGID_ALLOC_EXTENT:
        DisplayAE(iobuf,secsize*sectorsread,AEADType);
        break;

    case TAGID_INDIRECT:
        DisplayIE(iobuf );
        break;

    case TAGID_TERM_ENTRY:
        TitleLine("TERMINATING ENTRY (unimplemented)");
        hexdump(secsize*sectorsread,iobuf );
        break;

    case TAGID_FILE_ENTRY:
        DisplayFE(iobuf,secsize);
        break;

    case TAGID_EXT_ATTR:
        TitleLine("EXTENDED ATTRIBUTES (unimplemented)");
        hexdump(secsize*sectorsread,iobuf );
        break;

    case TAGID_UNALLOC_SP_ENTRY:
        DisplayUST(iobuf,secsize);
        break;

    case TAGID_SPACE_BMAP:
        DisplayBMAP(iobuf,secsize*sectorsread);
        break;

    case TAGID_PART_INTEGRITY:
        DisplayPIT( iobuf );
        break;

    default:
        TitleLine("UNRECOGNIZED STRUCTURE");
        hexdump(secsize*sectorsread,iobuf );
        break;

    } /* switch */

    exit(0);
}

/********************************************************************/
/* Display the contents of an ISO/IEC 13346 Tag Descriptor          */
/********************************************************************/

void
DisplayTag( char *iobuf )
{
    struct Tag *TagPtr;

    /* structure has been translated by the calling routine */
    /* so no call to xlate struct is required */

    TagPtr = (struct Tag *) iobuf;
    TitleLine("TAG DESCRIPTOR BODY");
    printf("000 002 Tag ID:              %10u (0x%4.4x)\n",
           TagPtr->uTagID,
           TagPtr->uTagID);
    printf("002 002 Descriptor Ver:      %10u (0x%4.4x)\n",
           TagPtr->uDescriptorVersion,
           TagPtr->uDescriptorVersion);
    printf("004 001 Tag Checksum         %10u (0x%2.2x)\n",
           TagPtr->uTagChecksum,
           TagPtr->uTagChecksum);
    printf("005 001 Reserved Field:      %10u (0x%2.2x)\n",
           TagPtr->uReserved,
           TagPtr->uReserved);
    printf("006 002 Tag Serial Num:      %10u (0x%4.4x)\n",
           TagPtr->uTagSerialNum, 
           TagPtr->uTagSerialNum);
    printf("008 002 Tag CRC:             %10u (0x%4.4x)\n",
           TagPtr->uDescriptorCRC,
           TagPtr->uDescriptorCRC);
    printf("00A 002 Tag CRC Length:      %10u (0x%4.4x)\n",
           TagPtr->uCRCLen,
           TagPtr->uCRCLen);
    printf("00C 004 Tag Location:        %10lu (0x%8.8lx)\n",
           TagPtr->uTagLoc,
           TagPtr->uTagLoc);

}
/********************************************************************/
/* Display the contents of an ISO/IEC 13346 ICBTag Descriptor       */
/********************************************************************/


void
DisplayICBTag( char *iobuf )
{
    struct ICBTag *ICBTagPtr;

    /* structure has been translated by the calling routine */
    /* so no call to xlate struct is required */

    ICBTagPtr = (struct ICBTag *) iobuf;
    TitleLine("ICBTAG DESCRIPTOR BODY");
    printf("010 004 Prior Directs:       %10lu (0x%8.8lx)\n",
           ICBTagPtr->PriorDirects, ICBTagPtr->PriorDirects);
    printf("014 002 Strategy Type:       %10u (0x%4.4x)\n",
           ICBTagPtr->StrategyType,
           ICBTagPtr->StrategyType);
    printf("016 001 Strategy Parameter:  %10u (0x%4.4x)\n",
           (UINT16) ICBTagPtr->StrategyParm[0],
           (UINT16) ICBTagPtr->StrategyParm[0]);
    printf("017 002 Max Number of Ents:  %10u (0x%4.4x)\n",
           ICBTagPtr->NumberEntries,
           ICBTagPtr->NumberEntries);
    printf("019 001 Reserved:            %10u (0x%2.2x)\n",
           ICBTagPtr->Reserved, ICBTagPtr->Reserved);
    printf("01A 001 File Type:           %10u (0x%2.2x) ",
           ICBTagPtr->FileType, ICBTagPtr->FileType);
    switch ( ICBTagPtr->FileType ) {
    case FILE_TYPE_UNSPECIFIED: printf("unspec\n"); break;
    case FILE_TYPE_UNALLOC_SP_ENTRY: printf("unalloc sp entry\n"); break;
    case FILE_TYPE_PARTITION_INTEGRITY: printf("part integ\n"); break;
    case FILE_TYPE_INDIRECT_ENTRY: printf("indirect\n"); break;
    case FILE_TYPE_DIRECTORY: printf("directory\n"); break;
    case FILE_TYPE_RAW: printf("raw\n"); break;
    case FILE_TYPE_BDEV: printf("bdev\n"); break;
    case FILE_TYPE_CDEV: printf("cdev\n"); break;
    case FILE_TYPE_EXT_ATTR: printf("ext attr\n"); break;
    case FILE_TYPE_FIFO: printf("fifo\n"); break;
    case FILE_TYPE_SOCKET: printf("socket\n"); break;
    case FILE_TYPE_TERMINAL_ENTRY: printf("term entry\n"); break;
    case FILE_TYPE_SYMLINK: printf("symbolic link\n"); break;
    default: printf("unknown\n"); break;
    }
    printf("01B 006 Parent ICB Address:\n");
    DisplayLb_addr(ICBTagPtr->sParentICB_LBN,ICBTagPtr->sParentICB_PartNo);
    printf("021 002 Flags:               %10u (0x%4.4x)\n",
           ICBTagPtr->Flags,
           ICBTagPtr->Flags);

    printf("        ADType:              ");
    switch (ICBTagPtr->Flags & ADTYPEMASK) {
    case ADSHORT: printf("short\n"); break;
    case ADLONG: printf("long\n"); break;
    case ADEXTENDED: printf("extended\n"); break;
    case ADNONE: printf("embedded data\n"); break;
    default: printf("*** UNKNOWN ***\n");
    }

    printf("        Characteristics:\n");
    if (ICBTagPtr->Flags & SORTED_DIRECTORY)
        printf("                         sorted dir\n");
    if (ICBTagPtr->Flags & NON_RELOCATABLE )
        printf("                         non relocatable\n");
    if (ICBTagPtr->Flags & ARCHIVE         )
        printf("                         archive\n");
    if (ICBTagPtr->Flags & ICBF_S_ISUID    )
        printf("                         isuid\n");
    if (ICBTagPtr->Flags & ICBF_S_ISGID    )
        printf("                         isgid\n");
    if (ICBTagPtr->Flags & ICBF_C_ISVTX    )
        printf("                         isvtx\n");
    if (ICBTagPtr->Flags & CONTIGUOUS      )
        printf("                         contiguous\n");
    if (ICBTagPtr->Flags & FLAG_SYSTEM     )
        printf("                         system\n");
    if (ICBTagPtr->Flags & FLAG_TANSFORMED )
        printf("                         transformed\n");
    if (ICBTagPtr->Flags & FLAG_MULTIVERS  )
        printf("                         multi-version\n");
}

/****************************************************************************/
/* display the contents of an ISO/IEC 13346 Primary Volume Descriptor       */
/****************************************************************************/

void
DisplayPVD( char *iobuf )
{

    struct PrimaryVolDes *PVDPtr;

    /* need to translate the structures */

#if !IBM
    ot_xlate(TAGID_PVD, FROM_OP, iobuf,0,0);
#endif

    TitleLine("PRIMARY VOLUME DESCRIPTOR");
    DisplayTag( iobuf );

    TitleLine("DESCRIPTOR BODY");

    PVDPtr = (struct PrimaryVolDes *) iobuf;

    printf("010 004 Desc. Sequence #:    %10lu (0x%8.8lx)\n",
           PVDPtr->uVolDescSeqNum, PVDPtr->uVolDescSeqNum);
    printf("014 004 PVD Number:          %10lu (0x%8.8lx)\n",
           PVDPtr->uPrimVolDesNum, PVDPtr->uPrimVolDesNum);
    printf("018 032 Volume Identifier:            ");
    printDstring( PVDPtr->aVolID, 32, 38);
    printf("038 002 VSN:                 %10u (0x%4.4x)\n",
           PVDPtr->uVSN, PVDPtr->uVSN);
    printf("03A 002 Max VSN:             %10u (0x%4.4x)\n",
           PVDPtr->uMaxVSN, PVDPtr->uMaxVSN);
    printf("03C 002 Interchange Level:   %10u (0x%4.4x)\n",
           PVDPtr->uInterchangeLev, PVDPtr->uInterchangeLev);
    printf("03E 002 Max Interchange Lev: %10u (0x%4.4x)\n",
           PVDPtr->uMaxInterchangeLev, PVDPtr->uMaxInterchangeLev);
    printf("040 004 Char Set List:       %10lu (0x%8.8lx)\n",
           PVDPtr->uCharSetList, PVDPtr->uCharSetList);
    printf("045 004 Max Char Set List:   %10lu (0x%8.8lx)\n",
           PVDPtr->uMaxCharSetList, PVDPtr->uMaxCharSetList);

    printf("048 128 Volume Set ID:                ");
    printDstring( PVDPtr->aVolSetID, 128, 38 );
 
    printf("0C8 064 Descriptor Charset:           ");
    DisplayCharSet( PVDPtr->sDesCharSet);

    printf("108 064 Explanatory Charset:          ");
    DisplayCharSet( PVDPtr->sExplanatoryCharSet);

    printf("148 008 Volume Abstract Loc:\n");
    Displayextent_ad( PVDPtr->sVolAbstract );

    printf("150 008 Volume Copyright Loc:\n");
    Displayextent_ad( PVDPtr->sVolCopyrightNotice );

    printf("158 032 Application ID:               ");
    DisplayRegID( & PVDPtr->sApplicationID);

    printf("178 012 Recording Time:               ");
    DisplayTimestamp( PVDPtr->sRecordingTime, 38);

    printf("184 032 Implementation ID:            ");
    DisplayImplID( & PVDPtr->sImplementationID);

    printf("1A4 064 Implementation Use:           ");
    DisplayByteStream((char *)PVDPtr->aImplementationUse, 64, FALSE, 38);

    printf("1E4 004 Predecessor VDS Loc: %10lu (0x%8.8lx)\n",
           PVDPtr->uPredecessorVDSLoc, PVDPtr->uPredecessorVDSLoc);
    printf("1E8 002 Flags:               %10u (0x%4.4x)",
           PVDPtr->uFlags, PVDPtr->uFlags);

    if ( PVDPtr->uFlags & COMMON_VOLSET_IDENTIFICATION) {
        printf(" (COMMON VolumeSet ID)\n");
    } else {
        printf(" (NOT COMMON VolumeSet ID) ** NON-UDF **\n");
    }

    printf("1EA 022 Reserved:                     ");
    DisplayByteStream((char *)PVDPtr->aReserved, 22, FALSE, 38);
    
    TitleLine("END");

}

/****************************************************************************/
/* Routine to display the contents of an ISO/IEC 13346 File Set Descriptor  */
/****************************************************************************/

void
DisplayFSD( char *iobuf )
{

    struct FileSetDesc *FSDPtr;

    /* need to translate the structures */

#if !IBM
    ot_xlate(TAGID_FSD, FROM_OP, iobuf, 0,0);
#endif

    TitleLine("FILE SET DESCRIPTOR");
    DisplayTag(iobuf );

    TitleLine("DESCRIPTOR BODY");

    FSDPtr = (struct FileSetDesc *) iobuf;

    printf("010 012 Recording Time:               ");
    DisplayTimestamp( FSDPtr->sRecordingTime, 38);

    printf("01C 002 InterchangeLevel:    %10lu (0x%8.8lx)\n",
           FSDPtr->uInterchangeLev,
           FSDPtr->uInterchangeLev);
    printf("01E 002 Max Interchange Lev: %10u (0x%4.4x)\n",
           FSDPtr->uMaxInterchangeLev,
           FSDPtr->uMaxInterchangeLev);
    printf("020 004 Char Set List:       %10lu (0x%8.8lx)\n",
           FSDPtr->uCharSetList,
           FSDPtr->uCharSetList);
    printf("024 004 Max Char Set List:   %10lu (0x%8.8lx)\n",
           FSDPtr->uMaxCharSetList,
           FSDPtr->uMaxCharSetList);
    printf("028 004 File Set Number:     %10lu (0x%8.8lx)\n",
           FSDPtr->uFileSetNum,
           FSDPtr->uFileSetNum);
    printf("02C 004 File Set Desc Num:   %10lu (0x%8.8lx)\n",
           FSDPtr->uFileSetDescNum,
           FSDPtr->uFileSetDescNum);
 
    printf("030 064 Log Vol ID Charset:           ");
    DisplayCharSet( FSDPtr->sLogVolIDCharSet);

    printf("070 128 Log Vol ID:                   ");
    printDstring( FSDPtr->aLogVolID, 128, 38);
    
    printf("0f0 064 File Set Charset:             ");
    DisplayCharSet( FSDPtr->sFileSetCharSet);
    
    printf("130 032 File Set ID:                  ");
    printDstring( FSDPtr->aFileSetID, 32, 38);

    printf("150 032 Copyright File Identifier:   ");
    printDstring( FSDPtr->aCopyrightFileID, 32, 38);
    printf("170 032 Abstract  File Identifier:   ");
    printDstring( FSDPtr->aAbstractFileID, 32, 38);
    
    printf("190 016 Root dir ICB:\n");
    DisplayLong_AD(&FSDPtr->sRootDirICB);

    printf("1A0 032 Domain ID:                    ");
    DisplayDomainID( & FSDPtr->DomainID);

    printf("1C0 016 Next Extent:\n");
    DisplayLong_AD(&FSDPtr->sNextExtent);

    printf("1D0 048 Reserved:                     ");
    DisplayByteStream((char *)FSDPtr->aReserved, 48, FALSE, 38);
    
    TitleLine("END");
}

/********************************************************************/
/*  Display a Volume Descriptor Pointer                             */
/********************************************************************/

void
DisplayPointer( char *iobuf )
{

    struct VolDescPtr *PtrPtr;

    /* need to translate the structures */

#if !IBM
    ot_xlate(TAGID_POINTER, FROM_OP, iobuf, 0,0);
#endif

    TitleLine("VOLUME DESCRIPTOR POINTER");
    DisplayTag( iobuf );

    TitleLine("DESCRIPTOR BODY");

    printf("010 004 VDS Number:          %10lu (0x%8.8lx)\n",
           PtrPtr->uVolDescSeqNum,
           PtrPtr->uVolDescSeqNum);

    printf("014 008 Next VDS Extent:\n");
    Displayextent_ad(PtrPtr->sNextVDS);

    printf("01C 484 Reserved:            ");
    DisplayByteStream((char *)PtrPtr->aReserved, 484, FALSE, 38);
} 

/********************************************************************/
/*  Display an Implementation Use Descriptor                        */
/********************************************************************/
void
DisplayImpUse( char *iobuf )
{

    struct ImpUseDesc *IUDPtr;

    /* need to translate the structures */

#if !IBM
    ot_xlate(TAGID_IUD, FROM_OP, iobuf, 0,0);
#endif

    TitleLine("IMPLEMENTATION USE VOLUME DESCRIPTOR");
    DisplayTag( iobuf );

    TitleLine("DESCRIPTOR BODY");

    IUDPtr = (struct ImpUseDesc *) iobuf;

    printf("010 004 VDS Number:          %10lu (0x%8.8lx)\n",
           IUDPtr->uVolDescSeqNum,
           IUDPtr->uVolDescSeqNum);

    printf("014 032 ImplementIdentifier: ");
    DisplayUdfID( & IUDPtr->sImplementationIdentifier);

    if (strncmp((char *)IUDPtr->sImplementationIdentifier.aID,
                OSTA_LVINFO_ID, 23) == 0) {
        struct LVInformation * lvip = (struct LVInformation *)IUDPtr->aReserved;

        printf("034 064 LVICharset:          ");
        DisplayCharSet( lvip->sLVICharset );

        printf("074 128 LogVolume Id:        ");
        printDstring(lvip->aLogicalVolumeIdentifier,128,38);
        
        printf("0F4 036 LogVolume Info[1]:   ");
        printDstring(lvip->aLVInfo1,36,38);
        
        printf("118 036 LogVolume Info[2]:   ");
        printDstring(lvip->aLVInfo2,36,38);
        
        printf("13C 036 LogVolume Info[3]:   ");
        printDstring(lvip->aLVInfo3,36,38);
        
        printf("160 032 Implementation Id:   ");
        DisplayImplID(& lvip->sImplementationID);
        
        printf("180 128 Implementation Use:  ");
        DisplayByteStream((char *)lvip->aImplementationUse,128,
                          FALSE,38);
    } else {
        printf("01C 460 Reserved:            ");
        DisplayByteStream((char *)IUDPtr->aReserved, 460, FALSE, 38);
    }
} 

/********************************************************************/
/*  Display and Anchor Volume Descriptor Pointer                    */
/********************************************************************/

void
DisplayAnchor( char *iobuf )
{

    struct AnchorVolDesPtr *AVDPtr;

    /* need to translate the structures */
#if !IBM
    ot_xlate(TAGID_ANCHOR, FROM_OP, iobuf,0,0);
#endif    


    TitleLine("ANCHOR VOLUME DESCRIPTOR POINTER");
    DisplayTag( iobuf );

    TitleLine("DESCRIPTOR BODY");

    AVDPtr = (struct AnchorVolDesPtr *) iobuf;
    
    printf("010 008 Main Sequence Address:\n");
    Displayextent_ad( AVDPtr->sMainVDSAdr);
    printf("018 008 Reserve Sequence Address:\n");
    Displayextent_ad( AVDPtr->sReserveVDSAdr);
    printf("020 008 Reserved:                 ");
    DisplayByteStream((char *)AVDPtr->Reserved, 480, FALSE, 38);
    TitleLine("END");
}

/********************************************************************/
/*  Display a Partition Descriptor                                  */
/********************************************************************/

void
DisplayPD( char *iobuf )
{
    struct PartDesc *PDPtr;
    struct PartHeaderDesc *PHDp;

    /* need to translate the structures */
#if !IBM
    ot_xlate(TAGID_PD, FROM_OP, iobuf, 0,0);
#endif    


    TitleLine("PARTITION DESCRIPTOR");
    DisplayTag( iobuf );

    TitleLine("DESCRIPTOR BODY");

    PDPtr = (struct PartDesc *) iobuf;

    printf("010 004 VDS Number:          %10lu (0x%8.8x)\n",
           PDPtr->uVolDescSeqNum, PDPtr->uVolDescSeqNum);

    printf("014 002 Parition Flags:      %10u (0x%4.4x)",
           PDPtr->uPartFlags, PDPtr->uPartFlags);
    if ( PDPtr->uPartFlags & PARTITION_ALLOCATED) {
        printf(" (Space Allocated)\n");
    } else {
        printf(" (Space NOT Allocated) ?? NON-UDF ??\n");
    }

    printf("016 002 Partition Number:    %10u (0x%4.4x)\n",
           PDPtr->uPartNumber, PDPtr->uPartNumber);

    printf("018 032 Partition Contents:           ");
    DisplayRegID( & PDPtr->sPartContents);
 
    printf("038 128 Partition Use:\n");
    PHDp = (struct PartHeaderDesc *)PDPtr->aPartContentsUse;
    
    printf("038 008 UST:\n");
    DisplayShort_AD(&PHDp->UST);
    printf("040 008 USB:\n");
    DisplayShort_AD(&PHDp->USB);
    printf("048 008 PIT:\n");
    DisplayShort_AD(&PHDp->PIT);
    printf("050 008 FST:\n");
    DisplayShort_AD(&PHDp->FST);
    printf("058 008 FSB:\n");
    DisplayShort_AD(&PHDp->FSB);
    printf("060 088 Reserved:                     ");
    DisplayByteStream((char *)PHDp->aReserved, 88, FALSE, 38);

    printf("0B8 004 Access Type:         %10lu (0x%8.8x)",
           PDPtr->uAccessType, PDPtr->uAccessType);

    switch ( PDPtr->uAccessType ) {
    case ACCESS_UNSPECIFIED:  printf(" (Unspecified)\n"); break;
    case ACCESS_READ_ONLY:    printf(" (ReadOnly)\n"); break;
    case ACCESS_WORM:         printf(" (WriteOnce)\n"); break;
    case ACCESS_REWRITABLE:   printf(" (Rewritable)\n"); break;
    case ACCESS_OVERWRITABLE: printf(" (Overwritable)\n"); break;
    default:                  printf(" (***Non-ISO***)\n"); break;
    }
        
    printf("0BC 004 Starting Location:  %10lu (0x%8.8x)\n",
           PDPtr->uPartStartingLoc, PDPtr->uPartStartingLoc);
    printf("0C0 004 Length:              %10lu (0x%8.8x)\n",
           PDPtr->uPartLength, PDPtr->uPartLength);

    printf("0C4 032 Implementation ID:            ");
    DisplayImplID( & PDPtr->sImplementationID);

    printf("0E4 128 Implementation Use:           ");
    DisplayByteStream((char *)PDPtr->aImplementationUse, 128, FALSE, 38);

    printf("164 156 Reserved:                     ");
    DisplayByteStream((char *)PDPtr->aReserved, 156, FALSE, 38);

    TitleLine("END");
}

/********************************************************************/
/*  Display a Partition Integrity Entry                             */
/********************************************************************/
void
DisplayPIT( char *iobuf )
{

    struct PartIntegrityEntry *pie;

    /* need to translate the structures */
#if !IBM
    ot_xlate(TAGID_PART_INTEGRITY, FROM_OP, iobuf, 0,0);
#endif    


    TitleLine("PARTITION INTEGRITY ENTRY");
    DisplayTag( iobuf );
    DisplayICBTag( iobuf + sizeof( struct Tag));

    TitleLine("DESCRIPTOR BODY");

    pie = (struct PartIntegrityEntry *) iobuf;

    printf("        Recording Time:               ");
    DisplayTimestamp(pie->sRecordingTime, 38);
    
    printf("        IntegrityType :      %10lu (0x%8.8x)\n",
           pie->IntegrityType,
           pie->IntegrityType);
    
    TitleLine("END");
    
}

/********************************************************************/
/*  Display an Indirect Entry                                       */
/********************************************************************/
void
DisplayIE(char *iobuf )
{

    struct IndirectEntry *ie;

    /* need to translate the structures */
#if !IBM
    ot_xlate(TAGID_INDIRECT, FROM_OP,iobuf, 0,0);
#endif    


    TitleLine("INDIRECT ENTRY");
    DisplayTag( iobuf );
    DisplayICBTag( iobuf + sizeof( struct Tag));

    ie = (struct IndirectEntry *) iobuf;

    TitleLine("INDIRECT ICB");

    printf("024 016 Indirect ICB:\n");
    DisplayLong_AD(&ie->sIndirectICB);
    
    TitleLine("END");
    
}

/********************************************************************/
/*  Display a Logical Volume Descriptor                             */
/********************************************************************/
void
DisplayLVD( char *iobuf, UINT32 bufLen) 
{
    struct LogVolDesc *LVDPtr;
    int i;

    /* need to translate the structures */
#if !IBM
    ot_xlate(TAGID_LVD, FROM_OP, iobuf, bufLen,0);
#endif    

    TitleLine("LOGICAL VOLUME DESCRIPTOR");
    DisplayTag( iobuf );

    TitleLine("DESCRIPTOR BODY");

    LVDPtr = (struct LogVolDesc *) iobuf;

    printf("010 004 VDS Number:          %10lu (0x%8.8x)\n",
           LVDPtr->uVolDescSeqNum
           , LVDPtr->uVolDescSeqNum);

    printf("014 064 Descriptor Charset:           ");
    DisplayCharSet( LVDPtr->sDesCharSet);

    printf("054 128 Logical Volume ID:            ");
    printDstring( LVDPtr->uLogVolID, 128, 38);

    printf("0D4 004 Logical Block Size:  %10lu (0x%8.8lx)\n",
           LVDPtr->uLogBlkSize,
           LVDPtr->uLogBlkSize);

    printf("0D8 032 Domain ID:                    ");
    DisplayDomainID( & LVDPtr->sDomainID);

    printf("0F8 016 Logical Volume Use:\n");
    DisplayLong_AD((struct long_AD *)&LVDPtr->uLogVolUse);

    printf("108 004 Map Table Length:    %10lu (0x%8.8x)\n",
           LVDPtr->uMapTabLen,
           LVDPtr->uMapTabLen);
    printf("10C 004 Number Part Maps:    %10lu (0x%8.8x)\n",
           LVDPtr->uNumPartMaps,
           LVDPtr->uNumPartMaps);
   
    printf("110 032 Implementation ID:            ");
    DisplayImplID( & LVDPtr->sImplementationID);

    printf("130 128 Implementation Use:           ");
    DisplayByteStream((char *) LVDPtr->aImplementationUse, 128, FALSE, 38);

    printf("1B0 008 Integrity Sequence Address:\n");
    Displayextent_ad( LVDPtr->integritySeqExtent);
    
    {
        /* Make sure the number of partition maps is reasonable */
        UINT16 npm = LVDPtr->uNumPartMaps;
        UINT16 MAXnpm = (bufLen - LVD_HdrLen)/sizeof(struct PartMap1);
        npm = MIN(npm,MAXnpm);

        printf("%3.3X %3.3u Partition Map Table:\n",
               LVD_HdrLen,LVDPtr->uMapTabLen);
        for(i=0; i<npm; i++) {
            printf("%3.3X %3.3u   Map[%2.2u]: VolSeqNum:%10lu (0x%8.8x)\n",
                   LVD_HdrLen+(i*sizeof(struct PartMap1)),
                   sizeof(struct PartMap1),i,
                   (LVD_PM1Start(LVDPtr))[i].uVSN,
                   (LVD_PM1Start(LVDPtr))[i].uVSN);
            printf("                  PartNum:   %10lu (0x%8.8x)\n",
                   (LVD_PM1Start(LVDPtr))[i].uPartNum,
                   (LVD_PM1Start(LVDPtr))[i].uPartNum);
        }
    }
    
    TitleLine("END");
}    

/********************************************************************/
/*  Display an Unallocated Space Descriptor                         */
/********************************************************************/
void
DisplayUSD( char * iobuf )
{
    struct UnallocSpDesHead *USDPtr;
    struct extent_ad *EADPtr;
    int i;

    /* need to translate the structures */
#if !IBM
    ot_xlate(TAGID_USD, FROM_OP, iobuf, 0,0);
#endif    


    TitleLine("UNALLOCATED SPACE DESCRIPTOR");
    DisplayTag( iobuf );

    TitleLine("DESCRIPTOR BODY");

    USDPtr = (struct UnallocSpDesHead *) iobuf;

    printf("010 004 Desc. Sequence #:    %10lu (0x%8.8lx)\n",
           USDPtr->uVolDescSeqNum,
           USDPtr->uVolDescSeqNum);

    printf("010 004 Num Allocation Des:  %10lu (0x%8.8x)\n",
           USDPtr->uNumAllocationDes,
           USDPtr->uNumAllocationDes);

    for(i=0; i<USDPtr->uNumAllocationDes; i++) {
        EADPtr = ((struct extent_ad *) (iobuf + sizeof( struct UnallocSpDesHead))) + i;
        printf("%3x 008 Descriptor %2d\n", 20+(8*i), i);
        Displayextent_ad( *EADPtr);
    }
 
    TitleLine("END");

}

/********************************************************************/
/*   Display a Terminator Descriptor                                */
/********************************************************************/
void
DisplayTerm( char *iobuf )
{
    struct Terminator *TermPtr;

    /* need to translate the structures */
#if !IBM
    ot_xlate(TAGID_TERM_DESC, FROM_OP, iobuf, 0,0);
#endif    


    TitleLine("TERMINATING DESCRIPTOR");
    DisplayTag( iobuf );

    TitleLine("DESCRIPTOR BODY");

    TermPtr = (struct Terminator *) iobuf;

    printf("010 496 Reserved Bytes:               ");
    DisplayByteStream((char *) TermPtr->aReserved, 496, FALSE, 38);

    TitleLine("END");

}

/********************************************************************/
/*  Display a Logical Volume Integrity Descriptor                   */
/********************************************************************/
void
DisplayLVID( char *iobuf, UINT32 bufLen ) 
{
    struct LogicalVolumeIntegrityDesc *LVIDPtr;
    int i;

    /* need to translate the structures */
#if !IBM
    ot_xlate(TAGID_LVID, FROM_OP, iobuf, bufLen,0);
#endif    


    TitleLine("LOGICAL VOLUME INTEGRITY DESCRIPTOR");
    DisplayTag( iobuf );

    TitleLine("DESCRIPTOR BODY");

    LVIDPtr = (struct LogicalVolumeIntegrityDesc *) iobuf;

    printf("010 012 Recording Date:               ");
    DisplayTimestamp( LVIDPtr->sRecordingTime, 38);

    printf("01C 004 Integrity Type:      %10lu (0x%8.8x)",
           LVIDPtr->integrityType,
           LVIDPtr->integrityType);
    switch ( LVIDPtr->integrityType ) {
    case INTEGRITY_OPEN:  printf(" (OPEN)\n"); break;
    case INTEGRITY_CLOSE: printf(" (CLOSED)\n"); break;
    default:              printf(" (INVALID) ** NON-ISO/UDF **\n"); break;
    }

    printf("020 008 Next Integrity Extent:\n");
    Displayextent_ad( LVIDPtr->nextIntegrityExtent);

    printf("028 004 Unique ID   (Low):   %10lu (0x%8.8lx)\n",
           LVIDPtr->UniqueIdL,LVIDPtr->UniqueIdL);
    printf("02C 004 Unique ID  (High):   %10lu (0x%8.8lx)\n",
           LVIDPtr->UniqueIdH,LVIDPtr->UniqueIdH);

    printf("030 024 reserved:                     ");
    DisplayByteStream( (char *)LVIDPtr->reserved, 24, FALSE, 38);

    printf("048 004 Number of Partitions:%10lu (0x%8.8lx)\n",
           LVIDPtr->N_P,LVIDPtr->N_P);
    printf("04C 004 Length of Implement Use%8lu (0x%8.8lx)\n",
           LVIDPtr->L_IU,LVIDPtr->L_IU);

    {
        /* NOTE that ANSI C standard says that pointer subtraction if
           not in the same array is undefined, so must go to char *
           domain. */
        
        /* Get offset and pointer to freeSpace table */
        UINT32 *fs      = LVID_FSTStart(LVIDPtr);
        UINT32 fsOffset = ((char *)fs - (char *)LVIDPtr);
        INT32  maxnfs   = (bufLen-fsOffset)/sizeof(UINT32);
        UINT32 nfs      = MIN(LVIDPtr->N_P,(maxnfs>0?maxnfs:0));

        /* Get offset and pointer to size table */
        UINT32 *s       = LVID_SZTStart(LVIDPtr);
        UINT32 sOffset  = ((char *)s - (char *)LVIDPtr);
        INT32  maxns    = (bufLen-sOffset)/sizeof(UINT32);
        UINT32 ns       = MIN(LVIDPtr->N_P,(maxns>0?maxns:0));

        /* Get offset and pointer to implement use area */
        UCHAR8 *iu      = LVID_IUStart(LVIDPtr);
        UINT32 iuOffset = ((char *)iu - (char *)LVIDPtr);
        INT32  maxliu   = bufLen-iuOffset;
        UINT32 liu      = MIN(LVIDPtr->L_IU,(maxliu>0?maxliu:0));

        /* Get pointer to structured representation of implUse area */
        struct LVIDImplUse * LVIDIUPtr = (struct LVIDImplUse *)iu;

        /* Get offset and pointer to implUseImplUse area */
        UCHAR8 *iuiu      = LVIDIU_IUStart(LVIDIUPtr);
        UINT32 iuiuOffset = iuOffset + LVIDIU_HdrLen;
        INT32  maxliuiu   = bufLen - iuiuOffset;
        INT32  L_IUIU     = LVIDPtr->L_IU - LVIDIU_HdrLen;
        UINT32 liuiu      = MIN((L_IUIU>0?L_IUIU:0),(maxliuiu>0?maxliuiu:0));
        
        printf("%3.3X %3.3u Free Space Table:\n",
               fsOffset,LVIDPtr->N_P * sizeof(UINT32));
        for(i=0; i<nfs; i++) {
            printf("%3.3X 004   Free Space [%2.2d]:   %10lu (0x%8.8lx)\n",
                   (fsOffset + 4*i),i,
                   fs[i],fs[i]);
        }

        printf("%3.3X %3.3u Size Table:\n",
               sOffset,LVIDPtr->N_P * sizeof(UINT32));
        for(i=0; i<ns; i++) {
            printf("%3.3X 004   Size [%2.2d]:         %10lu (0x%8.8lx)\n",
                   (sOffset + 4*i),i,
                   s[i],s[i]);
        }

        printf("%3.3X 032 Implementation ID:            ",iuOffset);
        if( liu >= sizeof(struct regid) ) {
            DisplayImplID( & LVIDIUPtr->implementationID );
        } else {
            printf(".......... (unavailable)\n");
        }

        printf("%3.3X 004 NumFiles:            ",iuOffset+32);
        if( liu >= 32 + 4 ) {
            printf("%10lu (0x%8.8lx)\n",
                   LVIDIUPtr->numFiles,
                   LVIDIUPtr->numFiles);
        } else {
            printf(".......... (unavailable)\n");
        }
        
        printf("%3.3X 004 NumDirectories:      ",iuOffset+36);
        if( liu >= 32 + 4 + 4 ) {
            printf("%10lu (0x%8.8lx)\n",
                   LVIDIUPtr->numDirectories,
                   LVIDIUPtr->numDirectories);
        } else {
            printf(".......... (unavailable)\n");
        }

        printf("%3.3X %3.3u Implementation Use:           ",
               iuiuOffset,(L_IUIU>0?L_IUIU:0));
        if (liuiu > 0) {
            DisplayByteStream( (char *)iuiu, liuiu, FALSE, 38);
        } else {
            printf("\n");
        }
    }
    TitleLine("END");
}    

/********************************************************************/
/* Display an ISO/IEC 13346 Directory Entry                         */
/********************************************************************/

void
DisplayFID( char *iobuf, UINT32 length )
{
    struct FileIDDesc *fidp;
    INT16  padlen, i;
    UINT16 fidLen;

    /* Used to calculate field locations and lengths */
    char * idStart = 0;         /* identifier field info */
    UINT16 idOffs  = 0;
    UINT16 idLen   = 0;
    char * iuStart = 0;         /* implementation use field info */
    UINT16 iuOffs  = 0;
    UINT16 iuLen   = 0;

    
    fidp = (struct FileIDDesc *) iobuf;
    i=0;
    
    TitleLine("ISO/IEC 13346 FILE ID DESCRIPTOR");
    
    /* need to translate the structures */
    
#if !IBM
    ot_xlate(TAGID_FILE_ID, FROM_OP, (char *) fidp, length, 0);
#endif    

    fidLen = ( FILE_ID_DESC_CONSTANT_LEN +
               fidp->L_FI +
               fidp->L_IU);
    fidLen = ((fidLen + 3)<<2)>>2; /* round up to mult of four */
    
    while (fidp->sTag.uTagID == TAGID_FILE_ID &&
           (char *)fidp + fidLen <= iobuf + length ) {
        
        TitleLine("");
        printf("Directory entry %d\n", i++);
        DisplayTag( (char *)fidp );
        TitleLine("DIRECTORY BODY");
        
        printf("010 002 File Version:        %10u (0x%4.4x)\n",
               fidp->VersionNum, fidp->VersionNum);
        printf("012 001 Characteristics:     %10u (0x%2.2x)",
               fidp->Characteristics, fidp->Characteristics);
        if( fidp->Characteristics & PARENT_ATTR) printf(" (PARENT)");
        if( fidp->Characteristics & DIR_ATTR) printf(" (DIRECTORY)");
        if( fidp->Characteristics & HIDDEN_ATTR) printf(" (HIDDEN)");
        if( fidp->Characteristics & DELETE_ATTR) printf(" (DELETED)");
        printf("\n");

        /* Field lengths */
        iuLen = fidp->L_IU;
        idLen = fidp->L_FI;
        /* Offsets into the fid */
        iuOffs = FILE_ID_DESC_CONSTANT_LEN;
        idOffs = iuOffs + iuLen;
        /* Addresses */
        iuStart = (char *)fidp + iuOffs;
        idStart = (char *)fidp + idOffs;

        printf("013 001 Length of File ID:   %10u (0x%2.2x)\n",
               idLen,idLen);

        printf("014 016 ICB Location:\n");
        DisplayLong_AD(& (fidp->ICB));

        printf("024 002 Length of Imp. Use:  %10u (0x%4.4x)\n",
               iuLen,iuLen);

        if( iuLen != 0 ) {
            if( iuLen >= 32 ) {
                /* print the implEntityID */
                struct implEntityId *ieip=(struct implEntityId * ) iuStart;
                printf("%3.3X 032 Implementation Use ID:        ",iuOffs);
                DisplayImplID(ieip);
                if( iuLen > 32 ) {
                    /* Print extra bytes after RegID */
                    iuLen   -= 32;
                    iuOffs  += 32;
                    iuStart += 32;
                    printf("%3.3X %3.3u Implementation Use:           ",
                           iuOffs,iuLen);
                    DisplayByteStream(iuStart,iuLen, FALSE, 38);
                }
            } else {
                /* Print all bytes (Non-UDF-compliant) */
                printf("%3.3X %3.3u Implementation Use:           ",
                       iuOffs,iuLen);
                DisplayByteStream(iuStart,iuLen, FALSE, 38);
            }
        }

        printf("%3.3X %3.3u File Identifier:      ",idOffs,idLen);
        printDchars(idStart,idLen, 38);

        padlen = 4 * ((idOffs + idLen + 3)/4) - (idOffs + idLen);

        if( padlen != 0 ) {
            printf("%3.3X %3.3u Pad:                          ",
                   idOffs+idLen,padlen);
            DisplayByteStream(idStart+idLen,padlen, FALSE, 38);
        }
        
        /* prepare for next entry */
        
        fidp = (struct FileIDDesc *)(idStart+idLen+padlen);
#if !IBM
        if (fidp->sTag.uTagID == TAGID_FILE_ID) {
            ot_xlate( TAGID_FILE_ID, FROM_OP, (char *) fidp,
                      (iobuf+length) - (char *)fidp, 0 );
        }
#endif
        fidLen = ( FILE_ID_DESC_CONSTANT_LEN +
                   fidp->L_FI +
                   fidp->L_IU);
        fidLen = ((fidLen + 3)<<2)>>2; /* round up to mult of four */
    }
    /* Display raw bytes of everything left */
    if( (char *)fidp < (iobuf+length) ) {
        TitleLine("INCOMPLETE FIDS OR UNUSED BYTES");
        printf("%3.3X %3.3u Leftovers:                    ",
               (iobuf+length) -(char *)fidp,
               (iobuf+length) -(char *)fidp);
        DisplayByteStream((char *)fidp,
                          (iobuf+length) -(char *)fidp,
                          FALSE, 38);
    }
    TitleLine("END");
}


/**********************************/
/* Display allocation descriptors */
/**********************************/
void
DisplayADs(UINT32 ADType,UINT32 length,void *fp)
{
    int i;
    int sizeofADs = ot_sizeOfAD(ADType);
    int numADs = length/sizeofADs;
    struct short_AD *sadp = (struct short_AD *)fp;
    struct long_AD *ladp = (struct long_AD *)fp;
    struct ext_AD *eadp = (struct ext_AD *)fp;

    TitleLine("ALLOCATION DESCRIPTORS");
    
    switch(ADType) {
    case ADSHORT:
        for( i=0; i< numADs; i++) 
        {
            printf("    %03d AD[%2d]: [%c] EL:%6d Loc:%6d \n",
                   sizeofADs, i,
                   ETypeStr[sadp->ExtentLength.bf.Type],
                   sadp->ExtentLength.bf.Length,
                   sadp->Location);
            sadp++;
        }
        break;
    case ADLONG:
        for( i=0; i< numADs; i++) 
        {
            printf("    %03d AD[%2d]: [%c] EL:%6d LBN:%6d PN:%5d %s IU:",
                   sizeofADs, i,
                   ETypeStr[ladp->ExtentLength.bf.Type],
                   ladp->ExtentLength.bf.Length,
                   ladp->Location_LBN, ladp->Location_PartNo,
                   (ladp->uUdfFlags & EXT_ERASED)? "(Erased)" : "(Unerased)");
            DisplayByteStream((char *) ladp->aImpUse, 4, FALSE, 38);
            ladp++;
        }
        break;
    case ADEXTENDED:
        for( i=0; i< numADs; i++) 
        {
            printf("     %03d AD[%2d]: [%c] EL:%6d RL:%6d IL:%6d LBN:%6d PN:%5d IU:",
                   sizeofADs, i,
                   ETypeStr[eadp->ExtentLength.bf.Type],
                   eadp->ExtentLength.bf.Length,
                   eadp->RecordedLength,eadp->InfoLength,
                   eadp->Location_LBN,eadp->Location_PartNo);
            DisplayByteStream((char *) eadp->aImpUse, 2, FALSE, 38);
            eadp++;
        }
        break;
    }
    TitleLine("END");
}

/********************************************************************/
/* Display an ISO/IEC 13346 Space Bitmap                            */
/********************************************************************/
void
DisplayBMAP(char *iobuf,int nbytes)
{
    struct SpaceBitmapEntry *bmap = (struct SpaceBitmapEntry *)iobuf;

    TitleLine("ISO/IEC 13346 SPACE BITMAP ENTRY");
    /* need to translate the structures */
    
#if !IBM
    ot_xlate(TAGID_SPACE_BMAP, FROM_OP, (char *) iobuf, 0,0);
#endif    

    DisplayTag( iobuf );

    printf("       N_Bits:                %10lu (0x%8.8lx)\n",
           bmap->hdr.N_Bits,bmap->hdr.N_Bits);
    printf("      N_Bytes:                %10lu (0x%8.8lx)\n",
           bmap->hdr.N_Bytes,bmap->hdr.N_Bytes);

    hexdump(nbytes-sizeof(bmap->hdr),(char *)bmap->bitmap);
}

/********************************************************************/
/* Display an ISO/IEC 13346 Unallocated Space Map Entry             */
/********************************************************************/
void
DisplayUST(char *iobuf,UINT32 length)
{
    struct UnallocSpEntry *ust = (struct UnallocSpEntry *)iobuf;

    TitleLine("ISO/IEC 13346 UNALLOCATED SPACE ENTRY");
    /* need to translate the structures */
    
#if !IBM
    ot_xlate(TAGID_UNALLOC_SP_ENTRY, FROM_OP, (char *) iobuf,length,0);
#endif    

    DisplayTag(iobuf );
    DisplayICBTag(iobuf + sizeof(struct Tag));
    printf("        L_AD:                %10lu (0x%8.8lx)\n",
           ust->L_AD, ust->L_AD);
    length = MIN(length-USE_HdrLen,ust->L_AD);
    DisplayADs(ust->sICBTag.Flags & ADTYPEMASK,ust->L_AD,USE_ADStart(ust));
}

/********************************************************************/
/* Display an ISO/IEC 13346 File Entry                              */
/********************************************************************/
void
DisplayFE(char *iobuf, UINT32 length)
{
    /* For displaying the type of extent in the ADs */
    /* Recorded/Allocated/Unallocated/eXtended */

    struct FileEntry *fep;
    INT16 padlen;
    
    fep = (struct FileEntry *) iobuf;
    
    TitleLine("ISO/IEC 13346 FILE ENTRY");
    
    /* need to translate the structures */
    
#if !IBM
    ot_xlate(TAGID_FILE_ENTRY, FROM_OP, (char *) fep, length,0);
#endif    

    DisplayTag( iobuf );
    DisplayICBTag( iobuf + sizeof( struct Tag));

    TitleLine("FILE ENTRY BODY");

    printf("024 004 UID:                 %10lu (0x%8.8lx)\n",
           fep->UID, fep->UID);
    printf("028 004 GID:                 %10lu (0x%8.8lx)\n",
           fep->GID, fep->GID);
    printf("02C 004 Permissions:         %10lu (0x%8.8lx)\n",
           fep->Permissions, fep->Permissions);
    printf("030 002 LinkCount:           %10u (0x%4.4x)\n",
           fep->LinkCount, fep->LinkCount);
    printf("032 001 Record Format:       %10u (0x%2.2x)\n",
           fep->RecFormat, fep->RecFormat);
    printf("033 001 Display Attributes:  %10u (0x%2.2x)\n",
           fep->RecDisplayAttr, fep->RecDisplayAttr);
    printf("034 004 Record Length:       %10lu (0x%8.8lx)\n",
           fep->RecLength, fep->RecLength);
    printf("038 004 Info Length  (Low):  %10lu (0x%8.8lx)\n",
           fep->InfoLengthL, fep->InfoLengthL);
    printf("03C 004 Info Length (High):  %10lu (0x%8.8lx)\n",
           fep->InfoLengthH, fep->InfoLengthH);
    printf("040 004 Log Blocks   (Low):  %10lu (0x%8.8lx)\n",
           fep->LogBlocksL, fep->LogBlocksL);
    printf("044 004 Log Blocks  (High):  %10lu (0x%8.8lx)\n",
           fep->LogBlocksH, fep->LogBlocksH);

    printf("048 012 Access Time:                  ");
    DisplayTimestamp( fep->sAccessTime, 38);

    printf("054 012 Modification Time:            ");
    DisplayTimestamp( fep->sModifyTime, 38);

    printf("060 012 Attribute Change Time:        ");
    DisplayTimestamp( fep->sAttrTime, 38);

    printf("06C 004 Checkpoint:          %10lu (0x%8.8lx)\n",
           fep->Checkpoint, fep->Checkpoint);

    printf("070 016 Extended Attr. ICB:\n");
    DisplayLong_AD(&fep->sExtAttrICB);

    printf("07F 032 Implementation Use ID:        ");
    DisplayImplID( & fep->sImpID);

    printf("0A0 004 Unique ID   (Low)  : %10lu (0x%8.8lx)\n",
           fep->UniqueIdL, fep->UniqueIdL);
    printf("0A4 004 Unique ID  (High)  : %10lu (0x%8.8lx)\n",
           fep->UniqueIdH, fep->UniqueIdH);

    printf("0A8 004 Length Ext. Attr.  : %10lu (0x%8.8lx)\n",
           fep->L_EA, fep->L_EA);
    printf("0AC 004 Length Alloc. Desc.: %10lu (0x%8.8lx)\n",
           fep->L_AD, fep->L_AD);

    if( fep->L_EA) {
        printf("0B0 LEA Extended Attr. Area:          ");
        DisplayByteStream(FE_EAStart(fep), fep->L_EA, FALSE, 38);
    }
    length = MIN(length-FE_HdrLen(fep),fep->L_AD);
    DisplayADs(fep->sICBTag.Flags & ADTYPEMASK,length,FE_ADStart(fep));
}
/*
 * DisplayAE()
 *
 * This displays an Allocation Extent. Note that the ADType must be
 * supplied to this function, as the AE itself does NOT hold this
 * information.
 */
void
DisplayAE( void *iobuf,UINT32 length,UINT32 ADType)
{
    struct AllocationExtentDesc *AEp = iobuf;
    TitleLine("ISO/IEC 13346 ALLOCATION EXTENT");
#if !IBM
    ot_xlate(TAGID_ALLOC_EXTENT, FROM_OP, (char *) AEp,length,ADType);
#endif
    DisplayTag( iobuf );

    TitleLine("ALLOCATION EXTENT BODY");
    printf("010 004 Previous AE Loc:     %10lu (0x%8.8lx)\n",
           AEp->prevAllocExtLoc,AEp->prevAllocExtLoc);
    printf("014 004 Length Alloc. Desc.: %10lu (0x%8.8lx)\n",
           AEp->L_AD, AEp->L_AD);
    length = MIN(length-AE_HdrLen,AEp->L_AD);
    DisplayADs(ADType,length,AE_ADStart( iobuf ));
}

/********************************************************************/
/* Display a string centered between dashes                         */
/********************************************************************/
void
TitleLine(char *title)
{
    int numleft, numright, titlesize,i;
    
    titlesize=strlen( title);
    numleft = (TITLELEN - titlesize)/2;
    numright = TITLELEN - numleft - titlesize;

    for(i=0; i < numleft; printf("-"), i++);
    printf("%s", title);
    for(i=0; i < numright; printf("-"), i++);
    printf("\n");
}

/********************************************************************/
/*   Display an ISO/IEC 13346 extent_ad structure                   */
/********************************************************************/
void
Displayextent_ad( struct extent_ad address)
{
    /* display an extent_ad address type */
    printf("        Length:              %10lu (0x%8.8lx)\n",
           address.Length, address.Length);
    printf("        Location:            %10lu (0x%8.8lx)\n",
           address.Location, address.Location);
}

/********************************************************************/
/*   Display an ISO/IEC 13346 lb_addr structure                     */
/********************************************************************/
void
DisplayLb_addr(UINT32 lbn, UINT16 pn)
{
    /* display an lb_addr address type */
    /* presumes cursor position is already set, terminates with a newline */
    printf("        LogicalBlockNum:     %10lu (0x%8.8lx)\n", lbn, lbn);
    printf("        PartitionRefNum:     %10lu (0x%8.8lx)\n", pn, pn);
}

/********************************************************************/
/*   Display an ISO/IEC 13346 Short_AD structure                    */
/********************************************************************/
void
DisplayShort_AD(struct short_AD *ADp)
{
    /* display an Short_AD address type */
    printf("    004 ExtentLength     [%c] %10lu (0x%8.8lx)\n",
           ETypeStr[ADp->ExtentLength.bf.Type],
           ADp->ExtentLength.bf.Length,
           ADp->ExtentLength.Length32);
    printf("    004 ExtentLocation:      %10lu (0x%8.8lx)\n",
           ADp->Location ,ADp->Location);
}

/********************************************************************/
/*   Display an ISO/IEC 13346 long_AD structure                     */
/********************************************************************/
void
DisplayLong_AD(struct long_AD *ADp)
{
    /* display an long_AD address type */
    printf("    004 ExtentLength     [%c] %10lu (0x%8.8lx)\n",
           ETypeStr[ADp->ExtentLength.bf.Type],
           ADp->ExtentLength.bf.Length,
           ADp->ExtentLength.Length32);
    DisplayLb_addr(ADp->Location_LBN,ADp->Location_PartNo);

    printf("    002 UDF Flags:           %10u (0x%4.4x)",
           ADp->uUdfFlags,ADp->uUdfFlags);
    if ( ADp->uUdfFlags & EXT_ERASED ) {
        printf(" (Erased)");
        if ( ADp->ExtentLength.bf.Type != E_ALLOCATED )
            printf("** NON-UDF ** - (type must be Allocated)");
    } else {
        printf(" (Unerased)");
    }
    printf("\n");

    printf("    004 ImplementationUse:            ");
    DisplayByteStream((char *) ADp->aImpUse, 4, FALSE, 38);
}

/********************************************************************/
/*  Display a length bytes.  Start with a newline if newline is     */
/*  and indent indent spaces after a newline.  If a a byte value    */
/*  is repeated more than twice in a row, type <XXX> of YY where    */
/*  XXX is the number of times repeated and YY is the byte value    */
/********************************************************************/
void
DisplayByteStream(char *bytestream, int length, int newline, int indent)
{
    int i;
    int repeatcount, columncount;
    UCHAR8 repeatbyte;

    columncount = 0;

    if( newline) {
        printf("\n");
        printspace( indent);
    }

    /* display length bytes */
    /* assumes cursor position is set, terminates with a newline */

    for(i=1, repeatbyte = *(bytestream), repeatcount=1; i<length; i++) {
        if(*(++bytestream) == repeatbyte) {
            repeatcount++;
        } else {
            /* flush the bytestream */
            if (((columncount == MAXCOLUMNS) && (repeatcount == 1)) ||
                ((columncount >= MAXCOLUMNS - 3) && (repeatcount != 1))) {
                printf("\n");
                printspace( indent);
                columncount = 0;
            }
            if (repeatcount == 1) {
                printf("%2.2X ", repeatbyte);
                columncount++;
            } else if (repeatcount == 2) {
                printf("%2.2X %2.2X ", repeatbyte, repeatbyte);
                columncount+=2;
            } else {
                printf("<%3.3u> of %2.2X ", repeatcount, repeatbyte);
                columncount+=3;
            }
            /* reset the counts */
            repeatcount = 1;
            repeatbyte = *bytestream;
        }
    }

    /* flush the bytestream */

    if (((columncount == MAXCOLUMNS) && (repeatcount == 1)) ||
        ((columncount >= MAXCOLUMNS - 3) && (repeatcount != 1))) {
        printf("\n");
        printspace( indent);
        columncount = 0;
    }

    if (repeatcount == 1) {
        printf("%2.2X\n", repeatbyte);
    } else if (repeatcount == 2) {
        printf("%2.2X %2.2X\n", repeatbyte, repeatbyte);
    } else {
        printf("<%3.3u> of %2.2X\n", repeatcount, repeatbyte);
    }
}

/********************************************************************/
/* Display a string pointed to by displaystring in a field display- */
/* columns wide.  If the string is longer than displaywidth, then   */
/* print the first and last (displaywidth-3)/2 characters separated */
/* ellipses (...)                                                   */
/********************************************************************/
void
DisplayLongString( char* displaystring, int displaywidth)

{
    int startlen, endlen, stringlen;
    char * stringp;

    /* presumes cursor is properly position, ends with a newline */
    /* NOTE! Not dstring compliant */

    /* get the length of the string */

    stringlen = strlen( displaystring);

    /* if length is less than or equal to display length then just print it */

    if( stringlen <= displaywidth) {
        printf("%s\n", displaystring);
    } else {
        /* its not, so get some working space */
        stringp = (char *) malloc( (size_t) displaywidth+1);

        startlen = (displaywidth - 3)/2;
        endlen = displaywidth - startlen - 3;

        stringp = strncpy( stringp,  displaystring, (size_t) startlen);
        stringp = strcat( stringp, "...");
        stringp = strncat( stringp,
                           displaystring+stringlen-1-startlen,
                           (size_t) endlen);

        printf("%s\n", stringp);
        free (stringp);
    }

    return;
}


/********************************************************************/
/* Display an ISO/IEC 13346 charspec structure                      */
/********************************************************************/
void
DisplayCharSet( struct charspec CharSet)
{
    /* assumes cursor is positioned, ends with a newline */
    printf("CSType = 0x%1x\n", CharSet.uCharSetType);
    printf("        CharacterSetInfo:             '%s'\n",CharSet.aCharSetInfo);
    printf("        (in hex):                     ");
    DisplayByteStream((char *) CharSet.aCharSetInfo, 63, FALSE, 38);
}

/********************************************************************/
/* Display only the first part of a regid                           */
/********************************************************************/
void
DisplayRegIDID( struct regid * RegIDp)
{
    /* assumes character is positioned */
    /* Make a null-terminated version of the Identifier field. */
    char id[24];
    memcpy(id,RegIDp->aRegisteredID,23);
    id[23] = '\000';

    printf("Flags: 0x%1x", RegIDp->uFlags);
    if ( RegIDp->uFlags & DIRTYREGID )
        printf(" (Dirty **NON-UDF**)");
    if ( RegIDp->uFlags & PROTECTREGID )
        printf(" (Protected **NON-UDF**)");
    printf("\n");
    
    printf("        Identifier:                   '%.23s'\n",id);
    printf("        (in hex):                     ");
    DisplayByteStream((char *) RegIDp->aRegisteredID, 23, FALSE, 38);
}

/********************************************************************/
/* Display an ISO/IEC 13346 regid                                   */
/********************************************************************/
void
DisplayRegID( struct regid * RegIDp)
{
    /* Display the Identifier and flags field first */
    DisplayRegIDID(RegIDp);

    /* Then display the suffix (raw) */
    printf("        Identifier Suffix:            ");
    DisplayByteStream((char *) RegIDp->aIDSuffix, 8, FALSE, 38);
}

/**********************************/
/* Display an UDF Domain EntityID */
/**********************************/
void
DisplayDomainID( struct domainEntityId * deip)
{
    /* Display the Identifier and flags field first */
    DisplayRegIDID((struct regid *)deip);

    /* Then display the suffix */
    printf("        UDF Revision:        %10u (0x%4.4x)\n",
           deip->uUDFRevision, deip->uUDFRevision);

    printf("        Domain Flags:        %10u (0x%2.2x)",
           deip->uDomainFlags, deip->uDomainFlags);
    if (deip->uDomainFlags & WRPROTECT_HARD )
        printf(" (HARD Write Protect)");
    if (deip->uDomainFlags & WRPROTECT_SOFT )
        printf(" (SOFT Write Protect)");
    printf("\n");

    printf("        Reserved:                     ");
    DisplayByteStream((char *) deip->aReserved, 5, FALSE, 38);
}

void
printOSInfo( UCHAR8 osClass, UCHAR8 osIdentifier )
{
    printf("        OS Class:            %10u (0x%2.2x)",
           osClass,osClass);
    switch (osClass) {
    case OSCLASS_UNDEF: printf(" (Undefined)\n"); break;
    case OSCLASS_DOS:   printf(" (DOS)\n");       break;
    case OSCLASS_OS2:   printf(" (OS/2)\n");      break;
    case OSCLASS_MAC:   printf(" (Macintosh)\n"); break;
    case OSCLASS_UNIX:  printf(" (UNIX)\n");      break;
    default:            printf(" (Illegal) ** NON-UDF1.01 **\n");
    }

    printf("        OS Identifier:       %10u (0x%2.2x)",
           osIdentifier, osIdentifier);
    if (osClass == OSCLASS_UNIX) {
        switch (osIdentifier) {
        case OSID_GENERIC:     printf(" (Undefined)"); break;
        case OSID_IBM_AIX:     printf(" (DOS)");       break;
        case OSID_SUN_SOLARIS: printf(" (OS/2)");      break;
        case OSID_HPUX:        printf(" (Macintosh)"); break;
        case OSID_SGI_IRIX:    printf(" (UNIX)");      break;
        default:               printf(" (Unknown) ** NON-UDF1.01 **");
        }
    } else {
        if (osIdentifier != 0)
            printf(" ** NON-UDF1.01 **");
    }
    printf("\n");
}

/**************************/
/* Display a UDF EntityID */
/**************************/
void
DisplayUdfID( struct udfEntityId * ueip)
{
    /* Display the Identifier and flags field first */
    DisplayRegIDID((struct regid *)ueip);

    /* Then display the suffix */
    printf("        UDF Revision:        %10u (0x%4.4x)\n",
           ueip->uUDFRevision, ueip->uUDFRevision);

    printOSInfo(ueip->uOSClass,ueip->uOSIdentifier);

    printf("        Reserved:                     ");
    DisplayByteStream((char *) ueip->aReserved, 4, FALSE, 38);
}

/*****************************************/
/* Display a UDF Implementation EntityID */
/*****************************************/
void
DisplayImplID( struct implEntityId * ieip)
{
    /* Display the Identifier and flags field first */
    DisplayRegIDID((struct regid *)ieip);

    /* Then display the suffix */
    printOSInfo(ieip->uOSClass,ieip->uOSIdentifier);

    printf("        HP Implementation Use:\n");
    printf("        Descriptor Version:  %10u (0x%2.2x)\n",
           ieip->uDescVersion ,ieip->uDescVersion );
    printf("        Library Version:     %10u (0x%2.2x)\n",
           ieip->uLibrVersion ,ieip->uLibrVersion );
    printf("        Impl Registration No:%10u (0x%4.4x)\n",
           ieip->uImplRegNo   ,ieip->uImplRegNo   );
    printf("        Impl Version:        %10u (0x%2.2x)\n",
           ieip->uImplVersion ,ieip->uImplVersion );
    printf("        Impl Revision:       %10u (0x%2.2x)\n",
           ieip->uImplRevision,ieip->uImplRevision);
}

/********************************************************************/
/* Display an ISO/IEC 13346 timestamp structure                     */
/********************************************************************/
void
DisplayTimestamp( struct timestamp Time, int indent)
{
    /* assumes character is positioned, ends with newline */

    int tp = GetTSTP(Time.uTypeAndTimeZone);
    int tz = GetTSTZ(Time.uTypeAndTimeZone);

    printf("Type: %s (%d)\n",
           (tp == 0 ? "Coordinated Universal Time" :
            (tp == 1 ? "Local Time" :
             "Non-ISO")),
           tp);
    printspace( indent);
    printf("TimeZone: %d %s\n",
           tz,(tz == -2047 ? "(No timezone specified)" :
               (tz <= 1440 && tz >= -1440 ? "minutes from UCT" :
                "(***INVALID timezone value***)")));
    printspace( indent);
    printf("YYYY-MM-DD: %4.4d-%2.2d-%2.2d\n",
           Time.iYear, Time.uMonth, Time.uDay); 
    printspace( indent);
    printf("HH-MM-SS: %2.2d-%2.2d-%2.2d\n",
           Time.uHour, Time.uMinute, Time.uSecond);
    printspace( indent);
    printf("CS-HM-MS: %2.2d-%2.2d-%2.2d\n",
           Time.uCentiseconds,Time.uHundredMicroseconds,Time.uMicroseconds); 

}


/********************************************************************/
/* print spaces blanks (character 20)                               */
/********************************************************************/
void
printspace( int spaces)
{
    int i;

    for(i=0; i<spaces; printf(" "), i++);
}


/*********************************************************************/
/* prints a Dstring field -- no guaranteed trailing null, and length */
/* in last byte.                                                     */
/* Supports OSTA Compressed Unicode, but if no Compression algorithm */
/* supplied, will print out ASCII string.                            */
/*********************************************************************/
void
printDstring( char *start, UCHAR8 fieldLen, UCHAR8 indent)
{
    /* First, grab the length of the string */
    UCHAR8 dstringLen = (UCHAR8)start[fieldLen - 1];

    /* Then, hand it all off to Dchars */
    printDchars(start,dstringLen,indent);
    
    /* Then, print out the string length */
    printf("    001 Length:              %10lu (0x%8.8lx)\n",
           dstringLen,dstringLen);

    /* Finally, dump it all as a ByteStream */
    printf("    %3.3u (last bytes in hex)           ",fieldLen-dstringLen);
    DisplayByteStream(start+dstringLen, fieldLen-dstringLen, FALSE, indent);
    
    return;
}
    
/*********************************************************************/
/* prints a Dchars field -- no guaranteed trailing null, and length  */
/* of Dchars field is max length of string.                          */
/* Supports OSTA Compressed Unicode, but if no Compression algorithm */
/* supplied, will print out ASCII string.                            */
/*********************************************************************/
void
printDchars( char *start, UCHAR8 length, UCHAR8 indent)
{
    /* Some (one) local variable(s) */
    UINT16 i;                       /* Index  */
    UINT16 unichar;                 /* Unicode character */
    UINT16 numCols;                 /* For wrapping on unicode display */

    char tbuff[257];                /* Buffer for non-unicode */

    char *dispStart = start;        /* to allow for hex dump at end */
    UCHAR8 dispLen = length;
    
    /* First, grab the Compressed Algorithm Number. */
    UCHAR8 alg = *(UCHAR8 *)dispStart;

    /* Throw out the algorithm byte, if any, and print out compression
       algorithm */
    switch (alg) {
    case 16:
        /* 16-bit Unicode, but ignore the first byte */
        printf("(OCU16) ");
        dispStart++;
        dispLen--;
        break;
    case 8:
        /* ASCII, but ignore the first byte */
        printf("(OCU08) ");
        dispStart++;
        dispLen--;
        break;
    default:
        /* ASCII, including the first byte */
        printf("(ASCII) ");
        break;
    }

    /* Print out the characters. */
    if ( alg == 16 ) {
        /* 16-bit unicode gets swapped and printed out in HEX. */
        for (i=0;i<dispLen/2;i++) {
            if (++numCols > MAXCOLUMNS)
                printspace(indent);
            unichar = *((UINT16 *)dispStart + i );
            swapUINT16( & unichar );
            printf("%4.4X ",unichar);
        }
        printf("\n");
    } else {
        /* Now make a copy of all the bytes, to make sure we have a null
           termination */
        strncpy( tbuff, dispStart, dispLen); /* copy over just enough characters */
        tbuff[dispLen]=0;           /* null terminate the string */
        printf("'%s'\n", tbuff);    /* print it out */
    }

    /* Now print it all out in hex */
    printf("    %3.3u (in hex)                      ",length);
    DisplayByteStream(start, length, FALSE, 38);
    return;
}
    
