/********************************************************************/
/*  makefs.c - NSR volume initialization routines                   */
/*                                                                  */
/*  29.Oct.90 15:00                                                 */
/*      written by Karen Klemm                                      */
/*                                                                  */
/*  Module name:    makefs.c                                        */
/*                                                                  */
/*  Description:                                                    */
/*      Routine to initialize an NSR volume                         */
/*      Routine to fill an onode                                    */
/*                                                                  */
/*  Functions in this module:                                       */
/*      makefs                                                      */
/*                                                                  */
/*  Modification history:                                           */
/*                                                                  */
/********************************************************************/
#include <string.h>
#include "Debug.h"
#include "def.h"
#include "o_vfsops.h"
#include "nsr.h"
#include "mkvol.h"
#include "makefs.h"

extern INT32 setFsId();

CHAR8   buf[1024]; 
CHAR8   empty[1024]; 

const UCHAR8 bitmap_bits[8] = {0xff,0xfe,0xfc,0xf8,0xf0,0xe0,0xc0,0x80};

/* prototype definitions */

INT32 makefs(struct fs_info *x,struct PartDesc *pd, struct LogVolDesc *lvd);
void ot_NewsetOtTime( struct timestamp *x);
void FillICBTag(struct ICBTag *icbtagp,
                struct otfs *otfsp, UINT32 ParentLBN,
                UINT16 FileType, UINT16 ADType);
void InitAD(struct otfs *fsp,
            union icbe *icbep,
            UINT16 ExtType,
            UINT32 addr, UINT32 len);
void FillRootICB(struct otfs *fs,UINT32 rootLocation,union icbe *op);

/********************************************************************/
/*  write_fsd_sequence                                              */
/*      write the file set descriptor sequence                      */
/*                                                                  */ 
/*  input parameter:                                                */
/*                                                                  */ 
/*  returns:                                                        */
/*      0       success                                             */
/*                                                                  */
/*  notes:                                                          */
/********************************************************************/
INT32 write_fsd_sequence(struct fs_info *fsi,struct otfs *fs, 
                         struct LogVolDesc *lvd,
                         struct PartDesc *pd,
                         UINT32 *rootLocation)
{
    int                   start_loc;    /* physical start of partition */
    int                   logical_loc;  /* logical location in partition */
    int                   physical_loc; /* physical location */
    struct FileSetDesc    *fsd;
    struct Terminator     *td;
    INT32                 error = 0;
    struct PartHeaderDesc *PHDp;

    physical_loc = pd->uPartStartingLoc;
    logical_loc = 0;
    
    fsd = (struct FileSetDesc *)buf;
    bzero((caddr_t)fsd,sizeof(struct FileSetDesc));
    FillTag(TAGID_FSD,1,1,NOCRC,0,logical_loc++,&fsd->sTag);
    ot_setOtTime(&(fsd->sRecordingTime));
    fsd->uInterchangeLev = 1;
    fsd->uMaxInterchangeLev = 1;
    fsd->uCharSetList = CHARSETLIST;
    fsd->uMaxCharSetList = CHARSETLIST;
    fsd->uFileSetNum = 0;
    fsd->uFileSetDescNum = 0;
    fsd->sLogVolIDCharSet.uCharSetType = CS5;
    strncpy(&fsd->aLogVolID[0],fsi->VolumeName,sizeof(fsd->aLogVolID));
    fsd->sFileSetCharSet.uCharSetType = CS5;
    strncpy(&fsd->aFileSetID[0],"file set identifier goes here",sizeof(fsd->aFileSetID));
    PHDp = (struct PartHeaderDesc *) pd->aPartContentsUse;
    if ( fsi->Type >= fsMO ) {
        fsd->sRootDirICB.Location_LBN = PHDp->FSB.Location + fsi->SectorsInBMap;
    }
    else {
        fsd->sRootDirICB.Location_LBN = PHDp->FST.Location + ICB_BLOCK_SIZE_WORM;
    }
    *rootLocation = fsd->sRootDirICB.Location_LBN;

    fsd->sRootDirICB.ExtentLength.bf.Length = fsi->SectorSize;
    fsd->sRootDirICB.ExtentLength.bf.Type   = E_RECORDED;
    fsd->sRootDirICB.Location_PartNo = 0;

    fsd->sNextExtent.ExtentLength.bf.Length = fsi->SectorSize;
    fsd->sNextExtent.ExtentLength.bf.Type   = E_RECORDED;
    fsd->sNextExtent.Location_LBN = logical_loc;
    fsd->sNextExtent.Location_PartNo = 0;
#if !IBM
    ot_xlate(TAGID_FSD, TO_OP, (caddr_t) fsd,0,0);
#endif
    error=ot_mopwrite(fs,NULL,(caddr_t)fsd,physical_loc++,fs->SectorSize);
    if ( ! error && fsi->Type >= fsMO ) {
        /* write a terminating descriptor */
        td = (struct Terminator *)buf;
        bzero((char *)td,sizeof(struct Terminator));
        FillTag(TAGID_TERM_DESC,1,1,NOCRC,0,logical_loc++,&td->sTag);
#if !IBM
        ot_xlate(TAGID_TERM_DESC, TO_OP, (caddr_t) td,0,0);
#endif
        error=ot_mopwrite(fs,NULL,(caddr_t)td,physical_loc++,fs->SectorSize);
    }
    return error;
}

/********************************************************************/
/*  write_partition_ICBs                                            */
/*      NSR utility                                                 */
/*                                                                  */ 
/*  input parameter:                                                */
/*      fsi    :    fs_info structure containing volume             */
/*                  initialization parameters                       */
/*                                                                  */ 
/*  returns:                                                        */
/*      0       success                                             */
/*                                                                  */
/*  notes:                                                          */
/********************************************************************/
INT32 write_partition_ICBs(struct fs_info *fsi,struct otfs *fs,
                           struct PartDesc *pd)
{
    INT32                         error = 0;
    struct PartHeaderDesc         *PHDp;
    struct PartIntegrityEntry     *pie;
    struct UnallocSpEntry         *ust;
    struct UnallocSpEntry         *fst;
    struct SpaceBitmapEntry       *ubmap;
    struct SpaceBitmapEntry       *fbmap;
    UINT32                        writtenSectors;
    UINT32                        bytesInBmap;
    UINT32                        locA;
    UCHAR8                        *iobuf;
    INT32                         idx;
    
    /* 
    * Allocate space for bitmaps and space tables. 
    * ISSUE: Since the bitmaps can be quite large, especially 
    * for 2x and later 4x media, the entire bit map should not 
    * be allocated at once.
    */
    bytesInBmap = fsi->SectorsInBMap*fs->SectorSize;
    iobuf = kmem_alloc(bytesInBmap,BUFRAM);
    
    /*
    * write the partition integrity entry
    */
    bzero((caddr_t)iobuf, (UINT32) sizeof(iobuf));
    pie = (struct PartIntegrityEntry *)iobuf;
    PHDp = (struct PartHeaderDesc *) pd->aPartContentsUse;
    FillTag(TAGID_PART_INTEGRITY,1,1,NOCRC, 0,PHDp->PIT.Location,&pie->sTag);
    FillICBTag(&pie->sICBTag,fs,PHDp->PIT.Location,
               FILE_TYPE_PARTITION_INTEGRITY,ADSHORT);
    pie->IntegrityType = INTEGRITY_CLOSE;
#if !IBM
    ot_xlate(TAGID_PART_INTEGRITY, TO_OP, (caddr_t) pie,0,0);
#endif
    error=ot_mopwrite(fs, fsi->VolSection[0].devvp,(caddr_t)pie,
                      PHDp->PIT.Location+START_FSD, fs->SectorSize);
    /* 
    * Fill in the partition unallocated space. 
    * This corresponds to OSA "free" space.
    * This is for sectors that are "erased".
    */
    bzero((caddr_t)iobuf, (UINT32) sizeof(iobuf));
    if ( fs->MediaType <= fsWORM ) {
        /* using a space table */
        ust = (struct UnallocSpEntry *)iobuf;
        FillTag(TAGID_UNALLOC_SP_ENTRY,1,1,NOCRC,0,PHDp->UST.Location,&ust->sTag);
        FillICBTag(&ust->sICBTag,fs,PHDp->UST.Location,
                   FILE_TYPE_UNALLOC_SP_ENTRY,fsi->ADType);
        /*
        * calculate number of written sectors 
        */
        writtenSectors = FSDS_LEN+PIT_LEN_WORM+UST_LEN+FST_LEN;
        writtenSectors += ICB_BLOCK_SIZE_WORM; /* root dir entry */
        writtenSectors += 1;                   /* root dir contents */

        InitAD(fs,(union icbe *)ust,E_UNALLOCATED,writtenSectors,
               (pd->uPartLength-writtenSectors) * fs->SectorSize);;

#if !IBM
        ot_xlate(TAGID_UNALLOC_SP_ENTRY,TO_OP,
                 (caddr_t)ust,fs->SectorSize,0);
#endif
        if (!error) {
            error=ot_mopwrite(fs, fsi->VolSection[0].devvp, (caddr_t)ust,
                          PHDp->UST.Location+START_FSD, fs->SectorSize);
        }
#if !IBM
        ot_xlate(TAGID_UNALLOC_SP_ENTRY,FROM_OP,
                 (caddr_t)ust,fs->SectorSize,0);
#endif
    } else {
        /* using a space bitmap */
        ubmap = (struct SpaceBitmapEntry *)iobuf;
        FillTag(TAGID_SPACE_BMAP,1,1,NOCRC,0,PHDp->USB.Location,&ubmap->hdr.sTag);
        /*
        * Calculate number of written sectors in the partition. 
        * Note that this excludes the volume descriptor sequence
        * and the anchors.
        */
        writtenSectors = FSDS_LEN+PIT_LEN_MO+(fsi->SectorsInBMap<<1);
        writtenSectors += ICB_BLOCK_SIZE_MO;  /* root dir file entry */
        writtenSectors++; /* root dir file identifiers */

        /* number of bits in bitmap */
        ubmap->hdr.N_Bits = pd->uPartLength; 
        
        /* number of bytes in bitmap */
        ubmap->hdr.N_Bytes = ((pd->uPartLength) >> 3) + (pd->uPartLength % 8 ? 1 : 0); 
        
        /*
        * Fill in unallocated space bitmap with appropriate values.
        */
        locA = writtenSectors>>3;
        memset(&ubmap->bitmap[0],0,locA);
        if ( ubmap->hdr.N_Bytes + sizeof(struct SpaceBitmapHdr) <= fs->SectorSize ) {
            /* space bitmap all fits in one sector. Initialize bytes to 1's. */
            memset(&ubmap->bitmap[locA],0xff,ubmap->hdr.N_Bytes-locA);
            /* Fix up the last byte.  */
            ubmap->bitmap[ubmap->hdr.N_Bytes-1] = bitmap_bits[ubmap->hdr.N_Bytes % 8];
        }
        else {
            /*
            * There is more than one sector in this bitmap. Initialize all
            * remaining bytes to 0xff.
            */
            memset(&ubmap->bitmap[locA],0xff,fs->SectorSize-locA);
        }
        /* fix up first non-zero byte */
        if ( writtenSectors % 8 ) {
            ubmap->bitmap[locA] = bitmap_bits[writtenSectors % 8];
        }
#if !IBM
        ot_xlate(TAGID_SPACE_BMAP,TO_OP,(caddr_t)ubmap,0,0);
#endif
        error=ot_mopwrite(fs, fsi->VolSection[0].devvp, (caddr_t)ubmap,
                          PHDp->USB.Location+START_FSD,fs->SectorSize);
#if !IBM
        ot_xlate(TAGID_SPACE_BMAP,FROM_OP,(caddr_t)ubmap,0,0);
#endif  
        /* 
        * fill in the remaining sectors of the unallocated space bitmap
        */
        memset(&buf[0],0xff,fs->SectorSize);
        for ( idx = 1; idx < fsi->SectorsInBMap && !error; idx++) 
        {
            if ( idx == (fsi->SectorsInBMap-1) ) {
                locA = (ubmap->hdr.N_Bytes + sizeof(struct SpaceBitmapHdr)) % fs->SectorSize;
                memset(buf+locA,0x0,fs->SectorSize-locA);
                if ( ubmap->hdr.N_Bits % 8 ) {
                    /* fix up last non-zero byte */
                    buf[locA-1] = ~bitmap_bits[ubmap->hdr.N_Bits % 8];
                }
            }
            error=ot_mopwrite(fs, fsi->VolSection[0].devvp, (caddr_t)buf,
                          START_FSD+PHDp->USB.Location+idx,fs->SectorSize);
        }
    }
    
    /*
     * Fill in the partition freed space entry.
     */
    bzero((caddr_t)iobuf, (UINT32) sizeof(iobuf));
    if ( fs->MediaType <= fsWORM ) {
        /* using a space table */
        fst = (struct UnallocSpEntry *)iobuf;
        FillTag(TAGID_UNALLOC_SP_ENTRY,1,1,NOCRC,0,PHDp->FST.Location,&fst->sTag);
        FillICBTag(&fst->sICBTag,fs,PHDp->FST.Location,
                   FILE_TYPE_UNALLOC_SP_ENTRY,fsi->ADType);
        /* Leave ADs zeroed. */
        /* InitAD(fs,(union icbe *)fst,0,0,0); */
#if !IBM
        ot_xlate(TAGID_UNALLOC_SP_ENTRY,TO_OP,
                 (caddr_t)fst,fs->SectorSize,0);
#endif
        error=ot_mopwrite(fs, fsi->VolSection[0].devvp, (caddr_t)fst,
                          PHDp->FST.Location+START_FSD, fs->SectorSize);
#if !IBM
        ot_xlate(TAGID_UNALLOC_SP_ENTRY,FROM_OP,
                 (caddr_t)fst,fs->SectorSize,0);
#endif
    }
    else {
        /* using a space bitmap */
        fbmap = (struct SpaceBitmapEntry *)iobuf;
        FillTag(TAGID_SPACE_BMAP,1,1,NOCRC,0,PHDp->FSB.Location,&fbmap->hdr.sTag);
        
        /* number of bits in bitmap */
        fbmap->hdr.N_Bits = pd->uPartLength; 
        
        /* number of bytes in bitmap */
        fbmap->hdr.N_Bytes = (pd->uPartLength>>3) + (pd->uPartLength % 8 ? 1 : 0); 
        
        memset(&fbmap->bitmap[0],0,fbmap->hdr.N_Bytes);
#if !IBM
        ot_xlate(TAGID_SPACE_BMAP,TO_OP,(caddr_t)fbmap,0,0);
#endif
        error=ot_mopwrite(fs, fsi->VolSection[0].devvp, (caddr_t) fbmap,
                          PHDp->FSB.Location+START_FSD, bytesInBmap);
#if !IBM
        ot_xlate(TAGID_SPACE_BMAP,FROM_OP,(caddr_t)fbmap,0,0);
#endif
        /* 
        * fill in the remaining sectors of the free space bitmap
        */
        memset(&buf[0],0x00,fs->SectorSize);
        for ( idx = 1; idx < fsi->SectorsInBMap && !error; idx++) 
        {
            error=ot_mopwrite(fs, fsi->VolSection[0].devvp, (caddr_t)buf,
                          START_FSD+PHDp->FSB.Location+idx,fs->SectorSize);
        }
    }

    kmem_free((caddr_t)iobuf,sizeof(iobuf),BUFRAM);
    return error; 
}

/********************************************************************/
/*  makefs                                                          */
/*      NSR volume initialization utility                           */
/*                                                                  */ 
/*  input parameter:                                                */
/*      fsi    :    fs_info structure containing volume             */
/*                  initialization parameters                       */
/*                                                                  */ 
/*  returns:                                                        */
/*      0       success                                             */
/*                                                                  */
/*  notes:                                                          */
/*      assumes the veneer gets sector size and disk size, and puts */
/*      them into the fs_info structure passed in.  The device has  */
/*      been opened, if applicable.                                 */
/********************************************************************/
INT32 makefs(struct fs_info *fsi, struct PartDesc *pd, struct LogVolDesc *lvd)
{
    UINT32            error = 0; 
    union icbe        *op;
    struct FileIDDesc *ent;
    struct otfs       fs;
    UINT32            rootLocation;

    DMLPRINTF(oMntLev)("makefs: got to makefs()\n");

    /* Make sure incoming parameters are okay. */
    switch (fsi->Type) {
    case fsMO:
        /* MO uses an onode block size of 2 only. */
        if (!(fsi->ICBsize==ICB_BLOCK_SIZE_MO 
           && fsi->IndirectSize==ICB_INDIRECT_SIZE_MO))
            error=EINVAL;
        break;
    case fsWORM:
    case fsWORMonMO:
        /* Check for valid onode block configuration. */
        if (!(fsi->ICBsize!=0 && 
            fsi->IndirectSize <= fsi->ICBsize))
            error=EINVAL;
        break;
    default:
        error=EINVAL;
    }
    /* These are limitations we've imposed temporarily. */
    if (!(fsi->FileSystemType==0 && fsi->StripeSize==0 && 
        fsi->VolSeqNumber==0 && fsi->VolumeSize==1)) {
        error=EINVAL;
    }

    DMLPRINTF(oMntLev)("makefs: checked parms\n");

    if (!error) {
        /* Clear out structures before initializing anything. */
        bzero((caddr_t) &fs, (UINT32) sizeof(struct otfs));

        /* Fill otfs structure with parameters 
         * passed in from veneer in the fs_info structure.
         */
        fs.SectorSize = fsi->SectorSize;
        fs.DiskSectors = fsi->DiskSectors;
        ot_copy((caddr_t) &(fsi->VolSection[0]),
            (caddr_t) &(fs.VolSection[0]), sizeof(struct otfsSection));
        ot_copy((caddr_t) &(fsi->vf), (caddr_t) &(fs.vf),
            sizeof(otfsVendorField));
        fs.ICBsize = fsi->ICBsize;
        fs.IndirectSize = fsi->IndirectSize;

        fs.MediaType = fsi->Type;
    
        /* write the file set descriptor sequence */
        error = write_fsd_sequence(fsi,&fs,lvd,pd,&rootLocation);
        
        /* Fill in necessary fields in otfs structure. */
        fs.rootAddr = rootLocation;

        /* write the partition integrity, free space info, unalloc space info */
        if ( ! error ) {
            error = write_partition_ICBs(fsi,&fs,pd);
        }

        /* Fill Root Directory Onode Block. */
        bzero(buf, (UINT32) sizeof(buf));
        op = (union icbe *) buf;
        FillTag(TAGID_FILE_ENTRY,1,1,NOCRC,0, rootLocation,(struct Tag *) op);
        FillICBTag(&op->fe.sICBTag,&fs, fs.rootAddr,
                   FILE_TYPE_DIRECTORY,fsi->ADType);
        FillRootICB(&fs,rootLocation,op);
#if !IBM
        ot_xlate(TAGID_FILE_ENTRY, TO_OP, (caddr_t) op,fs.SectorSize,0);
#endif
        if (!error)
            error=ot_mopwrite(&fs, fs.VolSection[0].devvp, (caddr_t) op, 
                              rootLocation+START_FSD, fs.SectorSize);

        /* Fill Root Directory Data. */

        /* zero out the buffer and cast it */
        bzero(buf, (UINT32) sizeof(buf));
        ent = (struct FileIDDesc *) buf;

        /* Now fill the tag entry for this directory entry */

        FillTag(TAGID_FILE_ID, 1, 1, NOCRC, 0, rootLocation+fs.ICBsize,
                (struct Tag *) buf);

        ent->VersionNum = 1;
        ent->Characteristics = DIR_ATTR;
        ent->L_FI = 2;
        ent->L_IU = sizeof(UINT32);

        ent->ICB.ExtentLength.Length32 = 0;
        ent->ICB.Location_LBN = rootLocation + fs.ICBsize;
        ent->ICB.Location_PartNo = 0;
        bzero((char *)ent->ICB.aImpUse,sizeof(ent->ICB.aImpUse));

        memmove( (void *) (buf + FILE_ID_DESC_CONSTANT_LEN + ent->L_IU) , (void *) "..", 2);
#if !IBM
        ot_xlate(TAGID_FILE_ID, TO_OP, (caddr_t) buf,1024,0);
#endif
        if (!error)
            error=ot_mopwrite(&fs, fs.VolSection[0].devvp, (caddr_t) buf, 
                              rootLocation+START_FSD+fs.ICBsize, 
                              fs.SectorSize);
    }
    
    DMLPRINTF(oMntLev)("makefs: returning error %d\n", error);

    return error;
}


void ot_NewsetOtTime( struct timestamp *timeval)
{
    SetTSTP(timeval->uTypeAndTimeZone,1);
    SetTSTZ(timeval->uTypeAndTimeZone,-7*60);
    timeval->iYear = 1992;
    timeval->uMonth = 11;
    timeval->uDay = 22;
    timeval->uHour = 0;
    timeval->uMinute = 0;
    timeval->uSecond = 0;
    timeval->uCentiseconds = 0;
    timeval->uHundredMicroseconds = 0;
    timeval->uMicroseconds = 0;
}

void FillICBTag(struct ICBTag *icbtagp,
                struct otfs *otfsp, UINT32 ParentLBN,
                UINT16 FileType, UINT16 ADType)
{
    /* fill in the ICB tag structure pointed to by icbtagp */
    icbtagp->PriorDirects = 0;

    if( otfsp->MediaType == fsMO) {
        icbtagp->StrategyType = 4;
        *((UINT16 *)icbtagp->StrategyParm) = 1;
    } else {
        icbtagp->StrategyType = 1;
        *((UINT16 *)icbtagp->StrategyParm) = otfsp->ICBsize- otfsp->IndirectSize;
    }

    icbtagp->NumberEntries = otfsp->ICBsize;
    icbtagp->Reserved = 0;
    icbtagp->FileType = FileType;

    icbtagp->sParentICB_LBN = ParentLBN;
    icbtagp->sParentICB_PartNo = 0;

    icbtagp->Flags = ValidADType(ADType) ? ADType : 0;
}

/*
 * initAD(fsp,icbep,extType,addr,len)
 *
 * This initializes the AD portion of a USE or FE to a single entry of
 * the supplied extent type, length, and address.
 *
 * Note that it looks in the supplied icbe to find which type of ADs to
 * use (short, long, or extended).
 */
void InitAD(struct otfs *fsp,
            union icbe *icbep,
            UINT16 ExtType,
            UINT32 addr, UINT32 len) {

    struct short_AD *sadp;
    struct long_AD  *ladp;
    struct ext_AD   *xadp;
    void            *ptr;

    const UINT16    ADType = icbep->ge.sICBTag.Flags & ADTYPEMASK;
    const UINT16    TagID  = icbep->ge.sTag.uTagID;
    UINT32          L_AD;
    
    if (TagID == TAGID_FILE_ENTRY) {
        ptr              = FE_ADStart(&(icbep->fe));
        L_AD             = fsp->SectorSize - FE_HdrLen(&(icbep->fe));
        memset(ptr,0,L_AD); /* zero AD Field */
        L_AD            -= L_AD % ot_sizeOfAD(ADType);
        icbep->fe.L_AD   = L_AD;
    } else {  /* TAGID_UNALLOC_SP_ENTRY */
        ptr              = USE_ADStart(icbep);
        L_AD             = fsp->SectorSize - USE_HdrLen;
        memset(ptr,0,L_AD); /* zero AD Field */
        L_AD            -= L_AD % ot_sizeOfAD(ADType);
        icbep->uspe.L_AD = L_AD;
    }
    
    switch (ADType) {
    case ADSHORT:
        ((struct short_AD *)ptr)->ExtentLength.bf.Length = len;
        ((struct short_AD *)ptr)->ExtentLength.bf.Type = ExtType;
        ((struct short_AD *)ptr)->Location = addr;
        break;
    case ADLONG:
        ((struct long_AD *)ptr)->ExtentLength.bf.Length = len;
        ((struct long_AD *)ptr)->ExtentLength.bf.Type = ExtType;
        ((struct long_AD *)ptr)->Location_LBN = addr;
        ((struct long_AD *)ptr)->Location_PartNo = 0;
        break;
    case ADEXTENDED:
        ((struct ext_AD *)ptr)->InfoLength =
            ((struct ext_AD *)ptr)->RecordedLength =
                ((struct ext_AD *)ptr)->ExtentLength.bf.Length = len;
        ((struct ext_AD *)ptr)->ExtentLength.bf.Type = ExtType;
        ((struct ext_AD *)ptr)->Location_LBN = addr;
        ((struct ext_AD *)ptr)->Location_PartNo = 0;
    }
}

/*
 * FillRootICB()
 *
 * This fills in the root dir's icbe, including its AD, to allow a single
 * extent a single sector long immediately following the supplied address
 * plus the ICBsize. NOTE that the address which should be supplied is
 * that of the ICB.
 */
void FillRootICB(struct otfs *fsp,UINT32 address, union icbe *icbep)
{
    int i;
    struct long_AD *ADp;

    icbep->fe.InfoLengthH = 0;
    icbep->fe.InfoLengthL = FILE_ID_DESC_CONSTANT_LEN+ROOTDIR_INFO_LENGTH;
    icbep->fe.UID = 0;
    icbep->fe.GID = 0;
    icbep->fe.Permissions = 0;
    icbep->fe.LinkCount = 0;
    icbep->fe.RecFormat = 0;
    icbep->fe.RecDisplayAttr = 0;
    icbep->fe.RecLength = 0;
    icbep->fe.LogBlocksH = 0;
    icbep->fe.LogBlocksL = 1;
        
    ot_NewsetOtTime(&(icbep->fe.sModifyTime));
    ot_NewsetOtTime(&(icbep->fe.sAccessTime));
    ot_NewsetOtTime(&(icbep->fe.sAttrTime));
        
    icbep->fe.Checkpoint = 1;
        
    icbep->fe.sExtAttrICB.ExtentLength.Length32 = 0;
    icbep->fe.sExtAttrICB.Location_LBN = 0;
    icbep->fe.sExtAttrICB.Location_PartNo = 0;

    /* ISSUE do we want to leave in the REGIDs? */
    memcpy( &(icbep->fe.sImpID.aRegisteredID), SFS_REGID, sizeof( SFS_REGID));
    memcpy( &(icbep->fe.sImpID.aIDSuffix), SFS_FE_REGID, sizeof( SFS_FE_REGID));
    icbep->fe.sImpID.uFlags=0x1;
    icbep->fe.L_EA = 0;

    InitAD(fsp,icbep,E_RECORDED,
           address + fsp->ICBsize,fsp->SectorSize);
}
