/******************************************************************/
/* mkvol.c - make a NSR volume and partion structure              */
/* Copyright (C) 1993 The Hewlett-Packard Company                 */
/*                                                                */
/* Limits of this version:                                        */
/*   Edits of some fields in PVD and PDLVD OK, but can't change   */
/*      size of a partition.  Can only erase and rewrite.         */
/*   2 Anchors: N-256 and 256                                     */
/*   Only one extent of PVD                                       */
/*   1024B of USD (About 125 entries)                             */
/*                                                                */
/* Modification History                                           */
/*                                                                */
/******************************************************************/
#include "def.h"

/* includes */

/* IBM version */
#if IBM
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <errno.h>
#include <conio.h>
#include <process.h>
#include "d:\pc-nsr\include\times.h"
#include "d:\corecode\src\nsrhdrs\vnode.h"
#include "d:\corecode\src\nsrhdrs\o_proto.h"
#include "d:\nitro\include\nsr.h"
#include "d:\nitro\include\newnsrer.h"
#include "d:\nitro\makefs\makefs.h"
#include "mkvol.h"
#endif

#if HP
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "../nsrHdrs/o_vfsops.h"
#include "../nsrHdrs/nsr.h"
#include "../nsrHdrs/makefs.h"
#include "../nsrHdrs/mkvol.h"
#include "../nsrHdrs/o_proto.h"
#include "../nsrHdrs/nsrerrno.h"
#include "Debug.h"
#endif

UINT32 mkvol(struct vol_info *vol_data, 
             struct fs_info *fsi,
             struct PartDesc *pd,
             struct LogVolDesc *lvd)
{
    INT16 result;
    struct otfs fsp = {0};        /* keep all the goodies in here */
    struct {
        struct UnallocSpDesHead SpaceHead;
        struct extent_ad SpaceRecord[MAX_USD];
    } USD = {0};
    struct FindVDSFlags sFindVDSFlags; 
    struct ReadVDSFlags sReadVDSFlags;
    struct AnchorVolDesPtr *sAnchorData;
    struct VolStructDesc *volDesc;
    unsigned char found_USD = FALSE;
    struct extent_ad sPrimaryAdr, sReserveAdr;
    struct extent_ad sPVDAdr, sPDAdr, sLVDAdr, sUSDAdr, sIUDAdr;
    UINT32 WrittenVDSLen = 0;              /* in sectors */
    UINT32 dummy32;
    INT32 Error=SUCCESS;
    UINT32 USDsectors;
    UINT32 BytesInMap;  
    UINT32 location;
    UINT32 seqnum;

    fsp.SectorSize = vol_data->SectorSize;
    fsp.DiskSectors = vol_data->DiskSectors;
    fsp.MediaType = vol_data->MediaType;
    fsp.VolSection[0] = vol_data->VolSection[0];

    /* Default values for Location of Primary VDS */    
    sPrimaryAdr.Location = START_PVDS;
    sPrimaryAdr.Length = VDS_LEN * fsp.SectorSize;
    
    sReserveAdr.Location = fsp.DiskSectors-256;
    sReserveAdr.Length = VDS_LEN * fsp.SectorSize;

    USDsectors = sizeof(USD)/fsp.SectorSize;
    if ( sizeof(USD) % fsp.SectorSize ) {
        USDsectors = (sizeof(USD)+fsp.SectorSize-1)/fsp.SectorSize;
    }
    
    if ( vol_data->MediaType <= fsWORM ) {
        /* 
        * Volume initialized before? If we find a VDS, get the
        * addresses of the prevailing instances of each of the 
        * descriptors found.
        */
        result = FindVDS(&fsp,&sFindVDSFlags,&sPrimaryAdr,&sReserveAdr);
        if( result == SUCCESS) {
            result = ReadVDS(&fsp,sPrimaryAdr,&sReadVDSFlags,
                             &sPVDAdr, &sPDAdr, &sLVDAdr, &sUSDAdr, &sIUDAdr,
                             &WrittenVDSLen);
        }

        if ( (result != SUCCESS) && (result != ENOVDS) && (result != EBADVDS)) {
            /* unexpected error */
            Error = result;
        }
        if ( !Error && result == SUCCESS && sReadVDSFlags.uUASDValid ) { 
            /* Read the last Unallocated Space Descriptor */ 
            Error = ot_read( &fsp, sUSDAdr.Location, (caddr_t) &USD, &sUSDAdr.Length);
            if ( ! Error ) {
#if !IBM
                ot_xlate(TAGID_USD,FROM_OP,(caddr_t)&USD,0,0);
#endif
                if ( ot_goodTag(TAGID_USD,(struct Tag *)&USD,sUSDAdr.Location) == SUCCESS ) {
                    found_USD = TRUE;
                    NextVolDesSeqNum(PUT,USD.SpaceHead.uVolDescSeqNum + 1);
                }
            }
        }
        if ( !found_USD ) {
            /* assume this is virgin media.  */
            WrittenVDSLen = 0;
        }
        /*
        * Read the last partition descriptor and use the contents to
        * override relevant vol_data fields.
        */
        if ( !Error && sReadVDSFlags.uPDValid ) {
            Error = ot_read( &fsp, sPDAdr.Location, (caddr_t) pd, &sPDAdr.Length);
            if ( ! Error ) {
#if !IBM
                ot_xlate(TAGID_PD,FROM_OP,(caddr_t)pd,0,0);
#endif
                if ( ot_goodTag(TAGID_PD,(struct Tag *)pd,sPDAdr.Location) == SUCCESS ) {
                    vol_data->PartStart = pd->uPartStartingLoc;
                    vol_data->PartLen = pd->uPartLength;
                }
            }
        }
    }

    /*
    * If MO media or virgin WORM media:
    * - reserve volume space for system area and for the anchors.
    * - initialize volume descriptor sequence number
    * - write the anchors
    * - write the bea,nsr,tea descriptors
    */
    if ( vol_data->MediaType >= fsMO || !found_USD ) {
        /*
        * set up a single USD Record for totally unallocated media 
        * Only works for media of 2**30 bytes or less, otherwise we 
        * need to create multiple records for each GB 
        */
        USD.SpaceHead.uNumAllocationDes = 1;
        USD.SpaceRecord[0].Location = 0;
        USD.SpaceRecord[0].Length = fsp.DiskSectors*fsp.SectorSize;

        /* start the Volume Descriptor Sequence Numbers at 1 */
        dummy32 = NextVolDesSeqNum( PUT, 1);

        /* Reserve sectors for PVDS  */
        Error = GetVolumeSpace( START_PVDS,VDS_LEN, REQUIRELOC,
                                &USD.SpaceHead.uNumAllocationDes, 
                                USD.SpaceRecord,
                                fsp.SectorSize);

        /* Ask for sectors ANCHOR_0 and N-256 for anchors */
        if (!Error) 
            Error = GetVolumeSpace( ANCHOR_0, 1, REQUIRELOC, 
                                    &USD.SpaceHead.uNumAllocationDes, 
                                    USD.SpaceRecord, fsp.SectorSize);
        if (!Error) 
            Error = GetVolumeSpace( fsp.DiskSectors-1-256, 1, REQUIRELOC, 
                                    &USD.SpaceHead.uNumAllocationDes, 
                                    USD.SpaceRecord, fsp.SectorSize);

        /* Write the anchors */
        if (!Error) {
            sAnchorData = (struct AnchorVolDesPtr *) kmem_alloc( fsp.SectorSize, BUFRAM);
            bzero( (char *) sAnchorData, fsp.SectorSize);

            Error = WriteAnchor( &fsp,ANCHOR_0, sAnchorData, START_PVDS,
                                 VDS_LEN*fsp.SectorSize,sReserveAdr.Location, VDS_LEN*fsp.SectorSize);
            if ( !Error )
                WriteAnchor( &fsp, fsp.DiskSectors-1-256, sAnchorData, START_PVDS,
                             VDS_LEN*fsp.SectorSize,sReserveAdr.Location, VDS_LEN*fsp.SectorSize);

                kmem_free( (char *) sAnchorData, (UINT32) fsp.SectorSize, 0);
        }

        /*
        * Write the bea, nsr, tea descriptors, starting at byte 32768. 
        * Sectors 0-255 were already reserved above.
        */
        if ( !Error ) {
            volDesc = (struct VolStructDesc *) kmem_alloc(sizeof(struct VolStructDesc), BUFRAM);
            bzero( (char *) volDesc, sizeof(struct VolStructDesc));
            volDesc->structType = 0;
            strncpy((char *)volDesc->stdIdentifier,"BEA01",5);
            volDesc->structVersion = 1;
            location = 32768/fsp.SectorSize;
            Error = ot_write( fsp,location,(caddr_t)volDesc,sizeof(struct VolStructDesc));
            if ( ! Error ) {
                strncpy((char *)volDesc->stdIdentifier,"NSR01",5);
                location += sizeof(struct VolStructDesc)/fsp.SectorSize;
                Error = ot_write( fsp,location,(caddr_t)volDesc,sizeof(struct VolStructDesc));
            }
            if ( ! Error ) {
                strncpy((char *)volDesc->stdIdentifier,"TEA01",5);
                location += sizeof(struct VolStructDesc)/fsp.SectorSize;
                Error = ot_write( fsp,location,(caddr_t)volDesc,sizeof(struct VolStructDesc));
            }
            /* write an invalid descriptor to implicitly end the sequence */
            if ( ! Error ) {
                strncpy((char *)volDesc->stdIdentifier,"XYZZY",5);
                location += sizeof(struct VolStructDesc)/fsp.SectorSize;
                Error = ot_write( fsp,location,(caddr_t)volDesc,sizeof(struct VolStructDesc));
            }
            kmem_free( (char *)volDesc, sizeof(struct VolStructDesc), 0);
        }
    }

    /*
    * Calculate number of sectors required for bitmaps. This calculation
    * will include the bitmaps and any headers prior to the bitmap.
    */
    if ( vol_data->MediaType >= fsMO ) {
        /* will use bitmaps for space management */
        BytesInMap = sizeof(struct SpaceBitmapHdr)
                   + vol_data->PartLen/8
                   + (vol_data->PartLen%8 ? 1 : 0);
        fsp.SectorsInBMap = BytesInMap/vol_data->BlkSize + (BytesInMap%vol_data->BlkSize ? 1 : 0);
        fsi->SectorsInBMap = fsp.SectorsInBMap;
    }

    if ( !Error ) {
        if ( vol_data->MediaType >= fsMO ) {
            /* MO. Just overwrite old VDS */
            WrittenVDSLen = 0;
        } 
        else {
            /*
            * WORM. We have to write starting at the end of the last VDS written, 
            * providing that there is room.
            */
            if (WrittenVDSLen + 3 + USDsectors > START_PVDS + VDS_LEN ) {
                DPRINTF("mkvol: no room for vds\n");
                return (ENOVDSRM);
            }
        }

        /* Write the Primary Volume Descriptor */
        seqnum = NextVolDesSeqNum( GET, 0);
        Error = WritePVD( &fsp,START_PVDS + WrittenVDSLen, vol_data, seqnum);
        if ( ! Error )
            Error = WritePVD( &fsp,sReserveAdr.Location+WrittenVDSLen,vol_data,seqnum);

        /*
        * Reserve space for the partition. 
        * NOTE: If worm, don't reserve the volume space if it has already been reserved.
        */
        if (!Error && !found_USD) {
            vol_data->PartStart = START_FSD;
            Error = GetVolumeSpace(vol_data->PartStart,
                                   vol_data->PartLen, REQUIRELOC,
                                   &USD.SpaceHead.uNumAllocationDes,
                                   USD.SpaceRecord, fsp.SectorSize);
        }
        
        /* Write the Partition Descriptor */
        if(!Error) {
            seqnum = NextVolDesSeqNum( GET, 0);
            Error = WritePD( &fsp, START_PVDS + WrittenVDSLen + 1,
                             vol_data, pd, seqnum);
            if ( !Error ) 
                Error = WritePD( &fsp, sReserveAdr.Location + WrittenVDSLen + 1,
                                 vol_data, pd, seqnum);
        }

        /* Write the Logical Volume Descriptor */        
        if(!Error) {
            seqnum = NextVolDesSeqNum( GET, 0);
            Error = WriteLVD( &fsp, START_PVDS + WrittenVDSLen + 2, vol_data, lvd,seqnum);
            if ( ! Error )
                Error = WriteLVD( &fsp, sReserveAdr.Location + WrittenVDSLen + 2,
                                  vol_data, lvd,seqnum);
        }

        /* 
        * Write the USD & terminator last 
        */
        if(!Error) {
            seqnum = NextVolDesSeqNum( GET, 0);
            Error = WriteUSD( &fsp, START_PVDS + WrittenVDSLen + 3, (void *) &USD,seqnum);
            if ( ! Error )
                Error = WriteUSD( &fsp, sReserveAdr.Location+ WrittenVDSLen + 3,
                                  (void *) &USD,seqnum);
            if ((fsp.MediaType >= fsMO) && (!Error))  {
                Error = WriteTerminator(&fsp,START_PVDS+WrittenVDSLen+3+USDsectors);
                if ( ! Error )
                    Error = WriteTerminator(&fsp,sReserveAdr.Location+WrittenVDSLen+3+USDsectors);
            }
        }
    }
    return( Error);
}

#pragma page()
/******************************************************************/
/*  FindVDS - Return the location of primary and reserve volume   */
/*            descriptor sequences.                               */
/*                                                                */
/*  Input - none                                                  */
/*                                                                */
/*  Output -                                                      */
/*  -sFlags. Structure that contains two bitfields which          */
/*           are 1 if the Primary and Reserve Sequences           */
/*           are valid.                                           */
/*  -sPrimary, sReserve.  Pointers to extent_ad structs           */
/*           that contain the location and length of the Primary  */
/*           and Reserve Sequences.                               */
/*                                                                */
/*  Returns - 0 if successful, EBADPRI For a bad primary sequence */
/*            EBADRES for a bad reserve sequence, ENOSEQ if no    */
/*            sequences were found.                               */
/*                                                                */
/******************************************************************/
UINT16 FindVDS( struct otfs *fsp,
                struct FindVDSFlags *sFlags,
                struct extent_ad *sPrimary,
                struct extent_ad *sReserve)
{
    struct AnchorVolDesPtr *AnchorData;
    INT32 Error = SUCCESS;
    INT32 bsize = fsp->SectorSize;
    INT32 i;
    INT32 found = 0;
    INT32 anchor;
    
    AnchorData = (struct AnchorVolDesPtr *)kmem_alloc(fsp->SectorSize,BUFRAM);
    /* 
    * Attempt to read the anchor at location 256 and N - the anchor pts 
    * to the first extents of the primary and secondary volume 
    * descriptor sequences.
    */
    for ( i = 0; i <= 61 && !found && ! Error ; i++ )
    {
        switch (i) {
        case 0: anchor = 256; break;
        case 1: anchor = fsp->DiskSectors-1; break;
        case 2: anchor = fsp->DiskSectors-1-256; break;
        default: anchor = ((i-3)*(fsp->DiskSectors-1))/59; break;
        }
        Error = (INT16) ot_read( fsp, anchor, (caddr_t) AnchorData, &bsize);
        if ( ! Error ) {
#if !IBM
            ot_xlate(TAGID_ANCHOR, FROM_OP, (char *) AnchorData,0,0);
#endif
            if ( ot_goodTag(TAGID_ANCHOR,(struct Tag *)AnchorData,anchor) == SUCCESS ) {
                sPrimary->Length = AnchorData->sMainVDSAdr.Length;
                sPrimary->Location = AnchorData->sMainVDSAdr.Location;
                sReserve->Length = AnchorData->sMainVDSAdr.Length;
                sReserve->Location = AnchorData->sMainVDSAdr.Location;
                if ( sPrimary->Location < 0 || sPrimary->Location >= fsp->DiskSectors ) {
                    Error = EBADVDS;
                }
                else {
                    found = 1;
                }
            }
        }
    }
    kmem_free( (char *) AnchorData, (UINT32) fsp->SectorSize, BUFRAM);
    
    if ( found )
        return SUCCESS;
    else 
        if ( Error == SUCCESS ) Error = ENOVDS;

    return(Error);
}
#pragma page()
/******************************************************************/
/*  ReadVDS - Returns addresses of the prevailing instances of    */
/*            Primary Volume Descriptors,                         */
/*            Partition Descriptors, Logical                      */
/*            Volume Descriptors and Unallocated Space Descrip-   */
/*            tors.                                               */
/*                                                                */
/*  Input -                                                       */
/*  - sPrimary = Location of primary volume descriptor            */
/*  - fsp = otfs structure (to get sector size)                   */
/*                                                                */
/*  Output -                                                      */
/*  - ReadVDS will fill in the addresses of lvd,pd,pvd,etc.       */
/*                                                                */
/*  Returns - SUCCESS or ENOVDS                                   */
/******************************************************************/
UINT16 ReadVDS( struct otfs *fsp,
                struct extent_ad sPrimary,
                struct ReadVDSFlags *sFlags,
                struct extent_ad *sPVDAdr,
                struct extent_ad *sPDAdr,
                struct extent_ad *sLVDAdr,
                struct extent_ad *sUSDAdr,
                struct extent_ad *sIUDAdr,
                UINT32 *VDSLen)
{
    unsigned char *iobuf;                /* buffer to hold disk data */
    INT32 bsize = fsp->SectorSize;       /* how many bytes to read */
    UINT32 location = sPrimary.Location; /* sector number to read */
    UINT16 Error = 0;                    
    struct UnallocSpDesHead *usd; 
    unsigned char terminator_found = FALSE;
    UINT32 vdseqnum;
    UINT32 last_vdseqnum = 0;
    UINT32 pLVD,pPVD,pUSD,pIUD,pPD;      /* prevailing vol des seq num's */
  
    /* initialization */
    sPDAdr->Location = 0;
    sPDAdr->Length = 0;
    sLVDAdr->Location = 0;
    sLVDAdr->Length = 0;
    sPVDAdr->Location = 0;
    sPVDAdr->Length = 0;
    sUSDAdr->Location = 0;
    sUSDAdr->Length = 0;
    sIUDAdr->Location = 0;
    sIUDAdr->Length = 0;
    *VDSLen = 0;
    
    /* initialize prevailing instances */
    pLVD = pPVD = pUSD = pIUD = pPD = 0;
    
    /* initialize flags */
    bzero((char *)sFlags,sizeof(struct ReadVDSFlags));
    
    iobuf = kmem_alloc( fsp->SectorSize, BUFRAM);
    
    /* 
    * Read the next descriptor - don't assume we know
    * which one it will be.
    */
    while ( ! Error && ! terminator_found ) 
    {
        bsize = fsp->SectorSize;
        Error = (INT16) ot_read(fsp,location,(caddr_t)iobuf, &bsize);
        if ( ! Error && bsize == fsp->SectorSize ) {
#if !IBM
            ot_xlate(TAGID_NONE,FROM_OP,(char *)iobuf,fsp->SectorSize,0);
#endif
            if (!Error && (ot_goodTag(TAGID_NONE,(struct Tag *)iobuf, location) == SUCCESS )) {
                /*
                * check for increasing sequence numbers per ECMA 3/8.4.1
                */
                if (((struct Tag *)iobuf)->uTagID != TAGID_TERM_DESC ) {
                    vdseqnum = ((struct PrimaryVolDes *)iobuf)->uVolDescSeqNum;
                    if ( vdseqnum <= last_vdseqnum || (last_vdseqnum == 0 && vdseqnum != 1)) {
                        Error = ENOVDS;
                    }
                    last_vdseqnum = vdseqnum;
                }
                
                switch (((struct Tag *)iobuf)->uTagID ) {
                case TAGID_PD:
                    if ( vdseqnum > pPD ) {
                        sPDAdr->Location = location;
                        sPDAdr->Length = sizeof(struct PartDesc);
                        sFlags->uPDValid = 1;
                    }
                    location++;
                    break;
                case TAGID_LVD:
                    if ( vdseqnum > pLVD ) {
                        sLVDAdr->Location = location;
                        sLVDAdr->Length = sizeof(struct LogVolDesc);
                        sFlags->uLVDValid = 1;
                    }
                    location++;
                    break;
                case TAGID_PVD:
                    if ( vdseqnum > pLVD ) {
                        sPVDAdr->Location = location;
                        sPVDAdr->Length = sizeof(struct PrimaryVolDes);
                        sFlags->uPVDValid = 1;
                    }
                    location++;
                    break;
                case TAGID_USD:
                    if ( vdseqnum > pUSD ) {
                        sUSDAdr->Location = location;
                        usd = (struct UnallocSpDesHead *)iobuf;
                        sUSDAdr->Length = sizeof(struct UnallocSpDesHead)
                                        + sizeof(struct extent_ad)*MAX_USD;
                        /* round length up to a multiple of fsp->SectorSize */
                        if ( sUSDAdr->Length % fsp->SectorSize != 0 ) {
                            sUSDAdr->Length = fsp->SectorSize*((sUSDAdr->Length+fsp->SectorSize-1)/fsp->SectorSize);
                        }
                        sFlags->uUASDValid = 1;
                    }
                    location += sUSDAdr->Length/fsp->SectorSize;
                    break;
                case TAGID_IUD:
                    if ( vdseqnum > pIUD ) {
                        sIUDAdr->Location = location;
                        sIUDAdr->Length = sizeof(struct ImpUseDesc);
                        sFlags->uIUVDValid = 1;
                    }
                    location++;
                    break;
                case TAGID_TERM_DESC:
                    terminator_found = TRUE;
                    location++;
                    break;
                default:
                    Error = ENOVDS;
                    break;
                }
            }
            else {
                Error = ENOVDS;
            }
        }
        else {
            if ( ! Error && bsize != fsp->SectorSize ) {
                terminator_found = TRUE;
            }
        }
    }
    
    kmem_free((caddr_t)iobuf,fsp->SectorSize,BUFRAM);

    /* Return length of vds to calling routine through VDSLen.  */
    *VDSLen = location - sPrimary.Location; 
    
    /* all we have to have is a primary vol desc */
    if ( sFlags->uPVDValid && terminator_found ) {
        return SUCCESS;
    }

    return( ENOVDS);
}
#pragma page()
/******************************************************************/
/*                                                                */
/* Returns                                                        */
/*                                                                */
/*                                                                */
/******************************************************************/
UINT16 GetVolumeSpace( UINT32 Location, UINT32 Length, UINT16 Flags,
       UINT32 *USDLen, struct extent_ad *SpaceMap, UINT32 SectorSize)

/* handy macros to convert sector lengths to byte lengths and vice versa */
/* manipulating extent_ads whose lengths are in BYTES */
/* Important to remember the input parameter Length is expressed in SECTORS */

#define SECLEN( Bytes) ((Bytes)/SectorSize)
#define BYTELEN( Sectors) ((Sectors)*SectorSize)

{
    UINT16 SpaceFound = FALSE;
    UINT16 i;
    UINT16 Error = SUCCESS;
    struct extent_ad Infront = {0L,0L};
    struct extent_ad Behind = {0L,0L};
    struct extent_ad *Hit = NULL; 

    /* search for an extent that fits */
    /* we presume that the map contains no records that together */
    /* define contiguous unallocated extents on the media */
    
    for(i=0; (i < (UINT16) *USDLen) && (Hit == NULL); i++ ) {
        if( (SpaceMap[i].Location <= Location) &&
            (SpaceMap[i].Location+SECLEN(SpaceMap[i].Length) >= Location+Length)) {
            SpaceFound = i;
            Hit = &(SpaceMap[SpaceFound]);
        };
    };

    if (Hit != NULL) {

        /* does the record exactly match the requested extent? */

        if(( Hit->Location == Location) && (Hit->Location+SECLEN(Hit->Length) == Location+Length)) {

                 /* just get rid of the record, we've used it all */
                 for(i=SpaceFound; i < ((UINT16) *USDLen)-1; SpaceMap[i] = SpaceMap[i+1], i++);
                 (*USDLen)--;
                 
        } else {

            /* is there spare space in front? */
            if( Hit->Location < Location) {
                Infront.Location = Hit->Location;
                Infront.Length = BYTELEN(Location - Hit->Location);
            };

            /* is there spare space behind? */
            if( Hit->Location + SECLEN(Hit->Length) > Location+Length) {
                Behind.Location = Location+Length;
                Behind.Length = BYTELEN(Hit->Location + SECLEN(Hit->Length) - (Length + Location));
            };

            /* Add the new records */
            /* But make sure we have room to do the work before starting anything */

            if (Infront.Length) {
                if (Behind.Length) {
                    if( *USDLen >= MAX_USD) {
                        Error = EUSDOVF;
                    } else {
                        
                        /* make space for the new record */
                         for(i=((UINT16) *USDLen)+1; i>SpaceFound; SpaceMap[i] = SpaceMap[i-1], i--);
                        SpaceMap[SpaceFound+1] = Behind;
                        (*USDLen)++;
                    }
                }
                SpaceMap[SpaceFound] = Infront;
            } else if (Behind.Length)
                SpaceMap[SpaceFound] = Behind;
        }

    /* space not found */
    
    } else {
        Error = ENOUAS;
    }
        
    return( Error);
}
#pragma page()
/******************************************************************/
/*                                                                */
/* Write an Anchor record at Location                             */
/*                                                                */
/*                                                                */
/******************************************************************/
INT32 WriteAnchor( struct otfs *fsp, UINT32 Location, struct AnchorVolDesPtr *Anchor,
                  UINT32 mainStart, UINT32 mainLength, UINT32 reserveStart, UINT32 reserveLength)
{
    INT32 Error;
    
    /* write an anchor descriptor at Location */
    /* fill in the tag values */
    FillTag(TAGID_ANCHOR, 1, 1, NOCRC, 0, Location, (struct Tag *) Anchor);

    Anchor->sMainVDSAdr.Location=mainStart;
    Anchor->sMainVDSAdr.Length=mainLength;
    Anchor->sReserveVDSAdr.Location=reserveStart;
    Anchor->sReserveVDSAdr.Length=reserveLength;

    /* translate the structures */
#if !IBM
    ot_xlate(TAGID_ANCHOR, TO_OP, (char *) Anchor,0,0);
#endif    

    /* write the data */
    Error = ot_write( fsp, Location, (caddr_t) Anchor, (UINT32) fsp->SectorSize);

    return( Error);
}
#pragma page()
/******************************************************************/
/*                                                                */
/* Write a Primary Volume Descriptor at Location                  */
/*                                                                */
/*                                                                */
/******************************************************************/
INT32 WritePVD( struct otfs *fsp, UINT32 Location, struct vol_info *vol_data,UINT32 seqnum)
{
    struct PrimaryVolDes *sPVDData;
    INT32 Error;

    sPVDData = (struct PrimaryVolDes *) kmem_alloc( fsp->SectorSize, BUFRAM);
    bzero( (char *) sPVDData, fsp->SectorSize);
        
    /* write a Primary Volume Descriptor at Location */

    /* now fill in the PVD fields */
    /* some fields are shown only by comment.  They are 0 or NULL */
    sPVDData->uVolDescSeqNum = seqnum;
    sPVDData->uPrimVolDesNum = 0;
    str2dstr( sPVDData->aVolID, 32, vol_data->VolID);
    sPVDData->uVSN=1;
    sPVDData->uMaxVSN=1;
    sPVDData->uInterchangeLev=1;
    sPVDData->uMaxInterchangeLev=1;
    sPVDData->uCharSetList = CHARSETLIST;
    sPVDData->uMaxCharSetList = CHARSETLIST;
    str2dstr( sPVDData->aVolSetID, 128, vol_data->VolID);
    sPVDData->sDesCharSet.uCharSetType = vol_data->CharSet[0];
    memcpy( (void *) &(sPVDData->sDesCharSet.aCharSetInfo), (void *) &(vol_data->CharSet[1]), 63);
    sPVDData->sExplanatoryCharSet.uCharSetType=sPVDData->sDesCharSet.uCharSetType;
    /* explantory character set information */
    /* abstract location */
    /* copyright notice location */
    /* application identifier */
    Error = ot_setOtTime(&(sPVDData->sRecordingTime));
    sPVDData->sImplementationID.uFlags = 0x0;
    memcpy( &(sPVDData->sImplementationID.aRegisteredID), SFS_REGID, sizeof( SFS_REGID));
    memcpy( &(sPVDData->sImplementationID.aIDSuffix), SFS_PVD_REGID, sizeof( SFS_PVD_REGID));
    sPVDData->uPredecessorVDSLoc = 0;
    sPVDData->uFlags=0x1;

    /* fill in the tag values */
    FillTag(TAGID_PVD, 1, 1, NOCRC, 0, Location, (struct Tag *) sPVDData);

    /* translate the structures */
#if !IBM
    ot_xlate(TAGID_PVD, TO_OP, (char *) sPVDData,0,0);
#endif

    /* write the data */
    Error = ot_write( fsp, Location, (caddr_t) sPVDData, (UINT32) fsp->SectorSize);

    kmem_free( (caddr_t) sPVDData, fsp->SectorSize, BUFRAM);

    return( Error);

}
#pragma page()
/******************************************************************/
/*                                                                */
/*  Write an Unallocated Space Descriptor at Location             */
/*                                                                */
/******************************************************************/
INT32 WriteUSD( struct otfs *fsp, UINT32 Location, void *USD,UINT32 seqnum)
{
    INT32 Error;
    UINT32 bsize;
    
    ((struct UnallocSpDesHead *)USD)->uVolDescSeqNum = seqnum;
    
    /* fill in the tag values */
    FillTag(TAGID_USD, 1, 1, NOCRC, 0, Location, (struct Tag *) USD);
    /* translate the structures */
#if !IBM
    ot_xlate(TAGID_USD, TO_OP, (char *) USD,0,0);
#endif

    bsize = sizeof(struct UnallocSpDesHead)
          + sizeof(struct extent_ad)*MAX_USD;
    
    /* round up to nearest number of sectors */
    if ( (bsize % fsp->SectorSize) != 0 ) {
        bsize = fsp->SectorSize*((bsize+fsp->SectorSize-1)/fsp->SectorSize);
    }
    
    /* write the data */
    Error = ot_write( fsp, Location, (caddr_t) USD, bsize);

#if !IBM
    ot_xlate(TAGID_USD,FROM_OP, (char *) USD,0,0);
#endif
    return( Error);
}
#pragma page()
/******************************************************************/
/*                                                                */
/* Fetch the next volume descriptor number and return in VDSN     */
/* op is either the define GET or PUT.  Put sets the counter      */
/* for initialization and GET gets the next number in the seq.    */
/*                                                                */
/******************************************************************/
UINT32 NextVolDesSeqNum( INT16 op, UINT32 VDSN)
{
    static UINT32 NextVDSN = 1;

    if ( op == GET ) {
        return( NextVDSN++);
    }
    else {
        if (op== PUT) {
            NextVDSN = VDSN;
            return( NextVDSN);
        } 
        else {
            return (-1);
        }
    }
}

#pragma page()
/******************************************************************/
/*                                                                */
/*                                                                */
/*  Write terminating descriptor at Location                      */
/*                                                                */
/******************************************************************/
INT32 WriteTerminator( struct otfs *fsp, UINT32 Location)
{
    struct Terminator *Term;
    INT32 Error;
    
    Term = (struct Terminator *) kmem_alloc( fsp->SectorSize, BUFRAM);
    bzero( (char *) Term, fsp->SectorSize);

    /* write a Terminator Descriptor at Location */
    /* fill in the tag values */
    FillTag(TAGID_TERM_DESC, 1, 1, NOCRC, 0, Location, (struct Tag *) Term);

    /* translate the structures */
#if !IBM
    ot_xlate(TAGID_TERM_DESC, TO_OP, (char *) Term,0,0);
#endif

    /* write the data */
    Error = ot_write( fsp, Location, (caddr_t) Term, fsp->SectorSize);

    kmem_free( (caddr_t) Term, fsp->SectorSize, 0);

    return( Error);
}

#pragma page()
/******************************************************************/
/*                                                                */
/*                                                                */
/* Write a partition descriptor at location                       */
/*                                                                */
/******************************************************************/
INT32 WritePD( struct otfs *fsp, UINT32 Location, 
               struct vol_info *vol_data,
               struct PartDesc *sPDData,
               UINT32 seqnum)
{
    struct PartHeaderDesc *PHDp;
    INT32 Error;

    bzero( (char *) sPDData, sizeof(*sPDData));
        
    /* write a Partition Descriptor at Location */

    /* load up the rest of the data */
    sPDData->uVolDescSeqNum = seqnum;
    sPDData->uPartFlags = 0x1;
    sPDData->uPartNumber = 0;
    /* Suffix field of regid for partition contents */
    /* The makefilesystem part must fill in the aPartContentsUse field */
    /* access type */
    sPDData->uPartStartingLoc = vol_data->PartStart;
    sPDData->uPartLength = vol_data->PartLen;
    sPDData->sImplementationID.uFlags = 0x0;
    memcpy( &(sPDData->sImplementationID.aRegisteredID), SFS_REGID, sizeof( SFS_REGID));
    memcpy( &(sPDData->sImplementationID.aIDSuffix), SFS_PD_REGID, sizeof( SFS_PD_REGID));

    /* the following structures are normally defined in the make file system
       code, but for the time being are going to be set here */

    /* partition contents */
    /* always the "+NSR01" descriptor */
    sPDData->sPartContents.uFlags = 0x0;
    memcpy( &(sPDData->sPartContents.aRegisteredID), NSRPARTCONTENTS, sizeof(NSRPARTCONTENTS));
    /* partition contents use - where are space maps and so on */
    PHDp = (struct PartHeaderDesc *) sPDData->aPartContentsUse;
    /* partition integrity entry */
    PHDp->PIT.Location = START_PIT-sPDData->uPartStartingLoc;
    /* set up tables or bitmaps and Access Type */
    /* note that we don't handle the case of fsMagnetic */
    if( vol_data->MediaType <= fsWORM) {
        PHDp->PIT.ExtentLength.bf.Length = ICB_BLOCK_SIZE_WORM * vol_data->BlkSize;
        PHDp->PIT.ExtentLength.bf.Type   = E_RECORDED;
        sPDData->uAccessType = ACCESS_WORM;
        PHDp->UST.ExtentLength.bf.Length = ICB_BLOCK_SIZE_WORM * vol_data->BlkSize;
        PHDp->UST.ExtentLength.bf.Type   = E_UNALLOCATED;
        PHDp->UST.Location = START_UST-sPDData->uPartStartingLoc;
        PHDp->FST.ExtentLength.bf.Length = ICB_BLOCK_SIZE_WORM * vol_data->BlkSize;
        PHDp->FST.ExtentLength.bf.Type   = E_UNALLOCATED;
        PHDp->FST.Location = START_FST-sPDData->uPartStartingLoc;
    } else if (vol_data->MediaType == fsMO) {
        PHDp->PIT.ExtentLength.bf.Length = ICB_BLOCK_SIZE_MO * vol_data->BlkSize;
        PHDp->PIT.ExtentLength.bf.Type   = E_RECORDED;
        sPDData->uAccessType = ACCESS_REWRITABLE;
        PHDp->USB.ExtentLength.bf.Length = fsp->SectorsInBMap * vol_data->BlkSize;
        PHDp->USB.ExtentLength.bf.Type   = E_RECORDED;
        PHDp->USB.Location = START_USB-sPDData->uPartStartingLoc;
        PHDp->FSB.ExtentLength.bf.Length = PHDp->USB.ExtentLength.Length32;
        PHDp->FSB.Location = START_USB + fsp->SectorsInBMap - sPDData->uPartStartingLoc;
     }

    /* fill in the tag values */
    FillTag(TAGID_PD, 1, 1, NOCRC, 0, Location, (struct Tag *) sPDData);
       
    /* translate the structures */
#if !IBM
    ot_xlate( TAGID_PD, TO_OP, (char *) sPDData,0,0);
#endif

    /* write the data */
    Error = ot_write( fsp, Location, (caddr_t) sPDData, (UINT32) fsp->SectorSize);

#if !IBM
    /* xlate it back since use data later */
    ot_xlate( TAGID_PD, FROM_OP, (char *) sPDData,0,0);
#endif
    
    return( Error);
}
#pragma page()
/******************************************************************/
/*                                                                */
/*                                                                */
/*  Write a logical volume descriptor at Location                 */
/*                                                                */
/******************************************************************/
INT32 WriteLVD( struct otfs *fsp, UINT32 Location, struct vol_info *vol_data,
                struct LogVolDesc *sLVDData,UINT32 seqnum)
{
    INT32 Error;

    bzero( (char *) sLVDData, sizeof(*sLVDData));
        
    /* write a Logical Volume Descriptor at Location */

    /* fill in the details */ 
    sLVDData->uVolDescSeqNum = seqnum;
    sLVDData->sDesCharSet.uCharSetType = vol_data->CharSet[0];
    memcpy( (void *) &(sLVDData->sDesCharSet.aCharSetInfo), (void *) &(vol_data->CharSet[1]), 63);
    str2dstr( sLVDData->uLogVolID, 128, vol_data->LogID);
    sLVDData->uLogBlkSize = vol_data->BlkSize;
    /* Domain registered ID */
    
    /* the following two lines along with the setting of sPartMaps later
       limit support to 1 partition per logical volume */
    sLVDData->uMapTabLen = 6;
    sLVDData->uNumPartMaps = 1;
    sLVDData->sImplementationID.uFlags = 0;
    memcpy( &(sLVDData->sImplementationID.aRegisteredID), SFS_REGID, sizeof( SFS_REGID));
    memcpy( &(sLVDData->sImplementationID.aIDSuffix), SFS_LVD_REGID, sizeof( SFS_LVD_REGID));
    
    /* only one partition in the Logical Volume */
    sLVDData->sPartMaps[0].uPartMapType=1;
    sLVDData->sPartMaps[0].uPartMapLen=6;
    sLVDData->sPartMaps[0].uVSN=1;
    sLVDData->sPartMaps[0].uPartNum=0;

    /* the following fields are normally set by the make file system code,
       but for the time being they will be defined here */

    /*
    * Logical Volume Contents Use 
    * (points to the File Set Descriptor)
    */
    ((struct long_AD *) (sLVDData->uLogVolUse))->ExtentLength.bf.Length = vol_data->BlkSize * FSDS_LEN;
    ((struct long_AD *) (sLVDData->uLogVolUse))->ExtentLength.bf.Type   = E_RECORDED;
    ((struct long_AD *) (sLVDData->uLogVolUse))->Location_LBN = 0;
    ((struct long_AD *) (sLVDData->uLogVolUse))->Location_PartNo = 0;

    /* fill in the tag values */
    FillTag(TAGID_LVD, 1, 1, NOCRC, 0, Location, (struct Tag *) sLVDData);

    /* translate the structures */
#if !IBM
    ot_xlate(TAGID_LVD, TO_OP, (char *) sLVDData, 0,0);
#endif

    /* write the data */
    Error = ot_write( fsp, Location, (caddr_t) sLVDData, (UINT32) fsp->SectorSize);

#if !IBM
    /* xlate the structure back since later use the Location_LBN */
    ot_xlate(TAGID_LVD, FROM_OP, (char *) sLVDData, 0,0);
#endif
    
    return( Error);
}
